Module class_CoordinateMap     
!! Coordinate maps describe the distribution of
!! gridpoint locations along a coordinate axis.
!! Notation: The coordinate x is parameterized by s, 
!! i.e. x = x(s). The discrete arrays [x] [s] [dxds] 
!! etc. are related through a discrete sampling of
!! an analytically (generally) defined mapping. 

 implicit none
 public         :: NewCoordinateMap
 public         :: DescribeCoordinateMap
 public         :: DeleteCoordinateMap
 
 type CoordinateMap
    character (len=80)                      :: label  
    integer                                 :: npts
    real(kind=8), pointer, dimension(:)     :: x      
    real(kind=8), pointer, dimension(:)     :: s      
    real(kind=8), pointer, dimension(:)     :: dxds   
    real(kind=8), pointer, dimension(:)     :: d2xds2 
    real(kind=8)                            :: ds,L     
    real(kind=8), dimension(2)              :: xlimits
    logical                                 :: uniform_dx
 end type CoordinateMap
 

Contains

 function NewCoordinateMap              &
           (N,label,x,s,dxds,d2xds2,L)   &
  result(map)
  
  implicit none 
  type( CoordinateMap )             :: map
  integer            ,intent(in)    :: N
  integer                           :: i
  character (len=80 ),intent(in)    :: label
  real(kind=8)       ,intent(in)    :: x     (N)
  real(kind=8)       ,intent(in)    :: s     (N)
  real(kind=8)       ,intent(in)    :: dxds  (N)
  real(kind=8)       ,intent(in)    :: d2xds2(N)
  real(kind=8)       ,intent(in)    :: L
    
  !! Allocate CoordinateMap arrays
  allocate( map%x     (N) ) 
  allocate( map%s     (N) ) 
  allocate( map%dxds  (N) )
  allocate( map%d2xds2(N) )
  
  map%npts        = N
  map%label       = label
  map%x(1:N)      = x(1:N)
  map%s(1:N)      = s(1:N)
  map%xlimits(1)  = x(1)
  map%xlimits(2)  = x(N)
  if(N>1) then
   map%ds          = s(2)-s(1)
  else
   map%ds=1.d0
  endif
  map%dxds(1:N)   = dxds  (1:N)
  map%d2xds2(1:N) = d2xds2(1:N)
  map%L = L  !! dimensionless
  
  map%uniform_dx = .TRUE.
  do i=2,N
   if( dxds(i) /= dxds(1) ) map%uniform_dx = .FALSE.
  enddo
  
 end function NewCoordinateMap
  
  
 subroutine DescribeCoordinateMap(map,docfile)
  implicit none
  type( CoordinateMap ), intent(in)   :: map
  character (len=80),    intent(in)   :: docfile
  integer                             :: i
  
   open(1,file=docfile,position='append')
    write(1,*) 'Map Label: ',trim(map%label)
    write(1,*) 'uniform_dx: ',map%uniform_dx
    write(1,*) '=================================================================================================================='
    write(1,*) '                   s                              x                           dx/ds                    d2x/ds2    '
    write(1,*) '=================================================================================================================='
    do i=1,map%npts
     write(1,100) i,map%s(i),map%x(i),map%dxds(i),map%d2xds2(i)
    enddo
    write(1,*) '   '
    write(1,*) '   '  
   close(1)
 100 format(1x,i3,4e22.10)
 end subroutine DescribeCoordinateMap
 
 
 subroutine DeleteCoordinateMap(map)
  implicit none
  type( CoordinateMap ),intent(inout) :: map
  deallocate(map%x)
  deallocate(map%s)
  deallocate(map%dxds)
  deallocate(map%d2xds2)
 end subroutine DeleteCoordinateMap
 
end Module class_CoordinateMap

