Module class_Derivative         !============================================
  implicit none
  public  :: NewDerivative
  public  :: DeleteDerivative
  public  :: EqualDerivs
  private :: CreatePlan
  private :: SetupAndFactor
  private :: DefineCompactScheme   
 
 !!==========================================================================
 !! A class containing data and algorithms for differentiating
 !! a 3d data "cube" in a single specified direction.
 !!==========================================================================
 type Derivative                      
  integer                                :: DimIndex     !! Dimension to be differentiated
  integer                                :: StorageDim   !! Size of array in differentiation dimension.
  integer                                :: DataDim      !! Corresponding number of data values.
  real(kind=8)                           :: h            !! Grid spacing w/ respect to coordinate's parameter.  
  character (len=80)                     :: MethodLabel  !! Label defining differentiion algorithm.
  character (len=80)                     :: Order        !! Order of convergence.
  integer(kind=8)                        :: forward_plan !! FFTW3 plan for forward "MethodLabel" transform.
  integer(kind=8)                        :: inverse_plan !! FFTW3 plan for inverse "MethodLabel" transform.
  integer                                :: rank_trans
  integer                                :: n_trans
  integer                                :: howmany_rank
  integer                                :: is_trans
  integer                                :: is_loop
  integer                                :: os_trans
  integer                                :: os_loop
  integer                                :: n_loop
  integer                                :: fft_kind
  integer                                :: sign,myflag  !! keep enough info to regenerate plan  
  real(kind=8), pointer, dimension(:)    :: kx           !! Wavenumber array used for spectral differentiation.
  real(kind=8), pointer, dimension(:)    :: kfilter      !! Wavenumber array used for spectral smoothing filter.
  
  real(kind=8), pointer, dimension(:,:)  :: A_Diff       !! Compact differentiation matrix, LAPACK format.
  integer,      pointer, dimension(:)    :: ipiv         !! LU pivot array, LAPACK format.  
  real(kind=8)                           :: a1,b1,c1     !! compact scheme coefficients
  
  real(kind=8), pointer, dimension(:,:)  :: A_Filt       !! Compact filtering matrix, LAPACK format.
  integer,      pointer, dimension(:)    :: jpiv         !! LU pivot array, LAPACK format.  
  real(kind=8)                           :: alpha,beta   !! compact filter scheme coefficients
  real(kind=8)                           :: a,b,c,d      !! compact filter scheme coefficients  
 end type Derivative
 !!==========================================================================
 !!==========================================================================
 
 
 !!==========================================================================
 interface assignment (=)
  module procedure EqualDerivs
 end interface
 !!==========================================================================
 
Contains           

 subroutine DeleteDerivative(Deriv)
  implicit none
  type( Derivative ),intent(inout)  :: Deriv
  if( associated(Deriv%kx) ) deallocate(Deriv%kx)
  if( associated(Deriv%kfilter) ) deallocate(Deriv%kfilter)
  if( associated(Deriv%A_Diff) ) deallocate(Deriv%A_Diff)
  if( associated(Deriv%ipiv) ) deallocate(Deriv%ipiv)
  if( associated(Deriv%A_Filt) ) deallocate(Deriv%A_Filt)
  if( associated(Deriv%jpiv) ) deallocate(Deriv%jpiv)
 end subroutine DeleteDerivative
 
 subroutine EqualDerivs(new,old)
 implicit none
  type( Derivative ), intent(inout) :: new
  type( Derivative ), intent(in)    :: old  
  integer                           :: n
  
  call DeleteDerivative(new)
   
  n=old%DataDim
  new%DimIndex = old%DimIndex
  new%StorageDim = old%StorageDim
  new%DataDim = old%DataDim
  new%h = old%h
  new%MethodLabel = old%MethodLabel
  new%Order = old%Order
  new%forward_plan = old%forward_plan
  new%inverse_plan = old%inverse_plan
  new%rank_trans = old%rank_trans
  new%n_trans = old%n_trans
  new%is_trans = old%is_trans
  new%os_trans = old%os_trans
  new%howmany_rank = old%howmany_rank
  new%n_loop = old%n_loop
  new%is_loop = old%is_loop
  new%os_loop = old%os_loop
  new%fft_kind = old%fft_kind
  new%sign = old%sign
  new%myflag = old%myflag
  new%a1 = old%a1
  new%b1 = old%b1
  new%c1 = old%c1
  new%a = old%a
  new%b = old%b
  new%c = old%c
  new%d = old%d

  if(old%MethodLabel=='fourier') then     !! Set Fourier wavenumber array, complex-->complex storage format.
   allocate( new%kx(n) )
   allocate( new%kfilter(n) )
   new%kx=old%kx
   new%kfilter=old%kfilter
  elseif(old%MethodLabel=='cos' .or. old%MethodLabel=='sin' ) then   !! sin/cos real-->real format
   allocate( new%kx(n) )
   allocate( new%kfilter(n) )
   new%kx=old%kx
   new%kfilter=old%kfilter
  elseif( old%MethodLabel=='compact' ) then
   allocate( new%A_diff(7,n), new%ipiv(n) ) 
   allocate( new%A_filt(7,n), new%jpiv(n) )
   new%A_Diff = old%A_Diff
   new%ipiv = old%ipiv
   new%A_Filt = old%A_Filt
   new%jpiv = old%jpiv
  endif 
 end subroutine EqualDerivs


 subroutine NewDerivative(DimIndex,DataMap,Grid,Deriv,second_deriv)
  use class_DataDecomposition
  use class_TensorProductGrid
  implicit none
   
  type( Derivative ),intent(inout)      :: Deriv
  type( TensorProductGrid ),intent(in)  :: Grid
  type( DataDecomposition ),intent(in)  :: DataMap
  
   
  character (len=80 )        :: MethodLabel,InvLabel,MapLabel
  character (len=80 )        :: docfile
  integer                    :: StorageDims(3),DataDims(3)
  integer                    :: DimIndex,i,j,k,Sign,n,nb2
  integer                    :: myid,numprocs,ierr
  real(kind=8)               :: dk,pi,kstar,kmax,taper
  logical                    :: second_deriv
  include 'mpif.h'
  
  pi=4.d0*atan(1.d0)
  docfile='classlog'
  call mpi_comm_rank(MPI_COMM_WORLD,myid,ierr)      
  call mpi_comm_size(MPI_COMM_WORLD,numprocs,ierr)
  
  if( associated( Deriv%kx ) .or.       &
      associated( Deriv%A_Diff ) .or.   &
      associated( Deriv%ipiv ) )        &   
      call DeleteDerivative( Deriv )  
   
  Deriv%DimIndex = DimIndex
  Deriv%h = Grid%SpaceDims(DimIndex)%ds
  
  !!==========================================
  !! Parse the requested deriv method label
  !! for a given variable, this depends on
  !! whether 1 first or second deriv object
  !! is being constructed.
  !!===========================================
  if( .NOT. second_deriv ) then
   Deriv%MethodLabel = DataMap%DerivTypes(DimIndex)
  elseif( second_deriv ) then
   Deriv%MethodLabel = DataMap%DerivTypes(DimIndex+3)
  endif
  MethodLabel=deriv%MethodLabel
  
  
  !!===================================================
  !! Storage and Data dimensions as local variables, 
  !! all 3 dimensions. Needed to pass to create_plans.
  !!===================================================
  do i=1,3
   if(DimIndex==3) then   
    StorageDims(i)= DataMap%ColDims(i,myid)
    DataDims(i)   = DataMap%ColVals(i,myid)
   else
    StorageDims(i)= DataMap%SlabDims(i,myid)
    DataDims(i)   = DataMap%SlabVals(i,myid)
   endif
  enddo
  
  !!===================================================
  !! Derivative object only retains info about 
  !! the differentiation dimension.
  !!===================================================
  if(DimIndex==3) then
   Deriv%StorageDim = DataMap%ColDims(DimIndex,myid)
   Deriv%DataDim =    DataMap%ColVals(DimIndex,myid)
  else
   Deriv%StorageDim = DataMap%SlabDims(DimIndex,myid)
   Deriv%DataDim =    DataMap%SlabVals(DimIndex,myid)
  endif
  n=deriv%DataDim
  nb2=n/2
  
  !!=======================================
  !! Forward and Inverse transform labels.
  !!=======================================
  if( MethodLabel=='fourier' )  then
          InvLabel='fourier'
  elseif( MethodLabel=='cheby') then
          InvLabel='cheby'
  elseif( MethodLabel=='cos' )  then
          InvLabel='sin'
  elseif( MethodLabel=='sin' )  then
          InvLabel='cos'
  endif
  
  !!=======================================
  !! convergence character
  !!=======================================
  if(    MethodLabel=='fourier' .or.        &
         MethodLabel=='sin'     .or.        &
         MethodLabel=='cos'     .or.        &
         MethodLabel=='cheby'    ) Deriv%Order='exponential'
  if(    MethodLabel=='compact'  ) Deriv%Order='SixthCompact'
    
        
    Sign=1 
    !!======================================================================
    !! For sin, the for/inv data/storage arrays passed to fftw are 
    !! slightly smaller (end zeros not needed).
    !! Adjust the sizes in the transform direction, make the plan, 
    !! then adjust back to original size.
    !!======================================================================
    if( trim(MethodLabel)=='sin' ) then
      DataDims(DimIndex) = DataDims(DimIndex) - 2   
      StorageDims(DimIndex) = StorageDims(DimIndex) - 2
      
      call CreatePlan(DataDims,StorageDims,Sign,DimIndex,  &
                      MethodLabel,deriv%forward_plan,deriv,docfile)
       
      StorageDims(DimIndex) = StorageDims(DimIndex) + 2
      DataDims(DimIndex) = DataDims(DimIndex) + 2
    else
       call CreatePlan(DataDims,StorageDims,Sign,DimIndex,  &
                       MethodLabel,deriv%forward_plan,deriv,docfile)
    endif
                  
    Sign=-1 
    !!======================================================================
    !! For sin, the for/inv data/storage arrays passed to fftw are 
    !! slightly smaller (end zeros not needed).
    !! Adjust the sizes in the transform direction, make the plan, 
    !! then adjust back to original size.
    !!======================================================================     
    if(trim(InvLabel)=='sin') then
      DataDims(DimIndex) = DataDims(DimIndex) - 2   
      StorageDims(DimIndex) = StorageDims(DimIndex) - 2
          call CreatePlan(DataDims,StorageDims,Sign,DimIndex,  &
                          InvLabel,deriv%inverse_plan,deriv,docfile)

      StorageDims(DimIndex) = StorageDims(DimIndex) + 2
      DataDims(DimIndex) = DataDims(DimIndex) + 2    
    else
          call CreatePlan(DataDims,StorageDims,Sign,DimIndex,  &
                          InvLabel,deriv%inverse_plan,deriv,docfile)
    endif
  
  if( MethodLabel=='compact' ) then
  
   allocate( deriv%A_diff(7,n), deriv%ipiv(n) )
   allocate( deriv%A_filt(7,n), deriv%jpiv(n) )
   if(n>=4)then
    call DefineCompactScheme(deriv%h,n,deriv%A_diff,deriv%ipiv, &
                             deriv%a1,deriv%b1,deriv%c1)  
    call DefineCompactFilter(n,deriv%A_filt,deriv%jpiv,deriv%alpha, &
                             deriv%beta,deriv%a,deriv%b,deriv%c,deriv%d)
   else
    write(0,*) 'skipping construction of compact diff matrices, n<4 ',n
   endif
   
  endif
  
  if(MethodLabel=='fourier' ) then
  
    pi=4.d0*atan(1.)
    dk = 2.d0*pi   !! assume s in [0,1]  i.e. 2.d0*pi/L w/ L=1    
    kmax = nb2+1
    kstar = (8./9.)*kmax
    allocate( deriv%kx(n) )
    allocate( deriv%kfilter(n) )
    
    do k=1,nb2+1     
     if(k>=kstar) then
      taper = 0.5*(1. + cos(pi*(i-kstar)/(kmax-kstar)))
     else
      taper=1.d0
     endif
     deriv%kx(k)=(k-1.)*dk
     deriv%kfilter(k) = taper
    enddo
     
    do k=2*nb2,nb2+2,-1
     deriv%kx(k) = -deriv%kx( n-k+2 )     
     deriv%kfilter(k) = deriv%kfilter(n-k+2)
    enddo
    
    if(n==1) then
     deriv%kx(1)=0.d0
     deriv%kfilter(1)=1.d0
    endif
       
  elseif(MethodLabel=='cos' .or. MethodLabel=='sin') then 
    
    allocate( deriv%kx(n) )
    allocate( deriv%kfilter(n) )
    dk = pi/1.  !! assume s in [0,1] 
    kmax = n
    kstar = (8./9.)*kmax  
    do i=1,n
     if(i>=kstar) then
      taper = 0.5*(1. + cos(pi*(i-kstar)/(kmax-kstar)))
     else
      taper=1.d0
     endif
     deriv%kx(i)=(i-1.)*dk
     deriv%kfilter(i) = taper
    enddo
    
    if(n==1) then
     deriv%kx(1)=0.d0
     deriv%kfilter(1)=1.d0
    endif
        
  elseif( MethodLabel=='cheby') then
   !! we use recursion, wavenumber arrays not needed
  endif 
 end subroutine NewDerivative
 
! 
!=========================================================
 subroutine CreatePlan(DataDims,StorageDims,Sign,  &
                       Dir,TransType,plan,deriv,docfile)
 !!=======================================================
 !!Assumes in and out arrays are reduced in size by 2 
 !!in the transform direction for sin transforms
 !!=======================================================
      implicit none                
      include 'fftw3.f'
      include 'mpif.h'
      
      type( Derivative )                     :: deriv
      real(kind=8)    , allocatable          :: in(:,:,:)
      real(kind=8)    , allocatable          :: out(:,:,:)
      real(kind=8)    , allocatable          :: testin(:,:,:)
      real(kind=8)    , allocatable          :: testout(:,:,:)
      real(kind=8)    , allocatable          :: kx(:)
      real(kind=8)                           :: tmax(2)
      integer*8                              :: plan
      integer                                :: DataDims(3)
      integer                                :: StorageDims(3)
      integer                                :: Sign,Dir
      integer                                :: rank_trans
      integer                                :: howmany_rank
      integer                                :: n_trans,n_loop
      integer                                :: fft_kind,myflag
      integer                                :: is_trans,is_loop
      integer                                :: os_trans,os_loop
      integer                                :: i,nx,ny,nz,n,nb2 
      integer                                :: myid,numprocs,ierr 
      character (len=80)                     :: TransType
      character (len=80)                     :: label,kind_label
      character (len=80)                     :: reality_in
      character (len=80)                     :: reality_out
      character (len=80)                     :: docfile
      
      call mpi_comm_rank(MPI_COMM_WORLD,myid,ierr)      
      call mpi_comm_size(MPI_COMM_WORLD,numprocs,ierr)
      
      rank_trans=1      !! Transforms work on 1 dimension at a time.
      howmany_rank=1    !! Loops over the other 2 dimensions treated as 1d in memory
      n = DataDims(Dir) !! Data size in transform direction, measured in real words.
      nb2 = n/2 
            
      if(TransType=='fourier') then
      
        if(Sign==1) then
         reality_in='real'
         reality_out='halfcomplex'
        elseif(Sign==-1) then
         reality_in='halfcomplex'
         reality_out='real'
        endif
        call get_kind(reality_in,reality_out,TransType,fft_kind,label) 
                     
       if(Dir==1) then  
     
            !!describe the transforms  in dim 1
            n_trans=DataDims(1)      ! length of transforms
            is_trans=1               ! stride btwn elements in a single transform
            os_trans=is_trans
        
            !!describe the loops over dims 2,3
            n_loop=DataDims(2)*DataDims(3)   ! 
            is_loop=DataDims(1)              ! stride betwn transforms
            os_loop=is_loop

       elseif(Dir==2) then !! must loop over k because stride 
                           !! is not constant otherwise      
                        
            !!describe the transforms in dim 2
            n_trans=DataDims(2)            !! transform length = ny
            is_trans=StorageDims(1)        !! stride of 1d transform 
            os_trans=is_trans
        
            !!describe the loop over dim 1
            n_loop=DataDims(1)             !! loop over i direction only
            is_loop=1                      !! idim defined by a stride of 1
            os_loop=is_loop
                                
       elseif(Dir==3) then  
            
            !!describe the transforms in dim 3
            n_trans=DataDims(3)              !! length of 1d transforms
            is_trans= StorageDims(1) *     & !! stride of 1d transform --> 
                      StorageDims(2)         !! xy plane counted in real words
            os_trans=is_trans
         
            !!describe the loops over dims 1,2
            n_loop=DataDims(1)*DataDims(2)  ! loop over i,j directions (real)
            is_loop=1                       ! which is defined by a stride of 1
            os_loop=os_loop
            
       endif
  
     elseif(TransType=='cos' .or. TransType=='sin' .or. TransType=='cheby') then    
      reality_in=  'real'
      reality_out= 'real'
      call get_kind(reality_in,reality_out,TransType,fft_kind,label)
  
      if(Dir==1) then                               
          !!describe the transforms  in dim 1
          n_trans=DataDims(1)            !! length of transform in real words     
          is_trans=1                     !! stride of 1d transform --> in x
          os_trans=is_trans
        
          !!describe the loops over dims 2,3
          n_loop=DataDims(2)*DataDims(3)  !! loop over j/k directions 
          is_loop=StorageDims(1)          !! defined by a stride of StorageDims(1)
          os_loop=is_loop
    
       elseif(Dir==2) then              
         !!describe the transforms in dim 2, SET UP FOR 1 K PLANE AT A TIME
         n_trans=DataDims(2)        
         is_trans=StorageDims(1)    ! stride of 1d transform --> StorageDims(1)
         os_trans=is_trans
        
        !!describe the loop over dim 1
         n_loop=DataDims(1)         ! loop over i direction (len=DataDims(1))
         is_loop=1                  !   which is defined by a stride of 1
         os_loop=is_loop
                 
       elseif(Dir==3) then
         !!describe the transforms
         n_trans=DataDims(3)            ! length of transforms
         is_trans=StorageDims(1) *  &
                  StorageDims(2)        ! stride=size of xy planes, in real words
         os_trans=is_trans
         
         !!describe the loops
         n_loop=DataDims(1)*DataDims(2)  ! loop over i/j directions
         is_loop=1                       ! which is defined by a stride of 1
         os_loop=is_loop
         
      endif
         
         
     endif
     
     !!================================
     !! Keep fftw info in deriv object
     !!================================
     deriv%rank_trans = rank_trans
     deriv%n_trans = n_trans
     deriv%is_trans = is_trans
     deriv%os_trans = os_trans
     deriv%howmany_rank = howmany_rank
     deriv%n_loop = n_loop
     deriv%is_loop = is_loop
     deriv%os_loop = os_loop
     deriv%fft_kind = fft_kind
     deriv%myflag = myflag
     
     kind_label=label
     label = 'FFTW_EXHAUSTIVE'
          
     !! FFTW_PATIENT+FFTW_UNALIGNED  !! in Fortran  p.50 6.2 of manual     
     myflag = FFTW_EXHAUSTIVE   !! the integer parameter
      
          
  !! Allocate arrays 
     allocate(  in(StorageDims(1),StorageDims(2),StorageDims(3)) )
     allocate( out(StorageDims(1),StorageDims(2),StorageDims(3)) )
     !!allocate(  testin(StorageDims(1),StorageDims(2),StorageDims(3)) )
     !!allocate( testout(StorageDims(1),StorageDims(2),StorageDims(3)) )
           
     call dfftw_plan_guru_r2r(plan,              &
                              rank_trans,        &
                              n_trans,           &
                              is_trans,          &
                              os_trans,          &
                              howmany_rank,      &
                              n_loop,            &
                              is_loop,           &
                              os_loop,           &
                              in,                &
                              out,               &
                              fft_kind,          &
                              myflag)
                                                                           
     deallocate(in)
     deallocate(out)
     !!deallocate(testin)
     !!deallocate(testout)
       
          
     if(myid==0) then
     open (1,file=docfile,position='append')
      write(1,*) ' '
      write(1,*) ' '
      write(1,*) '=================================================================' 
      write(1,*) ' Transform for dimension ',Dir,' isign = ',sign,' transform type:   ',TransType
      write(1,*) '  The following parameters are counts in terms of real words.               '
      write(1,*) '  The number of data pts in the transform dimension                     n = ',n_trans
      write(1,*) '  Number of words btwn successive elements in transform direction         = ',is_trans
      write(1,*) '  Number of words btwn successive 1st elements in loops over transforms   = ',is_loop
      write(1,*) '  Number of 1d transforms in single call, i.e. product of loop lengths    = ',n_loop
      write(1,*) '  StorageDims, i.e. array size for in/out                                   ',StorageDims
      write(1,*) '  DataDims,    i.e. how much of it is filled w/ data                        ',DataDims
      write(1,*) '  FFTW3 plan quality                                                        ',trim(label)
      write(1,*) '  FFT_KIND , label                                                          ',fft_kind,trim(kind_label)
      write(1,*) '  FFTW3 plan                                                                ',plan
      write(1,*) '  Actual Bounds on in/out arrays used by fftw3 to define plan '
      write(1,*) '  i dimension-- in    : ',LBOUND(in,1), UBOUND(in,1)
      write(1,*) '               out    : ',LBOUND(out,1), UBOUND(out,1)
      write(1,*) '  j dimension-- in    : ',LBOUND(in,2), UBOUND(in,2)
      write(1,*) '               out    : ',LBOUND(out,2), UBOUND(out,2)
      write(1,*) '  k dimension-- in    : ',LBOUND(in,3), UBOUND(in,3)
      write(1,*) '               out    : ',LBOUND(out,3), UBOUND(out,3)
      write(1,*) '  in,out arrays have KINDs of ',KIND(in),KIND(out)
      write(1,*) '================================================================='
      write(1,*) ' '
      write(1,*) ' ' 
     close(1)
     endif
       
    end subroutine CreatePlan
    
    !======================================================================================
      subroutine get_kind(in,out,type,fft_kind,label)
       implicit none                
       include 'fftw3.f'
       integer,intent(out)               :: fft_kind
       character (len=80),intent(out)    :: label
       character (len=11)                :: in
       character (len=11)                :: out
       character (len=11)                :: type
       
       
       if( trim(type)=='fourier' ) then
        if( in=='real' .and. out=='complex') then
         fft_kind=FFTW_R2HC
         label='no flag needed'
        elseif( in=='complex' .and. out=='real') then
         fft_kind=FFTW_HC2R
         label='no flag needed'
        elseif( in=='real' .and. out=='halfcomplex') then
         fft_kind=FFTW_R2HC
         label='FFTW_R2HC'
        elseif( in=='halfcomplex' .and. out=='real') then
         fft_kind=FFTW_HC2R
         label='FFTW_HC2R'
        elseif( in=='complex' .and. out=='complex') then
         fft_kind=-999
         label='no flag needed'
        else
         write(0,*) 'problem getting kind of fourier transform',in,out
         stop
        endif        
       endif
       
       if( trim(type)=='sin' ) then
        if( in=='real' .and. out=='real') then
         fft_kind=FFTW_RODFT00
         label='FFTW_RODFT00'
        elseif( in=='complex' .and. out=='complex') then
         fft_kind=-999
         label='no label needed'
        else
         write(0,*) 'problem getting kind of fourier transform',in,out
         stop
        endif        
       endif
       
       if( trim(type)=='cos' .or. trim(type)=='cheby' ) then
        if( in=='real' .and. out=='real') then
         fft_kind=FFTW_REDFT00
         label='FFTW_REDFT00'
        elseif( in=='complex' .and. out=='complex') then
         fft_kind=-999
         label='no label needed'
        else
         write(0,*) 'problem getting kind of fourier transform',in,out
         stop
        endif        
       endif         
      
      end subroutine get_kind
      
    subroutine DefineCompactScheme(dx,nx,A_diff,ipiv,a1,b1,c1)
    !! Routine to create & factor the various differentiation matrices
    !! corresponding to the optimized 6th order compact difference 
    !! operators from Lui & Lele 2001.
    implicit none

    integer,intent(inout)               :: ipiv(nx)
    real(kind=8),intent(inout)          :: A_diff(7,nx)
    real(kind=8)                        :: dx
    real(kind=8)                        :: alpha_1,beta_1,a1,b1,c1
    real(kind=8)                        :: alpha
    integer                             :: nx
    integer                             :: scheme
    
    alpha=0.4975        ! filtering parameter
    scheme=1

   if(scheme==1) then
   ! set parameters of the 6th order scheme described in
   ! the Lui & Lele 2001 conference paper
    alpha_1 = 0.5381301
    beta_1  = 0.0666332
    a1=1.3675777
    b1=0.8234282
    c1=0.0185208
   elseif(scheme==2) then
   ! spectral like scheme of Lele 1992
   ! formally 4th order but supposedly superior resolution characteristics
    alpha_1 = 0.5771439
    beta_1  = 0.0896406
    a1=1.3025166
    b1=0.9935500
    c1=0.03750245
  endif

! 1st derivative operator
  call SetupAndFactor(nx,A_diff,ipiv,alpha_1,beta_1)

end subroutine DefineCompactScheme

subroutine SetupAndFactor(nx,AB,ipiv,alpha_1,beta_1)
 !! SetupAndFactor uses double precision LAPACK routine DGBTRF.
  implicit none
  
  integer                    :: nx,i,j,ierr,KL,KU
  integer,intent(inout)      :: ipiv(nx)
  real(kind=8),intent(inout) :: AB(7,nx)
  real(kind=8)               :: alpha_1,beta_1

  KL=2
  KU=2
  AB=0.
! 1st 2 rows of banded matrix array AB are reserved for fill in
! 1st/last k elements for the kth super/sub diagonal are not accessed
! superdiagonal 2
   AB(3,3:5)=0.0
   AB(3,6:nx-1)=beta_1
   AB(3,nx)=0.0
! superdiagonal 1
   AB(4,2)= 2.     !!!! had 0
   AB(4,3)=1./2.   
   AB(4,4)=1./3.
   AB(4,5:nx-2)=alpha_1
   AB(4,nx-1)=1./3.
   AB(4,nx)=1./6.   
! diagonal 
   AB(5,1:nx)=1.0
! subdiagonal 1
  do i=1,nx
   AB(6,i)=AB(4,nx-i+1)
  enddo
! subdiagonal 2
  do i=1,nx
   AB(7,i)=AB(3,nx-i+1)
  enddo

! factor the banded matrix AB using LAPACK routine DGBTRF
! AB is overwritten with its LU factorization

     call DGBTRF(nx,nx,KL,KU,AB,7,ipiv,ierr)

! check error condition
  if( ierr .ne. 0 ) then
    write(0,*) ' factorization error ',ierr, ' size: ',nx
    write(0,*) alpha_1, beta_1
    if(ierr > 0 ) then
     write(0,*) ' EXACT SINGULARITY DETECTED,'  
     write(0,*) ' CHECK INPUT TO DGBTRF CALL IN SUBROUTINE SetupAndFactor'
    endif
    if(ierr < 0 ) then
     write(0,*) ierr ,'th input argument to DGBTRF'
     write(0,*) ' in subroutine SetupAndFactor illegal'
    endif
    STOP
  endif

end subroutine SetupAndFactor

subroutine DefineCompactFilter(nx,A_filt,jpiv,alpha,beta,a,b,c,d)
    !! Routine to create & factor the various differentiation matrices
    !! corresponding to the optimized 6th order compact difference 
    !! operators from Lui & Lele 2001.
    implicit none

    integer,intent(inout)               :: jpiv(nx)
    real(kind=8),intent(inout)          :: A_filt(7,nx)
    real(kind=8)                        :: alpha,beta,a,b,c,d
    integer                             :: nx, scheme=3
    
   if( scheme==1) then
   ! set parameters of the scheme described in
   ! the Lui & Lele JCP paper Fig 20 c
    alpha = 0.6522474
    beta  = 0.1702929
    a=0.9891856
    b=1.321180
    c=0.3333548
    d=.00135985
   elseif(scheme==2) then
   ! set parameters of the scheme described in
   ! the Lui & Lele JCP paper Fig 20 b
    alpha = 0.4627507
    beta  = 0.2265509
    a=0.847063
    b=1.166845
    c=0.3422386
    d=.02245659
   elseif(scheme==3) then
   ! set parameters of the scheme described in
   ! the Lui & Lele JCP paper Fig 20 a
    alpha = 0.450
    beta  = (3.-2.*alpha)/10.
    a=(2.+3.*alpha)/4.
    b=(6.+7.*alpha)/8.
    c=(6.+alpha)/20.
    d=(2.-3.*alpha)/40.
   endif

  call SetupAndFactorFilter(nx,A_filt,jpiv,alpha,beta)
end subroutine DefineCompactFilter
 
subroutine SetupAndFactorFilter(nx,AB,jpiv,alpha,beta)
 !! SetupAndFactorFilter uses double precision LAPACK routine DGBTRF.
  implicit none
  
  integer                    :: nx,i,j,ierr,KL,KU,seven=7
  integer,intent(inout)      :: jpiv(nx)
  real(kind=8),intent(inout) :: AB(7,nx)
  real(kind=8)               :: alpha,beta

  KL=2
  KU=2
  AB=0.
! 1st 2 rows of banded matrix array AB are reserved for fill in
! 1st/last k elements for the kth super/sub diagonal are not accessed
! superdiagonal 2
   AB(3,3:5)=0.0
   AB(3,6:nx-1)=beta
   AB(3,nx)=0.0
! superdiagonal 1
   AB(4,2:4)= 0.
   AB(4,5:nx-2)=alpha
   AB(4,nx-1:nx)=0.
! diagonal 
   AB(5,1:nx)=1.0
! subdiagonal 1
  do i=1,nx
   AB(6,i)=AB(4,nx-i+1)
  enddo
! subdiagonal 2
  do i=1,nx
   AB(7,i)=AB(3,nx-i+1)
  enddo
! factor the banded matrix AB using LAPACK routine DGBTRF
! AB is overwritten with its LU factorization

     call DGBTRF(nx,nx,KL,KU,AB,seven,jpiv,ierr)

! check error condition
  if( ierr .ne. 0 ) then
    write(0,*) ' factorization error ',ierr, ' size: ',nx
    if(ierr > 0 ) then
     write(0,*) ' EXACT SINGULARITY DETECTED,'  
     write(0,*) ' CHECK INPUT TO DGBTRF CALL IN SUBROUTINE SetupAndFactorFilter'
    endif
    if(ierr < 0 ) then
     write(0,*) ierr ,'th input argument to DGBTRF'
     write(0,*) ' in subroutine SetupAndFactor illegal'
    endif
    STOP
  endif

end subroutine SetupAndFactorFilter
end Module class_Derivative   !============================================





