!!=============================================================!!
!!    Statified flow solver in in 2 or 3 space dimensions,     !!
!!    defined over a tensor product grid. [i.e. 2 or 3         !!
!!    independent vectors, each uniformly spaced wrt           !!
!!    computational coordinates s, but arbitrarily mapped      !!
!!    (x<-->s) to physical space coordinates completeley       !!
!!    define the gridpoints.]                                  !!
!!=============================================================!!
!!    Features:                                                !!
!!                                                             !!
!!    Distributed memory parallel version using MPI, f90/95.   !!
!!                                                             !!
!!    Two scalars, which can be configured as combinations of  !!
!!    density, temperature, salinity or passive tracers.       !!
!!                                                             !!
!!    Lagrangian particle tracking based on cubic spline       !!
!!    interpolation.                                           !!
!!                                                             !!
!!    Spatial differentiation via fourier, sin, cos, or        !!
!!    Chebyshev expansion methods or using 6th order compact   !!
!!    differencing.                                            !!
!!                                                             !!
!!    Depending on boundary conditions and whether the grid    !!
!!    is stretched or not, the pressure solves prior to the    !!
!!    projection step are done in wavenumber space when        !!
!!    possible, either exactly for uniform gridding or         !!
!!    iteratively for stretched grids. For inhomogeneous       !!
!!    BCs, a matrix free iterative scheme, preconditioned      !!
!!    by incomplete factorization of the associated 2nd order  !!
!!    FD operator is used. All outer iterations are based on   !!
!!    Krylov methods implemented using Petsc.                  !!
!!                                                             !!
!!    2nd, 3rd or 4th order Adams Bashforth time stepping      !!
!!    for the inertia, buoyancy and rotation terms.            !!
!!                                                             !!
!!    3rd or 4th order Adams Moulton implicit time stepping    !!
!!    for the diffusion terms.                                 !!
!!                                                             !!
!!                                                             !!
!!    Immersed boundary at non-gridpoint surface z=h(x,y)      !!
!!                                                             !!
!!=============================================================!!
!!    Heavy lifting done using FFTW3, PETSC & LAPACK/BLAS.     !!
!!    Output data written in NETCDF format. Browse w/ ncview.  !!
!!=============================================================!!
!!                                                             !!
!!    Author: K. B. Winters                                    !!
!!            Scripps Institution of Oceanography              !!
!!                         &                                   !!
!!            Mechanical and Aerospace Engineering             !!
!!            University of California, San Diego              !!
!!=============================================================!!  
program flow_solve
use etc
use intermediate_variables,  only: rhs_b



implicit none

            integer :: locnz
            locnz = ubound( rhs_b%SlabVals,3 )
            

call initialize
do istep=istart,iend 
 call write_results
  call ibrd   
! print*,' after ibrd: max=',maxval(rhs_b%SlabVals(:,:,locnz))
  call apply_forcing        
! print*,' after apply_forcing: max=',maxval(rhs_b%SlabVals(:,:,locnz))
    call explicit_time_step
! print*,' after explicit_time_step: max=',maxval(rhs_b%SlabVals(:,:,locnz))
     call implicit_time_step
! print*,' after implicit_time_step: max=',maxval(rhs_b%SlabVals(:,:,locnz))
      call pressure_rhs
! print*,' after pressure_rhs: max=',maxval(rhs_b%SlabVals(:,:,locnz))
       call pressure_solve
! print*,' after pressure_solve: max=',maxval(rhs_b%SlabVals(:,:,locnz))
        call project
! print*,' after project: max=',maxval(rhs_b%SlabVals(:,:,locnz))
         call immersed_bdry
! print*,' after immersed_bdry: max=',maxval(rhs_b%SlabVals(:,:,locnz))
          call advance_particles 
! print*,' after advance_particles: max=',maxval(rhs_b%SlabVals(:,:,locnz))
           call toggle
! print*,' after toggle: max=',maxval(rhs_b%SlabVals(:,:,locnz))
            enddo 
             call finalize
              stop 
end program flow_solve

!!=================================================================
!!=================================================================
subroutine ibrd
 use mpi_params
 use user_data, only: do_immersed_boundary
 use etc,       only: MM0,N
 use intermediate_variables
 use immersed_boundary
 implicit none 
 integer  :: i
 
 if(myid==0) write(0,*) '... calculating rhs functions '
  call eos
  call ibrd_1
  call ibrd_2
  call ibrd_3
  call ibrd_4
  call ibrd_5
  
  !! zero out any accelerations beneath IB surface
  if( do_immersed_boundary ) then
   do i=1,5
    explicit_rhs(i,MM0)%SlabVals =                          &
                  explicit_rhs(i,MM0)%SlabVals*IB_Filter
    implicit_rhs(i,N)%SlabVals =                            &
                  implicit_rhs(i,N)%SlabVals*IB_Filter
   enddo
  endif
  
end subroutine ibrd
!!=================================================================
!!=================================================================


!!=================================================================
!!=================================================================
subroutine ibrd_1
!! Routine to calculate the inertia and rotation terms for the 
!! x momentum equation.
!!
!! N.B. beta_tilde is the dimensionless value of beta, the linear rate
!! of change of the coriolis parameter f (d'less f=Rot) with latitude.
!! When beta is nonzero, interpret positive {x,u} as northward, 
!! {y,v} as westward. This is not standard, but is more convenient
!! given the data decomposition in x and z and the convention that
!! in 2d, the y dimension is suppressed.
!! 
!! For constant f, i.e. beta=0, positive {x,u} can be taken as east
!! positive {y,v} can be taken as north in the usual way.
!!=================================================================
 use class_ScalarVariable   
 use dependent_variables      
 use intermediate_variables
 use user_data,          only: Rot,beta_tilde,Re,nx,Lx,Lz         
 use etc           
 use operators
 use mpi_params
 implicit none 
 logical,save               :: first_entry=.TRUE.
 integer                    :: sign=-1, id=1, istat
 real(kind=8),pointer,save  :: x(:)
 real(kind=8)               :: x0, xx
 
!!=================================================================
!!=================================================================
 if( first_entry .and. beta_tilde /= 0 ) then
  allocate( x(nx), stat=istat )
  if( istat /= 0  ) then
   write(0,*) 'Allocating x(nx) in ibrd_1'
   stop       'Allocation failed:  ibrd_1'
  endif
  x(:) = velocity(1)%Grid%SpaceDims(1)%x(:)  ! dimensionless x values
  x0 = 0.5*(Lx/Lz)                           ! dimensionless x0 
  first_entry=.FALSE.
 endif
!!=================================================================
if(myid==0) write(0,*) '..............................',id
!!=================================================================
 
!!=================================================================
!!=================================================================
!! Compute gradient of (u), using `implicit_rhs' for storing results.
!! Compute the dot product -( u dot grad(u) ).
!!=================================================================
!!================================================================= 
   call gradient(velocity(id),implicit_rhs(1,N))
  
   !! -u dot grad(u)
   call ScalarDotProduct( explicit_rhs(id,MM0),sign,           &
                          velocity(1),                         &
                          velocity(2),                         &
                          velocity(3),                         &
                          implicit_rhs(1,N),                   &
                          implicit_rhs(2,N),                   &
                          implicit_rhs(3,N) ) 
   
!!=================================================================
!!=================================================================
!! Add in  rotation term:  
!!     explicit_rhs(1,MM0) <=== explicit_rhs(1,MM0) + Rot*(v)
!!
!! or  explicit_rhs(1,MM0) <=== explicit_rhs(1,MM0) 
!!                              + (Rot + beta_tilde*(x-x0))*(v)
!!=================================================================
!!=================================================================    
   
   if( Rot /= 0 .and. beta_tilde == 0 ) then
   
    !! explicit_rhs <== Rot*v + explicit_rhs
    call MyAxpy(Rot,velocity(2),explicit_rhs(id,MM0))
   
   elseif( Rot /= 0 .and. beta_tilde /= 0 ) then
   
    do i=1,nx
     explicit_rhs(id,MM0)%SlabVals(i,:,:) =                      &
                          explicit_rhs(id,MM0)%SlabVals(i,:,:)   &
                      + ( Rot + beta_tilde*(x(i)-x0) )           &
                      *   velocity(2)%SlabVals(i,:,:)
    enddo
    
   endif
  
     
!!============================================================
!! Since we have grad(u) stored in implicit_rhs(1-3,N)
!! we can take its divergence and store the resulting
!! Laplacian for later use in the implicit time step.
!! Set the deriv objects w/ the correct 2nd derivs wrt x,y,z.
!!============================================================
   do i=1,3
    implicit_rhs(i,N)%Derivs(i) = velocity(id)%Derivs(i+3)
    implicit_rhs(i,N)%gridpoint_values =     .TRUE.
    implicit_rhs(i,N)%wavenumber_values =    .FALSE.
    implicit_rhs(i,N)%forward_transform_only=.FALSE.
    implicit_rhs(i,N)%inverse_transform_only=.FALSE.
   enddo
   
!!==========================================
!! Calculate Div Grad(u), store result in
!! implicit_rhs(id,N)
!! scratch(1) is work space
!!========================================== 
   call Divergence(implicit_rhs(1,N),implicit_rhs(id,N),scratch(1))
         
!!=============================================
!! mult by 1/Re  
!!=============================================
   xx = 1./(Re)
   call MyScal(xx,implicit_rhs(id,N))
   
end subroutine ibrd_1
!!=================================================================
!!=================================================================



!!=================================================================
!!=================================================================
subroutine ibrd_2
!! See comments for ibrd_1.
!!=================================================================
 use class_ScalarVariable 
 use dependent_variables   
 use intermediate_variables 
 use user_data,          only: Rot,beta_tilde,Re,nx,Lx,Lz      
 use etc   
 use mpi_params
 implicit none
 logical,save               :: first_entry=.TRUE.
 integer                    :: sign=-1, id=2, istat
 real(kind=8),pointer,save  :: x(:)
 real(kind=8)               :: x0,xx
 
 if( first_entry .and. beta_tilde /= 0 ) then
  allocate( x(nx), stat=istat )
  if( istat /= 0  ) then
   write(0,*) 'Allocating x(nx) in ibrd_2'
   stop       'Allocation failed:  ibrd_2'
  endif
  x(:) = velocity(1)%Grid%SpaceDims(1)%x(:)  ! dimensionless x values
  x0 = 0.5*(Lx/Lz)                           ! dimensionless x0 
  first_entry=.FALSE.
 endif

!!=================================================================
if(myid==0) write(0,*) '..............................',id
!!=================================================================
!! Compute gradient of (v), using `implicit_rhs' for storing results.
!! Compute the dot product -( u dot grad(v) ).
!!=================================================================
!!================================================================= 
   call gradient(velocity(id),implicit_rhs(1,N))
  
   !! -u dot grad(v)
   call ScalarDotProduct( explicit_rhs(id,MM0),sign,           &
                          velocity(1),                         &
                          velocity(2),                         &
                          velocity(3),                         &
                          implicit_rhs(1,N),                   &
                          implicit_rhs(2,N),                   &
                          implicit_rhs(3,N) ) 


!!=================================================================
!!=================================================================
!! Add in  rotation term:  
!!     explicit_rhs(2,MM0) <=== explicit_rhs(2,MM0) - Rot*(u)
!!
!! or  explicit_rhs(2,MM0) <=== explicit_rhs(2,MM0) 
!!                              - (Rot + beta_tilde*(x-x0))*(u)
!!=================================================================
!!=================================================================    
   
   if( Rot /= 0 .and. beta_tilde == 0 ) then
   
    !! explicit_rhs <== -Rot*u + explicit_rhs
    xx = -Rot
    call MyAxpy(xx,velocity(1),explicit_rhs(id,MM0))
   
   elseif( Rot /= 0 .and. beta_tilde /= 0 ) then
   
    do i=1,nx
     explicit_rhs(id,MM0)%SlabVals(i,:,:) =  &
                                 explicit_rhs(id,MM0)%SlabVals(i,:,:)   &
                                 - ( Rot + beta_tilde*(x(i)-x0) )       &
                                 * velocity(1)%SlabVals(i,:,:)
    enddo
    
   endif
 
   !!============================================================
   !! since we have grad(v) stored in implicit_rhs(1-3,N)
   !! we can take its divergence and store the resulting
   !! Laplacian for later use in the implicit time step
   !! Set the derivs for the correct 2nd derivs wrt x,y,z
   !!============================================================
   do i=1,3
    implicit_rhs(i,N)%Derivs(i) = velocity(id)%Derivs(i+3)
    implicit_rhs(i,N)%gridpoint_values =     .TRUE.
    implicit_rhs(i,N)%wavenumber_values =    .FALSE.
    implicit_rhs(i,N)%forward_transform_only=.FALSE.
    implicit_rhs(i,N)%inverse_transform_only=.FALSE.
   enddo
   
   !!==========================================
   !! Calculate Div Grad(v), store result in
   !! implicit_rhs(id,N)
   !! scratch(1) is work space
   !!========================================== 
   call Divergence(implicit_rhs(1,N),implicit_rhs(id,N),scratch(1))
         
   !!=============================================
   !! mult by 1/Re  
   !!=============================================
   xx = 1./(Re)
   call MyScal(xx,implicit_rhs(id,N))
     
end subroutine ibrd_2
!!=================================================================
!!=================================================================


!!=================================================================
!!=================================================================
subroutine ibrd_3
!! Routine to calculate the inertia and buoyancy and dissipation  
!! terms for the z momentum equation. 
!!=================================================================
 use class_ScalarVariable    
 use dependent_variables     
 use intermediate_variables  
 use user_data,                 only: Ri,Re,Pr           
 use etc,                       only: N,MM0  
 use mpi_params
 implicit none 
 integer      :: sign=-1, id=3, i
 real(kind=8) :: alpha
!!=================================================================
!!=================================================================
!! Compute gradient of (w), using `implicit_rhs' for storing results.
!! Compute the dot product -( u dot grad(w) ).
!! Add in -(Ri pd)  
!!  ---> explicit_rhs(1,MM0) <= -(u dot grad(w)) - Ri(pd)
!!=================================================================
if(myid==0) write(0,*) '..............................',id
!!================================================================= 
   call gradient(velocity(id),implicit_rhs(1,N))
  
   !! -u dot grad(w)
   call ScalarDotProduct( explicit_rhs(id,MM0),sign,           &
                          velocity(1),                         &
                          velocity(2),                         &
                          velocity(3),                         &
                          implicit_rhs(1,N),                   &
                          implicit_rhs(2,N),                   &
                          implicit_rhs(3,N) ) 
                          
   !!=================================================================                         
   !! add in buoyancy term
   !! explicit_rhs <== -Ri*pd + explicit_rhs
   !!=================================================================
   alpha = -Ri
   if( Ri /= 0 ) &
    call MyAxpy(alpha,pd,explicit_rhs(id,MM0))     
   
   !!============================================================
   !! since we have grad(w) stored in implicit_rhs(1-3,N)
   !! we can take its divergence and store the resulting
   !! Laplacian for later use in the implicit time step
   !! Set the derivs for the correct 2nd derivs wrt x,y,z
   !!============================================================
   do i=1,3
    implicit_rhs(i,N)%Derivs(i) = velocity(id)%Derivs(i+3)
    implicit_rhs(i,N)%gridpoint_values =     .TRUE.
    implicit_rhs(i,N)%wavenumber_values =    .FALSE.
    implicit_rhs(i,N)%forward_transform_only=.FALSE.
    implicit_rhs(i,N)%inverse_transform_only=.FALSE.
   enddo
   
   !!==========================================
   !! Calculate Div Grad(v), store result in
   !! implicit_rhs(id,N)
   !! scratch(1) is work space
   !!========================================== 
   call Divergence(implicit_rhs(1,N),implicit_rhs(id,N),scratch(1))
         
   !!=============================================
   !! mult by 1/Re  
   !!=============================================
   alpha = 1./(Re)
   call MyScal(alpha,implicit_rhs(id,N))
   
end subroutine ibrd_3
!!=================================================================
!!=================================================================



!!=================================================================
!!=================================================================
subroutine ibrd_4
 use class_ScalarVariable    
 use dependent_variables    
 use intermediate_variables 
 use user_data,                 only: Re,Pr           
 use etc,                       only: N,MM0  
 use mpi_params
 implicit none 
 integer      :: sign=-1, id=4, i, k
 real(kind=8) :: alpha
 
!!=================================================================
!!=================================================================
!! Compute gradient of (s1), using `implicit_rhs' for storing results.
!! Compute the dot product -( u dot grad(s1) ).
!!=================================================================
if(myid==0) write(0,*) '..............................',id
!!================================================================= 
   call gradient(scalars(1),implicit_rhs(1,N))
  
   !! -u dot grad(s1)
   call ScalarDotProduct( explicit_rhs(id,MM0),sign,           &
                          velocity(1),                         &
                          velocity(2),                         &
                          velocity(3),                         &
                          implicit_rhs(1,N),                   &
                          implicit_rhs(2,N),                   &
                          implicit_rhs(3,N) ) 
      
   !!==============================================================
   !! Now add in -w * d/dz (s1_bar) to finish off the explicit RHS
   !!==============================================================
   do k=1,scalars(1)%DataMap%locnz
    explicit_rhs(id,MM0)%SlabVals(:,:,k)                       &
        = explicit_rhs(id,MM0)%SlabVals(:,:,k)                 &
        - velocity(3)%SlabVals(:,:,k)*ddz_s1_bar(k)
   enddo 
   
   
   !!============================================================
   !! since we have grad(s1) stored in implicit_rhs(1-3,N)
   !! we can take its divergence and store the resulting
   !! Laplacian for later use in the implicit time step
   !! Set the derivs for the correct 2nd derivs wrt x,y,z
   !!============================================================
   do i=1,3
    implicit_rhs(i,N)%Derivs(i) = scalars(id-3)%Derivs(i+3)
    implicit_rhs(i,N)%gridpoint_values =     .TRUE.
    implicit_rhs(i,N)%wavenumber_values =    .FALSE.
    implicit_rhs(i,N)%forward_transform_only=.FALSE.
    implicit_rhs(i,N)%inverse_transform_only=.FALSE.
   enddo
   
   !!==========================================
   !! Calculate Div Grad(s1), store result in
   !! implicit_rhs(id,N)
   !! scratch(1) is work space
   !!========================================== 
   call Divergence(implicit_rhs(1,N),implicit_rhs(id,N),scratch(1))
         
   !!=============================================
   !! mult by 1/PrRe  
   !!=============================================
   alpha = 1./(Pr(id-3)*Re)
   call MyScal(alpha,implicit_rhs(id,N))

   
end subroutine ibrd_4
!!=================================================================
!!=================================================================


!!=================================================================
!!=================================================================
subroutine ibrd_5
 use class_ScalarVariable    
 use dependent_variables     
 use intermediate_variables  
 use user_data,                 only: Re,Pr           
 use etc,                       only: N,MM0  
 use mpi_params
 implicit none 
 integer      :: sign=-1, id=5, i, k
 real(kind=8) :: alpha
 
!!=================================================================
!!=================================================================
!! Compute gradient of (s2), using `implicit_rhs' for storing results.
!! Compute the dot product -( u dot grad(s2) ).
!!=================================================================
if(myid==0) write(0,*) '..............................',id
!!================================================================= 
   call gradient(scalars(2),implicit_rhs(1,N))
  
   !! -u dot grad(s2)
   call ScalarDotProduct( explicit_rhs(id,MM0),sign,           &
                          velocity(1),                         &
                          velocity(2),                         &
                          velocity(3),                         &
                          implicit_rhs(1,N),                   &
                          implicit_rhs(2,N),                   &
                          implicit_rhs(3,N) ) 
                          
   !!==============================================================
   !! Now add in -w * d/dz (s2_bar) to finish off the explicit RHS
   !!==============================================================
   do k=1,scalars(2)%DataMap%locnz
    explicit_rhs(id,MM0)%SlabVals(:,:,k)                         &
        = explicit_rhs(id,MM0)%SlabVals(:,:,k)                   &
        - velocity(3)%SlabVals(:,:,k)*ddz_s2_bar(k)
   enddo 
   
     
   !!============================================================
   !! since we have grad(s2) stored in implicit_rhs(1-3,N)
   !! we can take its divergence and store the resulting
   !! Laplacian for later use in the implicit time step
   !! Set the derivs for the correct 2nd derivs wrt x,y,z
   !!============================================================
   do i=1,3
    implicit_rhs(i,N)%Derivs(i) = scalars(id-3)%Derivs(i+3)
    implicit_rhs(i,N)%gridpoint_values =     .TRUE.
    implicit_rhs(i,N)%wavenumber_values =    .FALSE.
    implicit_rhs(i,N)%forward_transform_only=.FALSE.
    implicit_rhs(i,N)%inverse_transform_only=.FALSE.
   enddo
   
   !!==========================================
   !! Calculate Div Grad(s1), store result in
   !! implicit_rhs(id,N)
   !! scratch(1) is work space
   !!========================================== 
   call Divergence(implicit_rhs(1,N),implicit_rhs(id,N),scratch(1))
         
   !!=============================================
   !! mult by 1/PrRe  
   !!=============================================
   alpha = 1./(Pr(id-3)*Re)
   call MyScal(alpha,implicit_rhs(id,N))
   
end subroutine ibrd_5
!!=================================================================
!!=================================================================


!!========================================================
!!========================================================
 subroutine apply_forcing
 !!Routine to (optionally) call user_forcing to 
 !!evaluate the any inhomogeneous terms as functions
 !!of position & time.
 !!========================================================
 !!========================================================
  use user_routines
  use user_data,              only: u0,Lz,s1_scale,      &
                                    s2_scale,do_forcing, &
                                    do_immersed_boundary
  use dependent_variables,    only: velocity
  use intermediate_variables, only: explicit_rhs,scratch  
  use etc
  use mpi_params
  use immersed_boundary,      only: IB_Filter
  
  implicit none  
  real(kind=8), save               :: scale_factor(5)
  real(kind=8), pointer, save      :: xvals(:),yvals(:),zvals(:)
  integer, save                    :: m1,m2,m3,k0,k1,npts,id
  logical, save                    :: first_entry = .TRUE.
  
  if( .not. do_forcing ) return
  if(myid==0) write(0,*) '... applying user forcing '
!!========================================================
!!========================================================
  if( first_entry ) then
   m1=velocity(1)%DataMap%SlabDims(1,myid)
   m2=velocity(1)%DataMap%SlabDims(2,myid)
   m3=velocity(1)%DataMap%SlabDims(3,myid)
   npts=m1*m2*m3
   allocate( xvals(m1) )
   allocate( yvals(m2) )
   allocate( zvals(m3) )
           
   k0 = velocity(1)%DataMap%my_global_kvals(1)
   k1 = k0+m3-1
   !! dimensional positions to pass to user
   zvals(:)=velocity(1)%Grid%SpaceDims(3)%x(k0:k1)*Lz
   xvals(:)=velocity(1)%Grid%SpaceDims(1)%x(:)*Lz
   yvals(:)=velocity(1)%Grid%SpaceDims(2)%x(:)*Lz
      
   !! scale factors to nondimensionalize the rhs
   !! terms supplied by the user
   scale_factor(1:3) = 1./(U0**2/Lz)
   scale_factor(4)   = 1./(s1_scale*U0/Lz)
   scale_factor(5)   = 1./(s2_scale*U0/Lz)
   first_entry = .FALSE.
  endif
  !!========================================================
  !!========================================================
  
  !!==========================================================
  !!==========================================================
  !! call user forcing routine providing the dimensional
  !! local spatial coordinates, the current time in seconds,
  !! a Fortran array to fill with values and an integer
  !! indicator as to which field's forcing is being requested
  !! explicit_rhs(i,MM0)%SlabVals = 
  !!    explicit_rhs(id,MM0)%SlabVals + scratch*scale_factor(i)
  !!==========================================================
  !!==========================================================
  do id=1,5 
    if(myid==0) write(0,*) '..............................',id
    call user_forcing(xvals,yvals,zvals,t_secs,                &
                      scratch(1)%SlabVals,id,m1,m2,m3)
    
    !! I'll zero out forces beneath IB so user doesn't have to bother
    if(do_immersed_boundary) then
     scratch(1)%SlabVals=scratch(1)%SlabVals*IB_Filter
    endif
    
    call daxpy(npts,scale_factor(id),scratch(1)%SlabVals,1,    &
                             explicit_rhs(id,MM0)%SlabVals,1)
  enddo

  return
 end subroutine apply_forcing
!=============================================================== 
!===============================================================





!!=================================================================
!!=================================================================
subroutine explicit_time_step
!!  Routine to take an explicit time step via Euler,
!!  2nd or third order Adams Bashforth method. The differential
!!  equations are as follows:
!!
!!  d/dt u   = rhs1
!!  d/dt v   = rhs2
!!  d/dt w   = rhs3
!!  d/dt s1  = rhs4
!!  d/dt s2  = rhs5
!!
!!  notation:
!!  dt   time increment
!!  MM0   integer index identifying the time t    storage locations
!!  MM1   integer index identifying the time t-dt storage locations
!!  MM2   integer index identifying the time t-2*dt storage locations
!!  MM3   integer index identifying the time t-3*dt storage locations
!!
!!  explicit_rhs(1,MM0),
!!  explicit_rhs(1,MM1),
!!  explicit_rhs(1,MM2),
!!  explicit_rhs(1,MM3)  
!!       ====>   rhs for u at t,t-dt,t-2*dt,,t-3*dt
!!  etc for explicit_rhs(2,:),
!!          explicit_rhs(3,:),
!!          explicit_rhs(4,:) 
!!      and explicit_rhs(5,:)

!!
!!  outputs:
!!  u_star(1) calculated values of u at time t+dt
!!  u_star(2) calculated values of v at time t+dt
!!  u_star(3) calculated values of w at time t+dt
!!  scalar(1) calculated values of scalar(1) at time t+dt (overwrite)
!!  scalar(2) calculated values of scalar(1) at time t+dt (overwrite)
!!=======================================================================
!!=======================================================================

 use dependent_variables
 use intermediate_variables
 use user_data
 use etc
 use operators 
 use mpi_params
 
 implicit none

 real(kind=8), save     :: dt2,dt12,dt24
 logical, save          :: first_entry = .TRUE.

 if( first_entry ) then
  dt2=dt/2.d0
  dt12=dt/12.d0
  dt24=dt/24.d0
  first_entry = .FALSE.
 endif
  
!!=================================================================
!!=================================================================
 if( step_flag == 'euler') then
 
  if(myid==0) write(0,*) '... 1st order Euler step '
 
  velocity(1)%SlabVals = velocity(1)%SlabVals      &
                       + dt*explicit_rhs(1,MM0)%SlabVals
                  
  velocity(2)%SlabVals = velocity(2)%SlabVals      &
                       + dt*explicit_rhs(2,MM0)%SlabVals
                  
  velocity(3)%SlabVals = velocity(3)%SlabVals      &
                       + dt*explicit_rhs(3,MM0)%SlabVals
                  
  scalars(1)%SlabVals  = scalars(1)%SlabVals        &
                       + dt*explicit_rhs(4,MM0)%SlabVals
                  
  scalars(2)%SlabVals  = scalars(2)%SlabVals        &
                       + dt*explicit_rhs(5,MM0)%SlabVals

!!=================================================================
!!=================================================================                     
  if(AB_order>1) step_flag='AB2'
  
 elseif( step_flag == 'AB2' ) then
  
  if(myid==0) write(0,*) '... 2nd order AB step '
  
  velocity(1)%SlabVals = velocity(1)%SlabVals          &
                  + dt2*( 3.*explicit_rhs(1,MM0)%SlabVals       &
                          -  explicit_rhs(1,MM1)%SlabVals )
 
  velocity(2)%SlabVals = velocity(2)%SlabVals          &
                  + dt2*( 3.*explicit_rhs(2,MM0)%SlabVals       &
                          -  explicit_rhs(2,MM1)%SlabVals )
 
  velocity(3)%SlabVals = velocity(3)%SlabVals          &
                  + dt2*( 3.*explicit_rhs(3,MM0)%SlabVals       &
                          -  explicit_rhs(3,MM1)%SlabVals )
 
  scalars(1)%SlabVals = scalars(1)%SlabVals            &
                     + dt2*( 3.*explicit_rhs(4,MM0)%SlabVals    &
                             -  explicit_rhs(4,MM1)%SlabVals )
                          
  scalars(2)%SlabVals = scalars(2)%SlabVals            &
                     + dt2*( 3.*explicit_rhs(5,MM0)%SlabVals    &
                             -  explicit_rhs(5,MM1)%SlabVals ) 
                             
  if(AB_order>2) step_flag='AB3'

!!=================================================================
!!=================================================================  
 elseif( step_flag == 'AB3' ) then
 
  if(myid==0) write(0,*) '... 3rd order AB step '
  
  velocity(1)%SlabVals = velocity(1)%SlabVals             &
                  + dt12*( 23.*explicit_rhs(1,MM0)%SlabVals        &
                          -16.*explicit_rhs(1,MM1)%SlabVals        &
                          +5.0*explicit_rhs(1,MM2)%SlabVals )
                          
  velocity(2)%SlabVals = velocity(2)%SlabVals             &
                  + dt12*( 23.*explicit_rhs(2,MM0)%SlabVals        &
                          -16.*explicit_rhs(2,MM1)%SlabVals        &
                          +5.0*explicit_rhs(2,MM2)%SlabVals )
                          
  velocity(3)%SlabVals = velocity(3)%SlabVals             &
                  + dt12*( 23.*explicit_rhs(3,MM0)%SlabVals        &
                          -16.*explicit_rhs(3,MM1)%SlabVals        &
                          +5.0*explicit_rhs(3,MM2)%SlabVals )
                          
  scalars(1)%SlabVals = scalars(1)%SlabVals               &
                  + dt12*( 23.*explicit_rhs(4,MM0)%SlabVals        &
                          -16.*explicit_rhs(4,MM1)%SlabVals        &
                          +5.0*explicit_rhs(4,MM2)%SlabVals )
                          
  scalars(2)%SlabVals = scalars(2)%SlabVals               &
                  + dt12*( 23.*explicit_rhs(5,MM0)%SlabVals        &
                          -16.*explicit_rhs(5,MM1)%SlabVals        &
                          +5.0*explicit_rhs(5,MM2)%SlabVals )
                          
  if(AB_order>3) step_flag='AB4'

!!=================================================================
!!=================================================================  
 elseif( step_flag == 'AB4' ) then

  if(myid==0) write(0,*) '... 4th order AB step '
    
  velocity(1)%SlabVals =  velocity(1)%SlabVals            &
                  + dt24*( 55.*explicit_rhs(1,MM0)%SlabVals        &
                          -59.*explicit_rhs(1,MM1)%SlabVals        &
                          +37.*explicit_rhs(1,MM2)%SlabVals        &
                          -9.0*explicit_rhs(1,MM3)%SlabVals)
 
  velocity(2)%SlabVals =  velocity(2)%SlabVals            &
                  + dt24*( 55.*explicit_rhs(2,MM0)%SlabVals        &
                          -59.*explicit_rhs(2,MM1)%SlabVals        &
                          +37.*explicit_rhs(2,MM2)%SlabVals        &
                          -9.0*explicit_rhs(2,MM3)%SlabVals)
 
  velocity(3)%SlabVals =  velocity(3)%SlabVals            &
                  + dt24*( 55.*explicit_rhs(3,MM0)%SlabVals        &
                          -59.*explicit_rhs(3,MM1)%SlabVals        &
                          +37.*explicit_rhs(3,MM2)%SlabVals        &
                          -9.0*explicit_rhs(3,MM3)%SlabVals)
 
  scalars(1)%SlabVals =       scalars(1)%SlabVals         &
                     + dt24*( 55.*explicit_rhs(4,MM0)%SlabVals     &
                             -59.*explicit_rhs(4,MM1)%SlabVals     &
                             +37.*explicit_rhs(4,MM2)%SlabVals     &
                             -9.0*explicit_rhs(4,MM3)%SlabVals)
 
  scalars(2)%SlabVals =       scalars(2)%SlabVals         &
                     + dt24*( 55.*explicit_rhs(5,MM0)%SlabVals     &
                             -59.*explicit_rhs(5,MM1)%SlabVals     &
                             +37.*explicit_rhs(5,MM2)%SlabVals     &
                             -9.0*explicit_rhs(5,MM3)%SlabVals)
 
 endif
 
 !!=================================================================
 !!=================================================================
 !! Handle the (1/PrRe) d2/dz2 ( s_bar ) terms analytically
 !! since they are time independent and shouldn't be advanced
 !! using an explicit method. Another way would be to add the
 !! terms to the Diffusive terms before doing the implicit solves,
 !! but again, since we can do them analytically, why not?
 !!=================================================================
 !!=================================================================
  do k=1,scalars(1)%DataMap%locnz
    scalars(1)%SlabVals(:,:,k) = scalars(1)%SlabVals(:,:,k)   &
                               + dt * ddzz_s1_bar(k)/(Re*Pr(1))
  enddo
 
  do k=1,scalars(1)%DataMap%locnz
    scalars(2)%SlabVals(:,:,k) = scalars(2)%SlabVals(:,:,k)   &
                               + dt * ddzz_s2_bar(k)/(Re*Pr(2))
  enddo  
  
  
 return
end subroutine explicit_time_step
!!=================================================================
!!=================================================================


!!=================================================================
!!=================================================================
subroutine implicit_time_step
 use mpi_params
 implicit none 
 integer  :: id
 
 if(myid==0) write(0,*) '... implicit (diffusion) time steps'
 do id=1,5
  if(myid==0) write(0,*) '..............................',id
  call implicit_step(id)
 enddo
    
end subroutine implicit_time_step
!!=================================================================
!!=================================================================


!!=================================================================
!!=================================================================
subroutine implicit_step(id)
  use class_ScalarVariable 
  use dependent_variables,     only: velocity,scalars
  use intermediate_variables,  only: implicit_rhs,rhs_b
  use immersed_boundary
  use operators,               only: h2_scale_factor
  use user_data,               only:do_immersed_boundary, &
                                    usolve_exp_coeffs,    &
                                    vsolve_exp_coeffs,    &
                                    wsolve_exp_coeffs,    &
                                    s1solve_exp_coeffs,   &
                                    s2solve_exp_coeffs,   &
                                    AM_order,dt
  use etc,                     only:field_id,N,NM1
  implicit none 
  integer            :: i,id
  logical,save       :: p_space(5),w_space(5),            &
                        uniform_spacing=.FALSE.,          &
                        stretched_grid=.TRUE.,            &
                        first_entry=.TRUE.
  real(kind=8),save  :: h2(5)

!!===============================================================  
!!=============================================================== 
  if(first_entry) then
  
   !! change default as needed
   if( velocity(1)%Grid%SpaceDims(1)%uniform_dx  .and.      &
       velocity(1)%Grid%SpaceDims(2)%uniform_dx  .and.      &
       velocity(1)%Grid%SpaceDims(3)%uniform_dx  ) then
    uniform_spacing = .TRUE.
    stretched_grid  = .FALSE.
   endif
    
   do i=1,5      
    h2(i) = h2_scale_factor
    p_space(i)=.TRUE.
    w_space(i)=.FALSE.
    !! change as needed
    if( i==1 .and. usolve_exp_coeffs )  w_space(i)=.TRUE.   
    if( i==2 .and. vsolve_exp_coeffs )  w_space(i)=.TRUE.   
    if( i==3 .and. wsolve_exp_coeffs )  w_space(i)=.TRUE.   
    if( i==4 .and. s1solve_exp_coeffs ) w_space(i)=.TRUE.   
    if( i==5 .and. s2solve_exp_coeffs ) w_space(i)=.TRUE. 
 
    if(w_space(i)) then
     h2(i)=1.d0
     p_space(i)=.FALSE.
    endif
   enddo
         
   first_entry=.FALSE.
  endif
 !!=============================================================== 
 !!=============================================================== 
  !! On entry the implicit_rhs terms are filled with
  !!  1/Re grad^2 field(id) at t_n etc
  !!  first, form the rhs for the implicit AM step 
  !!....
  !! rhs_b for 3rd order Adams Moulton ...
  !! rhs = h2*( -(2./dt)*f -invRe*grad^2 f ) 
  !!  
  !!
  !!  OR.....  
  !!
  !! rhs_b for 4th order Adams Moulton ...
  !! rhs = h2*(-12/(5*dt)*f -8/5*invRe*grad^2 f 
  !!                        + 1/5*invRe*grad^2 f_n-1 )
  !!  
  !! ==>    ** (scaled by h2) and using invPrRe for scalars
  !!            
  !!=========================================================
  !!===============================================================
  
   if( p_space(id) .or. stretched_grid .or. do_immersed_boundary ) then
    !!==========================================================
    !! we CANNOT do an analytical solve in wavenumber space...
    !! ===> set up the rhs for the AM implicit step...
    !!==========================================================
    if( id <= 3 ) then
 
     if(AM_order==3) then
      rhs_b%SlabVals = ( (-2./dt)*velocity(id)%SlabVals                 &
                     -   implicit_rhs(id,N)%SlabVals )*h2(id)
     elseif(AM_order==4) then
      rhs_b%SlabVals = ( (-12./(5.*dt))*velocity(id)%SlabVals           &
                     -   (8./5.)*implicit_rhs(id,N)%SlabVals            &
                     +   (1./5.)*implicit_rhs(id,NM1)%SlabVals )*h2(id)
     endif
 
    elseif( id > 3 ) then
 
     if(AM_order==3) then
      rhs_b%SlabVals = ( (-2./dt)*scalars(id-3)%SlabVals                &
                   -   implicit_rhs(id,N)%SlabVals )*h2(id)
     elseif(AM_order==4) then
      rhs_b%SlabVals = ( (-12./(5.*dt))*scalars(id-3)%SlabVals          &
                   -   (8./5.)*implicit_rhs(id,N)%SlabVals              & 
                   +   (1./5.)*implicit_rhs(id,NM1)%SlabVals )*h2(id)
     endif

    endif 
    
   elseif( w_space(id) .and. uniform_spacing   &
              .and. .not. do_immersed_boundary ) then
   !! for this one special case, we can do the diffusion update
   !! analytically in wavenumber space, operating on the transform
   !! of the dependent variable directly ===> put it in rhs_b
    
     if( id <= 3 ) then
      call CopySlabData_a2b(velocity(id),rhs_b)
     elseif( id > 3 ) then
      call CopySlabData_a2b(scalars(id-3),rhs_b)
     endif
     
   endif
    
    
  !!======================================================
  !! Solve the Ax=b implicit systems 
  !! setting the BC values if necessary
  !!======================================================

   field_id=id
   call SetBCs(id)
   call Diffusion_Solve

end subroutine implicit_step
!!======================================================
!!======================================================


!!============================================================
!!============================================================
subroutine pressure_rhs
!!The velocity [u,v,w] is the intermediate velocity u*
!!The eqn for pressure is L[p]=(1/dt)Div(u*) in interior
!!with Neumann conditions on all boundaries (except periodic).
!!============================================================
 use class_ScalarVariable
 use operators
 use user_data 
 use immersed_boundary
 use mpi_params
 use dependent_variables,     only: velocity
 use intermediate_variables,  only: rhs_b,                   &
                                    div_u,                   &
                                    scratch,                 &
                                    phi
 use etc,                     only: field_id,                &
                                    numzplanes_below,        &
                                    int_weights_x,           &
                                    int_weights_y,           &
                                    int_weights_z,istep
 
 implicit none 
 integer                      :: jerr,idim,id
 integer                      :: i,j,k,kglobal
 integer                      :: npts,mval
 integer,save                 :: locnz,one=1
 real(kind=8),save            :: h2,xx
 real(kind=8)                 :: S(2)
 real(kind=8)                 :: local_value,global_value
 logical,save                 :: first_entry=.TRUE.
 !!============================================================
 
 !!============================================================
 if( first_entry ) then
  locnz=velocity(1)%DataMap%SlabVals(3,myid)
  !! get the global scale factor used to define
  !! the elliptic operator L[]
  h2 = h2_scale_factor
  xx=h2*(1./dt)
  if( psolve_exp_coeffs ) xx=1.d0/dt
  first_entry=.FALSE.
 endif
 
 
 !!============================================================
 !! take div of u in physical space
 !!============================================================
 do id=1,3
  velocity(id)%gridpoint_values =     .TRUE.
  velocity(id)%wavenumber_values =    .FALSE.
  velocity(id)%forward_transform_only=.FALSE.
  velocity(id)%inverse_transform_only=.FALSE.
 enddo
   
 !!==========================================
 !! Calculate Div u*
 !! scratch(1) is work space
 !!========================================== 
 call Divergence(velocity,div_u,scratch(1))
         
 !!=============================================
 !! Scale RHS ===>>  h2 * (1/dt) Div u* 
 !!=============================================
  call CopySlabData_a2b(div_u,rhs_b)
  call MyScal(xx,rhs_b)
    
 !!=============================================
 !! if solving for p in wavenumber space, don't 
 !! need to remove the mean or impose explicit
 !! boundary constraints here
 !! ====> exit w/ rhs_b filled w/ grdpt values
 !!============================================= 
 if( psolve_exp_coeffs ) return 
  
   
 !!==========================================
 !! do boundary values separately....
 !!==========================================
 idim=1
 if( xdim_implicit_bcs ) goto 100  
  
 !! if normal velocity is to be specified, assume rigid wall
 !! and set pressure gradient in terms of ustar dot n
 if(BCtype(idim,idim,1)=='dirichlet' .and. BCtype(6,idim,1)=='neumann') then
  rhs_b%SlabVals(1,:,:)  = (1./dt)*velocity(idim)%SlabVals(1,:,:)
 elseif(BCtype(idim,idim,1)=='neumann' .and. BCtype(6,idim,1)=='neumann') then
  rhs_b%SlabVals(1,:,:)  = 0.d0
 else
  if(myid==0) write(0,*) 'p bc: ',trim( BCtype(6,idim,1) )
  if(myid==0) write(0,*) 'u bc: ',trim( BCtype(1,idim,1) )
  stop 'x pressure bcs confusion, pressure_rhs'
 endif
    
 
100 idim=2
 if( ydim_implicit_bcs ) goto 101

 !! if normal velocity is to be specified, assume rigid wall
 !! and set pressure gradient in terms of ustar dot n
 if(BCtype(idim,idim,1)=='dirichlet' .and. BCtype(6,idim,1)=='neumann') then
  rhs_b%SlabVals(:,1,:)  = (1./dt)*velocity(idim)%SlabVals(:,1,:)
 elseif(BCtype(idim,idim,1)=='neumann' .and. BCtype(6,idim,1)=='neumann') then
  rhs_b%SlabVals(:,1,:)  = 0.d0
 else
  stop 'y pressure bcs confusion, pressure_rhs'
 endif
  
 
101 idim=3
 if(myid==0) then
  if( zdim_implicit_bcs .or. phi_deriv_types(idim)=='cos' ) goto 102
  
  !! if normal velocity is to be specified, assume rigid wall
  !! and set pressure gradient in terms of ustar dot n
  if(BCtype(idim,idim,1)=='dirichlet' .and. BCtype(6,idim,1)=='neumann') then
   rhs_b%SlabVals(:,:,1) = (1./dt)*velocity(idim)%SlabVals(:,:,1)
  else
   stop 'z pressure bcs confusion, pressure_rhs bottom'
  endif
   
 endif

 
102 if(myid==numprocs-1) then
  if( zdim_implicit_bcs .or. phi_deriv_types(idim)=='cos' ) return

  !! if normal velocity is to be specified, assume rigid wall
  !! and set pressure gradient in terms of ustar dot n
  if(BCtype(idim,idim,1)=='dirichlet' .and. BCtype(6,idim,1)=='neumann') then
   rhs_b%SlabVals(:,:,locnz) = (1./dt)*velocity(idim)%SlabVals(:,:,locnz)
  else
   stop 'z pressure bcs confusion, pressure_rhs top'
  endif
     
  endif
  
end subroutine pressure_rhs
!!=====================================================
!!=====================================================


!!=====================================================
!!=====================================================
subroutine project
!!==============================================
!!  Grad Phi stored in explicit_rhs(1:3,M_oldest)
!!  (we just used these in explicit step and will
!!  overwrite them next step w/o using em again)
!!   u <- u* -dt*GradPhi(1)  
!!   v <- v* -dt*GradPhi(2)
!!   w <- w* -dt*GradPhi(3)
!!===============================================
 use class_ScalarVariable 
 use dependent_variables,       only: velocity
 use intermediate_variables,    only: GradPhi => explicit_rhs
 use user_data,                 only: dt,do_immersed_boundary
 use etc,                       only: m => M_oldest
 use immersed_boundary
 implicit none
 integer       :: k
 
 !!  velocity(k) <== (-dt)*GradPhi + velocity(k) 
 do k=1,3
  if( do_immersed_boundary)    &
   GradPhi(k,m)%SlabVals=GradPhi(k,m)%SlabVals*IB_Filter
  call MyAxpy(-dt,GradPhi(k,m),velocity(k)) 
 enddo
 

end subroutine project
!!=====================================================
!!=====================================================








!!=====================================================
!!=====================================================
subroutine diffusion(id)
 use class_ScalarVariable
 use dependent_variables,     only: velocity,scalars
 use intermediate_variables,  only: implicit_rhs,rhs_b,   &
                                    scratch
 use user_data,               only: Re,Pr,                &
                                    do_immersed_boundary, &
                                    usolve_exp_coeffs,    &
                                    vsolve_exp_coeffs,    &
                                    wsolve_exp_coeffs,    &
                                    s1solve_exp_coeffs,   &
                                    s2solve_exp_coeffs 
 use etc,                     only: N 
 implicit none 
 integer,optional         :: id
 integer                  :: i,istart,iend
 real(kind=8), SAVE       :: invRe,invPrRe(2)
 logical, SAVE            :: p_space(5), w_space(5)
 logical, SAVE            :: uniform_spacing=.FALSE.
 logical, SAVE            :: stretched_grid=.TRUE.
 logical, SAVE            :: first_entry = .TRUE.

!!=====================================================
!!=====================================================
 if( first_entry ) then
  invRe=1./Re
  invPrRe(1)=1./(Re*Pr(1))
  invPrRe(2)=1./(Re*Pr(2))  
  !! check any variable for stretched grid
  if( velocity(1)%Grid%SpaceDims(1)%uniform_dx  .and.      &
      velocity(1)%Grid%SpaceDims(2)%uniform_dx  .and.      &
      velocity(1)%Grid%SpaceDims(3)%uniform_dx  ) then
    uniform_spacing = .TRUE.
    stretched_grid  = .FALSE.
  endif
  
  do i=1,5
   p_space(i)=.TRUE.
   w_space(i)=.FALSE.
   !! change as needed
   if( i==1 .and. usolve_exp_coeffs )  w_space(i)=.TRUE.   
   if( i==2 .and. vsolve_exp_coeffs )  w_space(i)=.TRUE.   
   if( i==3 .and. wsolve_exp_coeffs )  w_space(i)=.TRUE.   
   if( i==4 .and. s1solve_exp_coeffs ) w_space(i)=.TRUE.   
   if( i==5 .and. s2solve_exp_coeffs ) w_space(i)=.TRUE.  
   if(w_space(i)) then
    p_space(i)=.FALSE.
   endif
  enddo
  
  first_entry = .FALSE.
 endif
!!=====================================================
!!=====================================================

 if( present(id) ) then  !! just do the selected scalar
   istart=id
   iend=id
  else                   !! do the three vel components
   istart=1
   iend=3
  endif
  
 !!=====================================================
 !!=====================================================
 do i=istart,iend
 
  if(w_space(i) .and. uniform_spacing .and. .not. do_immersed_boundary) then
    !! in this case, exact solve is possible in w_space 
    !! ====> put field values in the rhs vector
    if( i <= 3 ) call CopySlabData_a2b(velocity(i), rhs_b)
    if( i > 3 )  call CopySlabData_a2b(scalars(i-3),rhs_b)    
  else  
   !!======================================================
   !!  compute Grad^2 of the dependent variable &
   !!  multiply by (1/Re) or (1/Pr(i)Re)
   !!    grad^2 f ==> implicit_rhs(i,N), N toggled w/ NM1
   !!    grad(f)==>implicit_rhs(1-3), scratch(4)=work
   !!======================================================
   if( i <= 3 ) then
    velocity(i)%gridpoint_values =     .TRUE.
    velocity(i)%wavenumber_values =    .FALSE.
    velocity(i)%forward_transform_only=.FALSE.
    velocity(i)%inverse_transform_only=.FALSE.
    
    call Laplacian(velocity(i),scratch,implicit_rhs(i,N),scratch(4))
    call MyScal(invRe,implicit_rhs(i,N)) !!  x <== alpha*x 
  
   elseif( i > 3 ) then
    scalars(i-3)%gridpoint_values =     .TRUE.
    scalars(i-3)%wavenumber_values =    .FALSE.
    scalars(i-3)%forward_transform_only=.FALSE.
    scalars(i-3)%inverse_transform_only=.FALSE.
  
    call Laplacian(scalars(i-3),scratch,implicit_rhs(i,N),scratch(4))
    call MyScal(invPrRe(i-3),implicit_rhs(i,N))
   endif
  
 endif  
enddo    !! end loop over each variable i

end subroutine diffusion
!!=====================================================
!!=====================================================
 


 
 
subroutine SetBCs(id)
 use class_ScalarVariable
 use intermediate_variables,  only: rhs_b
 use operators 
 use user_data,               only: nx,ny,                &
                                    xdim_implicit_bcs,    &
                                    ydim_implicit_bcs,    &
                                    zdim_implicit_bcs,    &
                                    usolve_exp_coeffs,    &
                                    vsolve_exp_coeffs,    &
                                    wsolve_exp_coeffs,    &
                                    s1solve_exp_coeffs,   &
                                    s2solve_exp_coeffs    
 use etc,                     only: numzplanes_below                                   
 use mpi_params,              only: myid,numprocs
  
 implicit none 
 integer                  :: id,i,j,k,idim,endid 
 real(kind=8)             :: S
 integer, SAVE            :: locnz
 logical, SAVE            :: p_space(5), w_space(5)
 logical, SAVE            :: first_entry = .TRUE.

!!=====================================================
!!=====================================================
 if( first_entry ) then
  
  locnz = ubound( rhs_b%SlabVals,3 )
  do i=1,5
   p_space(i)=.TRUE.
   w_space(i)=.FALSE.
   !! change as needed
   if( i==1 .and. usolve_exp_coeffs )  w_space(i)=.TRUE.   
   if( i==2 .and. vsolve_exp_coeffs )  w_space(i)=.TRUE.   
   if( i==3 .and. wsolve_exp_coeffs )  w_space(i)=.TRUE.   
   if( i==4 .and. s1solve_exp_coeffs ) w_space(i)=.TRUE.   
   if( i==5 .and. s2solve_exp_coeffs ) w_space(i)=.TRUE. 
 
   if(w_space(i)) then
    p_space(i)=.FALSE.
   endif
  enddo
  
  first_entry = .FALSE.
 endif

!!======================================================  
!!======================================================
!! If the solves are carried out in wavenumber space,
!! the BCs will be handled implicitly via the expansions
!! ====> just return
   if( w_space(id) )  return
!!======================================================
!!======================================================
   
  
 !!=========================
 !!  EAST FACE
 !!=========================
 if(xdim_implicit_bcs) goto 100
 
 idim=1
 endid=1
 i=1
 
 do j=1,ny
  do k=1,locnz
  
   call GetSval(id,idim,endid,j,k,S)
   rhs_b%SlabVals(i,j,k) = S
          
  enddo
 enddo
 
 !!=========================
 !!  WEST FACE
 !!=========================
 idim=1
 endid=2
 i=nx
 
 do j=1,ny
  do k=1,locnz
  
   call GetSval(id,idim,endid,j,k,S)
   rhs_b%SlabVals(i,j,k) = S
          
  enddo
 enddo
100 continue
 
 !!=========================
 !!  FRONT FACE
 !!=========================
 if(ydim_implicit_bcs) goto 101
 
 idim=2
 endid=1
 j=1
 
 do i=1,nx
  do k=1,locnz
  
   call GetSval(id,idim,endid,i,k,S)
   rhs_b%SlabVals(i,j,k) = S
     
  enddo
 enddo  
 !!=========================
 !!  BACK FACE
 !!=========================
 idim=2
 endid=2
 j=ny
 
 do i=1,nx
  do k=1,locnz
  
   call GetSval(id,idim,endid,i,k,S)
   rhs_b%SlabVals(i,j,k) = S
          
  enddo
 enddo  
101 continue
 
 !!=========================
 !!  BOTTOM
 !!=========================
 if(zdim_implicit_bcs) goto 102
 if( myid==0 ) then
  idim=3
  endid=1
  k=1
 
  do i=1,nx
   do j=1,ny
   
    call GetSval(id,idim,endid,i,j,S)
    rhs_b%SlabVals(i,j,k) = S
          
   enddo
  enddo  
 endif

 !!=========================
 !!  TOP
 !!=========================
 if( myid==numprocs-1 ) then
  idim=3
  endid=2
  k=locnz
 

  do i=1,nx
   do j=1,ny
   
    call GetSval(id,idim,endid,i,j,S)
    rhs_b%SlabVals(i,j,k) = S
          
   enddo
  enddo  

  print*,'max(rhs_b%SlabVals)=',maxval(rhs_b%SlabVals(:,:,k))

  endif
102 continue
 
end subroutine SetBCs
!!=====================================================
!!=====================================================



!!=====================================================
!!=====================================================
subroutine immersed_bdry
 use mpi_params
 use user_data,               only: do_immersed_boundary
 use immersed_boundary
 implicit none 
  
 if( .not. do_immersed_boundary ) return
 if(myid==0) write(0,*) '... interpolating projected velocity onto IB surface'
 
 !!====================================================
 !! interpolate the projected, div free velocity field
 !! to the IB surface, mult by -1/dt and store in SS
 !!====================================================
 call IB_SourceStrength
 
 !!====================================================
 !! update the velocity with the IB correction
 !!====================================================
 call IB_Update
   
   
end subroutine immersed_bdry
!!=====================================================
!!=====================================================




