

!!=====================================================================
  module immersed_boundary 
!!====================================================================
   implicit none
   
   public  :: InitializeImmersedBoundary
   public  :: IB_SourceStrength
   public  :: IB_Update
   public  :: ForcingFilter
   public  :: ij_from_count
   public  :: count_from_ij
   public  :: unit_normal
   
   real(kind=8)                  :: eps(3),d_inf(3),delta_min
   real(kind=8),allocatable      :: x_IB(:,:),y_IB(:,:),h_IB(:,:)
   real(kind=8),allocatable      :: nhat(:,:,:),SS(:,:),f_IB(:,:,:)
   real(kind=8),allocatable      :: IB_Filter(:,:,:)
   integer                       :: n_i,n_j,n_k
   integer,allocatable           :: labels(:,:),kplus(:,:)

  Contains 
  
!===================================================================
!===================================================================
subroutine InitializeImmersedBoundary
 use class_DataDecomposition
 use independent_variables
 use dependent_variables
 use intermediate_variables
 use user_data
 use user_routines
 use etc
 use mpi_params
 
 implicit none
 integer          :: idim,locnz
 real(kind=8)     :: x(3),ds(3),h_x,h_y
 real(kind=8)     :: s1,s2,s3,ans(3)
 real(kind=8)     :: x_s,y_s,z_s
 real(kind=8)     :: dx,dy,dz,xx,acosh2
 real(kind=8),allocatable  :: work(:)
 external            acosh2
  
 if( .not. do_immersed_boundary ) return
        
  !! get 'ds' in each of the x y z directions ==> ds(1-3)
  if(ny==1) then
   ds(2)=0.d0
  else
   ds(2)=Grid(0)%SpaceDims(2)%s(2)-Grid(0)%SpaceDims(2)%s(1)
  endif
  ds(1)=Grid(0)%SpaceDims(1)%s(2)-Grid(0)%SpaceDims(1)%s(1)  
  ds(3)=Grid(0)%SpaceDims(3)%s(2)-Grid(0)%SpaceDims(3)%s(1)
  locnz = ubound(velocity(1)%SlabVals,3)
 
  allocate( x_IB(nx,ny),y_IB(nx,ny),h_IB(nx,ny) )
  allocate( nhat(nx,ny,3),IB_Filter(nx,ny,locnz) )
  allocate( f_IB(nx,ny,3) )
  allocate( SS(nx,ny))
  allocate( labels(nx,ny),kplus(nx,ny) )
  labels(:,:)=1
  
  
  
  do j=1,ny
   do i=1,nx
   
   x(1)=Grid(0)%SpaceDims(1)%x(i)
   x(2)=Grid(0)%SpaceDims(2)%x(j)
   
   call eval_surface(x,x(3),h_x,h_y)
   call unit_normal(x,ans)
   
   !! x,y,z on IB surface
   x_IB(i,j)=x(1)
   y_IB(i,j)=x(2)
   h_IB(i,j)=x(3)

   !! unit normal vector
   nhat(i,j,1)=ans(1)
   nhat(i,j,2)=ans(2)
   nhat(i,j,3)=ans(3)
      
   !! find first globally indexed k value ABOVE surface
   do k=1,nz
    kplus(i,j)=k
    if( Grid(0)%SpaceDims(3)%x(k) > h_IB(i,j) ) goto 99
   enddo
99 continue
   
   
   if( i==1 .and. j==1 ) then
   
   !! Get the param values s in each direction -> (s1,s2,s3)
   !! corresponding too the surface point x(1:3).
   !! Remember, this user routine has DIMENSIONAL inputs
   call point_map_x2s(x(1)*Lz,1,xdim_periodic,Lx,nx,s1)
   call point_map_x2s(x(2)*Lz,2,ydim_periodic,Ly,ny,s2)
   call point_map_x2s(x(3)*Lz,3,zdim_periodic,Lz,nz,s3)
    
   !! now get the corresponding dx/ds,dy/ds,dz/ds values -> (x_s,y_s,z_s)
   !! Remember, this user routine has DIMENSIONAL outputs
   idim=1
   call point_map_s2x(s1,idim,xdim_periodic,Lx,nx,xx,x_s)
    
   idim=2
   call point_map_s2x(s2,idim,ydim_periodic,Ly,ny,xx,y_s)
    
   idim=3
   call point_map_s2x(s3,idim,zdim_periodic,Lz,nz,xx,z_s)
    
   dx = (x_s/Lz) * ds(1)   !! local dxds * ds
   dy = (y_s/Lz) * ds(2)   !! local dyds * ds
   dz = (z_s/Lz) * ds(3)   !! local dzds * ds
   if(ny==1) dy=1.d0
     
    delta_min = 1.e-6
    eps(1) = 3.*dx
    eps(2) = 3.*dy
    eps(3) = 3.*dz
    
    d_inf(:) =  4.  !acosh2( 1./sqrt(2.*eps(:)*delta_min) )
    n_i = eps(1)*d_inf(1)/dx
    n_j = eps(2)*d_inf(2)/dy
    n_k = eps(3)*d_inf(3)/dz
   endif   
             
  enddo
 enddo
 
 !! construct spatial filter so user forcing can be specified
 !! more easily, i.e. globally w/o worrying about IB location
 call ForcingFilter
      
 return
end subroutine InitializeImmersedBoundary
!=================================================================
!=================================================================


  
  !!==============================================================
  !!==============================================================
  subroutine IB_SourceStrength
   use user_data,  only: nx,ny,dt
   use dependent_variables
   use intermediate_variables
   use etc, only: istep,N,MM0,M_oldest
   implicit none
   integer           :: i,j,k,m(3),npts
   real(kind=8)      :: dtinv

   npts=nx*ny
   m(:)=0
   dtinv=1.d0/dt
   
                      
   !!====================================================
   !! do the 3 interpolations
   !!====================================================
   do k=1,3                                          
                                          
    call spline_interp_value( x_IB,                 &    !x'
                              y_IB,                 &    !y'
                              h_IB,                 &    !z' on surface
                              npts,velocity(k),m,   &    !field to interp
                              f_IB(1,1,k),          &    ! ans
                              labels,               &    !==1
                              scratch(2)%SlabVals,  &    !work
                              scratch(3)%SlabVals,  &    !work
                              scratch(4)%SlabVals )      !work
   enddo
   
   !! now fill in the SS w/ dot product with nhat
   SS(:,:) = (f_IB(:,:,1)*nhat(:,:,1)  &
           +  f_IB(:,:,2)*nhat(:,:,2)  &
           +  f_IB(:,:,3)*nhat(:,:,3))*dtinv
   
   return
  end subroutine IB_SourceStrength
  !!====================================================================
  !!====================================================================
  
  
  
  !!====================================================================
  !!==================================================================== 
   subroutine IB_Update
    use user_data,              only: nx,ny,dt
    use etc,                    only: numzplanes_below,N,istep                                  
    use mpi_params,             only: myid
    use independent_variables
    use dependent_variables
    implicit none
    integer           :: i,j,k,ii,jj,kk,idim
    integer,save      :: locnz,npts,k0
    integer           :: istart,iend
    integer           :: jstart,jend
    real(kind=8),save :: ds(3)
    real(kind=8)      :: x(3),dxds,dyds,dzds
    real(kind=8)      :: dx,dy,dz,delta
    real(kind=8)      :: d1,d2,d3,xx
    logical,save      :: first_entry=.TRUE.
    external          delta
   
  !!====================================================================
  !!====================================================================
   if( first_entry ) then
    locnz=ubound(velocity(1)%SlabVals,3)
    npts=nx*ny
    k0=numzplanes_below(myid)
    if(ny==1) then
     ds(2)=0.d0
    else
     ds(2)=Grid(0)%SpaceDims(2)%s(2)-Grid(0)%SpaceDims(2)%s(1)
    endif
    ds(1)=Grid(0)%SpaceDims(1)%s(2)-Grid(0)%SpaceDims(1)%s(1)  
    ds(3)=Grid(0)%SpaceDims(3)%s(2)-Grid(0)%SpaceDims(3)%s(1)
    first_entry=.FALSE.
   endif
  !!====================================================================
  !!==================================================================== 
    
    !! loop over "all" the grid points on myid (ii,jj,kk)
    do kk=1,locnz
     x(3)=Grid(0)%SpaceDims(3)%x(k0+kk) !; write(0,*) 'K = ',kk
      do jj=1,ny
       x(2)=Grid(0)%SpaceDims(2)%x(jj) 
        do ii=1,nx       
         x(1)=Grid(0)%SpaceDims(1)%x(ii)
    
         !!========================================================
         !! if x(ii,jj,kk) is 'near' IB surface, compute integral, 
         !! otherwise, if x is too far from surface, do nothing
         !!========================================================
         if( abs( ( x(3) - h_IB(ii,jj) ) ) < d_inf(3)*eps(3) ) then
    
          !!================================================
          !! above/below x(ii,jj,kk), define a range of i,j
          !! values where d1 and d2 could be nonzero
          !!================================================
          istart = maxval( (/1,  ii - n_i/) )
          iend   = minval( (/nx, ii + n_i/) )
    
          jstart = maxval( (/1,  jj - n_j/) )
          jend   = minval( (/ny, jj + n_j/) )
    
          !!================================================
          !! loop over this portion of the surface
          !! and accumulate any contributions to F(x)
          !!================================================
          do j=jstart,jend
           dyds= Grid(0)%SpaceDims(2)%dxds(j) 
            dy = dyds*ds(2)
            if(ny==1) dy=1.
      
            do i=istart,iend
             dxds= Grid(0)%SpaceDims(1)%dxds(i) 
             dx = dxds*ds(1)
       
              k=kplus(i,j)
              dzds= Grid(0)%SpaceDims(3)%dxds(k) 
              dz = dzds*ds(3)
       
              d1 = delta( (x(1)-x_IB(i,j)),eps(1) )
              d2 = delta( (x(2)-y_IB(i,j)),eps(2) )
              d3 = delta( (x(3)-h_IB(i,j)),eps(3) )
              xx = SS(i,j)*d1*d2*d3*dx*dy*dz
       
              !! acumulate terms for the "idim" velocity component
              do idim=1,3
               velocity(idim)%SlabVals(ii,jj,kk)                         &
                                 =  velocity(idim)%SlabVals(ii,jj,kk)    &
                                 -  nhat(i,j,idim)*xx*dt
              enddo
       
            enddo
          enddo     !! end loop over section of surface for x(i,j)
         
         endif  !!  end process or no process for x(ii,jj,kk)
         
       enddo    !! next ii
      enddo     !! next jj     
     enddo      !! next kk, i.e. end loop over grid points on myid
       
   return
  end subroutine IB_Update
  !!====================================================================
  !!====================================================================
  
  !!====================================================================
  !!==================================================================== 
   subroutine ForcingFilter
    use user_data,              only: nx,ny
    use etc,                    only: numzplanes_below                                  
    use mpi_params,             only: myid
    use intermediate_variables, only: scratch
    use independent_variables,  only: Grid
    implicit none
    integer           :: i,j,k
    integer,save      :: locnz,k0
    real(kind=8)      :: x(3),step,delta,XX,YY,ZZ
    external             step,delta
   
  !!====================================================================
  !!====================================================================
    locnz=ubound(scratch(1)%SlabVals,3)
    k0=numzplanes_below(myid)
    
    !! loop over "all" the grid points on myid (i,j,k)
    do k=1,locnz
     x(3)=Grid(0)%SpaceDims(3)%x(k0+k) 
      do j=1,ny
       x(2)=Grid(0)%SpaceDims(2)%x(j) 
        do i=1,nx       
         x(1)=Grid(0)%SpaceDims(1)%x(i)
    
         
          !! assuming the vector from x(:), that intersects
          !! the IB surface normally, is not too far from
          !! the surface point just above or below x(:). 
          !! Use that point and it's unit normal to to get
          !! the x,y and z distances from x(:) to the
          !! actual closest point on IB to x(:).
          XX = (x(3) - h_IB(i,j))*nhat(i,j,1)/(3.*eps(1))
          YY = (x(3) - h_IB(i,j))*nhat(i,j,2)/(3.*eps(2))
          ZZ = (x(3) - h_IB(i,j))*nhat(i,j,3)/(3.*eps(3))
         
          if(ZZ>=0.d0)  then
           IB_Filter(i,j,k) = 1.d0 
          else
           IB_Filter(i,j,k) = exp(-XX**2) * exp(-YY**2) * exp(-ZZ**2)
          endif
       
                           
       enddo    
      enddo       
     enddo    
     
    
       
   return
  end subroutine ForcingFilter
  !!====================================================================
  !!====================================================================
   

  !!====================================================================
  !!====================================================================
   subroutine ij_from_count(i,j,count)
    use user_data, only: nx
    implicit none
    integer   :: i,j,count
    j = (count-1)/nx + 1
    i = count - (j-1)/nx
    return
   end subroutine ij_from_count
  !!====================================================================
  !!====================================================================
   
  !!====================================================================
  !!====================================================================    
   subroutine count_from_ij(i,j,count)
    use user_data, only: nx
    implicit none
    integer   :: i,j,count
    count = (j-1)/nx + i
    return
   end subroutine count_from_ij
  !!====================================================================
  !!====================================================================
  
  
!======================================================================
!======================================================================
subroutine unit_normal(x,n)
 use user_routines, only: eval_surface
 implicit none
 real(kind=8) :: x(2),n(3),h,h_x,h_y
 real(kind=8) :: mag
 
   call eval_surface(x,h,h_x,h_y)
   mag = sqrt( h_x**2 + h_y**2 + 1. )
   n(1)=-h_x/mag
   n(2)=-h_y/mag
   n(3)=1.d0/mag
   
 return
end subroutine unit_normal  ! end of function unit_normal
!======================================================================
!======================================================================



  
!!====================================================================================  
  end module immersed_boundary 
!!====================================================================================
