subroutine initialize
   
 use mpi_params
 use operators
 use etc
 use io_params 
 use particles
 use user_data
 use dependent_variables
 use intermediate_variables
 use immersed_boundary
   
 implicit none  
 character(len=80) :: message
 real(kind=8)      :: ds,dxds,h1(3),xx

 docfile='output/DetailsLog'
 memoryfile='output/MemoryLog'
 optionfile='input/PetscOptions'

!!===========================================================
!! Initialize mpi & petsc libraries
!!===========================================================
 call StartMPI(myid,numprocs,docfile,comm)
 call PetscInitialize(optionfile,ierr)

 if(myid==0) then
  write(0,*) 'Creating log file : ',trim(docfile)
  write(0,*) 'Using ',trim(optionfile),' for petsc user options.'
  open(1,file=trim(docfile),position='rewind')
   write(1,*) '============================================================ '
   write(1,*) '============================================================ '
   write(1,*) '                  INITIALIZATION REPORTS:'
   write(1,*) '============================================================ '
   write(1,*) '============================================================ '
   write(1,*) '  '
  close(1)
  
  open(2,file=trim(memoryfile),position='rewind')
   write(2,*) ' =========================================================== '
   write(2,*) ' =========================================================== '
   write(2,*) '                  Memory Usage Reports:'
   write(2,*) ' =========================================================== '
   write(2,*) ' =========================================================== '
   write(2,*) '  '
   write(2,*) '  time step ................................', istep
   write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
  close(2)
 endif
 
 
 !!===========================================================
 !! Read in user supplied parameter values 
 !!===========================================================
 if(myid==0) then
  message = '...  Calling Read_UserData...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif
 call Read_UserData 
 

 !!===========================================================
 !! Misc. preliminaries
 !!===========================================================
 if(myid==0) then
  message =  '...  Calling PreliminaryTasks...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif
 call PreliminaryTasks
 

 
 !!===========================================================
 !! Define
 !!   3 space dimensions, 
 !!   3 coordinate maps,
 !!   1 tensor product grid
 !!===========================================================
 if(myid==0) then
  message = '...  Calling SetupDomain ...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif
 call SetupDomain(myid)
 
 !!===========================================================
 !! Create {CFD} scalar variables 
 !!              velocity=[u,v,w] 
 !!               scalars=[s1,s2]
 !!       pressure variable [phi]
 !!===========================================================
 if(myid==0) then
  message = '...  Calling CreateScalarVariables ...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif
 call CreateScalarVariables 


 !!===========================================================
 !! Initialize the {CFD} dependent variables
 !!===========================================================
 if(myid==0) then
  message = '...  Calling InitialConditions ...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif
 call InitialConditions
 call eos                !! so pd will be written out in t=0 fields
                         !! choose the Divergence option to have 
                         !! ux,vy,wz returned in
                         !! scratch (2),(3) and (4)
 
 !!===========================================================
 !! Sanity: check divergence of ics
 !!===========================================================
 call Divergence(velocity,div_u,scratch(1),scratch(2))
 
 if(myid==0) write(0,*)   &
    '...  max u on myid=0 (m/s)',maxval(abs(velocity(1)%SlabVals))*U0
 if(myid==0) write(0,*)   &
    '...  max v on myid=0 (m/s)',maxval(abs(velocity(2)%SlabVals))*U0
 if(myid==0) write(0,*)   &
    '...  max w on myid=0 (m/s)',maxval(abs(velocity(3)%SlabVals))*U0
  
 if(myid==0) write(0,*) '... checking initial divergence '
  
 if(myid==0) write(0,*)   &
    '...  max div on myid=0 (1/s)',maxval(abs(div_u%SlabVals))*(U0/Lz)
 
 
 !!===========================================================
 !! Initialize particles
 !!===========================================================
 if(myid==0) then
  message = '... Calling InitializeParticles ...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif 
 call InitializeParticles
 
 !!===========================================================
 !! Initialize the immersed boundary
 !!===========================================================
 if(myid==0) then
  message = '... Calling InitializeImmersedBoundary ...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif 
 call InitializeImmersedBoundary
 
 !!===========================================================
 !! Construct the petsc operators for iterative solves
 !!===========================================================
 if(myid==0) then
  message = '... Calling ConstructOperators ...'
  write(0,*) message
  call LogMessage(message,docfile)
 endif 
 call ConstructOperators
 
 
 !!===========================================================
 !! Preliminaries....  
 !!===========================================================
 
 !!=====================================
 !! Eqns all have the general form...
 !! alpha(i) grad^2 [f]  + beta(i) * f  = rhs
 !! for i=1,2...6
 !! (though slightly modified if LES is chosen)
 !! Set the constants defining the eqns.
 !!=====================================
 !!
    alpha(1:3)=1.d0/Re       !! u,v,w viscous terms
    alpha(4)=1.d0/(Pr(1)*Re) !! s1 diffusive term
    alpha(5)=1.d0/(Pr(2)*Re) !! s2 diffusive term
    alpha(6)=1.d0            !! pressure (Laplacian)
 
    if(AM_order==3) then
     beta(1:5)=(-2.d0/dt)      !! helmholtz term for 3nd order AdamsMoulton
    elseif(AM_order==4) then
     beta(1:5)=-12.d0/(5.*dt)  !! helmholtz term for 4rd order AdamsMoulton
    else
     stop 'problem w/ AM_order: must be 3 or 4 ... initialize.f90'
    endif
 
    beta(6)=0.d0              !! pressure  (Laplacian)
 
 !!===========================================================
 !! Use a scaling factor like dx^2 for
 !! scaling A_FD if its needed
 !!===========================================================
   do i=1,3                     
    ds=L(0)%Grid%SpaceDims(i)%ds
    dxds= minval( L(0)%Grid%SpaceDims(i)%dxds(:) )
    h1(i) = dxds*ds
   enddo 
   h2_scale_factor = minval(h1)**2
 
  
 !!===========================================================
 !!==== PRESSURE ===> Matrix free L[f] 
 !!===========================================================
   field_id = 6 
   if(myid==0) then
    message = '... Building pressure solver: Create_L_Solver ...'
    write(0,*) message
    call LogMessage(message,docfile)
   endif  
   call Create_L_Solver ( ksp(field_id) )
 
 
 !!===========================================================
 !!==== VEL & SCALARS ===> Matrix free L[f]
 !!=========================================================== 
   if(myid==0) then
    message = '... other variables... Create_L_Solvers '
    write(0,*) message
    call LogMessage(message,docfile)
   endif 
  
 !!==========================================
 !! u,v,w can generally share an L operator
 !! scalars need their own if Pr is not 1,
 !! otherwise they can share as well.
 !!==========================================
   field_id=1
   call Create_L_Solver(ksp(field_id))
   ksp(2)=ksp(1)
   ksp(3)=ksp(1)
    
   if( Pr(1)==1 ) then
    ksp(4)=ksp(1)
   else
    field_id=4
    call Create_L_Solver(ksp(field_id))
   endif
  
   if( Pr(2)==Pr(1) ) then
    ksp(5)=ksp(4)
   else
    field_id=5
    call Create_L_Solver(ksp(field_id))
   endif
    
 if(myid==0) then
   message = ' -----> routine Initialize exiting normally <---------- '
   write(0,*) message
   call LogMessage(message,docfile)
  endif 
     
 !!===========================================================
 call mpi_barrier(comm,ierr)
 !!===========================================================
 
end subroutine initialize
!!===========================================================
!!===========================================================


!!===========================================================
!!===========================================================
subroutine PreliminaryTasks
 use user_data
 use etc
 use particles
 use io_params
 use mpi_params
 
 implicit none
 
 if(myid==0) then
  open(1,file=docfile,position='append')
  write(1,*) ' =========================================================== '
  write(1,*) ' =========================================================== '
  write(1,*) '                  PreliminaryTasks Report:'
  write(1,*) '============================================================ '
  write(1,*) '============================================================ '
  write(1,*) '  '
 endif
 
 
!!=========================================================== 
!! Pressure boundary conditions are
!! not meant to be a user option, set 
!! them here
!!===========================================================
!! First, set pressure assuming walls in each direction
 BCtype(6,:,:)='neumann'
 
 !! if periodic, ALL variables will be periodic
 if( xdim_periodic ) then
  BCtype(:,1,:)='periodic'
 endif
 
 if( ydim_periodic ) then
  BCtype(:,2,:)='periodic'
 endif
 
 if( zdim_periodic ) then
  BCtype(:,3,:)='periodic'
 endif
 
!!===========================================================
!! check to see whether pressure solve is in physical
!! or wavenumber space
!!===========================================================
!! assume wavenumber space soln is ok to begin
psolve_at_gridpoints=.FALSE.
psolve_exp_coeffs=.TRUE.
do_explicit_hydrostatic = .FALSE.

!!===========================================================
!! now see if there is reason to change these flags 
!! transform soln only allowable for trig expansions
!!===========================================================
 do idim=1,3
  if( trim(phi_deriv_types(idim)) .ne. 'sin' .and.    &
      trim(phi_deriv_types(idim)) .ne. 'cos' .and.    &
      trim(phi_deriv_types(idim)) .ne. 'fourier' ) then      
      psolve_at_gridpoints=.TRUE.
      psolve_exp_coeffs=.FALSE.
  endif
 enddo
 
!!===========================================================
!! spit out decision for pressure treatment 
!!===========================================================
 i=6
 if(myid==0) then
  write(1,*) '... psolve_at_gridpoints ',psolve_at_gridpoints
  write(1,*) '... psolve_exp_coeffs    ',psolve_exp_coeffs
  idim=1
  write(1,*) '... phi   x direction at 0,Lx  :  ',  &
     trim(BCtype(i,idim,1)),' ',trim(BCtype(i,idim,2)),' ', &
     trim(phi_deriv_types(idim))
  idim=2
  write(1,*) '... phi   y direction at 0,Ly  :  ',  &
     trim(BCtype(i,idim,1)),' ',trim(BCtype(i,idim,2)),' ', &
     trim(phi_deriv_types(idim))
  idim=3
  write(1,*) '... phi   z direction at 0,Lz  :  ',  &
     trim(BCtype(i,idim,1)),' ',trim(BCtype(i,idim,2)),' ', &
     trim(phi_deriv_types(idim))
 endif
 
!!===========================================================
!! check to see whether u solve is in physical
!! or wavenumber space
!!===========================================================
!! assume wavenumber space soln is ok to begin
 usolve_at_gridpoints=.FALSE.
 usolve_exp_coeffs=.TRUE.

!! now see if there is reason to switch 
 do idim=1,3
  if( trim(u_deriv_types(idim)) .ne. 'sin' .and.    &
      trim(u_deriv_types(idim)) .ne. 'cos' .and.    &
      trim(u_deriv_types(idim)) .ne. 'fourier' ) then
      
      usolve_at_gridpoints=.TRUE.
      usolve_exp_coeffs=.FALSE.
  endif
 enddo
 print*,'usolve_exp_coeffs=',usolve_exp_coeffs
 
 if(myid==0) then
  i=1
  write(1,*) '... usolve_at_gridpoints ',usolve_at_gridpoints
  write(1,*) '... usolve_exp_coeffs    ',usolve_exp_coeffs
  idim=1
  write(1,*) '... u   x direction at 0,Lx  :  ',  &
    trim(BCtype(i,idim,1)),' ',trim(BCtype(i,idim,2)),' ', &
    trim(u_deriv_types(idim))
  idim=2
  write(1,*) '... u   y direction at 0,Ly  :  ',  &
    trim(BCtype(i,idim,1)),' ',trim(BCtype(i,idim,2)),' ', &
    trim(u_deriv_types(idim))
  idim=3
  write(1,*) '... u   z direction at 0,Lz  :  ',  &
    trim(BCtype(i,idim,1)),' ',trim(BCtype(i,idim,2)),' ', &
    trim(u_deriv_types(idim))
 endif
 
!!===========================================================
!! check to see whether v solve is in physical
!! or wavenumber space
!!===========================================================
!! assume wavenumber space soln is ok to begin
vsolve_at_gridpoints=.FALSE.
vsolve_exp_coeffs=.TRUE.

!! now see if there is reason to switch 
 do idim=1,3
  if( trim(v_deriv_types(idim)) .ne. 'sin' .and.    &
      trim(v_deriv_types(idim)) .ne. 'cos' .and.    &
      trim(v_deriv_types(idim)) .ne. 'fourier' ) then
      
      vsolve_at_gridpoints=.TRUE.
      vsolve_exp_coeffs=.FALSE.
  endif
 enddo
  
!!===========================================================
!! check to see whether w solve is in physical
!! or wavenumber space
!!===========================================================
!! assume wavenumber space soln is ok to begin
wsolve_at_gridpoints=.FALSE.
wsolve_exp_coeffs=.TRUE.

!! now see if there is reason to switch 
 do idim=1,3
  if( trim(w_deriv_types(idim)) .ne. 'sin' .and.    &
      trim(w_deriv_types(idim)) .ne. 'cos' .and.    &
      trim(w_deriv_types(idim)) .ne. 'fourier' ) then
      
      wsolve_at_gridpoints=.TRUE.
      wsolve_exp_coeffs=.FALSE.
  endif
 enddo
 
 
!!===========================================================
!! check to see whether s1 solve is in physical
!! or wavenumber space
!!===========================================================
!! assume wavenumber space soln is ok to begin
s1solve_at_gridpoints=.FALSE.
s1solve_exp_coeffs=.TRUE.

!! now see if there is reason to switch 
 do idim=1,3
  if( trim(s1_deriv_types(idim)) .ne. 'sin' .and.    &
      trim(s1_deriv_types(idim)) .ne. 'cos' .and.    &
      trim(s1_deriv_types(idim)) .ne. 'fourier' ) then
      
      s1solve_at_gridpoints=.TRUE.
      s1solve_exp_coeffs=.FALSE.
  endif
 enddo
  
!!===========================================================
!! check to see whether s2 solve is in physical
!! or wavenumber space
!!===========================================================
!! assume wavenumber space soln is ok to begin
s2solve_at_gridpoints=.FALSE.
s2solve_exp_coeffs=.TRUE.

!! now see if there is reason to switch 
 do idim=1,3
  if( trim(s2_deriv_types(idim)) .ne. 'sin' .and.    &
      trim(s2_deriv_types(idim)) .ne. 'cos' .and.    &
      trim(s2_deriv_types(idim)) .ne. 'fourier' ) then
      
      s2solve_at_gridpoints=.TRUE.
      s2solve_exp_coeffs=.FALSE.
  endif
 enddo
 

 if(myid==0) then
  write(1,*) ' =========================================================== '
  write(1,*) ' =========================================================== '
  write(1,*) '                  Boundary Conditions Flags etc :'
  write(1,*) '============================================================ '
  write(1,*) '============================================================ '
  idim=1
  write(1,*) '  '
  write(1,*) 'u   x direction at 0,Lx  :  ',  trim(BCtype(1,idim,1)),' ',  &
     trim(BCtype(1,idim,2)),' ',trim(u_deriv_types(idim))
  write(1,*) 'v   x direction at 0,Lx  :  ',  trim(BCtype(2,idim,1)),' ',  &
     trim(BCtype(2,idim,2)),' ',trim(v_deriv_types(idim))
  write(1,*) 'w   x direction at 0,Lx  :  ',  trim(BCtype(3,idim,1)),' ',  &
     trim(BCtype(3,idim,2)),' ',trim(w_deriv_types(idim))
  write(1,*) 's1  x direction at 0,Lx  :  ',  trim(BCtype(4,idim,1)),' ',  &
     trim(BCtype(4,idim,2)),' ',trim(s1_deriv_types(idim))
  write(1,*) 's2  x direction at 0,Lx  :  ',  trim(BCtype(5,idim,1)),' ',  &
     trim(BCtype(5,idim,2)),' ',trim(s2_deriv_types(idim))
  write(1,*) ' xdim_periodic ',xdim_periodic
  write(1,*) '  '
  idim=2
  write(1,*) 'u   y direction at 0,Lx  :  ',  trim(BCtype(1,idim,1)),' ',  &
     trim(BCtype(1,idim,2)),' ',trim(u_deriv_types(idim))
  write(1,*) 'v   y direction at 0,Lx  :  ',  trim(BCtype(2,idim,1)),' ',  &
     trim(BCtype(2,idim,2)),' ',trim(v_deriv_types(idim))
  write(1,*) 'w   y direction at 0,Lx  :  ',  trim(BCtype(3,idim,1)),' ',  &
     trim(BCtype(3,idim,2)),' ',trim(w_deriv_types(idim))
  write(1,*) 's1  y direction at 0,Lx  :  ',  trim(BCtype(4,idim,1)),' ',  &
     trim(BCtype(4,idim,2)),' ',trim(s1_deriv_types(idim))
  write(1,*) 's2  y direction at 0,Lx  :  ',  trim(BCtype(5,idim,1)),' ',  &
     trim(BCtype(5,idim,2)),' ',trim(s2_deriv_types(idim))
  write(1,*) ' ydim_periodic ',ydim_periodic
  write(1,*) '  '
  idim=3
  write(1,*) 'u   z direction at 0,Lx  :  ',  trim(BCtype(1,idim,1)),' ',  &
     trim(BCtype(1,idim,2)),' ',trim(u_deriv_types(idim))
  write(1,*) 'v   z direction at 0,Lx  :  ',  trim(BCtype(2,idim,1)),' ',  &
     trim(BCtype(2,idim,2)),' ',trim(v_deriv_types(idim))
  write(1,*) 'w   z direction at 0,Lx  :  ',  trim(BCtype(3,idim,1)),' ',  &
     trim(BCtype(3,idim,2)),' ',trim(w_deriv_types(idim))
  write(1,*) 's1  z direction at 0,Lx  :  ',  trim(BCtype(4,idim,1)),' ',  &
     trim(BCtype(4,idim,2)),' ',trim(s1_deriv_types(idim))
  write(1,*) 's2  z direction at 0,Lx  :  ',  trim(BCtype(5,idim,1)),' ',  &
     trim(BCtype(5,idim,2)),' ',trim(s2_deriv_types(idim))
  write(1,*) ' zdim_periodic ',zdim_periodic
  write(1,*) '  '
  write(1,*) '  '
  write(1,*) '  '
 endif
 

!!===========================================================
!! explicit methods start @ 1st order
!!===========================================================
 step_flag='euler'
 particle_step_flag='euler'

!!===========================================================
!! create mpi derived types
!! set to 'no' once all types
!! created
!!===========================================================
 DoMPITypes(:)='yes'

!!===========================================================
!! I'll set the explicit & implicit
!! timestepping orders for now
!!===========================================================
 AB_order=3         !! k=3 step method error=O(h^4)
 AM_order=AB_order  !! k=2 step method error=O(h^4)

!!===========================================================
!! initial values for cyclic counters
!!===========================================================
 MM0=1;MM1=2;MM2=3;MM3=4  !! for AB timestepping
 PM0=1;PM1=2;PM2=3;PM3=4  !! for AB stepping particles
 N=1;NM1=2                !! for AM timestepping
 M_oldest=AB_order
 P_oldest=PM3
 N_oldest=AM_order-2
 time_counter(:)=1
 iend = nint( (t_end-t_start)/dt )
 t_secs = t_start
 
 !!===========================================================
 !!Dimensional scales
 !!===========================================================
 time_scale = Lz/u0
 p_scale = rho_0*u0**2
 bfreq = sqrt((g/rho_0)*DGRAD)
 
 !!===========================================================
 !!Dimensionless parameters.
 !!
 !! N.B.
 !! if a scalar is pd, scale it consistently with
 !! how I do the perturbation density in the w eqn, 
 !! ====> in this case, ignore user supplied scale
 !!===========================================================
 if( scalar_kind(1)=='r' ) then
  s1_scale=Lz*dgrad
 endif
 if( scalar_kind(2)=='r' ) then
  s2_scale=Lz*dgrad
 endif
 if( scalar_kind(1)=='r' .and. scalar_kind(2)=='r' ) then
  stop 'cant have both scalars set to density'
 endif
 Rot= f*Lz/U0
 beta_tilde = beta_dimensional*(Lz*Lz/u0)
 N2 = (g/rho_0)*dgrad
 Ri = N2*Lz**2/(U0**2)
 Re = U0*Lz/nu
 Pr(1) = nu/kappa_1
 Pr(2) = nu/kappa_2
 
 !!===========================================================
 !!   ===> from here on... dt is DIMENSIONLESS
 !!===========================================================
 dt = dt/time_scale   !! time step now dimensionless
 !!===========================================================
 !!===========================================================
 
 !!===========================================================
 !!    Write out parameters characterizing simulation.
 !!===========================================================
 if( myid == 0 ) then  
  write(1,*) ' bulk velocity scale U                   [m/s]:  ',U0
  write(1,*) ' characteristic density gradient scale [kg/m4]:  ',dgrad
  write(1,*) ' characteristic range of scalar 1 [e.g. deg C]:  ',s1_scale
  write(1,*) ' characteristic range of scalar 2   [e.g. psu]:  ',s2_scale
  write(1,*) ' characteristic density                [kg/m3]:  ',rho_0
  write(1,*) ' coriolis parameter f                    [1/s]:  ',f
  write(1,*) ' gravitational acceleration g           [m/s2]:  ',g
  write(1,*) ' molecular diffusivity of scalar 1      [m2/s]:  ',kappa_1
  write(1,*) ' molecular diffusivity of scalar 2      [m2/s]:  ',kappa_2
  write(1,*) ' viscosity                              [m2/s]:  ',nu
  write(1,*) ' '
  write(1,*) ' '
  write(1,*) ' time step dt                             [s] :  ',dt*time_scale
  write(1,*) ' initial time                             [s] :  ',t_start
  write(1,*) ' final time                               [s] :  ',t_end
  write(1,*) ' Lz/U0                                    [s] :  ',time_scale
  write(1,*) ' number of steps                              :  ',iend-istart+1
  write(1,*) ' '
  write(1,*) ' '
  write(1,*) ' inverse Rossby number                fL/U    :  ',Rot
  write(1,*) ' bulk Richardson number (NL/U)^               :  ',Ri
  write(1,*) ' bulk Reynolds number UL/nu                   :  ',Re
  write(1,*) ' Prandtl number  nu/kappa_1                   :  ',Pr(1)
  write(1,*) ' Prandtl number  nu/kappa_2                   :  ',Pr(2)
  write(1,*) ' N/f                                          :  ',sqrt(N2)/f
  write(1,*) ' sqrt(nu/f)                               [m] :  ',sqrt(nu)/f
  write(1,*) ' '
  write(1,*) ' '
  write(1,*) ' -----> PreliminaryTasks routine exiting normally  <---------- '
  write(0,*) ' -----> PreliminaryTasks routine exiting normally  <---------- '
 close(1)
 endif 
 
!!===========================================================
!!===========================================================
!! this needs to be thought through again...
!!===========================================================
!!===========================================================
 if( LES ) then
  p_les(:) = 3  !! i.e. k^6
  T_les = 2.*pi * minval( (/ 1./Rot, 1./Ri, 1.d0 /) )  
 
  !! require eg. that alpha_les(1) * k^2p * t = O(1) when
  !!   k = .85*k_max  and 
  !!   t = T_les, the shortest time scale I find w/ scaling
  !!   (everything dimensionless here)
 
  alpha_les(1) = 1./( T_les * (pi*(nx)*(Lz/Lx) )**(2.*p_les(1)) ) !! x direction
  alpha_les(2) = 1./( T_les * (pi*(ny)*(Lz/Ly) )**(2.*p_les(2)) ) !! y direction
  alpha_les(3) = 1./( T_les * (pi*(nz)*(Lz/Lz) )**(2.*p_les(3)) ) !! z direction
 
  if(ny==1) alpha_les(2)=0.d0
 elseif( DNS ) then
  alpha_les(:)=0.d0
  p_les=0.d0
 endif
!!===========================================================
!!===========================================================
 
end subroutine PreliminaryTasks
!!===========================================================
!!===========================================================



!!===========================================================
!!===========================================================
subroutine SetupDomain
 use class_CoordinateMap
 use class_TensorProductGrid
 use independent_variables
 use etc
 use mpi_params
 use user_data
 use user_routines

 implicit none
 type( CoordinateMap )      :: CoordMap(3)
 character (len=80)         :: MapLabel
 character (len=80)         :: CoordName
 integer                    :: level,job,rem,nmax,idim
 integer,dimension(5)       :: jerr
 integer,dimension(3)       :: npts,npts_coarse=5
 integer,dimension(3)       :: DirIndex=(/1,2,3/)
 real(kind=8)               :: limits(2)
 real(kind=8)               :: ds,Lvals(3)
 real(kind=8),allocatable   :: x(:),s(:),dxds(:),d2xds2(:)
 logical,dimension(3)       :: periodic
!=========================================================================== 
!=========================================================================== 
   
 if(myid==0) then
  open(1,file=docfile,position='append')
  write(1,*) ' =========================================================== '
  write(1,*) ' =========================================================== '
  write(1,*) '                  SetupDomain Report:'
  write(1,*) ' =========================================================== '
  write(1,*) ' =========================================================== '
  write(1,*) '  '
  
  open(2,file=memoryfile,position='append')
  write(2,*) ' =========================================================== '
  write(2,*) ' =========================================================== '
  write(2,*) '                  Memory Usage SetupDomain:'
  write(2,*) ' =========================================================== '
  write(2,*) ' =========================================================== '
  write(2,*) '  '
  write(2,*) '  time step ................................', istep
  write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
 endif

!!===========================================================  
!!Construct coordinate maps and grid at each multigrid level.
!!This coarsening algorithm should match petsc's refine/coarsen.
!!===========================================================
 npts=(/nx,ny,nz/)
 nmax=MAXVAL( (/nx,ny,nz/) )
 periodic=(/xdim_periodic,ydim_periodic,zdim_periodic/)
 Lvals=(/Lx,Ly,Lz/)

 if(myid==0) write(1,*) ' Constructing coordinate maps, space dimensions, ' 
 if(myid==0) write(1,*) ' & grids for each level of the multigrid heirarchy.'

 allocate( mgrefine(3,0:1-1) ) 
 mgrefine(:,:) = -999 
 if( xdim_periodic ) npts_coarse(1)=-999
 if( ydim_periodic ) npts_coarse(2)=-999

do level=0,1-1   
  
!!===================================================================
!! scaled x dimension:         x in [0,Lx/Lz] ----> CoordMap(1)
!! scaled y dimension:         y in [0,Ly/Lz] ----> CoordMap(2)
!! scaled, global z dimension: z in [0,Lz/Lz] ----> CoordMap(3)
!!===================================================================


do idim=1,3

 allocate(      x(npts(idim)), stat=jerr(1) )
 allocate(      s(npts(idim)), stat=jerr(2) )
 allocate(   dxds(npts(idim)), stat=jerr(3) )
 allocate( d2xds2(npts(idim)), stat=jerr(4) )
 
 !!===========================================================
 !! Check for allocation errors
 !!===========================================================
 if(jerr(1)/=0)   &
  stop 'problem allocating memory for ...x in SetupDomain/initialize.f90'
 if(jerr(2)/=0)   &
  stop 'problem allocating memory for ...s in SetupDomain/initialize.f90'
 if(jerr(3)/=0)   &
  stop 'problem allocating memory for .x_s in SetupDomain/initialize.f90'
 if(jerr(4)/=0)   &
  stop 'problem allocating memory for x_ss in SetupDomain/initialize.f90'

 !!===========================================================
 !! Call user routine to get gridpoint values via s-->x map
 !!===========================================================
 call map_s2x_at_gridpoints(idim,periodic(idim),s,Lvals(idim), &
                            x,dxds,d2xds2,MapLabel)
 if(npts(idim)==1) then
  x=0.d0
  s=0.d0
  dxds=1.d0  !! I divide by dxds to get dsdx
  d2xds2=0.d0
 endif
  
 !!===========================================================
 !! Scale the results using Lz [m], s is dimensionless
 !!===========================================================
 x=x/Lz   
 dxds=dxds/Lz 
 d2xds2=d2xds2/Lz
 
 CoordMap(idim)    &
   = NewCoordinateMap(npts(idim),MapLabel,x,s,      &
                      dxds,d2xds2,Lvals(idim)/Lz)
 
 !!======================================================== 
 !! Record details
 !!======================================================== 
 if( level==0 .and. myid==0 ) then
  close(1)
  call DescribeCoordinateMap(CoordMap(idim),docfile)
  open(1,file=docfile,position='append')
 endif
 
 !!======================================================== 
 !! Output 1d gridpoint & stretching vectors
 !!========================================================
 if(idim==1) open(10,file='output/xvec',position='rewind')
 if(idim==2) open(10,file='output/yvec',position='rewind')
 if(idim==3) open(10,file='output/zvec',position='rewind')
 do k=1,npts(idim)
  write(10,*) k,x(k)*Lz,dxds(k)*Lz,d2xds2(k)*Lz
 enddo
 close(10)
 
 !!========================================================
 !! coarsen grid in dimension idim for next level
 !!  ... historical really, we're fixed at 1 level
 !!========================================================
 rem        = mod( npts(idim),mgrefine(idim,level) )
 npts(idim) = ( npts(idim)-rem)/(mgrefine(idim,level) ) + rem
 npts(idim) = maxval( (/npts_coarse(idim),npts(idim)/) ) 
 
 deallocate(     x)
 deallocate(     s)
 deallocate(  dxds)
 deallocate(d2xds2)
 
enddo

!!======================================================== 
!! Link the coordinate maps together to form a T.P. grid
!!======================================================== 
 call NewTensorProductGrid(CoordMap,Grid(level))
  
enddo    !! end loop over multigrid levels
 
!!=================================================================
!! setup weightings for easy spatial integration
!! first, set all weights to dz = dz/ds * ds  etc
!!=================================================================
allocate( int_weights_x(nx) )
allocate( int_weights_y(ny) )
allocate( int_weights_z(nz) )

idim=1
int_weights_x(:) = CoordMap(idim)%dxds(:) * CoordMap(idim)%ds
idim=2
int_weights_y(:) = CoordMap(idim)%dxds(:) * CoordMap(idim)%ds
idim=3
int_weights_z(:) = CoordMap(idim)%dxds(:) * CoordMap(idim)%ds

!!=================================================================
!! now incorporate the numerical weights for Simpsons
!! closed extended rule w/ error O(1/N^4)
!! except for periodic directions where weights are all equal
!!=================================================================
if( .not. xdim_periodic .and. nx>3) then
 int_weights_x(1) = int_weights_x(1) * (3.d0/8.d0)
 int_weights_x(2) = int_weights_x(2) * (7.d0/6.d0)
 int_weights_x(3) = int_weights_x(3) * (23.d0/24.d0)

 int_weights_x(nx-2) = int_weights_x(nx-2) * (23.d0/24.d0)
 int_weights_x(nx-1) = int_weights_x(nx-1) * (7.d0/6.d0)
 int_weights_x(nx) = int_weights_x(nx) * (3.d0/8.d0)
endif

if( .not. ydim_periodic .and. ny>3) then
 int_weights_y(1) = int_weights_y(1) * (3.d0/8.d0)
 int_weights_y(2) = int_weights_y(2) * (7.d0/6.d0)
 int_weights_y(3) = int_weights_y(3) * (23.d0/24.d0)

 int_weights_y(ny-2) = int_weights_y(ny-2) * (23.d0/24.d0)
 int_weights_y(ny-1) = int_weights_y(ny-1) * (7.d0/6.d0)
 int_weights_y(ny) = int_weights_y(ny) * (3.d0/8.d0)
endif

if( ny==1 ) int_weights_y=1.d0

if( .not. zdim_periodic .and. nz>3 ) then
 int_weights_z(1) = int_weights_z(1) * (3.d0/8.d0)
 int_weights_z(2) = int_weights_z(2) * (7.d0/6.d0)
 int_weights_z(3) = int_weights_z(3) * (23.d0/24.d0)

 int_weights_z(nz-2) = int_weights_z(nz-2) * (23.d0/24.d0)
 int_weights_z(nz-1) = int_weights_z(nz-1) * (7.d0/6.d0)
 int_weights_z(nz) = int_weights_z(nz) * (3.d0/8.d0)
endif

deallocate(mgrefine)
if(myid==0) then
 write(1,*) ' -----> SetupDomain routine exiting normally  <---------- '
 write(0,*) ' -----> SetupDomain routine exiting normally  <---------- '
 close(1)
endif
end subroutine SetupDomain
!!===========================================================
!!===========================================================





subroutine CreateScalarVariables
!===========================================================================
!===========================================================================
!Classes
 use class_TensorProductGrid
 use class_DataDecomposition
 use class_ScalarVariable
 use class_CoordinateMap

!!Modules
 use mpi_params 
 use user_data
 use etc
 use dependent_variables
 use independent_variables
 use intermediate_variables
 use operators
!=========================================================================== 
!===========================================================================
 implicit none
 
 character(len=80)    :: label
 character(len=80)    :: scalarname
 
 integer              :: level,jerr(5),islice
 real(kind=8)         :: scale
 integer,allocatable  :: numplanes(:)
!===========================================================================
!===========================================================================



 if(myid==0) then
  open(1,file=docfile,position='append')
  write(1,*) ' =========================================================== '
  write(1,*) ' =========================================================== '
  write(1,*) '              CreateScalarVariables Report:'
  write(1,*) ' =========================================================== '
  write(1,*) ' =========================================================== '
  write(1,*) '  '
  close(1)
  open(2,file=memoryfile,position='append')
  write(2,*) ' =========================================================== '
  write(2,*) ' =========================================================== '
  write(2,*) '                  Memory Usage CreateScalarVariables:'
  write(2,*) ' =========================================================== '
  write(2,*) ' =========================================================== '
  write(2,*) '  '
  write(2,*) '  time step ................................', istep
  write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
 endif

!!=========================================================================== 
!!Use the number of processors, the grid parameters and the derivative types
!!to layout the distributed data arrays. Give each variable its own data
!!decomposition map.
!!===========================================================================
 level=0 !! finest level
 if(myid==0) write(0,*) '... creating NewDataDecompositions'
 
 label='U'
 call NewDataDecomposition(label,Grid(0),level,U_deriv_types,U_DataMap)
 label='V'
 call NewDataDecomposition(label,Grid(0),level,V_deriv_types,V_DataMap)
 label='W'
 call NewDataDecomposition(label,Grid(0),level,W_deriv_types,W_DataMap)
 label='S1'
 call NewDataDecomposition(label,Grid(0),level,S1_deriv_types,S1_DataMap)
 label='S2'
 call NewDataDecomposition(label,Grid(0),level,S2_deriv_types,S2_DataMap)
 label='Div_U'
  
 !! deriv types for divergence don't really matter since we won't
 !! ever differentiate div(u*)
 div_deriv_types=U_deriv_types
 label='Div_U'
 call NewDataDecomposition(label,Grid(0),level,div_deriv_types,DIV_DataMap)
  
 call DescribeDecomposition(U_DataMap,docfile)
 call DescribeDecomposition(V_DataMap,docfile)
 call DescribeDecomposition(W_DataMap,docfile)
 call DescribeDecomposition(S1_DataMap,docfile)
 call DescribeDecomposition(S2_DataMap,docfile)
  
 allocate( numzplanes(0:numprocs-1), stat=jerr(1) )
 allocate( numzplanes_below(0:numprocs-1), stat=jerr(2) )
 allocate( numplanes(0:numprocs-1), stat=jerr(3) )
 
 if(jerr(1)/=0)  &
   stop 'problem allocating numzplanes in CreateScalarVariables'
 if(jerr(2)/=0)  &
   stop 'problem allocating numzplanes_below in CreateScalarVariables'
 if(jerr(3)/=0)  &
   stop 'problem allocating numplanes in CreateScalarVariables'
 
 !!=====================================================
 !! count numbewr of grid points below processor myid
 !! in 'column' orientation and to the left for 'slabs'
 !!=====================================================
 numzplanes=0; numzplanes_below=0; numplanes=0

 if(myid==0) write(0,*) '... MPI AllReduce ...',myid,numprocs
 if(U_DataMap%locnz .ne. DIV_DataMap%locnz)    &
   stop 'datamaps are different u & div_u locnz'
   
 numzplanes(myid) = U_DataMap%locnz
 call mpi_allreduce(numzplanes,numplanes,numprocs,         &
                    mpi_integer,mpi_sum,comm,ierr )
 numzplanes=numplanes

 numzplanes_below(myid) = U_DataMap%my_global_kvals(1)-1
 call mpi_allreduce(numzplanes_below,numplanes,numprocs,   &
                    mpi_integer,mpi_sum,comm,ierr )
 numzplanes_below=numplanes
 
 allocate( numxplanes(0:numprocs-1), numxplanes_left(0:numprocs-1) )
 numxplanes=0; numxplanes_left=0; numplanes=0

 if(U_DataMap%locnx .ne. DIV_DataMap%locnx)    &
   stop 'datamaps are different u & div_u locnx'
   
 numxplanes(myid) = U_DataMap%locnx
 call mpi_allreduce(numxplanes,numplanes,numprocs,         &
                    mpi_integer,mpi_sum,comm,ierr )
 numxplanes=numplanes

 numxplanes_left(myid) = U_DataMap%my_global_ivals(1)-1
 call mpi_allreduce(numxplanes_left,numplanes,numprocs,   &
                    mpi_integer,mpi_sum,comm,ierr )
 numxplanes_left=numplanes
 deallocate(numplanes)
 
 !!===========================================================================
 !!allocate arrays for ambient scalar profiles
 !!each processor stores its local portion only
 !!===========================================================================
 allocate( s1_bar(numzplanes(myid)) )
 allocate( s2_bar(numzplanes(myid)) )
 allocate( ddz_s1_bar(numzplanes(myid)) )
 allocate( ddz_s2_bar(numzplanes(myid)) )
 allocate( ddzz_s1_bar(numzplanes(myid)) )
 allocate( ddzz_s2_bar(numzplanes(myid)) )
 
 if(myid==0) then
  NewAllocationGB = 4*numzplanes(myid)*8*GBperByte
  AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
  write(2,*) '  New allocations s1_bar,s2_bar etc ........',NewAllocationGB
  write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
  write(2,*) '   '
  close(2)
endif

!!=========================================================================== 
!!Create ScalarVariable objects that exist only on the finest gridlevel=0
!!===========================================================================

 if(myid==0) open(1,file=docfile,position='append')
 if(myid==0) write(1,*) ' Creating ScalarVariables on fine grid level for '
 if(myid==0) write(1,*) ' velocity(:), scalars(:) '
 if(myid==0) write(0,*) '... creating scalars velocity[],s1,s2,pd,rhs_b'
 if(myid==0) close(1)
 
 
 do i=1,1          
  scalarname = 'u'
  if(myid==0) write(0,*) '... creating scalar u'
  call NewScalar(scalarname,Grid(0),U_DataMap,u0,velocity(1))  
 
  scalarname = 'v'
  if(myid==0) write(0,*) '... creating scalar v'
  call NewScalar(scalarname,Grid(0),V_DataMap,u0,velocity(2))
 
  scalarname = 'w'
  if(myid==0) write(0,*) '... creating scalar w'
  call NewScalar(scalarname,Grid(0),W_DataMap,u0,velocity(3))
 
  scalarname = 's1'
  if(myid==0) write(0,*) '... creating scalar s1'
  call NewScalar(scalarname,Grid(0),S1_DataMap,s1_scale,scalars(1)) 
 
  scalarname = 's2'
  if(myid==0) write(0,*) '... creating scalar s2'
  call NewScalar(scalarname,Grid(0),S2_DataMap,s2_scale,scalars(2))
 enddo
 
  scalarname = 'pd'
   if(myid==0) write(0,*) '... creating scalar pd'
  call NewScalar(scalarname,Grid(0),S2_DataMap,s2_scale,pd)
 
  scalarname = 'div_u'
  if(myid==0) write(0,*) '... creating scalar div_u'
  scale=1/time_scale
  call NewScalar(scalarname,Grid(0),DIV_DataMap,scale,div_u)
 
  do j=1,AM_order-2  
   scalarname = 'implicit_rhs_u'
   if(myid==0) write(0,*) '... creating scalar implicit_rhs(1)'
   scale=1/time_scale
   call NewScalar(scalarname,Grid(0),U_DataMap,scale,implicit_rhs(1,j))
 
   scalarname = 'implicit_rhs_v'
   if(myid==0) write(0,*) '... creating scalar implicit_rhs(2)'
   scale=1/time_scale
   call NewScalar(scalarname,Grid(0),U_DataMap,scale,implicit_rhs(2,j))
 
   scalarname = 'implicit_rhs_w'
   if(myid==0) write(0,*) '... creating scalar implicit_rhs(3)'
   scale=1/time_scale
   call NewScalar(scalarname,Grid(0),U_DataMap,scale,implicit_rhs(3,j))
 
   scalarname = 'implicit_rhs_s1'
   if(myid==0) write(0,*) '... creating scalar implicit_rhs(4)'
   scale=1/time_scale
   call NewScalar(scalarname,Grid(0),U_DataMap,scale,implicit_rhs(4,j))
 
   scalarname = 'implicit_rhs_s2'
   if(myid==0) write(0,*) '... creating scalar implicit_rhs(5)'
   scale=1/time_scale
   call NewScalar(scalarname,Grid(0),U_DataMap,scale,implicit_rhs(5,j))
  enddo

   
  if(myid==0) write(0,*) '... creating ScalarVariable explicit_rhsx5x',AB_ORDER
  scale=(u0**2/Lz)
  do islice=1,AB_order
   scalarname = 'explicit_rhs_1'
   call NewScalar(scalarname,Grid(0),U_DataMap,scale,explicit_rhs(1,islice))
   scalarname = 'explicit_rhs_2'
   call NewScalar(scalarname,Grid(0),V_DataMap,scale,explicit_rhs(2,islice))
   scalarname = 'explicit_rhs_3'
   call NewScalar(scalarname,Grid(0),W_DataMap,scale,explicit_rhs(3,islice))
   scalarname = 'explicit_rhs_4'
   call NewScalar(scalarname,Grid(0),S1_DataMap,scale,explicit_rhs(4,islice))
   scalarname = 'explicit_rhs_5'
   call NewScalar(scalarname,Grid(0),S2_DataMap,scale,explicit_rhs(5,islice))
  enddo

!!===========================================================================  
!! Create DataDecomposition and 
!! ScalarVariable Phi for each grid level:
!!===========================================================================
 label = 'phi'
 do level=0,mglevels-1
  call NewDataDecomposition( label, Grid(level), level,            &
                             Phi_deriv_types, Phi_DataMap(level) )
  scalarname = 'Phi'
  if(myid==0) write(0,*) '... creating scalar Phi at grid level ',level
  
  call NewScalar( scalarname, Grid(level),            &
                  Phi_DataMap(level), U0, Phi(level) )  
  call DescribeDecomposition(Phi_DataMap(level),docfile)
  
  if(myid==0)  &
    write(0,*) '... Phi_deriv_types(1): ',trim( Phi(level)%DerivTypes(1) )
  if(myid==0)  &
    write(0,*) '... Phi_deriv_types(2): ',trim( Phi(level)%DerivTypes(2) )
  if(myid==0)  &
    write(0,*) '... Phi_deriv_types(3): ',trim( Phi(level)%DerivTypes(3) )
 enddo
 
 scalarname = 'rhs_b'
  if(myid==0) write(0,*) '... creating scalar rhs b'
  scale=(u0/Lz)**2   !! [p]=rho_0 u0^2  --> 1/rho_0 grad^2 p
  call NewScalar(scalarname,Grid(0),Phi_DataMap(0),scale,rhs_b)
  if(myid==0)  &
    write(0,*) '... rhs_b_deriv_types(1): ',trim( rhs_b%DerivTypes(1) )
  if(myid==0)  &
    write(0,*) '... rhs_b_deriv_types(2): ',trim( rhs_b%DerivTypes(2) )
  if(myid==0)  &
    write(0,*) '... rhs_b_deriv_types(3): ',trim( rhs_b%DerivTypes(3) )
 
 if(myid==0) write(0,*) '... creating scalar scratch x 4'
 scale=1
 scalarname = 'scratch(1)'
 call NewScalar(scalarname,Grid(0),Phi_DataMap(0),scale,scratch(1))
 scalarname = 'scratch(2)'
 call NewScalar(scalarname,Grid(0),Phi_DataMap(0),scale,scratch(2))
 scalarname = 'scratch(3)'
 call NewScalar(scalarname,Grid(0),Phi_DataMap(0),scale,scratch(3))
 scalarname = 'scratch(4)'
 call NewScalar(scalarname,Grid(0),Phi_DataMap(0),scale,scratch(4))
   
 !!call TestGatherScatter(velocity(1))

if(myid==0) then
 open(1,file=docfile,position='append')
 write(0,*) ' -----> CreateScalarVariables routine exiting normally  <---------- '
 write(1,*) ' -----> CreateScalarVariables routine exiting normally  <---------- '
 open(2,file=memoryfile,position='append')
  write(2,*) ' -----> CreateScalarVariables routine exiting normally  <---------- '
 close(2)
 close(1)
endif 
end subroutine CreateScalarVariables


!!=================================================================
!!=================================================================
 subroutine InitialConditions      
  !!Routine to call user_ics to initialize
  !!the dependent variables  
  use user_routines
  use user_data
  use etc
  use mpi_params
  use dependent_variables
  use independent_variables,  only: Grid
  use intermediate_variables
  
  implicit none
      
  real(kind=8)                     :: scale_factor(5),zpos
  real(kind=8)                     :: kx,kz,ff,dsdx,dsdz
  integer                          :: kg,xx
  integer                          :: m1,m2,m3,k0,k1,my_k
  character(len=80)                :: for_inv_flag
  
  if(myid==0) then
   open(1,file=docfile,position='append')
   write(1,*) ' =========================================================== '
   write(1,*) ' =========================================================== '
   write(1,*) '              InitialConditions Report:'
   write(1,*) ' =========================================================== '
   write(1,*) ' =========================================================== '
   write(1,*) '  '
  endif
 
  scale_factor(1:3) = 1./(U0)
  scale_factor(4)   = 1./(s1_scale)
  scale_factor(5)   = 1./(s2_scale)
  if(myid==0) then
   write(1,*) 'Scale factor for u ',scale_factor(1) 
   write(1,*) 'Scale factor for v ',scale_factor(2) 
   write(1,*) 'Scale factor for w ',scale_factor(3) 
   write(1,*) 'Scale factor for s1 ',scale_factor(4) 
   write(1,*) 'Scale factor for s2 ',scale_factor(5)
  endif
              
  m1=velocity(1)%DataMap%SlabDims(1,myid)
  m2=velocity(1)%DataMap%SlabDims(2,myid)
  m3=velocity(1)%DataMap%SlabDims(3,myid)
  
  !!=====================================================
  !! figure out which portion of the global 
  !! z vector is allocated to processor myid
  !!=====================================================
  k0 = velocity(1)%DataMap%my_global_kvals(1)
  k1 = k0+m3-1
  
  !!=====================================================
  !! use s1_bar(:) as a temp array here, its not set yet
  !!=====================================================
  s1_bar(:) = velocity(1)%Grid%SpaceDims(3)%x(k0:k1)

      
  !!==========================================================
  !! call user_ics routine providing the dimensional
  !! local spatial coordinates, the current time in seconds
  !! a Fortran array to fill with values and an integer
  !! indicator as to which field's ics are being requested
  !!==========================================================
  if(myid==0) write(0,*) '... Calling user_ics for u,v,w',m1,m2,m3

  
  do i=1,3
    call user_ics(Lz*velocity(1)%Grid%SpaceDims(1)%x,      &
                  Lz*velocity(1)%Grid%SpaceDims(2)%x,      &
                  Lz*s1_bar,  &
                  velocity(i)%SlabVals,i,m1,m2,m3)   
    velocity(i)%SlabVals = velocity(i)%SlabVals*scale_factor(i)  
  enddo
  
  
  if(myid==0) write(0,*) '... Calling user_ics again for scalars'
  do i=4,5
    call user_ics(Lz*velocity(1)%Grid%SpaceDims(1)%x,      &
                  Lz*velocity(1)%Grid%SpaceDims(2)%x,      &
                  Lz*s1_bar,  &
                  scalars(i-3)%SlabVals,i,m1,m2,m3)                          
    scalars(i-3)%SlabVals = scalars(i-3)%SlabVals*scale_factor(i)  
  enddo
  
  !do k=1,scalars(1)%DataMap%locnz
  ! my_k = velocity(1)%DataMap%my_global_kvals(k)
  ! zpos = Lz*Grid(0)%SpaceDims(3)%x(my_k) ! [m] 
  ! call eval_s1_bar(zpos,s1_bar(k),ddz_s1_bar(k),ddzz_s1_bar(k))
  ! call eval_s2_bar(zpos,s2_bar(k),ddz_s2_bar(k),ddzz_s2_bar(k))
  !enddo
  call eval_s1_bar(s1_bar,ddz_s1_bar,ddzz_s1_bar)
  call eval_s2_bar(s2_bar,ddz_s2_bar,ddzz_s2_bar)
  
  !!nondimensionalize
  s1_bar=s1_bar/(s1_scale)
  s2_bar=s2_bar/(s2_scale)
  ddz_s1_bar=ddz_s1_bar/(s1_scale/Lz)
  ddz_s2_bar=ddz_s2_bar/(s2_scale/Lz)
  ddzz_s1_bar=ddzz_s1_bar/(s1_scale/Lz**2)
  ddzz_s2_bar=ddzz_s2_bar/(s2_scale/Lz**2)
  
  !!Compute dimensionless z limits for each processor
  !!get k index of myid's lowest z gridpoint
  k = velocity(1)%DataMap%my_global_kvals(1)  
  my_z_limits(1) = Grid(0)%SpaceDims(3)%x(k)

  !!get k index of myid's uppermost z gridpoint
  k = velocity(1)%DataMap%my_global_kvals(velocity(1)%DataMap%locnz)
  
  !!if possible, extend region of responsibility
  !!(for particles, immersed bdry) upward 1 grdpoint
  if(myid<numprocs-1) k=k+1
  my_z_limits(2) = Grid(0)%SpaceDims(3)%x(k)
        
  if(myid==0) then
   open(1,file=docfile,position='append')
   write(1,*) ' -----> InitialConditions routine exiting normally  <---------- '
   write(0,*) ' -----> InitialConditions routine exiting normally  <---------- '
   close(1)
  endif
  return
 end subroutine InitialConditions
!=============================================================== 
!===============================================================



 
subroutine ConstructOperators
!=============================================================== 
!!    need to check & see how much of this is still used
!===============================================================
 use class_SAOperator
 use class_BoundaryOperator
 use class_BoundaryConditions
 use class_ScalarVariable
 use operators
 use dependent_variables
 use intermediate_variables
 use independent_variables
 use user_data
 use user_routines
 use etc
 use mpi_params
!===========================================================================
!===========================================================================
 implicit none
 integer, parameter        :: order=2  !! up to 5th order operators possible
 integer                   :: level,_n1,_n2,k0
 integer                   :: idim,jpos,rcode(2)=0
 integer                   :: locnx,locny,locnz
 character(len=80)         :: label,bdry_type
 real(kind=8),allocatable  :: vals(:,:)
 real(kind=8),allocatable  :: x1(:),x2(:)
 real(kind=8)              :: XX

 if(myid==0) then
   open(1,file=docfile,position='append')
   write(1,*) ' =========================================================== '
   write(1,*) ' =========================================================== '
   write(1,*) '              ConstructOperators Report:'
   write(1,*) ' =========================================================== '
   write(1,*) ' =========================================================== '
   write(1,*) '  '
  endif
 
 locnx=velocity(1)%DataMap%SlabDims(1,myid)
 locny=velocity(1)%DataMap%SlabDims(2,myid)
 locnz=velocity(1)%DataMap%SlabDims(3,myid)
 _n1=MAXVAL( (/nx,ny,nz/) )
 allocate( x1(_n1), stat=rcode(1) )
 allocate( x2(_n1), stat=rcode(2) )
 if( any( rcode /= 0 ) ) Stop 'Allocation failed:  ConstructOperators 1'
 
 do level=0,mglevels -1
 
   !!Get a set of homogeneous bdry operators, both D & N
   call InitializeBoundaryConditions(     &
        Grid(level),Phi_DataMap(level),   &
        order,DirBdryOps,PhiBdryOps  )
   
   !! Correct labels for periodic dimensions
   if( xdim_periodic ) then
    do i=1,2
     DirBdryOps(1,i)%label='periodic'
     PhiBdryOps(1,i)%label='periodic'
    enddo
   endif
   
   if( ydim_periodic ) then
    do i=1,2
     DirBdryOps(2,i)%label='periodic'
     PhiBdryOps(2,i)%label='periodic'
    enddo
   endif
           
   !!Create new SAOperator at each level, attach boundary operators
   call NewSAOperator(Phi(level),PhiBdryOps,L(level),mglevels)
   if(myid==0)  &
     write(1,*) '   SA operator L(i) created & linked w/ PhiBdryOps '
   if(myid==0)  &
     write(0,*) '   SA operator L(i) mg level, nrows ',level,L(level)%nrows
         
   !!Delete the BdryOps before re-initializing them at next coarser level
   if( level > 0 ) then
    do i=1,3
     do j=1,2
      call DeleteBoundaryOperator(DirBdryOps(i,j))
      call DeleteBoundaryOperator(NeuBdryOps(i,j))
     enddo
    enddo 
   endif
   
 enddo    !! end loop over multigrid levels
  

!! Construct a set of homogeneous neumann conditions
!! for each variable. These will then be updated to
!! conform to the user specifications.
 level=0
 call InitializeBoundaryConditions(      &
        Grid(level),Phi_DataMap(level),  &
        order,DirBdryOps,u_BdryOps )
        
 call InitializeBoundaryConditions(      &
        Grid(level),Phi_DataMap(level),  &
        order,DirBdryOps,v_BdryOps )
        
 call InitializeBoundaryConditions(      &
        Grid(level),Phi_DataMap(level),  &
        order,DirBdryOps,w_BdryOps )
        
 call InitializeBoundaryConditions(      &
        Grid(level),Phi_DataMap(level),  &
        order,DirBdryOps,s1_BdryOps )
        
 call InitializeBoundaryConditions(      &
        Grid(level),Phi_DataMap(level),  &
        order,DirBdryOps,s2_BdryOps )
 
!! update labels for periodic dimensions
if( xdim_periodic ) then
   BCtype(:,1,:)='periodic'
   u_BdryOps(1,1)%label='periodic'
   u_BdryOps(1,2)%label='periodic'
   v_BdryOps(1,1)%label='periodic'
   v_BdryOps(1,2)%label='periodic'
   w_BdryOps(1,1)%label='periodic'
   w_BdryOps(1,2)%label='periodic'
   s1_BdryOps(1,1)%label='periodic'
   s1_BdryOps(1,2)%label='periodic'
   s2_BdryOps(1,1)%label='periodic'
   s2_BdryOps(1,2)%label='periodic'
 endif
 if( ydim_periodic ) then
   BCtype(:,2,:)='periodic'
   u_BdryOps(2,1)%label='periodic'
   u_BdryOps(2,2)%label='periodic'
   v_BdryOps(2,1)%label='periodic'
   v_BdryOps(2,2)%label='periodic'
   w_BdryOps(2,1)%label='periodic'
   w_BdryOps(2,2)%label='periodic'
   s1_BdryOps(2,1)%label='periodic'
   s1_BdryOps(2,2)%label='periodic'
   s2_BdryOps(2,1)%label='periodic'
   s2_BdryOps(2,2)%label='periodic'
 endif
 
!! Now loop through all the operators and update
!! as necessary using data specified in user_bdry_data.
k0 = velocity(1)%DataMap%my_global_kvals(1)
label='data'
do field_id=1,5
 do idim=1,3
 
  if(idim==1) then
   _n1=velocity(1)%DataMap%SlabDims(2,myid)
   _n2=velocity(1)%DataMap%SlabDims(3,myid)  !! YZ
   x1(1:_n1)=velocity(1)%Grid%SpaceDims(2)%x
   x2(1:_n2)=velocity(1)%Grid%SpaceDims(3)%x(k0:k0+locnz-1)
  elseif(idim==2) then
   _n1=velocity(1)%DataMap%SlabDims(1,myid)
   _n2=velocity(1)%DataMap%SlabDims(3,myid)  !! XZ
   x1(1:_n1)=velocity(1)%Grid%SpaceDims(1)%x
   x2(1:_n2)=velocity(1)%Grid%SpaceDims(3)%x(k0:k0+locnz-1)
  elseif(idim==3) then
   _n1=velocity(1)%DataMap%SlabDims(1,myid)
   _n2=velocity(1)%DataMap%SlabDims(2,myid)  !! XY
   x1(1:_n1)=velocity(1)%Grid%SpaceDims(1)%x
   x2(1:_n2)=velocity(1)%Grid%SpaceDims(2)%x
  endif
  !! use dimensional values to pass to user routines
  x1(:)=x1*Lz
  x2(:)=x2*Lz
  
  allocate( vals(_n1,_n2), stat=rcode(1) )
  if( rcode(1) /= 0  ) then
   write(0,*) 'allocating vals(:,:) ',_n1,_n2
   Stop 'Allocation failed:  ConstructOperators  2'
  endif
 
  do jpos=1,2
   !! determine user requested bdry type and switch 
   !! from default neumann to user type if necessary
   bdry_type=BCtype(field_id,idim,jpos)
   if( trim(bdry_type) .ne. 'neumann' ) then
     if(field_id==1) call SwitchType( u_BdryOps(idim,jpos),bdry_type )
     if(field_id==2) call SwitchType( v_BdryOps(idim,jpos),bdry_type )
     if(field_id==3) call SwitchType( w_BdryOps(idim,jpos),bdry_type )
     if(field_id==4) call SwitchType( s1_BdryOps(idim,jpos),bdry_type )
     if(field_id==5) call SwitchType( s2_BdryOps(idim,jpos),bdry_type )
   endif
   
   !! get the user specified boundary values ==> vals
   call UserBdryData(field_id,idim,jpos,bdry_type,_n1,_n2,x1,x2,vals)
   
   !! nondimensionalize these values (depends on field_id and bc_type)
   if( field_id < 4 ) then
   
     if( trim(bdry_type)=='neumann' .or. trim(bdry_type)=='Neumann' ) then
       XX = Lz/U0
     elseif( trim(bdry_type)=='dirichlet' .or. trim(bdry_type)=='Dirichlet' ) then
       XX = 1./U0
     endif
     vals=XX*vals
     
     !! update the operator
     if(field_id==1) &
       call UpdateBoundaryOperator(u_BdryOps(idim,jpos),label,vals,_n1,_n2)
     if(field_id==2) &
       call UpdateBoundaryOperator(v_BdryOps(idim,jpos),label,vals,_n1,_n2)
     if(field_id==3) &
       call UpdateBoundaryOperator(w_BdryOps(idim,jpos),label,vals,_n1,_n2)
       
    elseif( field_id==4) then
    
     if( trim(bdry_type)=='neumann' .or. trim(bdry_type)=='Neumann' ) then
       XX = Lz/s1_scale
     elseif( trim(bdry_type)=='dirichlet' .or. trim(bdry_type)=='Dirichlet' ) then
       XX = 1./s1_scale
     endif
     vals=XX*vals
     
     !! update the operator
     call UpdateBoundaryOperator(s1_BdryOps(idim,jpos),label,vals,_n1,_n2)
     
    elseif( field_id==5) then
        
     if( trim(bdry_type)=='neumann' .or. trim(bdry_type)=='Neumann' ) then
       XX = Lz/s2_scale
     elseif( trim(bdry_type)=='dirichlet' .or. trim(bdry_type)=='Dirichlet' ) then
       XX = 1./s2_scale
     endif
     vals=XX*vals
     
     !! update the operator
     call UpdateBoundaryOperator(s2_BdryOps(idim,jpos),label,vals,_n1,_n2)
     
    endif
   
  enddo  !!end loop over jpos
  
  deallocate( vals )
  
 enddo  !! end loop over dimensions

enddo  !! end field_id loop

deallocate( x1,x2 )
!! keep a fresh set of homogeneous dirichlet & neumann for fine grid
level=0
call InitializeBoundaryConditions(      &
        Grid(level),Phi_DataMap(level),  &
        order,DirBdryOps,NeuBdryOps )

 
!! Correct labels for periodic dimensions
if( xdim_periodic ) then
 do i=1,2
  DirBdryOps(1,i)%label='periodic'
  PhiBdryOps(1,i)%label='periodic'
 enddo
endif
   
if( ydim_periodic ) then
 do i=1,2
  DirBdryOps(2,i)%label='periodic'
  PhiBdryOps(2,i)%label='periodic'
 enddo
endif
   
if(myid==0) then
  write(1,*) ' -----> ConstructOperators routine exiting normally  <---------- '
  write(0,*) ' -----> ConstructOperators routine exiting normally  <---------- '
 close(1)
endif
end subroutine ConstructOperators






