subroutine StartMPI(myid,numprocs,docfile,comm) 
  use mpi_times
  implicit none
  integer             :: ierr,comm,myid,numprocs
  character(len=80)   :: docfile
  include 'mpif.h'
 
  comm=MPI_COMM_WORLD
  call mpi_init(ierr)
  call mpi_comm_rank(comm,myid,ierr)
  call mpi_comm_size(comm,numprocs,ierr)
      
  if( myid .eq. 0 .and. ierr==MPI_SUCCESS ) then
    write(0,*) ' MPI initialized with ',numprocs,' live processors'
    open(1,file=docfile,position='rewind')
      write(1,*) ' '
      write(1,*) ' MPI initialized with ',numprocs,' live processors'
      write(1,*) '             comm =   ',comm
      write(1,*) ' MPI error flag   =   ',ierr
    close(1)
   endif  
   
   call MPI_TIME(0)

 end subroutine StartMPI

 subroutine quit_mpi
   implicit none
   integer   :: jerr
   include 'mpif.h'
   call mpi_finalize(jerr)
   print*,' A Bientot ...' 
   stop ' '
  end subroutine quit_mpi

      
 subroutine create_MPI_data_types(DataMap) 
!! Routine to create a specialized data type for MPI data swaps
!! between processors. A subslab is a section taken from a data SLAB.
!! Each processor's SLAB is made up of numprocs subslab. Each processor
!! will have to export (numprocs-1) subslabs to other processors while
!! receiving 1 subslab from each of the others. The received subslabs
!! form a data COLUMN, i.e. an array partitioned in x rather than z. Columns
!! are necessary to compute transforms in the z direction. Having a special
!! MPI derived type simplifies coding of the data swaps.
!! Subslab indexing: 
!!         subslab(level,i,j) where i,j defined such that
!!         SlabDims(3,i) defines the number of z planes 
!!         ColDims(1,j)  defines the number of x planes
!!                       all subslabs have SlabDims(2,:)=ColDims(2,:) y planes
!! All processors create and store all possible data types.

!!MPI_TYPE_HVECTOR (count, blocklength, stride, oldtype, newtype, ierr)

      !!IN count Number of blocks to be added
      !!IN blocklen Number of elements in block
      !!IN stride Number of BYTES between start of each block
      !!IN oldtype Datatype of each element
      !!OUT newtype Handle (pointer) for new derived type
  
  use class_DataDecomposition
  use user_data
  use mpi_params
  use etc,  only: docfile
  implicit none
      
  type( DataDecomposition )     :: DataMap
  integer                       :: ierr,level,i,j
  integer                       :: numblocks,elements_per_block,stride
  integer                       :: sizeofrealword, iflag  
  integer, dimension(3)         :: SlabDims,ColDims !! slab/col array sizes in real words
  integer, dimension(3)         :: SlabVals,ColVals !! slab/col data sizes in real words

  include 'mpif.h'
  
  level = DataMap%level
  if(DoMPITypes(level)=='no') return

  !if( myid .eq. 0 ) then
  !  open(1,file=docfile,position='append')
  !    write(1,*) ' '
  !    write(1,*) ' Creating specialized MPI data types for level ',DataMap%level
  !    write(1,*) '       myid, comm : ',myid,comm
  !  close(1)
  ! endif
    
  call mpi_initialized(iflag,ierr)
  call mpi_type_extent(mpi_double_precision,sizeofrealword,ierr)
   
  do j=0,numprocs-1       !! block id
   do i=0,numprocs-1      !! process id
   
   !! Define sizes of for subslab(j)  
   SlabDims(3) = DataMap%SlabDims(3,i)   !! # of z values
   SlabDims(2) = DataMap%SlabDims(2,i)   !! # of y values
   SlabDims(1) = DataMap%ColDims(1,j)    !! # of x values  in SUB-BLOCK J
   SlabVals=SlabDims
       
   !!Build type for a 1d data "pencil" spanning this "subslab" in the x direction
   numblocks = SlabDims(1)
   elements_per_block = 1
   stride = sizeofrealword
   call mpi_type_hvector(numblocks,elements_per_block,stride,     &
                         mpi_double_precision,pencil,ierr)

   !!Type for a 2d section of a subslab :  stack pencils ==> an xz plane ----> sheet
   !!The num of blocks and the stride (in bytes) between the beginning of 1-pencil blocks 
   !!depends on the total Slab size on processor i
   numblocks=SlabDims(3)       !! stack pencils vertically
   elements_per_block = 1      !! each block = 1 pencil
   stride = DataMap%SlabDims(1,i)*DataMap%SlabDims(2,i)*sizeofrealword  !! bytes btwn pencils
   call mpi_type_hvector(numblocks,elements_per_block,stride,  &
                         pencil,slab_sheet(i,j),ierr)
      
   if( ierr .ne. MPI_SUCCESS ) then
    write(0,*) ' mpi_type_hvector slab sheet failure, myid=',myid
    stop
   endif
   call mpi_type_commit(slab_sheet(i,j),ierr)

   numblocks=SlabDims(2)       !! i.e. ny sheets
   elements_per_block = 1      !! each block = 1 sheet
   stride = DataMap%SlabDims(1,i)*sizeofrealword  !! bytes btwn sheets
   call mpi_type_hvector(numblocks,elements_per_block,stride,      &
                        slab_sheet(i,j),subslab(i,j),ierr)    
  
   if( ierr .ne. MPI_SUCCESS ) then
    write(0,*) ' mpi_type_hvector subslab failure: myid=',myid
    stop
   endif
   call mpi_type_commit(subslab(i,j),ierr)
  
!  if( myid .eq. 0 ) then
!   open(1,file=docfile,position='append')
!     write(1,*) ' '
!     write(1,*) '       type subslab created (level,i,j) (',level,i,j,')'
!   close(1)
!  endif
   
  enddo
 enddo
  
 
  do j=0,numprocs-1       !! block id
   do i=0,numprocs-1      !! process id
   
   !! Define sizes of for subslab(j)  
   ColDims(3) = DataMap%SlabDims(3,j)   !! # of z values in SUB-BLOCK j
   ColDims(2) = DataMap%ColDims(2,i)    !! # of y values
   ColDims(1) = DataMap%ColDims(1,i)    !! # of x values  
   ColVals=ColDims
       
   !!Build type for a 1d data "pencil" spanning this "subslab" in the x direction
   numblocks = ColDims(1)
   elements_per_block = 1
   stride = sizeofrealword
   call mpi_type_hvector(numblocks,elements_per_block,stride,     &
                         mpi_double_precision,pencil,ierr)

   !!Type for a 2d section of a subcol :  stack pencils ==> an xz plane ----> sheet
   numblocks=ColDims(3)        !! stack pencils vertically
   elements_per_block = 1      !! each block = 1 pencil
   stride = DataMap%ColDims(1,i)*DataMap%ColDims(2,i)*sizeofrealword  !! bytes btwn pencils
   call mpi_type_hvector(numblocks,elements_per_block,stride,  &
                         pencil,col_sheet(i,j),ierr)
      
   if( ierr .ne. MPI_SUCCESS ) then
    write(0,*) ' mpi_type_hvector slab sheet failure, myid=',myid
    stop
   endif
   call mpi_type_commit(col_sheet(i,j),ierr)

   numblocks=ColDims(2)        !! i.e. ny sheets
   elements_per_block = 1      !! each block = 1 sheet
   stride = DataMap%ColDims(1,i)*sizeofrealword  !! bytes btwn sheets
   call mpi_type_hvector(numblocks,elements_per_block,stride,      &
                        col_sheet(i,j),subcol(i,j),ierr)    
  
   if( ierr .ne. MPI_SUCCESS ) then
    write(0,*) ' mpi_type_hvector subslab failure: myid=',myid
    stop
   endif
   call mpi_type_commit(subcol(i,j),ierr)
  
!  if( myid .eq. 0 ) then
!   open(1,file=docfile,position='append')
!     write(1,*) ' '
!     write(1,*) '       type subcol created (i,j) (',i,j,')'
!   close(1)
!  endif
   
  enddo
 enddo

 DoMPITypes(level)='no'
 call mpi_barrier(comm,ierr)
 return
 end subroutine create_MPI_data_types
      

 subroutine columns_to_slabs(x,array,n1,n2,n3)
!!Routine to swap data from "column" orientation to "slab"  orientation.
!!See data maps for details. Relies on the specialized data  structures
!!"subslab" (and could use "sheet") defined in "create_MPI_data_type".

  use mpi_params
  use class_ScalarVariable
  use etc, only: numzplanes_below,numxplanes_left,numzplanes

  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: x
  integer                :: n1,n2,n3
  integer                :: status_array(mpi_status_size,4*numprocs+1)
  integer                :: reqs(4*numprocs) 
  integer                :: kstart,istart,kend,iend,ierr
  real(kind=8)           :: array(n1,n2,n3)  !! myid's column allocation
  
  
  if(numprocs==1) then
   x%SlabVals = array
   return
  endif
  
  
!!Ship portions of my local data "column" x%ColVals needed by other nodes
!!and recieve remote data required to construct my local data "slab" x%SlabVals.

!!Post the receives first....immediate mode, ie return before data is actually transferred
  nreqs=0
  dest=myid
  numwords=1
  tag = 2
  
  do iproc=0,numprocs-1
   source=iproc
    istart=(numxplanes_left(source))+1  !! start index for receiving "subslab" data  x    
      
       if( dest .ne. source ) then !! bypass mpi when source=dest=myid
       
        nreqs=nreqs+1
        call  mpi_irecv(x%SlabVals(istart,1,1),numwords,subslab(dest,source),  &
                        source,tag,comm,reqs(nreqs),ierr)       
        if( ierr .ne. 0 ) then
         stop 'MPI IRECV ERROR IN COLS TO SLABS'
        endif  
        
       endif
       
  enddo
  
  
  call mpi_barrier(comm,ierr)

!!All processors have issued all their receives, now the sends can be done w/o buffering.
!!Use immediate-ready-send  meaning, return before transfer is finished and guarantee that
!!the target locations on the destination processors are ready for writing. Faster.
  source=myid
  numwords=1
  tag = 2

  do iproc=0,numprocs-1
  
   dest=iproc
    kstart=(numzplanes_below(dest))+1 !! start index for outgoing column data in x%ColVals    
      
        if( dest .ne. source ) then !! bypass mpi when source=dest=myid
        
         nreqs=nreqs+1
         call mpi_irsend(array(1,1,kstart),numwords,subcol(source,dest),   &
                          dest,tag,comm,reqs(nreqs),ierr)
         if( ierr .ne. 0 ) then
          stop 'MPI IRSEND ERROR IN COLS TO SLABS'
         endif
         
        endif
        
  enddo
  
  !! case where source=dest=myid
  istart = numxplanes_left(myid) + 1
  iend   = istart + n1 - 1
  kstart = numzplanes_below(myid) + 1
  kend   = kstart + numzplanes(myid) - 1
  x%SlabVals(istart:iend,:,:) = array(:,:,kstart:kend)
  
  !!make sure all communications are done before proceeding
  call mpi_waitall(nreqs,reqs(1),status_array(1,1),ierr)  
   
  return
 end subroutine columns_to_slabs





 subroutine slabs_to_columns(x,array,n1,n2,n3)
!!Routine to swap data from "slab" orientation to "column"  orientation.
!!See data maps for details. Relies on the specialized mpi data structures
!!"slab" and "sheet" defined in "create_MPI_data_type".

  use mpi_params
  use class_ScalarVariable
  use etc, only: numzplanes_below,numxplanes_left,numzplanes

  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: x
  integer                :: n1,n2,n3
  integer                :: status_array(mpi_status_size,4*numprocs+1)
  integer                :: reqs(4*numprocs)
  integer                :: kstart,istart,ierr
  integer                :: iend,kend
  real(kind=8)           :: array(n1,n2,n3)
  
  
  if(numprocs==1) then
   array = x%SlabVals
   return
  endif
  
    
!!All processors initiate nonblocking receives, establishing  buffers for the sends
  nreqs=0
  dest=myid
  numwords=1                       !! i.e. 1 word of type subslab
  tag = 1

  do iproc=0,numprocs-1
   source=iproc
    kstart=(numzplanes_below(source))+1 !! start index for recieving column data in x%ColVals
      
       if( dest .ne. source ) then !! bypass mpi when source=dest=myid
       
         nreqs=nreqs+1
         call mpi_irecv(array(1,1,kstart),numwords,subcol(dest,source),   &
                        source,tag,comm,reqs(nreqs),ierr )
         if( ierr .ne. 0 ) then
          stop 'MPI IRECV ERROR IN SLABS TO COLS'
         endif 
         
       endif
       
  enddo
  
  
  call mpi_barrier(comm,ierr)   ! make sure all processors receives are posted to avoid buffering


!!All processors now initiate nonblocking ready sends, writing to the pre-established dest locations
  source=myid
  numwords=1                       !! i.e. 1 word of type subslab
  tag = 1
      
  do iproc=0,numprocs-1
  
   dest=iproc
    istart=(numxplanes_left(dest))+1  !! start index for sending "subslab" data  x
      
        if( dest .ne. source ) then !! bypass mpi when source=dest=myid
        
         nreqs=nreqs+1
         call  mpi_irsend(x%SlabVals(istart,1,1),numwords,subslab(source,dest),  &
                          dest,tag,comm,reqs(nreqs),ierr)
         if( ierr .ne. 0 ) then
          stop 'MPI IRSEND ERROR IN SLABS TO COLS'
         endif
         
        endif
        
  enddo
  
  !! case where source=dest=myid
  kstart = numzplanes_below(myid) + 1
  kend   = kstart + numzplanes(myid) - 1
  istart = numxplanes_left(myid) + 1
  iend   = istart + n1 - 1
  array(:,:,kstart:kend) = x%SlabVals(istart:iend,:,:)
  
  call mpi_waitall(nreqs,reqs(1),status_array(1,1),ierr)

  return
 end subroutine slabs_to_columns


 subroutine TestGatherScatter(x)
  use mpi_params
  use class_ScalarVariable
  use etc, only: numzplanes_below,numxplanes_left

  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: x
  integer                :: s1,s2,s3
  integer                :: c1,c2,c3
  integer                :: i,j,k,n,istart,iend,ierr
  real(kind=8),allocatable   :: slab(:,:,:)
  real(kind=8),allocatable   :: col(:,:,:)

  s1=x%DataMap%SlabDims(1,myid)
  s2=x%DataMap%SlabDims(2,myid)
  s3=x%DataMap%SlabDims(3,myid)
  
  c1=x%DataMap%ColDims(1,myid)
  c2=x%DataMap%ColDims(2,myid)
  c3=x%DataMap%ColDims(3,myid)
  
  call mpi_barrier(comm,ierr)
  
  allocate( slab(s1,s2,s3),col(c1,c2,c3) )
  slab=-999.d0
  
  !!Fill slab in sections corresponding to "subslabs"
  !!with a different integer in each subslab
  do n=0,numprocs-1
   istart=numxplanes_left(n)+1
   iend=istart + x%DataMap%ColDims(1,n) - 1
   slab(istart:iend,:,:)=n + 0.d0
  enddo
  
  if(myid==0) then
  do i=1,s1
   do j=1,s2
    do k=1,s3
     if( slab(i,j,k) == -999 ) then
      do n=0,numprocs-1
       write(0,*) 'n, numxplanes_left(n) ',n, numxplanes_left(n)
       write(0,*) ' istart, iend, coldims(n) ',istart, iend,x%DataMap%ColDims(1,n)
      enddo
      stop 'part of slab not filled, mpi_utils.f90 426'
     endif
    enddo
   enddo
  enddo
  endif
  
  
  
  do n=0,numprocs-1  
   if(myid==n) then
    open(1,file='output/COLS_SLABS_TEST',position='append')
    write(1,*) ' ==================== '
    do i=1,s1
     write(1,*) 'myid=',myid,'row of slab vals: ',i,slab(i,1,1)
    enddo
    write(1,*) ' ==================== '
    do i=1,s3
     write(1,*) 'myid=',myid,'col of slab vals: ',i,slab(1,1,i)
    enddo
    write(1,*) ' ==================== '
    write(1,*) '  '
   endif
   close(1)
   call mpi_barrier(comm,ierr)  
  enddo
  
  call slabs_to_columns(x,col,c1,c2,c3)
  call mpi_barrier(comm,ierr)
  
  do n=0,numprocs-1  
   if(myid==n) then
    open(1,file='output/COLS_SLABS_TEST',position='append')
    write(1,*) ' ==================== '
    do i=1,c1
     write(1,*) 'myid=',myid,'row of col vals: ',i,col(i,1,1)
    enddo
    write(1,*) ' ==================== '
    do i=1,c3
     write(1,*) 'myid=',myid,'col of col vals: ',i,col(1,1,i)
    enddo
    write(1,*) ' ==================== '
    write(1,*) '   '
    close(1)
   endif
   call mpi_barrier(comm,ierr)  
  enddo
  
  call columns_to_slabs(x,col,c1,c2,c3)
  
  do n=0,numprocs-1  
   if(myid==n) then
    open(1,file='output/COLS_SLABS_TEST',position='append')
    write(1,*) ' ==================== '
    do i=1,s1
     write(1,*) 'myid=',myid,'row of SV slab vals: ',x%SlabVals(i,1,1)
    enddo
    write(1,*) ' ==================== '
    do i=1,s3
     write(1,*) 'myid=',myid,'col of SV slab vals: ',x%SlabVals(1,1,i)
    enddo
    write(1,*) ' ==================== '
    write(1,*) '  '
    close(1)
   endif
   call mpi_barrier(comm,ierr)  
  enddo
  
  !!check for nans
! do i=1,s1
!  do j=1,s2
!   do k=1,s3
!    if( ieee_is_nan(slab(i,j,k)) ) then
!     write(0,*) 'NaN detected in SLABS_COLS_TEST myid,i,j,k:',myid,i,j,k
!     stop ' stopped at line 458 mpi_utils.f90 '
!    endif
!   enddo
!  enddo
! enddo

  
  
  deallocate( slab,col )
  return
  end
