module columnslab

implicit none

contains

 subroutine slabs_to_columns(slab,col)
! In the cfd code, data is partitioned across processors in equal
! sub-blocks in dimension 3, i.e. in horizontal slabs. It is sometimes
! convenient to exchange data so that it is partitioned along dimension
! 1, i.e. in vertical columns. This routine redistributes the data
! array "slab" into column orientation "col".

  use mpi_parameters
  use boundary_information
  include '../input/problem_size.h'
  include 'mpif.h'

  integer              :: tag,ierr,status_array(MPI_STATUS_SIZE,4*numprocs+1)
  integer              :: dest,source,istart,iend,kstart,kend,reqs(4*numprocs)
  integer              :: sizeofreal,nreqs
  integer              :: iproc,numwords
  integer                               :: m1,m2,m3
  real,dimension(:,:,:),intent(in)      :: slab 
  real,dimension(:,:,:),intent(out)     :: col   

  if( boundary_type2(1,1) .ne. 'periodic' .or.  boundary_type4(1,1) .ne. 'periodic' ) then
   write(0,*) 'myid,boundary_type2(1,1) = ',myid,boundary_type2(1,1)
   write(0,*) '  USING PERIODIC ENDPOINT LOGIC IN SLABS_TO_COLUMNS W/O PERIODIC BCS !!!'
   stop
  endif

 m1=size(col,1)
 m2=size(slab,2)
 m3=size(slab,3)

!!*******************************************************************************
!!*************   DO THE DATA EXCHANGES  ****************************************
!!*******************************************************************************
!! Ship portions of my local data "slab" slab needed by other nodes 
!! and recieve remote data required to construct my local data "column" col .

 !! All processors initiatze nonblocking receives, establishing buffers for the
 !! sends
 nreqs=0
 dest=myid
 numwords=m1*m2*m3
 tag=1

  do iproc=0,numprocs-1
    source=iproc
    kstart=(iproc*m3)+1
    if ( dest .ne. source ) then  !! bypass mpi when source=dest=myid 
            nreqs=nreqs+1
            call MPI_IRECV(col(1,1,kstart),numwords,MPI_REAL,source,tag,comm,reqs(nreqs),ierr)
            if ( ierr .ne. 0 ) then
                 stop 'MPI IRECV ERROR IN SLABS TO COLS'   
            endif 
    endif        
  enddo

  call MPI_BARRIER(comm,ierr)    !! make sure all processors receives are posted
                                 !! to avoid buffering
    
 !! All processors now initiate nonblocking ready sends, writing to the
 !! pre-established dest locations 
 source=myid
 numwords=1 
 tag = 1

 do iproc=0,numprocs-1
   dest=iproc
   istart=(iproc*m1)+1
   if ( dest .ne. source ) then  !! bypass mpi when source=dest=myid 
       nreqs=nreqs+1    
       call MPI_IRSEND(slab(istart,1,1),numwords,subslice,dest,tag,comm,reqs(nreqs),ierr)
       if ( ierr .ne. 0 ) then
               stop 'MPI IRSEND ERROR IN SLABS TO COLS'
       endif
   endif
 enddo

 !! case where source=dest=myid
 kstart=myid*m3+1
 kend=kstart+m3-1
 istart=myid*m1+1
 iend=istart+m1-1
 col(:,:,kstart:kend)=slab(istart:iend,:,:)

 call MPI_WAITALL(nreqs,reqs(1),status_array,ierr)

end subroutine slabs_to_columns


subroutine columns_to_slabs(col,slab)
! In the cfd code, data is partitioned across processors in equal
! sub-blocks in dimension 3, i.e. in horizontal slabs. It is sometimes
! convenient to exchange data so that it is partitioned along dimension
! 1, i.e. in vertical columns. This routine redistributes the data
! array "col" into slab orientation "slab".

  use mpi_parameters
  use boundary_information
  include '../input/problem_size.h'
  include 'mpif.h'

  integer              :: tag,ierr,status_array(MPI_STATUS_SIZE,2*numprocs)
  integer              :: dest,source,istart,iend,kstart,kend,reqs(4*numprocs+1)
  integer              :: sizeofreal,nreqs,j
  integer              :: iproc,numwords
  integer                             :: m1,m2,m3
  real,dimension(:,:,:),intent(out)   :: slab
  real,dimension(:,:,:),intent(in)    :: col 

  if( boundary_type2(1,1) .ne. 'periodic' .or.  boundary_type4(1,1) .ne. 'periodic' ) then
   write(0,*) '  USING PERIODIC ENDPOINT LOGIC IN SLABS_TO_COLUMNS W/O PERIODIC BCS !!!'
   stop
  endif

  m1=size(col,1)
  m2=size(slab,2)
  m3=size(slab,3)

!!******************************************************************************
!!*************   DO THE DATA EXCHANGES  ***************************************
!!******************************************************************************
!!Ship portions of my local data "column" x%ColVals needed by other nodes
!!and recieve remote data required to construct my local data "slab" x%SlabVals.

!!Post the receives first....immediate mode, ie return before data is actually transferred
  nreqs=0
  dest=myid
  numwords=1
  tag = 2

  do iproc=0,numprocs-1
    source=iproc
    istart=iproc*m1+1  !! start index for receiving "subslab" data  x
      if( dest .ne. source ) then !! bypass mpi when source=dest=myid
        nreqs=nreqs+1
        call  MPI_IRECV(slab(istart,1,1),numwords,subslice,source,tag,comm,reqs(nreqs),ierr)
        if( ierr .ne. 0 ) then
           stop 'MPI IRECV ERROR IN COLS TO SLABS'
        endif
      endif
  enddo

  call MPI_BARRIER(comm,ierr)    !! make sure all processors receives are posted
                                 !! to avoid buffering

!!All processors have issued all their receives, now the sends can be done w/o buffering.
!!Use immediate-ready-send  meaning, return before transfer is finished and guarantee that
!!the target locations on the destination processors are ready for writing. Faster.
  source=myid
  numwords=m1*m2*m3
  tag = 2

  do iproc=0,numprocs-1
    dest=iproc
    kstart=iproc*m3+1 !! start index for outgoing column data in x%ColVals
      if( dest .ne. source ) then !! bypass mpi when source=dest=myid
         nreqs=nreqs+1
         call MPI_IRSEND(col(1,1,kstart),numwords,MPI_REAL,dest,tag,comm,reqs(nreqs),ierr)
         if( ierr .ne. 0 ) then
            stop 'MPI IRSEND ERROR IN COLS TO SLABS'
         endif
      endif
  enddo

  !! case where source=dest=myid
  istart = myid*m1 + 1
  iend   = istart + m1 - 1
  kstart = myid*m3 + 1
  kend   = kstart + m3 - 1
  slab(istart:iend,:,:) = col(:,:,kstart:kend)

  !!make sure all communications are done before proceeding
  call MPI_WAITALL(nreqs,reqs,status_array,ierr)
!******************************************************************************
!******************************************************************************
!******************************************************************************

end subroutine columns_to_slabs


end module



