subroutine equation_of_state

! routine to use the equation of state from
! MILLERO et.al. (1980) DEEP-SEA RES.
! to calculate the dimensionless perturbation density pd(x,y,z) 
! given the scalar values s1(x,y,z), s2(x,y,z), s1_bar(x,z) and s2_bar(x,z).
! The "perturbation" density is defined relative to the ambient state
! rho_bar(x,z) determined by s1_bar and s2_bar. As written, we interpret
! the scalar s1 to be temparature and the scalar s2 to be salinity.
! Salinity in g/kg or equivilently ppt (typical value ~ 35/1000)

use mpi_parameters
use grid_info
use dependent_variables
use dimensional_scales
use counters_flags_etc
implicit none
include '../input/problem_size.h'

double precision :: rho,xx,z_surf   ! rho is external function
  real, dimension(locnz)            ::  zz,rh

! local variables
! S,T,P all DOUBLE precision arguments of function rho:  P in MPa, T in deg C, S in ppt
double precision, allocatable         :: Temp(:,:),Salinity(:,:),P_Mpa(:,:),density(:,:)
allocate( Temp(nx,locnz),Salinity(nx,locnz),P_Mpa(nx,locnz),density(nx,locnz) )

!   pressure as function of depth, ignoring density perturbations
!   formula assumes surface height z is equal to L_scale

 z_surf=L_scale  ! surface height in [m]   ! top processor should broadcast value
!!! P_Mpa(:,:) = g*rho0*( z_surf - Grid(0)%z(:,:)*L_scale )  ! [kg/ms^2=Pa]
!!! P_Mpa(:,:) = P_Mpa/1.e6                                  ! Mpa
 P_Mpa(:,:) = 0.                                  ! Mpa

do j=1,ny

!declare interpretation of s1 and s2
 Temp(:,:) = (s1_bar(:,:) + s1(:,j,:)) - s1_scale        ! dimensional quantity: total Temp.
 Salinity(:,:) = (s2_bar(:,:) + s2(:,j,:))*s2_scale    ! dimensional quantity: total Salinity
 
  do k=1,locnz
   do i=1,nx
    density(i,k)=rho( Salinity(i,k),Temp(i,k),P_Mpa(i,k) ) ! total density [kg/m3]
    density(i,k) = (density(i,k)-rho0)/density_scale ! dimensionless "total"(rel. to rho0) density in j plane
   enddo
  enddo

!If not yet initialized, calculate rho_bar, otherwise its available via "dependent_variables" 
 if( rho_bar(1,1) == -999.9 ) then     
  Temp(:,:) = s1_bar(:,:) - s1_scale                 ! dimensional background Temp
  Salinity(:,:) = s2_bar(:,:)*s2_scale             ! dimensional background Salinity

  do k=1,locnz
   do i=1,nx
    xx=rho( Salinity(i,k),Temp(i,k),P_Mpa(i,k) )
    rho_bar(i,k)=(xx-rho0)/density_scale           ! dimensionless ambient profile, rel. to rho0
   enddo
  rh(k) = rho_bar(1,k)*density_scale + rho0
  zz(k) = Grid(0)%z(1,k)*L_scale
  enddo
!!!      write(0,'(/A8,I2.2,2F20.5/)') (('DENSPROF',myid,zz(k),rh(k)),k=1,locnz)
 endif
       

!store dimensionless values of perturbation densityin jth plane of pd
  do k=1,locnz
   do i=1,nx
     pd(i,j,k) = density(i,k)  - rho_bar(i,k)
   enddo
  enddo

enddo  ! end loop over j


deallocate( Temp,Salinity,density,P_Mpa )
end subroutine equation_of_state


!	"rho.f"

!	Parameter	:	equation of state
!	Source		:	MILLERO et.al. (1980) DEEP-SEA RES.
!				Vol 27A, pp 255-264
!	Units		:	Kg / m**3
!	Input		:	s (concentration units), t (deg C), p (MPa)
!	Range		:	s (0 to 0.040), t (-2 to 40), p (MPa)
!	Ck value	:	K (0, 0, 100)      = 2.297721e4
!				K (0.035, 0, 100)  = 2.499200e4
!				K (0, 25, 100)     = 2.54051e4
!				K (0.035, 25, 100) = 2.710895e4
!	Coded		:	Ngoc Dang, July 1982
	
        double precision function rho (s, t, p)

	double precision s, t, p
	double precision a, b, c, d, e, k, t1, t2, t3, t4, ta, tb
	double precision k0, k0w, alpha0, alpha, rho0, aw, bw

      if(s .lt. 0.) write(0,*) 's,t,p = ',s,t,p
!  convert input units

	s = 1000.0d0 * s
	p = 10.0d0 * p

!  compute alpha0 = alpha (s, t, 0)

	t1 = 9.99841594d-1 + t * (6.793952d-5 + t * (-9.095290d-6 + t *   &
           (1.001685d-7 + t * (-1.120083d-9 + t * 6.536332d-12))))
	t2 = 8.25917d-4 + t * (-4.4490d-6 + t * (1.0485d-7 + t *      &
           (-1.258d-9 + t * 3.315d-12)))
	t3 = -6.33761d-6 + t * (2.8441d-7 + t * (-1.6871d-8 + t *     &
      	     2.83258d-10))
	t4 = 5.4705d-7 + t * (-1.97975d-8 + t * (1.6641d-9 - t *      &
           3.1203d-11))
	rho0 = t1 + t2 * s + t3 * s**(1.5) + t4 * s**2
	alpha0 = 1.0d0 / rho0

!  compute the pressure coefficient

!	the first term
	k0w = 1.965221d4 + t * (1.484206d2 + t * (-2.327105d0 + t *     &
            (1.360477d-2 - t * 5.155288d-5)))
	ta = 5.46746d1 + t * (-6.03459d-1 + t * (1.09987d-2 - t *       &
           6.167d-5))
	tb = 7.944d-2 + t * (1.6483d-2 - t * 5.3009d-4)
	k0 = k0w + ta * s + tb * s**(1.5)

!	the second term
	aw = 3.239908d0 + t * (1.43713d-3 + t * (1.16092d-4 - t *       &
          5.77905d-7))
	c = 2.2838d-3 + t * (-1.0981d-5 - t * 1.6078d-6)
	d = 1.9107d-4
	a = aw + c * s + d * s**(1.5)

!	the third term
	bw = 8.50935d-5 + t * (-6.12293d-6 + t * 5.2787d-8)
	e = -9.9348d-7 + t * (2.0816d-8 + t * 9.1697d-10)
	b = bw + e * s

	k = k0 + a * p + b * p**2

!  compute density
	alpha = alpha0 * (1.0d0 - p / k)
	rho = 1.0d0 / alpha

!  convert to kg / m**3
	rho = 1000.0d0 * rho
	s = s / 1000.0d0
	p = p / 10.0d0

	return

	end
