subroutine divergence
#define DEBUG_LEVEL 1
!     Routine to take the divergence of a three dimensional
!     flow field expressed with U and W contravariant components
!     and a normal v component.
!     
!     Though the grid is stretched in the y direction, we use v rather
!     than V for the velocity, which affects the form of
!     the last term in the divergence. 
!      
!     "div_u"(U_vector) == { (det_J*U)_xi + (det_J*W)_zeta }/{det_J} + (1/y_eta) * d/d eta (v) 
!  
!     where   det_J  = x_xi*z_zeta - x_zeta*z_xi i.e. the determinant of the Jacobian matrix
!

 use mpi_parameters
 use grid_info
 use dimensional_scales
 use intermediate_variables, U=>u_star, v=>v_star, W=>w_star
 use boundary_information
 use counters_flags_etc
 use mpi_parameters
 use integration_weights
 use immersed_boundary
 implicit none

 real                     :: h,end_values(2)
 real                     :: xx, mean_value
 real, allocatable        :: work(:,:,:,:)
 integer                  :: flag
 include 'mpif.h'
 include '../input/problem_size.h'
 allocate ( work(nx,ny,locnz,2) )
 work(:,:,:,:) = 0.


#if DEBUG_LEVEL >= 1
 if(myid==0) write(0,*) 'hello world from divergence'
#endif

!    modified source such that G~ dot n^ = 0 at top Marshall et al JGR 97  5738
     if(myid==numprocs-1) W(:,:,locnz)=0.


! form J*U and take xi derivative, store result in work(:,:,:,1)
 do j=1,ny
  work(:,j,:,1)=Grid(0)%det_J(:,:)*U(:,j,:)
 enddo

 flag=1    ! as default, don't prescribe end values
 end_values=-1e19 !dummy values
 h=Grid(0)%dxi
 call compact_ddx(work(1,1,1,1),work(1,1,1,2),flag,end_values,nx,ny,locnz)

! store  (det_J*U)_xi /{det_J} in div_u(:,:,:,new)
 do j=1,ny
  div_u(:,j,:,new)= work(:,j,:,2)/Grid(0)%det_J(:,:)
 enddo

! form J*W and take zeta derivative, store result in work(:,:,:,1)
 do j=1,ny
  work(:,j,:,1)=Grid(0)%det_J(:,:)*W(:,j,:)
 enddo

 h=Grid(0)%dzeta
 call compact_ddz_mpi(work(1,1,1,1),work(1,1,1,2),flag,end_values,nx,ny,locnz)

! add (det_J*W)_zeta }/{det_J} to ans
 do j=1,ny
  div_u(:,j,:,new)= div_u(:,j,:,new) +   &
                       work(:,j,:,2)/Grid(0)%det_J(:,:)
 enddo

! Now compute the v part
! take eta derivative of v, store result in work(:,:,:,1), use work(:,:,:,2)
! as scratch space
       
 flag=1    ! don't prescribe end values
 h=Grid(0)%deta
 call compact_ddy(v,work(1,1,1,1),flag,end_values,nx,ny,locnz)

! divide by y_eta to obtain the y derivative and add to divergence
 do j=1,ny
  div_u(:,j,:,new) = ( div_u(:,j,:,new)     &
                     + work(:,j,:,1)/Grid(0)%y_eta(j) )
 enddo


! calculate the rms divergence
 xx=sum( div_u(:,:,:,new)**2 )/float(nx*ny*locnz)   ! local rms value
 call MPI_ALLREDUCE(xx,mean_value,1,MPI_REAL,MPI_SUM,comm,ierr)
 div_rms = mean_value/(numprocs) * U_scale/t_scale   !    dimensional units

 if( myid == 0 ) then
  open(1,file='../output/div_rms',position='append')
   write(1,*) t*t_scale,div_rms
  close(1)
 endif

deallocate( work )      
end subroutine divergence
