!   file containing the various routines to 
!   interpolate/restrict data between grids

!*******************************************************
subroutine interp1d(yc,yf,nc,nf,xc,xf)
! data=y(x), c=coarse, f=fine, x=monotonic independent variable
! n=number of data points

 implicit none
 integer       :: nc, nf, nwrk, ierr, i
 real          :: yc(nc), yf(nf), xc(nc), xf(nf)
 real, allocatable, dimension(:)   :: wrk, dc, df

 nwrk=2*nc
 allocate( wrk(nwrk), dc(nc), df(nf) )

!find interpolation derivs for coarse grid function
! .FALSE. means Hermite cubic interpolation  ( smoother )
! .TRUE.  means cubic spline interpolation

  call pchez(nc,xc,yc,dc,.TRUE.,wrk,nwrk,ierr)

 if( ierr < 0 ) then
  write(0,*) '*** PROBLEM IN INTERP1D, PCHEZ',nc,nf
  STOP
 endif

!Evaluate interpolating polynomial at the fine grid points
   call pchev(nc, xc, yc, dc, nf, xf, yf, df, ierr)

 if( ierr < 0 ) then
  write(0,*) '*** PROBLEM IN INTERP1D, PCHEV',nc,nf
  STOP
 endif

 deallocate( dc, df, wrk )
end subroutine interp1d
!*******************************************************

!*******************************************************
subroutine restrict1d(yc,yf,nc,nf,xc,xf)
! data=y(x), c=coarse, f=fine, x=monotonic independent variable
! n=number of data points

 use mpi_parameters      ! myid,nprocs
 use grid_info, only: restriction_method
 implicit none
 integer       :: nc, nf, nwrk, ierr, i
 real          :: yc(nc), yf(nf), xc(nc), xf(nf), hc, hf
 real, allocatable, dimension(:)   :: wrk, dc, df
 real, allocatable, dimension(:)   :: yl, yr, xl, xr
 character(len=80)                    :: method

 method=restriction_method

 nwrk=2*nf
 allocate( wrk(nwrk), dc(nc), df(nf) )

!find interpolation derivs for fine grid function
! .FALSE. means Hermite cubic interpolation 
! .TRUE.  means dubic spline interpolation

 call pchez(nf,xf,yf,df,.TRUE.,wrk,nwrk,ierr)
 if( ierr < 0 ) then
  write(0,*) '*** PROBLEM IN RESTRICT1D, PCHEZ',ierr
  STOP
 endif


!Evaluate interpolating polynomial at each coarse grid point
 call pchev(nf, xf, yf, df, nc, xc, yc, dc, ierr)
 if( ierr < 0 ) then
  write(0,*) '*** PROBLEM IN RESTRICT1D, PCHEV, coarse pts',nc,nf
  STOP
 endif

 if( method == 'injection' ) then
  deallocate( dc, df, wrk  )
  return
 endif

! full weighting logic   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 allocate( yl(nc), yr(nc), xl(nc), xr(nc) )

!nearby neighbor points
 hf=xf(2)-xf(1)
 xl=xc-hf
 xr=xc+hf

!GET VALUES yr(1:nc)= values just to the right of yc(1:nc)
! N.B. values are extrapolated if xr is outside range of xc
 call pchev(nf, xf, yf, df, nc, xr, yr, dc, ierr)  
 if( ierr < 0 ) then
  write(0,*) '*** PROBLEM IN RESTRICT1D, PCHEV, right pts',nc,nf
  STOP
 endif


!GET VALUES yl(1:nc)= values just to the left of yc(1:nc)
! N.B. values are extrapolated if xl is outside range of xc
 call pchev(nf, xf, yf, df, nc, xl, yl, dc, ierr)
 if( ierr < 0 ) then
  write(0,*) '*** PROBLEM IN RESTRICT1D, PCHEV, left pts',nc,nf
  STOP
 endif

!  full weighting, store result in dc for now
 do i=2,nc-1
  dc(i) = ( yl(i) + 2.*yc(i) + yr(i) )/4.
 enddo
 
! LEFT END
 if( xc(1) == xf(1) ) then
  dc(1) = yf(1)                              ! injection at endpoints
 elseif( xc(1) > xf(1) ) then
  dc(1) = ( yl(1) + 2.*yc(i) + yr(i) )/4.    ! FW
 elseif( xc(1) < xf(1) ) then
  write(0,*) ' problem in restrict1d: xc(1) < xf(1) '
  stop
 endif

! RIGHT END
 if( xc(nc) == xf(nf) ) then
  dc(nc) = yf(nf)                                ! injection at endpoints
 elseif( xc(nc) < xf(nf) ) then
  dc(nc) = ( yl(nc) + 2.*yc(nc) + yr(nc) )/4.    ! FW
 elseif( xc(nc) > xf(nf) ) then
  write(0,*) ' problem in restrict1d: xc(nc) > xf(nf) '
  stop
 endif


! return values in yc
  yc = dc


 deallocate( dc, df, wrk, yl, yr, xl, xr )
end subroutine restrict1d
!*******************************************************


!*******************************************************
subroutine restrict(yf, yc, jfine)
! full weighting restriction of the data yf at level jfine
! to the array yc at level (jfine+1)
! uses the 1d routine restrict1d(coarse,fine,nc,nf,hc,hf)
!
 use mpi_parameters
 use grid_info
 implicit none
 include 'mpif.h'

 integer    :: jfine
 real       :: yf( Grid(jfine)%n1, Grid(jfine)%n2, Grid(jfine)%locn3 )
 real       :: yc( Grid(jfine+1)%n1, Grid(jfine+1)%n2, Grid(jfine+1)%locn3 )

 integer                              :: i,j,k,nf,nc,n,npts
 integer                              :: nf_z,k1,k2
 integer                              :: status(MPI_STATUS_SIZE),ierr,tag
 real,dimension(:),allocatable        :: coarse, fine, xf, xc
 real,dimension(:,:,:),allocatable    :: temp

 n=max( Grid(jfine)%n1, Grid(jfine)%n2, Grid(jfine)%locn3+2 )
 allocate( fine(0:n-1), xf(0:n-1) )

 n=max( Grid(jfine+1)%n1, Grid(jfine+1)%n2, Grid(jfine+1)%locn3 )
 allocate( coarse(n), xc(n) )

 allocate( temp(Grid(jfine)%n1, Grid(jfine)%n2, 0:Grid(jfine)%locn3+1) )  ! 2 extra locations

!store input fine grid data in an array with 2 extra z planes
 do k=1,Grid(jfine)%locn3
  temp(:,:,k)=yf(:,:,k)
 enddo
 temp(:,:,0)=-9999.                      ! initially set xtra planes to junk
 temp(:,:,Grid(jfine)%locn3+1)=-9999.

! set the number of fine grid data points to use
if( nprocs == 1 ) then
 nf_z = Grid(jfine)%locn3
 k1=1
 k2=nf_z                     ! don't need extended data
elseif( nprocs > 1) then
  nf_z = Grid(jfine)%locn3+2
  k1=0
  k2= Grid(jfine)%locn3+1     ! use extended data both North and South
 if( myid==0 ) then
  nf_z = Grid(jfine)%locn3+1
  k1=1
  k2= Grid(jfine)%locn3+1    ! use extended data only to North
 endif
 if( myid==nprocs-1 ) then
  nf_z = Grid(jfine)%locn3+1
  k1=0
  k2= Grid(jfine)%locn3      ! use extended data only to South
 endif
endif


! DATA EXCHANGES
if( nprocs > 1 ) then    !fill extra planes with data from neighboring processors
  npts=Grid(jfine)%n1 * Grid(jfine)%n2
  tag = 777
  if( myid > 0 ) then
!    exchange data with south processor
    call MPI_SENDRECV(yf(1,1,1),npts,MPI_REAL,myid-1,tag, &
         temp(1,1,0),npts,MPI_REAL,myid-1,tag,comm,status,ierr)
  endif

  if( myid < nprocs-1 ) then
!   exchange data with north processor
    call MPI_SENDRECV(yf(1,1,Grid(jfine)%locn3),npts,MPI_REAL,myid+1,tag, &
         temp(1,1,Grid(jfine)%locn3+1),npts,MPI_REAL,myid+1,tag,comm,status,ierr)
   endif
endif

!x direction 
 nf=Grid(jfine)%n1
 nc=Grid(jfine+1)%n1
 do i=1,nf
  xf(i)=(i-1.)*Grid(jfine)%dxi
 enddo
 do i=1,nc
  xc(i)=(i-1.)*Grid(jfine+1)%dxi
 enddo

 do k=k1,k2
  do j=1,Grid(jfine)%n2
   fine(1:nf)=temp(1:nf,j,k)
   call restrict1d(coarse,fine(1),nc,nf,xc,xf(1))  ! coarse <- restrict(fine)
   temp(1:nc,j,k)=coarse(1:nc)
  enddo
 enddo

! y direction 
 nf=Grid(jfine)%n2
 nc=Grid(jfine+1)%n2
 do i=1,nf
  xf(i)=(i-1.)*Grid(jfine)%deta
 enddo
 do i=1,nc
  xc(i)=(i-1.)*Grid(jfine+1)%deta
 enddo

 do k=k1,k2
  do i=1,Grid(jfine+1)%n1    ! already restricted to coarse grid
   fine(1:nf)=temp(i,1:nf,k)
   call restrict1d(coarse,fine(1),nc,nf,xc,xf(1))  ! coarse <- restrict(fine)
   temp(i,1:nc,k)=coarse(1:nc)
  enddo
 enddo

! z direction 
 nc=Grid(jfine+1)%locn3
 do i=k1,k2
  xf(i)=(myid*Grid(jfine)%locn3 + i - 1.)*Grid(jfine)%dzeta
 enddo
 do i=1,nc
  xc(i)=(myid*Grid(jfine+1)%locn3 + i - 1.)*Grid(jfine+1)%dzeta
 enddo

 do j=1,Grid(jfine+1)%n2    ! already restricted to coarse grid
  do i=1,Grid(jfine+1)%n1    ! already restricted to coarse grid
   fine(k1:k2)=temp(i,j,k1:k2)
   call restrict1d(coarse,fine(k1),nc,nf_z,xc,xf(k1))  ! coarse <- restrict(fine)
   temp(i,j,1:nc)=coarse(1:nc)
  enddo
 enddo

! return results in the coarse array yc
 do k=1,Grid(jfine+1)%locn3
  do j=1,Grid(jfine+1)%n2
   do i=1,Grid(jfine+1)%n1
    yc(i,j,k)=temp(i,j,k)
   enddo
  enddo
 enddo

 deallocate( coarse, fine, temp )
end subroutine restrict
!*******************************************************


!*******************************************************
subroutine interpolate(yc, yf, jfine)
! Cubic Hermite/spline interpolation of the data yc at level (jfine+1)
! to the array yf at level jfine
! uses the 1d routine interp1d(coarse,fine,nc,nf,hc,hf)
!
 use mpi_parameters
 use grid_info
 implicit none
 include 'mpif.h'

 integer    :: jfine
 real       :: yf( Grid(jfine)%n1, Grid(jfine)%n2, Grid(jfine)%locn3 )
 real       :: yc( Grid(jfine+1)%n1, Grid(jfine+1)%n2, Grid(jfine+1)%locn3 )

 integer                              :: i,j,k,nf,nc,n,npts
 integer                              :: nc_z,k1,k2
 integer                              :: status(MPI_STATUS_SIZE),ierr,tag
 real,dimension(:),allocatable        :: coarse, fine, xc, xf
 real,dimension(:,:,:),allocatable    :: temp

 n=max( Grid(jfine)%n1, Grid(jfine)%n2, Grid(jfine)%locn3 )
 allocate( fine(n), xf(n) )

 n=max( Grid(jfine+1)%n1, Grid(jfine+1)%n2, Grid(jfine+1)%locn3 + 2 )
 allocate( coarse(0:n-1), xc(0:n-1) )

 allocate( temp(Grid(jfine)%n1, Grid(jfine)%n2, 0:Grid(jfine)%locn3+1) )
 temp=0.0

! set the number of z coarse grid data points to use
if( nprocs == 1 ) then
 nc_z = Grid(jfine+1)%locn3
 k1=1
 k2=nc_z                     ! don't need extended data
elseif( nprocs > 1) then
  nc_z = Grid(jfine+1)%locn3+2
  k1=0
  k2= Grid(jfine+1)%locn3+1     ! use extended data both North and South
 if( myid==0 ) then
  nc_z = Grid(jfine+1)%locn3+1
  k1=1
  k2= Grid(jfine+1)%locn3+1    ! use extended data only to North
 endif
 if( myid==nprocs-1 ) then
  nc_z = Grid(jfine+1)%locn3+1
  k1=0
  k2= Grid(jfine+1)%locn3      ! use extended data only to South
 endif
endif

!put yc in temp(:,:,1:locn3), leaving room for north & south neighbor planes
do k=1,Grid(jfine+1)%locn3
 do j=1,Grid(jfine+1)%n2
  do i=1,Grid(jfine+1)%n1
   temp(i,j,k)=yc(i,j,k)
  enddo
 enddo
enddo
! initially set xtra planes to junk
temp(:,:,0)=-9999. 
temp(:,:,Grid(jfine+1)%locn3+1)=-9999.

! put valid coarse grid data in these planes if necessary
if( nprocs > 1 ) then
 npts=Grid(jfine+1)%n1 * Grid(jfine+1)%n2
 tag=777
 if( myid > 0 ) then
 ! exchange data with south processor
  call MPI_SENDRECV(yc(1,1,1),npts,MPI_REAL,myid-1,tag, &
        temp(1,1,0),npts,MPI_REAL,myid-1,tag,comm,status,ierr)
 endif

 if( myid < nprocs-1 ) then
 ! exchange data with north processor
  call MPI_SENDRECV(yc(1,1,Grid(jfine+1)%locn3),npts,MPI_REAL,myid+1,tag, &
       temp(1,1,Grid(jfine+1)%locn3+1),npts,MPI_REAL,myid+1,tag,comm,status,ierr)
 endif
endif

! x direction 
 nf=Grid(jfine)%n1
 nc=Grid(jfine+1)%n1
 do i=1,nf
  xf(i)=(i-1.)*Grid(jfine)%dxi
 enddo
 do i=1,nc
  xc(i)=(i-1.)*Grid(jfine+1)%dxi
 enddo

 do k=k1,k2
  do j=1,Grid(jfine+1)%n2
   coarse(1:nc)=temp(1:nc,j,k)
   call interp1d(coarse(1),fine,nc,nf,xc(1),xf)  ! fine <- interp(coarse)
   temp(1:nf,j,k)=fine(1:nf)
  enddo
 enddo

! y direction 
 nf=Grid(jfine)%n2
 nc=Grid(jfine+1)%n2
 do i=1,nf
  xf(i)=(i-1.)*Grid(jfine)%deta
 enddo
 do i=1,nc
  xc(i)=(i-1.)*Grid(jfine+1)%deta
 enddo

 do k=k1,k2
  do i=1,Grid(jfine)%n1
   coarse(1:nc)=temp(i,1:nc,k)
   call interp1d(coarse(1),fine,nc,nf,xc(1),xf)  ! fine <- interp(coarse)
   temp(i,1:nf,k)=fine(1:nf)
  enddo
 enddo

! z direction 
 nf=Grid(jfine)%locn3
 do i=1,nf
  xf(i)=(myid*Grid(jfine)%locn3 + i - 1.)*Grid(jfine)%dzeta
 enddo
 do i=k1,k2
  xc(i)=(myid*Grid(jfine+1)%locn3 + i - 1.)*Grid(jfine+1)%dzeta
 enddo

 do j=1,Grid(jfine)%n2
  do i=1,Grid(jfine)%n1
   coarse(k1:k2)=temp(i,j,k1:k2)
   call interp1d(coarse(k1),fine,nc_z,nf,xc(k1),xf)  ! fine <- interp(coarse)
   temp(i,j,1:nf)=fine(1:nf)
  enddo
 enddo

! return results in the fine array yf
 yf=temp(:,:,1:nf)

 deallocate( coarse, fine, temp )
end subroutine interpolate
!*******************************************************

!*******************************************************

!*******************************************************
subroutine restrictXZ(yf, yc, jfine)
! full weighting restriction of the XZ data yf at level jfine
! to the array yc at level (jfine+1)
! uses the 1d routine restrict1d(coarse,fine,nc,nf,hc,hf)
!
 use mpi_parameters
 use grid_info
 implicit none
 include 'mpif.h'

 integer    :: jfine
 real       :: yf( Grid(jfine)%n1, Grid(jfine)%locn3 )
 real       :: yc( Grid(jfine+1)%n1, Grid(jfine+1)%locn3 )

 integer                              :: i,j,k,nf,nc,n,npts
 integer                              :: nf_z,k1,k2
 integer                              :: status(MPI_STATUS_SIZE),ierr,tag
 real,dimension(:),allocatable        :: coarse, fine, xf, xc
 real,dimension(:,:),allocatable      :: temp

 n=max( Grid(jfine)%n1, Grid(jfine)%locn3+2 )
 allocate( fine(0:n-1), xf(0:n-1) )

 n=max( Grid(jfine+1)%n1, Grid(jfine+1)%locn3 )
 allocate( coarse(n), xc(n) )

 allocate( temp(Grid(jfine)%n1, 0:Grid(jfine)%locn3+1) )  ! 2 extra locations

!store input fine grid data in an array with 2 extra z planes
 do k=1,Grid(jfine)%locn3
  temp(:,k)=yf(:,k)
 enddo
 temp(:,0)=-9999.                      ! initially set xtra planes to junk
 temp(:,Grid(jfine)%locn3+1)=-9999.

! set the number of fine grid data points to use
if( nprocs == 1 ) then
 nf_z = Grid(jfine)%locn3
 k1=1
 k2=nf_z                     ! don't need extended data
elseif( nprocs > 1) then
  nf_z = Grid(jfine)%locn3+2
  k1=0
  k2= Grid(jfine)%locn3+1     ! use extended data both North and South
 if( myid==0 ) then
  nf_z = Grid(jfine)%locn3+1
  k1=1
  k2= Grid(jfine)%locn3+1    ! use extended data only to North
 endif
 if( myid==nprocs-1 ) then
  nf_z = Grid(jfine)%locn3+1
  k1=0
  k2= Grid(jfine)%locn3      ! use extended data only to South
 endif
endif


! DATA EXCHANGES
 if( nprocs > 1 ) then    !fill extra planes with data from neighboring processors
  npts=Grid(jfine)%n1 * 1
  tag=777
  if( myid > 0 ) then
!   exchange data with south processor
    call MPI_SENDRECV(yf(1,1),npts,MPI_REAL,myid-1,tag, &
         temp(1,0),npts,MPI_REAL,myid-1,tag,comm,status,ierr)
  endif

  if( myid < nprocs-1 ) then
!   exchange data with north processor
    call MPI_SENDRECV(yf(1,Grid(jfine)%locn3),npts,MPI_REAL,myid+1,tag, &
         temp(1,Grid(jfine)%locn3+1),npts,MPI_REAL,myid+1,tag,comm,status,ierr)
   endif
 endif

!x direction 
 nf=Grid(jfine)%n1
 nc=Grid(jfine+1)%n1
 do i=1,nf
  xf(i)=(i-1.)*Grid(jfine)%dxi
 enddo
 do i=1,nc
  xc(i)=(i-1.)*Grid(jfine+1)%dxi
 enddo

 do k=k1,k2
   fine(1:nf)=temp(1:nf,k)
   call restrict1d(coarse,fine(1),nc,nf,xc,xf(1))  ! coarse <- restrict(fine)
   temp(1:nc,k)=coarse(1:nc)
 enddo

! z direction 
 nc=Grid(jfine+1)%locn3
 do i=k1,k2
  xf(i)=(myid*Grid(jfine)%locn3 + i - 1.)*Grid(jfine)%dzeta
 enddo
 do i=1,nc
  xc(i)=(myid*Grid(jfine+1)%locn3 + i - 1.)*Grid(jfine+1)%dzeta
 enddo

  do i=1,Grid(jfine+1)%n1    ! already restricted to coarse grid
   fine(k1:k2)=temp(i,k1:k2)
   call restrict1d(coarse,fine(k1),nc,nf_z,xc,xf(k1))  ! coarse <- restrict(fine)
   temp(i,1:nc)=coarse(1:nc)
  enddo

! return results in the coarse array yc
 do k=1,Grid(jfine+1)%locn3
   do i=1,Grid(jfine+1)%n1
    yc(i,k)=temp(i,k)
   enddo
 enddo

 deallocate( coarse, fine, temp, xc, xf )
end subroutine restrictXZ
!*******************************************************

!*******************************************************

!*******************************************************
subroutine restrictY(yf, yc, jfine)
! full weighting restriction of the Y data yf at level jfine
! to the array yc at level (jfine+1)
! uses the 1d routine restrict1d(coarse,fine,nc,nf,hc,hf)
!
 use mpi_parameters
 use grid_info
 implicit none
 include 'mpif.h'

 integer    :: jfine
 real       :: yf( Grid(jfine)%n2 )
 real       :: yc( Grid(jfine+1)%n2 )

 integer                              :: i,j,k,nf,nc,n,npts
 real,dimension(:),allocatable        :: coarse, fine, xf, xc

 n=Grid(jfine)%n2
 allocate( fine(n), xf(n) )

 n=Grid(jfine+1)%n2
 allocate( coarse(n), xc(n) )

!y direction 
 nf=Grid(jfine)%n2
 nc=Grid(jfine+1)%n2
 do i=1,nf
  xf(i)=(i-1.)*Grid(jfine)%deta
 enddo
 do i=1,nc
  xc(i)=(i-1.)*Grid(jfine+1)%deta
 enddo

 fine(1:nf)=yf(1:nf)
 call restrict1d(coarse,fine,nc,nf,xc,xf)  ! coarse <- restrict(fine)
 yc(1:nc)=coarse(1:nc)

 deallocate( coarse, fine, xf, xc )
end subroutine restrictY
!*******************************************************
