subroutine hydrostatic
#define DEBUG_LEVEL 1
!     Routine to calculate the hydrostatic pressure
!     associated with the density fluctuations pd about
!     a time independent reference density rho_bar(:,:)

use mpi_parameters                   !! myid,numprocs,comm etc
use pde_parameters                   !! Ri
use grid_info                        !! z_zeta(i,k)
use dependent_variables              !! pd, pressure
use counters_flags_etc               !! i,j,k etc
implicit none
include '../input/problem_size.h'    !! nx,ny,locnz
include 'mpif.h'

real                                 :: h                      ! step size dzeta
real, allocatable                    :: d_even(:,:),d_odd(:,:) ! correction terms
integer                              :: p_id                   ! pressure storage id
integer                              :: id,root

#if DEBUG_LEVEL >= 1
 if(myid==0) write(0,*) 'hello world from hydrostatic'
#endif

  allocate( d_even(nx,ny), d_odd(nx,ny) )
  h=Grid(0)%dzeta
  p_id=2            ! store data in 2nd block

!  step 1: Each processor computes that part of the hydrostatic pressure
!          that is independent of data stored in processors above.

 do j=1,ny

!   top 2 points
  if( myid == numprocs-1 ) then    ! no corrections to be added later
   pressure(:,j,locnz,p_id) = 0.0
   pressure(:,j,locnz-1,p_id) = pressure(:,j,locnz,p_id)                                &
                              + (h*Ri/2.)*( pd(:,j,locnz)*Grid(0)%z_zeta(:,locnz)       &
                                           +pd(:,j,locnz-1)*Grid(0)%z_zeta(:,locnz-1) )
  else  ! corrections from above will be added later
   pressure(:,j,locnz,p_id) = (h*Ri/3.)*pd(:,j,locnz)*Grid(0)%z_zeta(:,locnz)
   pressure(:,j,locnz-1,p_id) = (h*Ri/3.)*(4.*pd(:,j,locnz)*Grid(0)%z_zeta(:,locnz)      & 
                                        +  pd(:,j,locnz-1)*Grid(0)%z_zeta(:,locnz-1) )
  endif

!   compute values for the rest of the points, working downward
!   use 4th order, 3pt quadrature formula
  do k=locnz-2,1,-1
   pressure(:,j,k,p_id) = pressure(:,j,k+2,p_id) + (h*Ri/3.)*    &
    ( pd(:,j,k+2)*Grid(0)%z_zeta(:,k+2) +4.*pd(:,j,k+1)*Grid(0)%z_zeta(:,k+1) + pd(:,j,k)*Grid(0)%z_zeta(:,k) )
  enddo

 enddo    ! end loop over index j

!   step 2: Top processor kicks off the correction sweep by calculating the 2 correction
!           terms needed by its southern neighbor.

  if( myid == numprocs-1 .and. numprocs > 1 ) then
   do j=1,ny
    d_even(:,j) = pressure(:,j,2,p_id) + (h*Ri/3.)*(pd(:,j,2)*Grid(0)%z_zeta(:,2)      &
                                         +  4.*pd(:,j,1)*Grid(0)%z_zeta(:,1) )
    d_odd(:,j)  = pressure(:,j,1,p_id) + (h*Ri/3.)*(pd(:,j,1)*Grid(0)%z_zeta(:,1) )
   enddo
  endif
 
!  BCAST data
   root=numprocs-1
   call MPI_BCAST(d_even,nx*ny,MPI_REAL,root,comm,ierr)
   call MPI_BCAST(d_odd ,nx*ny,MPI_REAL,root,comm,ierr)

!   step 3: Working sequentially downward, each processor receives data from above,
!           corrects the locally computed pressure, calculates the correction terms
!           needed below and issues non-blocking sends to its south neighbor (if it exists)

 if( numprocs > 1 ) then
  do id=numprocs-2,0,-1      ! start w/ second from top, work downward
   if( myid == id ) then
    ! correct local pressure values
    do k=locnz,2,-2
      pressure(:,:,k,p_id)=pressure(:,:,k,p_id)+d_even(:,:)
    enddo
    do k=locnz-1,1,-2
      pressure(:,:,k,p_id)=pressure(:,:,k,p_id)+d_odd(:,:)
    enddo

    !calculate corrections required  by south neighbor
    do j=1,ny
     d_even(:,j) = pressure(:,j,2,p_id) + (h*Ri/3.)*(pd(:,j,2)*Grid(0)%z_zeta(:,2)      &
                                        +  4.*pd(:,j,1)*Grid(0)%z_zeta(:,1) )
     d_odd(:,j) = pressure(:,j,1,p_id) + (h*Ri/3.)*(pd(:,j,1)*Grid(0)%z_zeta(:,1) )
    enddo

   elseif(myid .ne. id) then
    ! its not my turn to calculate things
   endif

   !BCAST new corrections
   root=id
   call MPI_BCAST(d_even,nx*ny,MPI_REAL,root,comm,ierr)
   call MPI_BCAST(d_odd ,nx*ny,MPI_REAL,root,comm,ierr)

  enddo   ! continue down to next processor until finished
 endif

! hydrostatic pressure associated with pd now stored in pressure(:,:,:,p_id)
  deallocate( d_even, d_odd )

end subroutine hydrostatic
