recursive subroutine V_cycle(h)	!%%%%%%%%%%%%%%%%%%%%%%%%%%%
!   Do a V cycle starting/ending at grid level h.
!   Assume that the rhs vector f is stored in Grid(h)%f
!   and that the initial guess v is stored in Grid(h)%v,
!   both of which are available via the module grid_info.
!   On output, the solution vector Grid(h)%v will be updated.
 				!see pp.40-41 of A multigrid tutorial, Briggs et al.

use grid_info 			! finest grid level = 0, coarset grid level = (nlevels-1)
use counters_flags_etc
use mpi_parameters
implicit none

integer		      :: h, nu_1, nu_2

nu_1 = V_sweeps_pre	        !  number of  (pre) relaxation sweeps
nu_2 = V_sweeps_post	        !  number of  (post)relaxation sweeps
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

current_grid_level = h  ! matvec needs to know this value, share via module grid_info
if(myid==0) then
 !write(0,*)'    Starting V cycle at level ',h
endif

!(1) Relax nu_1 times on A u = f  @ level h with initial guess Grid(h)%v
     call relax( Grid(h)%v, Grid(h)%f, nu_1,  h )     ! overwrite v, rhs in f

!(2) If h is coarsest grid, then go to step 4
     if( current_grid_level == Grid(0)%n_mg_levels-1  ) goto 4

    !calculate A v  at level h, scratch <-- Av
     call matvec( Grid(h)%v , Grid(h)%scratch, Grid(h)%ipar )   

    ! use result to calculate f - A v at level h, overwrite scratch with the result
     Grid(h)%scratch = -1*Grid(h)%scratch + Grid(h)%f

    ! Restrict  ( f - A v ) at level h  to the next coarser grid (h+1), store in f
     call restrict( Grid(h)%scratch, Grid(h+1)%f,  h )

    ! set initial guess to zero at level (h+1)
     Grid(h+1)%v = 0.0

    !update Grid(h+1)%v by performing a V-cycle
     call V_cycle(h+1)

!(3) Correct Grid(h)%v  Grid(h)%v + interpolate( Grid(h+1)%v )
     call interpolate( Grid(h+1)%v, Grid(h)%scratch, h)
     Grid(h)%v = Grid(h)%v + Grid(h)%scratch


!(4) Relax nu_2 times on A u = f  @ level h with initial guess Grid(h)%v
4    continue
     current_grid_level = h  ! matvec needs to know this value, share via module grid_info
     call relax( Grid(h)%v, Grid(h)%f, nu_2,  h )

end subroutine V_cycle 


recursive subroutine FMG(h)   ! v <- FMG(f) ; level h
!   Do a full multigrid V-cycle starting/ending at grid level h.
!   Assume that the rhs vector f is stored in Grid(h)%f,
!   which is available via the module grid_info.
!   On output, the solution vector Grid(h)%v will be updated.
 				!see pp.43 of A multigrid tutorial, Briggs et al.

use grid_info 			! finest grid level = 0, coarset grid level = (nlevels  1)
use mpi_parameters, only:myid
use counters_flags_etc, nu_0=>n_mg_cycles
implicit none

integer			            :: h
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

current_grid_level=h
if(myid==0) then
 write(0,*)'  Starting FMG cycle at level ',h
 !write(0,*)'  coarsest level ',Grid(0)%n_mg_levels-1
endif

!(1) Check to see if h is coarsest grid
if( h == Grid(0)%n_mg_levels - 1 ) then
 Grid(h)%v = 0.0
 goto 3
else
 call restrict(Grid(h)%f, Grid(h+1)%f, h ) ! f at level (h+1)  restrict( f at level h )
 call FMG(h+1)                             ! v at level (h+1)  FMG( f at level (h+1) )
endif

!(2) Correct v at level h  interpolate( v at level h+1 )
 call interpolate( Grid(h+1)%v, Grid(h)%v, h)

!(3) do nu_0  level h V-cycles 
3 continue
 do i=1,nu_0
   call V_cycle(h)     ! v  Vcycle(v,f), level (h), nu_0 times
 enddo

 current_grid_level = h  
end subroutine FMG


subroutine relax(v, f, n, h)	! relax at level h, for n iterations
				! i.e. iterate n times on A v = f  (A<->matvec)
				! overwrite result onto v
!grid _info contains Grid(0:nlevels-1),  current_grid_level
use grid_info
use mpi_parameters
use counters_flags_etc
implicit none

!Local variables
real                             :: v(1), f(1) !only 1st location address referenced
integer                          :: n, h
real, dimension(:), allocatable  :: work

external matvec, diagl, mg_solve, pdumr, pssum, psnrm2, progress

if( current_grid_level .ne. h ) then
 write(0,*) 'Routine relax: problem with current_grid_level/h'
 write(0,*) current_grid_level,h
 stop
endif

 Grid(h)%ipar(10)=n       ! set the number of iterations
 Grid(h)%ipar(5)=n        ! set the number of basis vectors to match
 Grid(h)%spar(1)=1.e-29   ! impossibly small, force specified number of relaxation sweeps

 if( relax_precon == 'diagl' ) then
  Grid(h)%ipar(8)=1 
 elseif( relax_precon == 'mg' ) then
  Grid(h)%ipar(8)=1 
 elseif( relax_precon == 'none' ) then
  Grid(h)%ipar(8)=0 
 endif

! Allocate work vector needed by particular method to be used
 if( relaxation_method == 'rgmres' ) then
  allocate( work( (4+Grid(h)%ipar(5))*Grid(h)%ipar(4) ) )  

    if( relax_precon == 'diagl' ) then
     call PIMSRGMRES(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, diagl, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'none' ) then
     call PIMSRGMRES(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, pdumr, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'mg' ) then
     call PIMSRGMRES(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, mg_solve, pdumr, pssum, psnrm2, progress)
    else
     write(0,*) ' relax_precon string not valid in routine relax, multigrid.F90 '
    endif

 elseif( relaxation_method == 'rbicgstab' ) then
  allocate( work( (6+2*Grid(h)%ipar(5))*Grid(h)%ipar(4) ) ) 

    if( relax_precon == 'diagl' ) then
     call PIMSRBICGSTAB(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, diagl, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'none' ) then
     call PIMSRBICGSTAB(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, pdumr, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'mg' ) then
     call PIMSRBICGSTAB(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, mg_solve, pdumr, pssum, psnrm2, progress)
    else
     write(0,*) ' relax_precon string not valid in routine relax, multigrid.F90 '
    endif

 elseif( relaxation_method == 'bicgstab' ) then
  allocate( work( 10*Grid(h)%ipar(4) ) )  

    if( relax_precon == 'diagl' ) then
     call PIMSBICGSTAB(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, diagl, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'none' ) then
     call PIMSBICGSTAB(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, pdumr, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'mg' ) then
     call PIMSBICGSTAB(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, mg_solve, pdumr, pssum, psnrm2, progress)
    else
     write(0,*) ' relax_precon string not valid in routine relax, multigrid.F90 '
    endif

 elseif( relaxation_method == 'cg' ) then
  allocate( work( 6*Grid(h)%ipar(4) ) )  

    if( relax_precon == 'diagl' ) then
     call PIMSCG(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, diagl, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'none' ) then
     call PIMSCG(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, pdumr, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'mg' ) then
     call PIMSCG(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, mg_solve, pdumr, pssum, psnrm2, progress)
    else
     write(0,*) ' relax_precon string not valid in routine relax, multigrid.F90 '
    endif

 elseif( relaxation_method == 'cgs' ) then
  allocate( work( 10*Grid(h)%ipar(4) ) )  

    if( relax_precon == 'diagl' ) then
     call PIMSCGS(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, diagl, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'none' ) then
     call PIMSCGS(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, pdumr, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'mg' ) then
     call PIMSCGS(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, mg_solve, pdumr, pssum, psnrm2, progress)
    else
     write(0,*) ' relax_precon string not valid in routine relax, multigrid.F90 '
    endif

 elseif( relaxation_method == 'tfqmr' ) then
  allocate( work( 12*Grid(h)%ipar(4) ) )  

    if( relax_precon == 'diagl' ) then
     call PIMSTFQMR(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, diagl, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'none' ) then
     call PIMSTFQMR(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, pdumr, pdumr, pssum, psnrm2, progress)
    elseif( relax_precon == 'mg' ) then
     call PIMSTFQMR(v, f, work, Grid(h)%ipar, Grid(h)%spar,   &
                     matvec, mg_solve, pdumr, pssum, psnrm2, progress)
    else
     write(0,*) ' relax_precon string not valid in routine relax, multigrid.F90 '
    endif

 endif


if( Grid(h)%ipar(11) < -1 ) then
 if( myid ==0 ) write(0,*) 'problem in PIM relaxation, grid level ',h
 stop
endif

deallocate( work )
end subroutine relax

