subroutine setup_compact_schemes(h)
! Routine to create & factor the various differentiation matrices
! corresponding to the optimized 6th order compact difference 
! operators from Lui & Lele 2001

#define DEBUG_LEVEL 1

use grid_info
use mpi_parameters
implicit none

integer          :: h     ! grid level

#if DEBUG_LEVEL >= 1
  if(myid==0) then
    write(0,*)   'hello world from setup_compact_schemes'
    write(0,*) '       initializing for grid level: ',h
   endif
#endif

! set parameters of the 6th order scheme described in
! the Lui & Lele 2001 conference paper
  alpha_1 = 0.5381301
  beta_1  = 0.0666332
  a1=1.3675777
  b1=0.8234282
  c1=0.0185208
  alpha=0.475     ! filtering parameter

! spectral like scheme of Lele 1992
! formally 4th order but supposedly superior resolution characteristics
! alpha_1 = 0.5771439
! beta_1  = 0.0896406
! a1=1.3025166
! b1=0.9935500
! c1=0.03750245
! alpha=0.475

  if( Grid(h)%n1 < 4 .or. Grid(h)%n2 < 4 ) then
   write(0,*) 'skipping setup compact schemes at level ',h
   return
  endif

! 1st derivative operators
   call set_deriv1(Grid(h)%n1,Grid(h)%ddx,Grid(h)%piv_dx)
   call set_deriv1(Grid(h)%n2,Grid(h)%ddy,Grid(h)%piv_dy)
   call set_deriv1(Grid(h)%n3,Grid(h)%ddz,Grid(h)%piv_dz)
      if(myid == 0) then
      write(0,*) '********** h = ',h
      write(0,*) '********** %n3 = ',Grid(h)%n3
      write(0,*) '********** %ddz = ',Grid(h)%ddz
      write(0,*) '********** %piv_dz = ',Grid(h)%piv_dz
      endif

! compact filtering operators
   call set_filter4(Grid(h)%n1,Grid(h)%filt_x,Grid(h)%piv_fx,alpha)
   call set_filter4(Grid(h)%n2,Grid(h)%filt_y,Grid(h)%piv_fy,alpha)
   call set_filter4(Grid(h)%n3,Grid(h)%filt_z,Grid(h)%piv_fz,alpha)
	
end subroutine setup_compact_schemes

!************************************************************************
!************************************************************************
subroutine set_deriv1(nx,AB,ipiv)	
  use grid_info, only      :  alpha_1, beta_1
  use user_parameters, only:  LAPACK_FLAG
  implicit none
  integer              :: nx,i,j,ipiv(nx),ierr
!!!  double precision     :: AB(7,nx)
  real     :: AB(7,nx)

   AB=-9999.


! 1st 2 rows of banded matrix array AB are reserved for fill in
! 1st/last k elements for the kth super/sub diagonal are not accessed
! superdiagonal 2
 AB(3,3:5)=0.0
 AB(3,6:nx-1)=beta_1
 AB(3,nx)=0.0
! superdiagonal 1
 AB(4,2)= 0.   
 AB(4,3)=                      0*1./2.   ! make second eqn explicit FD too
 AB(4,4)=1./3.
 AB(4,5:nx-2)=alpha_1
 AB(4,nx-1)=1./3.
 AB(4,nx)=                     0*1./6.   ! make second eqn explicit FD too
! diagonal 
 AB(5,1:nx)=1.0
! subdiagonal 1
 do i=1,nx
  AB(6,i)=AB(4,nx-i+1)
 enddo
! subdiagonal 2
 do i=1,nx
  AB(7,i)=AB(3,nx-i+1)
 enddo

! factor the banded matrix AB using LAPACK routine SGBTF2
! AB is overwritten with its LU factorization

  if( LAPACK_FLAG == 'double' ) then
!!!   call DGBTRF(nx,nx,2,2,AB,7,ipiv,ierr)
      write(0,*) 'LAPACK_FLAG = double instead of single'
      stop
  elseif( LAPACK_FLAG == 'single' ) then
   call SGBTRF(nx,nx,2,2,AB,7,ipiv,ierr)
  endif

! check error condition
  if( ierr .ne. 0 ) then
   write(0,*) ' factorization error ',ierr, ' size: ',nx
   STOP
  endif

end subroutine set_deriv1

!************************************************************************
!************************************************************************
subroutine set_filter4(nx,AB,ipiv,alpha)	
  use user_parameters, only:  LAPACK_FLAG
  implicit none
  integer              :: nx,i,j,ipiv(nx),ierr
!!!  double precision     :: AB(4,nx)
  real     :: AB(4,nx)
  real                 :: alpha

  AB=-9999.

! 1st row of banded matrix array AB is reserved for fill in
! 1st/last k elements for the kth super/sub diagonal are not accessed
! superdiagonal 1
 AB(2,2:3)=0.0
 AB(2,4:nx-1)=alpha
 AB(2,nx)=0.0
! diagonal 
 AB(3,1:nx)=1.0
! subdiagonal 1
 do i=1,nx
  AB(4,i)=AB(2,nx-i+1)
 enddo

! factor the banded matrix AB using LAPACK routine SGBTF2
! AB is overwritten with its LU factorization

  if( LAPACK_FLAG == 'double' ) then
!!!   call DGBTRF(nx,nx,1,1,AB,4,ipiv,ierr)
      write(0,*) 'LAPACK_FLAG = double instead of single'
      stop
  elseif( LAPACK_FLAG == 'single' ) then
   call SGBTRF(nx,nx,1,1,AB,4,ipiv,ierr)
  endif

! check error condition
  if( ierr .ne. 0 ) then
   write(0,*) ' factorization error in filter matrix',ierr, ' size: ',nx
   STOP
  endif

end subroutine set_filter4
