
module specs_and_plans
!! structure used to define fftw3 transforms
 TYPE fft_specs_3d
  character (len=6)   :: precision       ! single or double
  character (len=3)   :: in_place        ! yes or no, i.e. in- or out-of-place transform
  character (len=11)  :: transform_type  ! type of transform; sin,cos,fourier
  integer             :: trans_dims(3)   ! 1 value for each transform direction
                                         ! 1=forward  -1=inverse  0=no transform
  character (len=11)  :: reality_in      ! type of storage array: real,complex,halfcomplex
  character (len=11)  :: reality_out     ! type of output array: real,complex,halfcomplex
                                         ! (half complex is real w/ a special structure)
  integer             :: storage_dims(3) ! size storage array in words of type matching
                                         ! the input array
  integer             :: data_dims(3)    ! size data field in words of type matching
                                         ! the input array
  double precision    :: normval         ! normalization factor to multiply by after inverse
 END TYPE fft_specs_3d
 type(fft_specs_3d)    ::  fft_spec(-1:1,-1:1,-2:2)
 integer*8             ::  fft_plan(-1:1,-1:1,-2:2)
end module specs_and_plans

      subroutine define_transforms
      
      use specs_and_plans
      use flags_counters_etc 
      
      implicit none      
      
      integer             :: i1,i2,i3       ! flags for transform type in each dimension  

!==============================================================      
!    Create specs for a 1d forward FFT in x of data 
!    on a single horizontal plane of a 3d data set.
!    The x dimension is treated as real --> 1/2 complex. 
!    All transforms out-of-place.
      i1 = 1
      i2 = 0
      i3 = 0
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'fourier'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=1   ! forward in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=0   ! nothing in y
      fft_spec(i1,i2,i3)%trans_dims(3)=0   ! nothing in z
      
      fft_spec(i1,i2,i3)%reality_in='real'
      fft_spec(i1,i2,i3)%reality_out='halfcomplex'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=nx+2
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=locnz
      
      fft_spec(i1,i2,i3)%data_dims(1)=nx
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=locnz      
      fft_spec(i1,i2,i3)%normval=1.d0/nx
      
      call CreatePlan(i1,i2,i3)



!==============================================================      
!    Create specs for a 1d inverse FFT in x of data 
!    on a single horizontal plane of a 3d data set.
!    The x dimension is treated as real --> 1/2 complex. 
!    All transforms out-of-place.
      i1 = -1
      i2 = 0
      i3 = 0
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'fourier'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=-1  ! inverse in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=0   ! nothing in y
      fft_spec(i1,i2,i3)%trans_dims(3)=0   ! nothing in z
      
      fft_spec(i1,i2,i3)%reality_in='halfcomplex'
      fft_spec(i1,i2,i3)%reality_out='real'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=nx+2
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=locnz
      
      fft_spec(i1,i2,i3)%data_dims(1)=nx
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=locnz
      fft_spec(i1,i2,i3)%normval=1.d0/nx
      
      call CreatePlan(i1,i2,i3)


!==============================================================      
!    Create specs for a 1d forward FFT in y of data 
!    on a single horizontal plane of a 3d data set.
!    The y dimension is treated as complex --> complex,
!    (nx+2)/2+1 transforms of length ny.
!    All transforms out-of-place.
      i1 = 0
      i2 = 1
      i3 = 0
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'fourier'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=0   ! nothing in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=1   ! forward in y
      fft_spec(i1,i2,i3)%trans_dims(3)=0   ! nothing in z
      
      fft_spec(i1,i2,i3)%reality_in='complex'
      fft_spec(i1,i2,i3)%reality_out='complex'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=nx/2+1
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=locnz
      
      fft_spec(i1,i2,i3)%data_dims(1)=nx/2+1
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=locnz      
      fft_spec(i1,i2,i3)%normval=1.d0/ny
      
      call CreatePlan(i1,i2,i3)

!==============================================================      
!    Create specs for a 1d inverse FFT in y of data 
!    on a single horizontal plane of a 3d data set.
!    The y dimension is treated as complex --> complex,
!    (nx+2)/2+1 transforms of length ny.
!    All transforms out-of-place.
      i1 = 0
      i2 = -1
      i3 = 0
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'fourier'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=0   ! nothing in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=-1  ! inverse in y
      fft_spec(i1,i2,i3)%trans_dims(3)=0   ! nothing in z
      
      fft_spec(i1,i2,i3)%reality_in='complex'
      fft_spec(i1,i2,i3)%reality_out='complex'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=nx/2+1
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=locnz
      
      fft_spec(i1,i2,i3)%data_dims(1)=nx/2+1
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=locnz
      fft_spec(i1,i2,i3)%normval=1.d0/ny
      
      call CreatePlan(i1,i2,i3)
      

     if( bc_flag == 'zslip' ) then
!==============================================================      
!    Create specs for a 1d cos transform in z of data 
!    on a single xz plane of a 3d data set.
!    All data is treated as real.
!    (nx+2) transforms of data with length nz+1.
!    All transforms out-of-place.
      i1 = 0
      i2 = 0       
      i3 = -1 !   ---->  (-1 --> cos)  (+1 --> sin)  (2/-2 --> for/inv fourier)
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'cos'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=0   ! nothing in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=0   ! nothing in y
      fft_spec(i1,i2,i3)%trans_dims(3)=-1  ! cos in z
      
      fft_spec(i1,i2,i3)%reality_in='real'
      fft_spec(i1,i2,i3)%reality_out='real'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=locnx*2
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=n3
      
      fft_spec(i1,i2,i3)%data_dims(1)=locnx*2
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=n3
      fft_spec(i1,i2,i3)%normval=1.d0/(2.*nz)
      
      call CreatePlan(i1,i2,i3)
      
!==============================================================      
!    Create specs for a 1d sin transform in z of data 
!    on a single xz plane of a 3d data set.
!    All data is treated as real.
!    (nx+2) transforms of data with length nz+1.
!    All transforms out-of-place.
      i1 = 0
      i2 = 0       
      i3 = 1    !   ---->  (-1 --> cos)  (+1 --> sin)  (2/-2 --> for/inv fourier)
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'sin'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=0   ! nothing in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=0   ! nothing in y
      fft_spec(i1,i2,i3)%trans_dims(3)=1   ! sin in z        
      
      fft_spec(i1,i2,i3)%reality_in='real'
      fft_spec(i1,i2,i3)%reality_out='real'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=locnx*2   
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=n3-2
      
      fft_spec(i1,i2,i3)%data_dims(1)=locnx*2
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=n3-2       ! exclude top/bottom zeros for sin    
      fft_spec(i1,i2,i3)%normval=1.d0/(2.*nz)
      
      call CreatePlan(i1,i2,i3)
      
      
      
     elseif( bc_flag == 'zperiodic' ) then
!==============================================================      
!    Create specs for a 1d fourier transform in z of 
!    data on a single xz plane of a 3d data set.
!    All data is treated as complex.
!    (locnx) transforms of data with length nz.
!    All transforms out-of-place.
      i1 = 0
      i2 = 0       
      i3 = 2    !   ---->  (-1 --> cos)  (+1 --> sin)  (2/-2 --> for/inv fourier)
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'fourier'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=0   ! nothing in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=0   ! nothing in y
      fft_spec(i1,i2,i3)%trans_dims(3)=2   ! forward fourier in z
      
      fft_spec(i1,i2,i3)%reality_in='complex'
      fft_spec(i1,i2,i3)%reality_out='complex'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=locnx  ! locnx cmplx words
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=n3
      
      fft_spec(i1,i2,i3)%data_dims(1)=locnx
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=n3
      fft_spec(i1,i2,i3)%normval=1.d0/(n3)
      
      call CreatePlan(i1,i2,i3)
      
      
      
 !==============================================================      
!    Create specs for a 1d inverse fourier transform in z of 
!    data on a single xz plane of a 3d data set.
!    All data is treated as complex.
!    (locnx) transforms of data with length nz.
!    All transforms out-of-place.
      i1 = 0
      i2 = 0       
      i3 =-2    !   ---->  (-1 --> cos)  (+1 --> sin)  (2/-2 --> for/inv fourier)
!==============================================================
      fft_spec(i1,i2,i3)%precision = 'double'
      fft_spec(i1,i2,i3)%in_place  = 'no'
      fft_spec(i1,i2,i3)%transform_type = 'fourier'
      
      fft_spec(i1,i2,i3)%trans_dims(1)=0   ! nothing in x  
      fft_spec(i1,i2,i3)%trans_dims(2)=0   ! nothing in y
      fft_spec(i1,i2,i3)%trans_dims(3)=-2  ! inverse fourier in z
      
      fft_spec(i1,i2,i3)%reality_in='complex'
      fft_spec(i1,i2,i3)%reality_out='complex'
      
      fft_spec(i1,i2,i3)%storage_dims(1)=locnx
      fft_spec(i1,i2,i3)%storage_dims(2)=ny
      fft_spec(i1,i2,i3)%storage_dims(3)=n3
      
      fft_spec(i1,i2,i3)%data_dims(1)=locnx
      fft_spec(i1,i2,i3)%data_dims(2)=ny
      fft_spec(i1,i2,i3)%data_dims(3)=n3
      fft_spec(i1,i2,i3)%normval=1.d0/(n3)
      
      call CreatePlan(i1,i2,i3)

     endif
      
   end subroutine define_transforms
      
      
! 
!======================================================================================
 subroutine CreatePlan(i1,i2,i3)

!! N.B. For sin transforms, this routine
!! assumes in and out arrays are reduced in size by 2 
!! in the transform direction.
      use mpi_params 
      use specs_and_plans
      implicit none                
      include 'fftw3.f'
      
      real(kind=8)    , allocatable          :: in(:,:,:),testin(:,:,:)
      real(kind=8)    , allocatable          :: out(:,:,:),testout(:,:,:)
      real(kind=8)                           :: tmax(2)
      integer                                :: i1,i2,i3
      integer                                :: DataDims(3)
      integer                                :: StorageDims(3)
      integer                                :: Sign
      integer                                :: Dir
      integer                                :: rank_trans,howmany_rank
      integer                                :: n,nb2,nfac
      integer                                :: n_trans,n_loop
      integer                                :: fft_kind,myflag
      integer                                :: is_trans,is_loop
      integer                                :: os_trans,os_loop
      integer                                :: i,nx,ny,nz,loc(2) 
      character (len=80)                     :: TransType
      character (len=80)                     :: label,kind_label
      character (len=80)                     :: reality_in
      character (len=80)                     :: reality_out
      character (len=80)                     :: count_type
      
      !!================================================================
      !! translate info from specs_and_plans to local variables
      !!================================================================
        
        ! get coordinate direction and sign (forward/inverse) of transform
        Sign=1
        if(i1 .ne. 0) then
         Dir=1
         if(i1<0) Sign=-1
        elseif(i2 .ne. 0) then
         Dir=2
         if(i2<0) Sign=-1
        elseif(i3 .ne. 0) then
         Dir=3
         if(i3<0) Sign=-1
        endif
        
        TransType=fft_spec(i1,i2,i3)%transform_type
        do i=1,3
         DataDims(i)=fft_spec(i1,i2,i3)%data_dims(i)
         StorageDims(i)=fft_spec(i1,i2,i3)%storage_dims(i)
        enddo
      
      !!================================================================
      !!================================================================
      
      rank_trans=1          !! Transforms work on 1 dimension at a time.
      howmany_rank=1        !! Loop over the other 2 dimensions treated as 1d loop
      n = DataDims(Dir)     !! Data size in transform direction, measured in real words.
      nb2 = n/2 
      nfac = 1
      if( abs(i3)==2 ) nfac=2  !! locnx complex elements but  use compatable real arrays in,out
      
      !Allocate arrays 
       allocate(  in(StorageDims(1)*nfac,StorageDims(2),StorageDims(3)) )
       allocate( out(StorageDims(1)*nfac,StorageDims(2),StorageDims(3)) )
       allocate(  testin(StorageDims(1)*nfac,StorageDims(2),StorageDims(3)) )
       allocate( testout(StorageDims(1)*nfac,StorageDims(2),StorageDims(3)) )
            
      if(TransType=='fourier') then
                     
       if(Dir==1) then  
     
            if(Sign==1) then
             count_type = 'REAL'
             reality_in='real'
             reality_out='halfcomplex'
            elseif(Sign==-1) then
             count_type = 'REAL'
             reality_in='halfcomplex'
             reality_out='real'
            endif
            ! get fftw kind flag
            call get_kind_label(reality_in,reality_out,    &
                                TransType,fft_kind,kind_label) 
        
           !describe the transforms  in dim 1
              n_trans=DataDims(1)           ! length of transforms
              is_trans=1                    ! stride btwn elements in a single transform
              os_trans=is_trans
        
            !describe the loops over dims 2,3
              n_loop=DataDims(2)*DataDims(3)        ! 
              is_loop=StorageDims(1)                   ! stride betwn transforms
              os_loop=is_loop
              

       elseif(Dir==2) then !! must loop over k because stride is not constant otherwise      
            
            count_type = 'COMPLEX'
            reality_in='complex'
            reality_out='complex'
            ! get fftw kind flag
            call get_kind_label(reality_in,reality_out,   &
                                TransType,fft_kind,kind_label)
        
     !      describe the transforms  in dim 2
            n_trans=DataDims(2)           ! ny cplx data vals
            is_trans=StorageDims(1)       ! stride of 1d transform --> StorageDims counts complex 
            os_trans=is_trans
        
     !      describe the loops over dims 1
            n_loop=DataDims(1)         ! loop over i direction only, (count in complex words)
            is_loop=1                  !   which is defined by a stride of 1
            os_loop=is_loop
        
       elseif(Dir==3) then
            
       !    describe the transforms
            n_trans=DataDims(3)             ! fourier -->=n3=nz cmplx data vals
            is_trans= StorageDims(1)   &
                    * StorageDims(2)        ! stride of 1d transform 
                                            ! --> 1 horiz plane counted in complex words
            os_trans=is_trans
         
       !    describe the loops
            n_loop=(DataDims(1))*DataDims(2)   ! loop over i,j direction (count in complex words)
            is_loop=1                          !   which is defined by a stride of 1
            os_loop=os_loop
            
       endif
  
     elseif(TransType=='cos' .or. TransType=='cheby' .or. TransType=='sin') then    
      count_type = 'REAL'
      reality_in='real'
      reality_out='real'
      ! get fftw kind flag
      call get_kind_label(reality_in,reality_out,   &
                          TransType,fft_kind,kind_label)     
  
      if(Dir==1) then                               
     ! describe the transforms  in dim 1
          n_trans=DataDims(1)                 ! length of transform in real words     
          is_trans=1                          ! stride of 1d transform --> in x
          os_trans=is_trans
        
     !    describe the loops    over dims 2,3
          n_loop=DataDims(2)*DataDims(3) ! loop over j/k directions 
          is_loop=StorageDims(1)         !   which is defined by a stride of StorageDims(1)
          os_loop=is_loop
    
       elseif(Dir==2) then              
         !!describe the transforms in dim 2, SET UP FOR 1 K PLANE AT A TIME
         n_trans=DataDims(2)        
         is_trans=StorageDims(1)            ! stride of 1d transform --> StorageDims(1)
         os_trans=is_trans
        
        !!describe the loops    over dim 1
         n_loop=DataDims(1)                 ! loop over i direction (len=DataDims(1))
         is_loop=1                          !   which is defined by a stride of 1
         os_loop=is_loop
                 
       elseif(Dir==3) then
     !   describe the transforms
         n_trans=DataDims(3)               ! length of transforms
         is_trans=StorageDims(1) *  &
                  StorageDims(2)           ! stride of size of xy planes counted in real words
         os_trans=is_trans
         
     !   describe the loops
         n_loop=DataDims(1)*DataDims(2)            ! loop over i direction (len=locnx)
         is_loop=1                     !   which is defined by a stride of 1
         os_loop=os_loop
         
      endif
         
         
     endif
          
     myflag = FFTW_PATIENT   !! the integer parameter
     label = 'FFTW_PATIENT'  !! it's label so I can see if the right one was chosen
           
     if(TransType=='fourier' .and. Dir==1) then
           call dfftw_plan_guru_r2r(fft_plan(i1,i2,i3),                            &
                                    rank_trans,n_trans,is_trans,os_trans,          &
                                    howmany_rank,                                  &
                                    n_loop,                                        &
                                    is_loop,                                       &
                                    os_loop,                                       &
                                    in,out,fft_kind,myflag)

          in = 1.0
          testin=1.0
          call dfftw_execute_r2r( fft_plan(i1,i2,i3), in, out )
          call dfftw_execute_r2r( fft_plan(i1,i2,i3), testin, testout )
                                   
     elseif(TransType=='fourier' .and. Dir>1) then
          call dfftw_plan_guru_dft(fft_plan(i1,i2,i3),                            &
                                   rank_trans,n_trans,is_trans,os_trans,          &
                                   howmany_rank,                                  &
                                   n_loop,                                        &
                                   is_loop,                                       &
                                   os_loop,                                       &
                                   in,out,sign,myflag) 
          in = 1.0
          testin=1.0
          call dfftw_execute_dft( fft_plan(i1,i2,i3), in, out )
          call dfftw_execute_dft( fft_plan(i1,i2,i3), testin, testout )
                                   
     elseif(TransType=='cos' .or. TransType=='cheby' .or. TransType=='sin') then 
          call dfftw_plan_guru_r2r(fft_plan(i1,i2,i3),                            &
                                   rank_trans,n_trans,is_trans,os_trans,          &
                                   howmany_rank,                                  &
                                   n_loop,                                        &
                                   is_loop,                                       &
                                   os_loop,                                       &
                                   in,out,fft_kind,myflag)  
                                   
          in = 1.0
          testin=1.0
          call dfftw_execute_r2r( fft_plan(i1,i2,i3), in, out )
          call dfftw_execute_r2r( fft_plan(i1,i2,i3), testin, testout )
     endif
     
     tmax(1)=maxval( abs(out) )
     tmax(2)=maxval( abs(testout) )

     if(myid==0) then
     open (1,file='output/fftw3plans',position='rewind')
      write(1,*) ' '
      write(1,*) ' '
      write(1,*) '=================================================================' 
      write(1,*) ' Transform for dimension ',Dir,' isign = ',sign,' transform type:    ',TransType
      write(1,*) '  The following parameters are counts in terms of ',trim(count_type),' words.'
      write(1,*) '  The number of data pts in the transform dimension                      n= ',n_trans
      write(1,*) '  Number of words btwn successive elements in transform direction         = ',is_trans
      write(1,*) '  Number of vectors to be transformed together, i.e. loop lengths         = ',n_loop
      write(1,*) '  The loop is defined by words btwn succesive 1st elements                = ',is_loop
      write(1,*) '  StorageDims, i.e. array size for in/out                                   ',StorageDims
      write(1,*) '  DataDims , i.e. how much of in is filled w/ data                          ',DataDims
      write(1,*) '  FFTW3 plan quality                                                        ',trim(label)
      write(1,*) '  FFT_KIND , label                                                          ',fft_kind,trim(kind_label)
      write(1,*) '  FFTW3 plan                                                                ',fft_plan(i1,i2,i3)
      write(1,*) '  Actual Bounds on in/out arrays used to define plan '
      write(1,*) '  i dimension-- in    : ',LBOUND(in,1), UBOUND(in,1)
      write(1,*) '               out    : ',LBOUND(out,1), UBOUND(out,1)
      write(1,*) '  j dimension-- in    : ',LBOUND(in,2), UBOUND(in,2)
      write(1,*) '               out    : ',LBOUND(out,2), UBOUND(out,2)
      write(1,*) '  k dimension-- in    : ',LBOUND(in,3), UBOUND(in,3)
      write(1,*) '               out    : ',LBOUND(out,3), UBOUND(out,3)
      write(1,*) '  in,out arrays have KINDs of ',KIND(in),KIND(out)
      write(1,*) '  transform executed with in=1.0  '
      write(1,*) '  maxval and maxloc of out array :  ',tmax(1),loc(1)
      write(1,*) '  testin,testout arrays are similar but distinct, should take same plan'
      write(1,*) '  transform executed with testin=1.0  '
      write(1,*) '  maxval and maxloc of testout array :  ',tmax(2),loc(2)
      write(1,*) '================================================================='
      write(1,*) ' '
      write(1,*) ' ' 
     close(1)
     endif
     
     if( allocated(in) ) deallocate(in)
     if( allocated(out) ) deallocate(out)
     if( allocated(testin) ) deallocate(testin)
     if( allocated(testout) ) deallocate(testout)
            
    end subroutine CreatePlan
    
    !======================================================================================
      subroutine get_kind_label(in,out,type,fft_kind,label)
       implicit none                
       include 'fftw3.f'
       integer,intent(out)               :: fft_kind
       character (len=80),intent(out)    :: label
       character (len=11)                :: in
       character (len=11)                :: out
       character (len=11)                :: type
       
       
       if( trim(type)=='fourier' ) then
        if( in=='real' .and. out=='complex') then
         fft_kind=FFTW_R2HC
         label='no flag needed'
        elseif( in=='complex' .and. out=='real') then
         fft_kind=FFTW_HC2R
         label='no flag needed'
        elseif( in=='real' .and. out=='halfcomplex') then
         fft_kind=FFTW_R2HC
         label='FFTW_R2HC'
        elseif( in=='halfcomplex' .and. out=='real') then
         fft_kind=FFTW_HC2R
         label='FFTW_HC2R'
        elseif( in=='complex' .and. out=='complex') then
         fft_kind=-999
         label='no flag needed'
        else
         write(0,*) 'problem getting kind of fourier transform',in,out
         stop
        endif        
       endif
       
       if( trim(type)=='sin' ) then
        if( in=='real' .and. out=='real') then
         fft_kind=FFTW_RODFT00
         label='FFTW_REDFT00'
        elseif( in=='complex' .and. out=='complex') then
         fft_kind=-999
         label='no label needed'
        else
         write(0,*) 'problem getting kind of fourier transform',in,out
         stop
        endif        
       endif
       
       if( trim(type)=='cos' .or. trim(type)=='cheby' ) then
        if( in=='real' .and. out=='real') then
         fft_kind=FFTW_REDFT00
         label='FFTW_RODFT00'
        elseif( in=='complex' .and. out=='complex') then
         fft_kind=-999
         label='no label needed'
        else
         write(0,*) 'problem getting kind of fourier transform',in,out
         stop
        endif        
       endif         
      
      end subroutine get_kind_label
      
      subroutine transform3d(in,out,ztype)      
      use specs_and_plans
      use flags_counters_etc
      
      implicit none      
      include 'fftw3.f'
      
      integer             :: i3          ! flags for transform type in each dimension
      character (len=11)  :: ztype       ! sin,cos or fourier, type of z transform
      
      real(kind=8)                         :: in(nx+2,ny,locnz)      ! slab
      double complex                       :: out(locnx,ny,n3)       ! column
      real(kind=8), allocatable            :: work(:,:,:),work2(:,:,:)
      double complex, allocatable          :: short1(:,:,:),short2(:,:,:)
      double complex, allocatable          :: cwork(:,:,:),cplxwork(:,:,:)  !! c-->col  cplx-->complex
      real(kind=8)                         :: xx
      
      
      allocate(work(nx+2,ny,locnz),work2(nx+2,ny,locnz))
      allocate(cwork(locnx,ny,n3),cplxwork((nx+2)/2,ny,locnz))  
      allocate(short1(locnx,ny,n3-2),short2(locnx,ny,n3-2))
      
!!!==================================================================
!!!============== X transforms first, REAL to HALF COMPLEX ==========
!!!==================================================================
       call dfftw_execute_r2r(fft_plan(1,0,0),in,work2) 
       call shuffleHC2C(work2,work,nx,ny,locnz)
!!!==================================================================
!!!==================================================================
       
!!!==================================================================
!!!============== Y transforms next, COMPLEX to COMPLEX =============
!!!==================================================================
        do k=1,locnz
         call dfftw_execute_dft(fft_plan(0,1,0),work(1,1,k),cplxwork(1,1,k))
        enddo
        !exchange data so z transforms can be done
        call slabs_to_columns(cplxwork,cwork)
!!!==================================================================
!!!==================================================================

                     
!!!==================================================================
!!!============== Z transforms last, data in COLUMN FMT =============
!!!==================================================================
       if(ztype=='fourier') then 
         i3=2   ! forward
         call dfftw_execute_dft( fft_plan(0,0,i3),cwork,out )
       elseif(ztype=='cos') then
         i3=-1
         call dfftw_execute_r2r( fft_plan(0,0,i3),cwork,out )
       elseif(ztype=='sin') then
         i3=1  
         short1(:,:,:)=cwork(:,:,2:n3-1)
         out(:,:,1)=0. ; out(:,:,n3)=0.
         call dfftw_execute_r2r( fft_plan(0,0,i3),short1,short2 ) 
         out(:,:,2:n3-1)=short2(:,:,1:n3-2)
       endif
!!!==================================================================
!!!==================================================================
                         
       deallocate(work,work2,cplxwork,cwork,short1,short2)
      end subroutine transform3d
!============================================================================================      
      
      subroutine inv_transform3d(in,out,ztype)      
      use specs_and_plans
      use flags_counters_etc
      
      implicit none      
      include 'fftw3.f'
      
      integer             :: i3          ! flags for transform type in each dimension
      character (len=11)  :: ztype       ! sin,cos or fourier, type of z transform
      
      double complex                       :: in(locnx,ny,n3)      ! column
      real(kind=8)                         :: out(nx+2,ny,locnz)   ! slab
      double complex, allocatable          :: cwork1(:,:,:)
      double complex, allocatable          :: short1(:,:,:),short2(:,:,:)
      real(kind=8), allocatable            :: work1(:,:,:)
      real(kind=8), allocatable            :: work2(:,:,:)
      real(kind=8)                         :: xx
            
      allocate( cwork1(locnx,ny,n3),work1(nx+2,ny,locnz) )
      allocate( work2(nx+2,ny,locnz) )
      allocate( short1(locnx,ny,n3-2),short2(locnx,ny,n3-2) )
            
!!!==================================================================
!!!============== Z transforms first, data in COLUMN FMT =============
!!!==================================================================
       if(trim(ztype) == 'fourier') then 
         i3=-2
         call dfftw_execute_dft(fft_plan(0,0,i3),in,cwork1)
       elseif(trim(ztype)=='cos') then
         i3=-1
         call dfftw_execute_r2r(fft_plan(0,0,i3),in,cwork1)
       elseif(trim(ztype)=='sin') then
         i3=1
         short1(:,:,1:n3-2)=in(:,:,2:n3-1)
         cwork1(:,:,1)=0. ; cwork1(:,:,n3)=0.
         call dfftw_execute_r2r(fft_plan(0,0,i3),short1,short2)
         cwork1(:,:,2:n3-1)=short2(:,:,1:n3-2)
       endif
!!!==================================================================
!!!==================================================================
        
!!!==================================================================
!!!============== Y transforms next, COMPLEX to COMPLEX =============
!!!==================================================================
       !exchange data so xy transforms can be done
       call columns_to_slabs(work1,cwork1)    !work1 is output array
       do k=1,locnz
        call dfftw_execute_dft(fft_plan(0,-1,0),work1(1,1,k),work2(1,1,k))
       enddo
!!!==================================================================
!!!==================================================================
 
        
!!!==================================================================
!!!============== X transforms last, HALF COMPLEX to REAL ===========
!!!==================================================================
       call shuffle2HC(work2,work1,nx,ny,locnz)
       call dfftw_execute_r2r(fft_plan(-1,0,0),work1,out)
!!!==================================================================
!!!==================================================================
 

                 
!!!==================================================================
!!!============== Normalize only after inverse transforms ===========
!!!==================================================================
       if(trim(ztype)=='fourier') then
        xx=fft_spec(-1,0,0)%normval * fft_spec(0,-1,0)%normval * fft_spec(0,0,-2)%normval
       else
        xx=fft_spec(-1,0,0)%normval * fft_spec(0,-1,0)%normval * fft_spec(0,0,-1)%normval
       endif
       out = xx*out
!!!==================================================================
!!!==================================================================
            
       deallocate(work1,work2,short1,short2,cwork1)
      end subroutine inv_transform3d

     subroutine shuffleHC2C(in,out,nx,ny,locnz)   
      !! pack work2 into a real array that could
      !! be exactly equivilanced to a cmplx array
      !! with leading dimension (nx+2)/2.
      
      implicit none
      integer         :: nx,ny,locnz,i,j
      real(kind=8)    :: in(nx+2,ny,locnz)
      real(kind=8)    :: out(nx+2,ny,locnz)
      
      do i=1,nx/2+1
       out(2*i-1,:,:)=in(i,:,:)
      enddo
      
      j=nx
      do i=nx/2+2,nx
       out(j,:,:)=in(i,:,:)
       j=j-2
      enddo
      
      out(2,:,:)=0.d0
      out(nx+2,:,:)=0.d0          
     end subroutine shuffleHC2C
      
     subroutine shuffle2HC(in,out,nx,ny,locnz)   
      !! pack work2 into a real array that could
      !! be exactly equivilanced to a cmplx array
      !! with leading dimension (nx+2)/2.
      
      implicit none
      integer         :: nx,ny,locnz,i,j
      real(kind=8)    :: in(nx+2,ny,locnz)
      real(kind=8)    :: out(nx+2,ny,locnz)
      
      do i=1,nx+1,2   !! all the real parts
       j=(i+1)/2
       out(j,:,:)=in(i,:,:)
      enddo
      
      !! im part of mode 1 is in slot 4
      !! don't deal with im of mode 0 or nyquist
      do i=4,nx,2   !! all the imag parts
       j=i/2-1  !!i.e. mode number
       out(nx-j+1,:,:)=in(i,:,:)
      enddo   
     end subroutine shuffle2HC

