!===========================================================================
!!    Free-Slip or Triply Periodic Spectral model in in 2/3 dimensions.
!!    2nd, 3rd or 4th order Adams Bashforth method for integration in time.
!!    Parallel version using MPI. Two scalars, can be configured as
!!    combinations of density, temperature, salinity or passive tracer.
!!    Lagrangian trajectories. Cubic spline spatial interpolation.
!=========================================================================== 
 subroutine flow_solve
 use flags_counters_etc,    only: istep,istart,iend,N
 use state_variables,       only: u,v,w
 use independent_variables, only: t_secs
 use particles,             only: positions
 use mpi_params,            only: comm,myid,nprocs
 implicit none

!pba
 write(*,*)'debut programme'

 call initialize_mpi(comm,myid,nprocs)
 call read_userdata
 call allocate_arrays
 call runtime_initialization

 do istep=istart,iend 
   call vort_fcs
      call ifft 
         call equation_of_state
             call advance_particles
                 call nonlin 
               call write_results
            call fft
         call rhs
      call step
   call toggle 
 enddo
 
 call quit_mpi
 end subroutine flow_solve
!=============================================================== 
!    End of top level flow solver routine
!=============================================================== 
 
 
 
 
!=============================================================== 
!=============================================================== 
 subroutine allocate_arrays
   use state_variables
   use independent_variables
   use dimensional_params
   use dimensionless_params
   use mpi_params
   use flags_counters_etc
   use intermediate_variables
   use io_params
   use particles

   implicit none
#include "mpif.h"
!=============================================================== 
!=============================================================== 
!!     Determine the appropriate sizes of arrays
!!     for multi processor runs, each processor stores
!!     locnz horizontal planes of a distributed array.
!!     Processor (numprocs-1) stores one extra plane
!!     if sin/cos expansions are used in z.
!=============================================================== 
!===============================================================     
      if( bc_flag == 'zslip' ) then
       n3=nz+1 
       locnz = nz/numprocs
       if(myid==numprocs-1) locnz=locnz+1
      elseif( bc_flag == 'zperiodic') then
       n3=nz
       locnz = nz/numprocs
      else
       write(0,*) 'problem specifying z boundary conditions'
      endif
      
      !! each processor store the locnz for all processors in
      !! the array numzplanes(0:numprocs-1)
      allocate( numzplanes(0:numprocs-1),itmp1d(0:numprocs-1) )
      allocate( numzplanes_below(0:numprocs-1) )
      
      itmp1d(:)=0   ;  itmp1d(myid)=locnz      
      call MPI_ALLREDUCE(itmp1d,numzplanes,numprocs,   &
                         MPI_INTEGER,MPI_SUM,comm,ierr)
      deallocate(itmp1d)
      
      nyplanes=ny
      if(ny==0) stop ' 2d problems have ny=1 '
      
      if( mod( (nx/2)/numprocs,2 ) .ne. 0 ) &
        stop ' (nx/2)/numprocs must be even '
      locnx=(nx/2)/numprocs
!=============================================================== 
!=============================================================== 

!!    Allocate storage for spatial positions x,y,z and zglobal.
      allocate( x(nx), y(ny), z(locnz), zglobal(n3) )
      
      
!!    Allocate space for the physical space representation    
!!    of the dependent variables at time t_n.
      allocate( u(nx+2,ny,locnz,2),v(nx+2,ny,locnz,2) )
      allocate( w(nx+2,ny,locnz,2),pd(nx+2,ny,locnz) )
      allocate( s1(nx+2,ny,locnz,2),s2(nx+2,ny,locnz,2) )
      allocate( ambient_density(nx+2,ny,locnz) )
      
      
!     Wavenumber space dependent variables: times t_n, t_n-1
      allocate( u_cplx(locnx,ny,n3,2), v_cplx(locnx,ny,n3,2) )
      allocate( s1_cplx(locnx,ny,n3,2), s2_cplx(locnx,ny,n3,2) )
      allocate( w_cplx(locnx,ny,n3,2) )

  
!     3 copies of transfer terms: t_n, t_n-1, t_n-2
      allocate( Tr1(locnx,ny,n3,3), Tr2(locnx,ny,n3,3) )
      allocate( Tr3(locnx,ny,n3,3), Tr4(locnx,ny,n3,3) )
      allocate( Tr5(locnx,ny,n3,3) )

!     Intermediate complex arrays
      allocate(T1_cplx(locnx,ny,n3),T2_cplx(locnx,ny,n3),T3_cplx(locnx,ny,n3))
      allocate(u_dot_grad_s1_bar_cplx(locnx,ny,n3),u_dot_grad_s2_bar_cplx(locnx,ny,n3))
      allocate(s1_times_u_cplx(locnx,ny,n3),s1_times_v_cplx(locnx,ny,n3),  &
               s1_times_w_cplx(locnx,ny,n3),s2_times_u_cplx(locnx,ny,n3),  &
               s2_times_v_cplx(locnx,ny,n3),s2_times_w_cplx(locnx,ny,n3))
               
      allocate( dz_s1_bar_cplx(locnx,ny,n3),   &
                dz_s2_bar_cplx(locnx,ny,n3),   &
                dzz_s1_bar_cplx(locnx,ny,n3),  &
                dzz_s2_bar_cplx(locnx,ny,n3) )

!     Wavenumber arrays and integrating factors
      allocate( wnx(locnx), wny(ny), wnz(n3) )
      allocate( amp(locnx,ny,n3,3) )

!     Time independent, spatially variable quantities.
      allocate( dz_s1_bar(nx+2,ny,locnz),   &
                dz_s2_bar(nx+2,ny,locnz),   &
                dzz_s1_bar(nx+2,ny,locnz),  &
                dzz_s2_bar(nx+2,ny,locnz) )

!     Storage for lagrangian particle trajectories.   
      if(ntrajectories > 0)  then
   
       !! allocate the ghosted velocity array, depends on z bc, myid and numprocs
       if( bc_flag=='zperiodic' ) then
    
        allocate( gvelocity(1:nx+1,1:ny+1,1:numzplanes(myid)+1,3) )
     
       elseif( bc_flag=='zslip' ) then
    
        if( numprocs>1 .and. myid > 0 .and. myid < numprocs-1 )  &
        allocate( gvelocity(1:nx+1,1:ny+1,1:numzplanes(myid)+1,3) )  !! interior nodes
     
        if( myid==numprocs-1 .and. numprocs>1 )   &
          allocate( gvelocity(1:nx+1,1:ny+1,1:numzplanes(myid),3) )  !! top
        
        if( myid==0 .and. numprocs>1 )   &
          allocate( gvelocity(1:nx+1,1:ny+1,1:numzplanes(myid)+1,3) ) !! bottom
        
        if( numprocs==1 )   &
          allocate( gvelocity(1:nx+1,1:ny+1,1:numzplanes(myid),3) ) !! only
        
       endif
    
       allocate( owners(ntrajectories) )
       allocate( tmp_owners(ntrajectories) )
       allocate( positions(ntrajectories,3) )
       allocate( uvels(ntrajectories,3) )
       allocate( vvels(ntrajectories,3) )
       allocate( wvels(ntrajectories,3) )
    
      endif 
   if(myid==0) write(0,*) 'arrays allocated in routine allocate_arrays, SpectralAlgorithm.f90'
 end subroutine allocate_arrays
!=============================================================== 
!===============================================================

!=============================================================== 
!===============================================================      
  subroutine runtime_initialization
    use state_variables
    use independent_variables
    use dimensional_params
    use dimensionless_params
    use mpi_params
    use flags_counters_etc
    use intermediate_variables
    use io_params
    use particles
    use user_module
    use Ziggurat

    implicit none
!===============================================================
!BEGIN RUNTIME INITIALIZATION
!===============================================================
    call tic(clock_start(runtime_initialization_id),rate)  ! start timer
    call zigset(4)           ! seed the random number generator
    maxprocs = 1024
    N=1 ; NM1=2              ! 2-cycle toggles
    M=1 ; MM1=2  ;  MM2=3    ! 3-cycle toggles
!!===============================================================
!   Gridpoint locations and allocations (store dimensional values)
!   e.g. for zslip:  
!    nz=16, n3=nz+1=17, 
!    locnz=16/numprocs except uppermost proc. has 1 more
!   e.g. for zperiodic:  
!    nz=16, n3=nz=16, 
!    locnz=16/numprocs

    dx = Lx/float(nx)
    dy = Ly/float(ny)
    dz = Lz/float(nz)
   
    x(:) = dx*(/ (i,i=0,nx-1) /)
    y(:) = dy*(/ (i,i=0,ny-1) /) 
    
!! for each processor below myid, count pts below
    numzplanes_below(0)=0
    do k=1,numprocs-1  
     numzplanes_below(k)=numzplanes_below(k-1)+numzplanes(k-1)
    enddo

    zglobal(:) =  dz*(/ (i,i=0,n3-1) /)
    z(:) = zglobal( numzplanes_below(myid) + 1 )  &
         + dz*(/ (i,i=0,numzplanes(myid)-1) /)    
    
    !! dimensionless limits for processors
    my_z_limits(1) = z(1)/Lz
    my_z_limits(2) = (z(numzplanes(myid))+dz)/Lz
    if(bc_flag=='zslip' .and. myid==numprocs-1) my_z_limits(2) = (z(numzplanes(myid)))/Lz
   
    !write(0,*) myid,numzplanes(myid),my_z_limits
    !if(myid==0) write(0,*) 'Spatial variables initialized.'
    !stop
!!===============================================================

!   Create specialized data structures for MPI data swapping
!   routines slabs_to_columns  and  columns_to_slabs.

  call create_MPI_data_type(nx,ny,nz,locnx,numzplanes(myid),   &
                            twoslice,subslice,myid,comm)

!   Set initial conditions and ambient scalar profiles s1_bar(z), s2_bar(z)    
  N=1 ; pi=4.*atan(1.)
  do k=1,numzplanes(myid)
   do j=1,ny
    do i=1,nx
   
    call user_defined_ics(x(i),y(j),z(k),Lx,Ly,Lz, &
                         u(i,j,k,N),v(i,j,k,N),w(i,j,k,N),  &
                         s1(i,j,k,N),s2(i,j,k,N),f,g,rho_0,DGRAD)
   
    call eval_s1_bar(z(k),Lz,s1_bar,s1_bar_z,s1_bar_zz)  
    call eval_s2_bar(z(k),Lz,s2_bar,s2_bar_z,s2_bar_zz)
 
    call nondimensionalize_s1(s1_bar,s1_bar_z,s1_bar_zz)
    call nondimensionalize_s2(s2_bar,s2_bar_z,s2_bar_zz)
 
  
    dz_s1_bar(i,j,k) = s1_bar_z
    dz_s2_bar(i,j,k) = s2_bar_z
    dzz_s1_bar(i,j,k) = s1_bar_zz
    dzz_s2_bar(i,j,k) = s2_bar_zz
  

  if( scalars .eq. 'ts' ) then

    Temperature = (s1_bar*s1_scale)  ! [deg C]
    Salinity    = (s2_bar*s2_scale)  ! [psu]
    Salinity=Salinity/1000.! to match fn "rho"
    pmPa=0.d0
    call rho(Salinity,Temperature,pMpa,density)
    rho_bar = density - rho_0   ! [kg/m3]
    ambient_density(i,j,k) = rho_bar/(Lz*DGRAD)   ! store dimensionless values
    
  elseif( scalars .eq. 'tp' .or. scalars .eq. 't0' ) then

    Temperature = (s1_bar*s1_scale)  ! [deg C]
    Salinity = 0.! [psu]
    pmPa=0.d0
    call rho(Salinity,Temperature,pMpa,density)
    rho_bar = density - rho_0   ! [kg/m3]
    ambient_density(i,j,k) = rho_bar/(Lz*DGRAD)   ! store dimensionless values
 
  elseif( scalars .eq. 'sp' .or. scalars .eq. 's0' ) then

    Temperature = 0.  ! [deg C]
    Salinity = (s2_bar*s2_scale)! [psu]
    Salinity=Salinity/1000.! to match fn "rho"
    pmPa=0.d0
    call rho(Salinity,Temperature,pMpa,density)
    rho_bar = density - rho_0   ! [kg/m3]
    ambient_density(i,j,k) = rho_bar/(Lz*DGRAD)   ! store dimensionless values
 
  elseif( scalars .eq. 'rp' .or. scalars .eq. 'r0' ) then

    ambient_density(i,j,k) = s1_bar! store dimensionless values
 
  elseif( scalars .eq. 'pp' ) then

    ambient_density(i,j,k) = 0.0   ! store dimensionless values
  
  endif 

  enddo
 enddo
enddo   

!Get initial particle positions
 call initialize_particles( positions,ntrajectories,Lx,Ly,Lz )  !! returned in [m]
 positions = positions/Lz  !! now dimensionless
 owners=myid   !! will be updated before use if numprocs>1

!Get initial maxima to estimate CFL condition.    
 umax = maxval(abs(u))
 vmax = maxval(abs(v))
 wmax = maxval(abs(w))
 
 time_scale = Lz/U0
 bfreq = sqrt((g/rho_0)*DGRAD)   ! char. buoyancy frequency [1/s]
 
!Nondimensionalize dependent variables:
 u(:,:,:,:) = u(:,:,:,:)/U0
 v(:,:,:,:) = v(:,:,:,:)/U0
 w(:,:,:,:) = w(:,:,:,:)/U0
 s1(:,:,:,:) = s1(:,:,:,:)/(s1_scale)
 s2(:,:,:,:) = s2(:,:,:,:)/(s2_scale)
 
!Write out parameters characterizing simulation.
!
 if( myid .eq. 0 ) then
  write(0,*) ' nx, ny, nz  ',nx,ny,nz
  write(0,*) ' boundary conditions:  ',bc_flag
  write(0,*) ' locnx, locnz, n3 ',locnx, locnz, n3
  write(0,*) ' '
  write(0,*) ' '
  
  write(0,*) ' domain lengths Lx, Ly, Lz  [m]:  ',Lx,Ly,Lz
  write(0,*) ' grid spacings dx, dy, dz   [m]:  ',dx,dy,dz
  write(0,*) ' bulk velocity scale U    [m/s]:  ',U0
  write(0,*) ' characteristic density gradient scale [kg/m4]:  ',DGRAD
  write(0,*) ' characteristic range of scalar 1 [e.g. deg C]:  ',s1_scale
  write(0,*) ' characteristic range of scalar 2 [e.g. psu]  :  ',s2_scale
  write(0,*) ' characteristic density [kg/m3]:  ',rho_0
  write(0,*) ' coriolis parameter f[1/s]:  ',f
  write(0,*) ' gravitational acceleration g [m/s2]:  ',g
  write(0,*) ' molecular diffusivity of scalar 1 [m2/s]:  ',kappa_1
  write(0,*) ' molecular diffusivity of scalar 2 [m2/s]:  ',kappa_2
  write(0,*) ' viscosity[m2/s]:  ',nu
  write(0,*) ' CFL_x = Umax dt/dx  :  ', UMAX*dt/dx
  write(0,*) ' CFL_y = Vmax dt/dy  :  ', VMAX*dt/dy
  write(0,*) ' CFL_z = Wmax dt/dz  :  ', WMAX*dt/dz
  write(0,*) ' '
  write(0,*) ' '

!Nondimensionalize the initial fields, the stratification  
!parameters and the time variables.

  write(0,*) ' time step dt    [s] :  ',dt
  write(0,*) ' initial time    [s] :  ',t_start
  write(0,*) ' final time [s] :  ',t_end
  write(0,*) ' Lz/U0 [s] :  ',time_scale
 endif

 istart = 0
 iend = floor( ( t_end - t_start ) / dt )
 i_stat = floor( t_stat/dt )

 t_secs = t_start
 dt = dt/time_scale
 t_start = t_start/time_scale
 t_end = t_end/time_scale
 t_stat = t_stat/time_scale

 
 if( myid .eq. 0 ) then
  write(0,*) ' '
  write(0,*) ' time increment counters set'
  write(0,*) ' number of steps:  ',iend-istart+1
  write(0,*) ' number of steps between energetics calculations   :  ',i_stat
 endif

!Compute the dimensionless parameters.
 Rot= f*Lz/U0
 N2 = (g/rho_0)*DGRAD
 Ri = N2*Lz**2/(U0**2)
 Re = U0*Lz/nu
 Pr_1 = nu/kappa_1
 Pr_2 = nu/kappa_2
 
 if( myid .eq. 0 ) then
  write(0,*) ' '
  write(0,*) ' '
  write(0,*) ' inverse Rossby number fL/U    :  ',Rot
  write(0,*) ' bulk Richardson number (NL/U)^2    :  ',Ri
  write(0,*) ' bulk Reynolds number UL/nu    :  ',Re
  write(0,*) ' Prandtl number  nu/kappa_1    :  ',Pr_1
  write(0,*) ' Prandtl number  nu/kappa_2    :  ',Pr_2
  write(0,*) ' '
  write(0,*) ' '
 endif
 
!Initialize wavenumber vectors etc
 call fill_wave_vectors
   
!Store detailed specifications of the various transforms
!and create fftw3 plans (guru plans that can be used on
!different arrays that are similar in layout.

 call define_transforms

 ctype='cos'
 stype='sin'
 ftype='fourier'

 if( myid .eq. 0 ) then
  write(0,*) ' ' 
  write(0,*) 'FORWARD TRANSFORMING THE INITIAL CONDITIONS'
 endif 
 
 if( trim(bc_flag) == 'zslip' ) then
  call transform3d(u(1,1,1,N),u_cplx(1,1,1,N),ctype)
   if(myid==0) write(0,*) ' ics for u transformed,  z---> ',ctype
   
  call transform3d(v(1,1,1,N),v_cplx(1,1,1,N),ctype)
   if(myid==0) write(0,*) ' ics for v transformed,  z---> ',ctype
   
  call transform3d(w(1,1,1,N),w_cplx(1,1,1,N),stype)
   if(myid==0) write(0,*) ' ics for w transformed,  z---> ',stype
   
  call transform3d(s1(1,1,1,N),s1_cplx(1,1,1,N),stype)
   if(myid==0) write(0,*) ' ics for s1 transformed,  z---> ',stype
   
  call transform3d(s2(1,1,1,N),s2_cplx(1,1,1,N),stype)
   if(myid==0) write(0,*) ' ics for s2 transformed,  z---> ',stype
   
 elseif( trim(bc_flag) == 'zperiodic' ) then
  call transform3d(u(1,1,1,N),u_cplx(1,1,1,N),ftype)
   if(myid==0) write(0,*) ' ics for u transformed,  z---> ',ftype
   
  call transform3d(v(1,1,1,N),v_cplx(1,1,1,N),ftype)
   if(myid==0) write(0,*) ' ics for v transformed,  z---> ',ftype
   
  call transform3d(w(1,1,1,N),w_cplx(1,1,1,N),ftype)
   if(myid==0) write(0,*) ' ics for w transformed,  z---> ',ftype
   
  call transform3d(s1(1,1,1,N),s1_cplx(1,1,1,N),ftype)
   if(myid==0) write(0,*) ' ics for s1 transformed,  z---> ',ftype
   
  call transform3d(s2(1,1,1,N),s2_cplx(1,1,1,N),ftype)
   if(myid==0) write(0,*) ' ics for s2 transformed,  z---> ',ftype
 endif
 
 
 pressure_flag='NOTdo_pressure_only'

100   format(1x,i3,4e24.10)
101   format(1x,i3,7e24.10)

 if( myid .eq. 0 ) write(0,*) 'Runtime initialization completed.'
 call MPI_BARRIER(comm,ierr)
 call toc(clock_start(runtime_initialization_id),  &
          rate,secs(runtime_initialization_id))
 cumsecs(runtime_initialization_id)=cumsecs(runtime_initialization_id) &
                                    +secs(runtime_initialization_id)
 
end subroutine runtime_initialization
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
 subroutine vort_fcs 
!==========================================================================
!
!  vort_fcs - This FORTRAN subroutine computes the frequency domain
!             compute transform of vorticity given transform of u,v,w 
!             Assumes the form of the transform is of the type
!             F/F/C  u,v and F/F/S for w,s1,s2 when bc_flag = 'zslip'
!             and F/F/F for all variables when bc_flag = 'zperiodic'.  

! derivative operators in transform space:
!
!    fourier directions:  i * wavenumber  i.e. d/dx --> i k
!    z derivative of w :                       d/dz --> +m
!    z derivative of u,v:                      d/dz --> -m
!==========================================================================

      use state_variables
      use intermediate_variables 
      use flags_counters_etc
      use independent_variables
      implicit none
      
      call tic(clock_start(vort_fcs_id),rate)  ! start timer
      allocate(vortx(locnx,ny,n3),   &
               vorty(locnx,ny,n3),   &
               vortz(locnx,ny,n3) )               
    
      if( trim(bc_flag) .eq. 'zslip' ) then  ! use sin/cos logic in z  
      
       do k=1,n3-1
        do j=1,ny
         do i=1,locnx
          vortx(i,j,k) = cmplxi*wny(j)*w_cplx(i,j,k,N) - (-wnz(k)*v_cplx(i,j,k,N))
          vorty(i,j,k) = -wnz(k)*u_cplx(i,j,k,N) - cmplxi*wnx(i)*w_cplx(i,j,k,N)
          vortz(i,j,k) = cmplxi*(wnx(i)*v_cplx(i,j,k,N) - wny(j)*u_cplx(i,j,k,N))
         enddo
        enddo
       enddo
      !zero out nyquist in kz
       vortx(:,:,n3)=0.0  ;  vorty(:,:,n3)=0.0  ;  vortz(:,:,n3)=0.0
      
     elseif( trim(bc_flag) .eq. 'zperiodic') then   ! use triply fourier logic
      
       do k=1,n3
        do j=1,ny
         do i=1,locnx
          vortx(i,j,k) = cmplxi*(wny(j)*w_cplx(i,j,k,N) - wnz(k)*v_cplx(i,j,k,N))
          vorty(i,j,k) = cmplxi*(wnz(k)*u_cplx(i,j,k,N) - wnx(i)*w_cplx(i,j,k,N))
          vortz(i,j,k) = cmplxi*(wnx(i)*v_cplx(i,j,k,N) - wny(j)*u_cplx(i,j,k,N)) 
         enddo
        enddo
       enddo
       !zero out nyquist in kz
       vortx(:,:,n3)=0.0  ;  vorty(:,:,n3)=0.0  ;  vortz(:,:,n3)=0.0
       
      endif
      
      call toc(clock_start(vort_fcs_id),rate,secs(vort_fcs_id))         
      cumsecs(vort_fcs_id)=cumsecs(vort_fcs_id) + secs(vort_fcs_id)
      numcalls(vort_fcs_id)=numcalls(vort_fcs_id)+1
                                    
      return
      end   
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
    subroutine ifft 
	
    use state_variables
    use intermediate_variables
    use flags_counters_etc
	
    implicit none
	
	call tic(clock_start(ifft_id),rate)  ! start timer
    allocate(zetax(nx+2,nyplanes,locnz),   &
             zetay(nx+2,nyplanes,locnz),   &
             zetaz(nx+2,nyplanes,locnz))
              
    ! inverse transforms (complex --> real)
    if(trim(bc_flag)=='zslip') then	     
    
     call inv_transform3d(u_cplx(1,1,1,N),u(1,1,1,N),ctype)
     call inv_transform3d(v_cplx(1,1,1,N),v(1,1,1,N),ctype)
     call inv_transform3d(w_cplx(1,1,1,N),w(1,1,1,N),stype)
     call inv_transform3d(s1_cplx(1,1,1,N),s1(1,1,1,N),stype)
     call inv_transform3d(s2_cplx(1,1,1,N),s2(1,1,1,N),stype)
    
     call inv_transform3d(vortx,zetax,stype)
     call inv_transform3d(vorty,zetay,stype)
     call inv_transform3d(vortz,zetaz,ctype)
     
    elseif(trim(bc_flag)=='zperiodic') then
    
     call inv_transform3d(u_cplx(1,1,1,N),u(1,1,1,N),ftype)
     call inv_transform3d(v_cplx(1,1,1,N),v(1,1,1,N),ftype)
     call inv_transform3d(w_cplx(1,1,1,N),w(1,1,1,N),ftype)
     call inv_transform3d(s1_cplx(1,1,1,N),s1(1,1,1,N),ftype)
     call inv_transform3d(s2_cplx(1,1,1,N),s2(1,1,1,N),ftype)
    
     call inv_transform3d(vortx,zetax,ftype)
     call inv_transform3d(vorty,zetay,ftype)
     call inv_transform3d(vortz,zetaz,ftype)   
     
    endif
    
    ! don't need transforms of vorticity components anymore
    deallocate(vortx,vorty,vortz)
    
    call toc(clock_start(ifft_id),rate,secs(ifft_id))         
    cumsecs(ifft_id)=cumsecs(ifft_id) + secs(ifft_id)
    numcalls(ifft_id)=numcalls(ifft_id)+1

    return
    end
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
	 subroutine nonlin 
	
!============================================================================
!  nonlin - Compute u cross vorticity (zeta) and add the appropriate 
!           rotation and buoyancy terms. Also compute velocity-scalar
!           products and the (negative of) velocity dot mean gradient terms.
!============================================================================


      use mpi_params
      use dimensionless_params
      use state_variables
      use intermediate_variables
      use flags_counters_etc
      implicit none
      
      call tic(clock_start(nonlin_id),rate)  ! start timer
      allocate( s1_times_u(nx+2,ny,locnz),  &
                s1_times_v(nx+2,ny,locnz),  &
                s1_times_w(nx+2,ny,locnz),  &
                s2_times_u(nx+2,ny,locnz),  &
                s2_times_v(nx+2,ny,locnz),  &
                s2_times_w(nx+2,ny,locnz),  &
                T1(nx+2,ny,locnz),          &
                T2(nx+2,ny,locnz),          &
                T3(nx+2,ny,locnz),          &
                u_dot_grad_s1_bar(nx+2,ny,locnz),  &
                u_dot_grad_s2_bar(nx+2,ny,locnz) )
        

!!       Compute the cross product of velocity and vorticity 
!!       and add the rotational and buoyancy  terms.         
         T1 = (v(:,:,:,N)*zetaz - w(:,:,:,N)*zetay) + Rot*v(:,:,:,N)
         T2 = (w(:,:,:,N)*zetax - u(:,:,:,N)*zetaz) - Rot*u(:,:,:,N)
         T3 = (u(:,:,:,N)*zetay - v(:,:,:,N)*zetax) - Ri*pd(:,:,:)         

!!       Compute velocity-scalar products for s1.         
         s1_times_u = s1(:,:,:,N)*u(:,:,:,N)
         s1_times_v = s1(:,:,:,N)*v(:,:,:,N)
         s1_times_w = s1(:,:,:,N)*w(:,:,:,N)

!!       Compute velocity-scalar products for s2.         
         s2_times_u = s2(:,:,:,N)*u(:,:,:,N)
         s2_times_v = s2(:,:,:,N)*v(:,:,:,N)
         s2_times_w = s2(:,:,:,N)*w(:,:,:,N)

!!       Compute the (negative of) velocity dot mean gradient terms.         
         u_dot_grad_s1_bar = -( w(:,:,:,N)*dz_s1_bar ) 
         u_dot_grad_s2_bar = -( w(:,:,:,N)*dz_s2_bar ) 
         
	    call mpi_barrier(comm,ierr)  ! processor numprocs-1 may have had an xtra plane to process
	    call toc(clock_start(nonlin_id),rate,secs(nonlin_id))         
        cumsecs(nonlin_id)=cumsecs(nonlin_id) + secs(nonlin_id)
        numcalls(nonlin_id)=numcalls(nonlin_id)+1
        return
        end subroutine nonlin
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
        subroutine equation_of_state

!     Routine to calculate d'less pd from values of d'less s1 and s2
!     and their reference profiles s1_bar(x,z) and s2_bar(x,z)
!     The "perturbation" density is defined relative to the reference
!     profile rho_bar(x,z). 
!======================================================================
!     scalars='TS'
!     EOS from MILLERO et.al. (1980) DEEP-SEA RES.
!     The scalar s1=temparature in deg C and 
!     the scalar s2 is salinity in psu (i.e. ~ 36)

      use dimensional_params
      use independent_variables
      use state_variables
      use mpi_params
      use flags_counters_etc
      use user_module
      

      implicit none
      real(kind=8)          :: xx,yy,total_density
      
      PMpa = 0.0   !   compute potential density, not in situ density

      if( trim(scalars) .eq. 'ts' ) then
        xx=1.0    ! use this to keep  Salinity
        yy=1.0    ! use this to keep Temperature
      elseif( trim(scalars) .eq. 'sp'  .or. trim(scalars) .eq. 'S0' ) then
        xx=1.0    ! use this to keep  Salinity
        yy=0.0    ! use this to neglect Temperature
      elseif( trim(scalars) .eq. 'tp'  .or. trim(scalars) .eq. 'T0' ) then
        yy=1.0    ! use this to keep Temperature
        xx=0.0    ! use this to neglect Salinity
      endif

      
      if( scalars .eq. 'ts'                                      &
          .or. scalars .eq. 'tp' .or. scalars .eq. 'sp'          &
             .or. scalars .eq. 't0' .or. scalars .eq. 's0' ) then

      do k=1,locnz
        do j=1,nyplanes
          do i=1,nx
           call eval_s1_bar(z(k),Lz,s1_bar,s1_bar_z,s1_bar_zz)
           call eval_s2_bar(z(k),Lz,s2_bar,s2_bar_z,s2_bar_zz)
           call nondimensionalize_s1(s1_bar,s1_bar_z,s1_bar_zz)
           call nondimensionalize_s2(s2_bar,s2_bar_z,s2_bar_zz)
 
           Temperature = yy*(s1_bar*s1_scale)    ! [deg C]
           Salinity = (s2_bar*s2_scale)          ! [psu]
           Salinity=xx*Salinity/1000.            ! for consistency w/ fortran routine rho
           call rho(Salinity,Temperature,Pmpa,density)
           rho_bar = density                     ! ambient density [kg/m3]

!          now repeat the calculation for total s1 & s2
           Temperature = yy*((s1_bar + s1(i,j,k,N))*s1_scale) ! dimensional quantity: total Temp.
 
           Salinity = ((s2_bar + s2(i,j,k,N))*s2_scale) ! dimensional quantity: total Salinity
                
           Salinity=xx*Salinity/1000.                 ! for consistency w/ fortran routine rho       

           call rho(Salinity,Temperature,PMpa,density) ! total density [kg/m3]
           total_density = density                    ! [kg/m3]

           pd(i,j,k) = total_density - rho_bar   ! perturbation in kg/m3
           pd(i,j,k) = pd(i,j,k)/(Lz*DGRAD)      ! dimensionless perturbation density

         enddo
        enddo
       enddo  

      elseif( scalars .eq. 'rp' .or. scalars .eq. 'r0' ) then  ! s1 is pd
       
        pd(:,:,:) = s1(:,:,:,N)
     
      elseif( scalars .eq. 'pp' ) then ! both scalars passive, set pd=0 

        pd(:,:,:) = 0.0

      endif

      end subroutine equation_of_state
!=============================================================== 
!===============================================================

        !	"rho.f"

        !	Parameter	:	equation of state
        !	Source		:	MILLERO et.al. (1980) DEEP-SEA RES.
        !				Vol 27A, pp 255-264
        !	Units		:	Kg / m**3
        !	Input		:	s (concentration units), t (deg C), p (MPa)
        !	Range		:	s (0 to 0.040), t (-2 to 40), p (MPa)
        !	Ck value	:	K (0, 0, 100)      = 2.297721e4
        !				K (0.035, 0, 100)  = 2.499200e4
        !				K (0, 25, 100)     = 2.54051e4
        !				K (0.035, 25, 100) = 2.710895e4
        !	Coded		:	Ngoc Dang, July 1982

        subroutine rho (s, t, p, density)

        double precision s, t, p
        double precision a, b, c, d, e, k, t1, t2, t3, t4, ta, tb
        double precision k0, k0w, alpha0, alpha, rho0, aw, bw, density 


        !  convert input units

        s = 1000.0d0 * s
        p = 10.0d0 * p

        !  compute alpha0 = alpha (s, t, 0)

        t1 = 9.99841594d-1 + t * (6.793952d-5 + t * (-9.095290d-6 + t * &
             (1.001685d-7 + t * (-1.120083d-9 + t * 6.536332d-12))))
     
        t2 = 8.25917d-4 + t * (-4.4490d-6 + t * (1.0485d-7 + t *    &   
             (-1.258d-9 + t * 3.315d-12)))
             
        t3 = -6.33761d-6 + t * (2.8441d-7 + t * (-1.6871d-8 + t *   &   
             2.83258d-10))
             
        t4 = 5.4705d-7 + t * (-1.97975d-8 + t * (1.6641d-9 - t *    &   
             3.1203d-11))
     
        rho0 = t1 + t2 * s + t3 * s**(1.5) + t4 * s**2
        alpha0 = 1.0d0 / rho0

        !  compute the pressure coefficient

        !	the first term
        k0w = 1.965221d4 + t * (1.484206d2 + t * (-2.327105d0 + t *   &  
              (1.360477d-2 - t * 5.155288d-5)))
              
        ta = 5.46746d1 + t * (-6.03459d-1 + t * (1.09987d-2 - t *     &    
             6.167d-5))
             
        tb = 7.944d-2 + t * (1.6483d-2 - t * 5.3009d-4)
        k0 = k0w + ta * s + tb * s**(1.5)

        !	the second term
        aw = 3.239908d0 + t * (1.43713d-3 + t * (1.16092d-4 - t *  &     
             5.77905d-7))
     
        c = 2.2838d-3 + t * (-1.0981d-5 - t * 1.6078d-6)
        d = 1.9107d-4
        a = aw + c * s + d * s**(1.5)

        !	the third term
        bw = 8.50935d-5 + t * (-6.12293d-6 + t * 5.2787d-8)
        e = -9.9348d-7 + t * (2.0816d-8 + t * 9.1697d-10)
        b = bw + e * s

        k = k0 + a * p + b * p**2

        !  compute density
        alpha = alpha0 * (1.0d0 - p / k)
        density = 1.0d0 / alpha

        !  convert to kg / m**3
        density = 1000.0d0 * density
        s = s / 1000.0d0
        p = p / 10.0d0

        return
        end
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
      subroutine add_forcing
      
!     Optional routine to call user_defined_forcing to evaluate the routine
!     defining the inhomogeneous terms as functions of position & time.

      use flags_counters_etc
      use independent_variables
      use intermediate_variables
      use dimensional_params
      use user_module
      
      implicit none
      
      double precision                     :: scale_factor(5)
      double precision, allocatable        :: phi(:,:,:)
      
      if( trim(force_flag) .ne. 'yes' ) return
      call tic(clock_start(forcing_id),rate)
      allocate( phi(nx+2,nyplanes,5) )
      
      
      !! time variable available:  t_secs
      scale_factor(1:3) = 1./(U0**2/Lz)
      scale_factor(4) = 1./(s1_scale*U0/Lz)
      scale_factor(5) = 1./(s2_scale*U0/Lz)
      
      do k=1,locnz
      
       do j=1,nyplanes
        do i=1,nx
           call user_defined_forcing(x(i), y(j), z(k), t_secs,         &
                                     phi(i,j,1),phi(i,j,2),phi(i,j,3), &
                                     phi(i,j,4),phi(i,j,5),            &
                                     Lx, Ly, Lz, f, g, rho_0 )                         
        enddo
       enddo
       
       T1(:,:,k) =  T1(:,:,k) + phi(:,:,1)*scale_factor(1)
       T2(:,:,k) =  T1(:,:,k) + phi(:,:,2)*scale_factor(2)
       T3(:,:,k) =  T1(:,:,k) + phi(:,:,3)*scale_factor(3)
       u_dot_grad_s1_bar(:,:,k) = u_dot_grad_s1_bar(:,:,k) + phi(:,:,4)*scale_factor(4)
       u_dot_grad_s2_bar(:,:,k) = u_dot_grad_s2_bar(:,:,k) + phi(:,:,5)*scale_factor(5)

      enddo

      deallocate(phi)
      call toc(clock_start(forcing_id),rate,secs(forcing_id))         
      cumsecs(forcing_id)=cumsecs(forcing_id) + secs(forcing_id)
      numcalls(forcing_id)=numcalls(forcing_id)+1
      return
      end subroutine add_forcing
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
  subroutine fft
	  
  use intermediate_variables
  use state_variables
  use flags_counters_etc
  use mpi_params
	  
  implicit none
   
  call tic(clock_start(fft_id),rate) 
  deallocate(zetax,zetay,zetaz)  ! no longer need these
      
  ! forward transforms ( real --> complex )
    if(trim(bc_flag)=='zslip') then	     
          
      call transform3d(T1,T1_cplx,ctype)
      call transform3d(T2,T2_cplx,ctype)
      call transform3d(T3,T3_cplx,stype)
     deallocate(T1,T2,T3)
     
     
      call transform3d(u_dot_grad_s1_bar,u_dot_grad_s1_bar_cplx,stype)
      call transform3d(u_dot_grad_s2_bar,u_dot_grad_s2_bar_cplx,stype)
    deallocate(u_dot_grad_s1_bar,u_dot_grad_s2_bar)
     
     
      call transform3d(s1_times_u,s1_times_u_cplx,stype)
      call transform3d(s1_times_v,s1_times_v_cplx,stype)
      call transform3d(s1_times_w,s1_times_w_cplx,ctype)
      call transform3d(s2_times_u,s2_times_u_cplx,stype)
      call transform3d(s2_times_v,s2_times_v_cplx,stype)
      call transform3d(s2_times_w,s2_times_w_cplx,ctype)
     deallocate(s1_times_u,s1_times_v,s1_times_w,  &
                s2_times_u,s2_times_v,s2_times_w)
                
     if(istep==istart) then
      call transform3d(dzz_s1_bar,dzz_s1_bar_cplx,stype)
      call transform3d(dzz_s2_bar,dzz_s2_bar_cplx,stype)
      deallocate(dzz_s1_bar,dzz_s2_bar)
     endif
     
     
    elseif(trim(bc_flag)=='zperiodic') then
       
      call transform3d(T1,T1_cplx,ftype)
      call transform3d(T2,T2_cplx,ftype)
      call transform3d(T3,T3_cplx,ftype)
     deallocate(T1,T2,T3)
     
      call transform3d(u_dot_grad_s1_bar,u_dot_grad_s1_bar_cplx,ftype)
      call transform3d(u_dot_grad_s2_bar,u_dot_grad_s2_bar_cplx,ftype)
     deallocate(u_dot_grad_s1_bar,u_dot_grad_s2_bar)
     
      call transform3d(s1_times_u,s1_times_u_cplx,ftype)
      call transform3d(s1_times_v,s1_times_v_cplx,ftype)
      call transform3d(s1_times_w,s1_times_w_cplx,ftype)
      call transform3d(s2_times_u,s2_times_u_cplx,ftype)
      call transform3d(s2_times_v,s2_times_v_cplx,ftype)
      call transform3d(s2_times_w,s2_times_w_cplx,ftype)
     deallocate(s1_times_u,s1_times_v,s1_times_w,  &
                s2_times_u,s2_times_v,s2_times_w)
                
     if(istep==istart) then
      call transform3d(dzz_s1_bar,dzz_s1_bar_cplx,ftype)
      call transform3d(dzz_s2_bar,dzz_s2_bar_cplx,ftype)
      deallocate(dzz_s1_bar,dzz_s2_bar)
     endif
     
    endif

    call mpi_barrier(comm,ierr)
    call toc(clock_start(fft_id),rate,secs(fft_id))         
    cumsecs(fft_id)=cumsecs(fft_id) + secs(fft_id)
    numcalls(fft_id)=numcalls(fft_id)+1
    return
    end subroutine fft
!=============================================================== 
!===============================================================



!=============================================================== 
!===============================================================
	subroutine rhs
	
	use intermediate_variables
	use flags_counters_etc
	use independent_variables
	use dimensionless_params
	use mpi_params
	use user_module
	
	implicit none
	real(kind=8)          :: kappa2,diff1,diff2
	
	
	 call tic(clock_start(rhs_id),rate)
     diff1 = 1./(Pr_1*Re)
     diff2 = 1./(Pr_2*Re)
    
    if( bc_flag .eq. 'zslip' ) then  ! sin/cos polarity for z derivatives
      isign = -1
          
	 do k=1,n3
	  do j=1,ny
	   do i=1,locnx
	   
	    kappa2 = (wnx(i)**2) + (wny(j)**2) + (wnz(k)**2) 
	    if(kappa2==0) kappa2=1.e33   ! avoid division by zero
	    
	    !  u, v & w equations: Tr1, Tr2, Tr3

	    Tr1(i,j,k,M) = ((wny(j)**2 + wnz(k)**2)/kappa2) * T1_cplx(i,j,k)     &
                     -  (wnx(i)*wny(j)/kappa2) * T2_cplx(i,j,k)              &
                     +  (cmplxi*wnx(i)*wnz(k)/kappa2) * T3_cplx(i,j,k)        

	    Tr2(i,j,k,M) = - (wnx(i)*wny(j)/kappa2) * T1_cplx(i,j,k)             &
                       +((wnx(i)**2 + wnz(k)**2)/kappa2) * T2_cplx(i,j,k)    &
                       + (cmplxi*wny(j)*wnz(k)/kappa2) * T3_cplx(i,j,k)


	    Tr3(i,j,k,M) = - (cmplxi*wnx(i)*wnz(k)/kappa2) * T1_cplx(i,j,k)      &
                       - (cmplxi*wny(j)*wnz(k)/kappa2) * T2_cplx(i,j,k)      &
                       +((wnx(i)**2 + wny(j)**2)/kappa2) * T3_cplx(i,j,k)
                                  
                       
        Tr4(i,j,k,M) = - cmplxi*wnx(i)*s1_times_u_cplx(i,j,k)                     &
                       - cmplxi*wny(j)*s1_times_v_cplx(i,j,k)                     &
                       - isign*wnz(k)*s1_times_w_cplx(i,j,k)                      &
                       + u_dot_grad_s1_bar_cplx(i,j,k)                            &
                       - (diff1/kappa2) * dzz_s1_bar_cplx(i,j,k)

	    Tr5(i,j,k,M) = - cmplxi*wnx(i)*s2_times_u_cplx(i,j,k)                     &
                       - cmplxi*wny(j)*s2_times_v_cplx(i,j,k)                     &
                       - isign*wnz(k)*s2_times_w_cplx(i,j,k)                      &
                       + u_dot_grad_s2_bar_cplx(i,j,k)                       &
                       - (diff2/kappa2) * dzz_s2_bar_cplx(i,j,k)
	    enddo
	   enddo
	  enddo
	 
	 elseif( bc_flag .eq. 'zperiodic' ) then  ! fourier in 3d, --> d/dx --> ik
     isign =  1
          
	 do k=1,n3
	  do j=1,ny
	   do i=1,locnx	 
	   	   
	    kappa2 = (wnx(i)**2) + (wny(j)**2) + (wnz(k)**2) 
	    if(kappa2==0) kappa2=1.e33   ! avoid division by zero
	    
	    !  u, v & w equations: Tr1, Tr2, Tr3

	    Tr1(i,j,k,M) = ((wny(j)**2 + wnz(k)**2)/kappa2) * T1_cplx(i,j,k)     &
                     -  (wnx(i)*wny(j)/kappa2) * T2_cplx(i,j,k)              &
                     -  (wnx(i)*wnz(k)/kappa2) * T3_cplx(i,j,k)        

	    Tr2(i,j,k,M) = - (wnx(i)*wny(j)/kappa2) * T1_cplx(i,j,k)             &
                       +((wnx(i)**2 + wnz(k)**2)/kappa2) * T2_cplx(i,j,k)    &
                       - (wny(j)*wnz(k)/kappa2) * T3_cplx(i,j,k)


	    Tr3(i,j,k,M) = - (wnx(i)*wnz(k)/kappa2) * T1_cplx(i,j,k)             &
                       - (wny(j)*wnz(k)/kappa2) * T2_cplx(i,j,k)             &
                       +((wnx(i)**2 + wny(j)**2)/kappa2) * T3_cplx(i,j,k)
                       
        Tr4(i,j,k,M) = - cmplxi*wnx(i)*s1_times_u_cplx(i,j,k)                     &
                       - cmplxi*wny(j)*s1_times_v_cplx(i,j,k)                     &
                       - cmplxi*wnz(k)*s1_times_w_cplx(i,j,k)                     &
                       + u_dot_grad_s1_bar_cplx(i,j,k)                       &
                       - (diff1/kappa2) * dzz_s1_bar_cplx(i,j,k)

	    Tr5(i,j,k,M) = - cmplxi*wnx(i)*s2_times_u_cplx(i,j,k)                     &
                       - cmplxi*wny(j)*s2_times_v_cplx(i,j,k)                     &
                       - cmplxi*wnz(k)*s2_times_w_cplx(i,j,k)                     &
                       + u_dot_grad_s2_bar_cplx(i,j,k)                       &
                       - (diff2/kappa2) * dzz_s2_bar_cplx(i,j,k)
	    enddo
	   enddo
	  enddo                       
	 endif

     if( myid .eq. 0 ) then              ! 1,1,1 location corresponds to (k,l,m)=(0,0,0)
      Tr1(1,1,1,M) = T1_cplx(1,1,1)      !  Rot*v
      Tr2(1,1,1,M) = T2_cplx(1,1,1)      ! -Rot*u
      Tr3(1,1,1,M) = 0.                  ! identically disallow net vertical velocity
     endif

     call toc(clock_start(rhs_id),rate,secs(rhs_id))         
     cumsecs(rhs_id)=cumsecs(rhs_id) + secs(rhs_id)
     numcalls(rhs_id)=numcalls(rhs_id)+1
      
     return
     end subroutine rhs
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
       subroutine step
       
       use state_variables
       use intermediate_variables
       use independent_variables
       use flags_counters_etc

!      Routine to take a single explicit time step via the 
!      Euler, AB2, AB3 methods. Dissipation/diffusion terms
!      are integrated analytically via the integrating factors
!      exp( 1/Re kappa^2 dt ) and exp( 1/PrRe kappa^2 dt ) 
!      The differential equations are as follows:
!
!                  d/dt (amp*u)   = amp*Tr1
!                  d/dt (amp*v)   = amp*Tr2
!                  d/dt (amp*w)   = amp*Tr3
!                  d/dt (amp*s1)  = amp*Tr4
!
!              M,MM1,MM2    integer variables pointing to 
!                           M, M-1, M-2 storage locations
!              N,NP1        integer variables pointing to 
!                           N, N+1 storage locations
!              Tr1(:,:,:,M),Tr1(:,:,:,MM1),Tr1(:,:,:,MM2)  Tr1 at t,t-dt,t-2*dt etc
!              u_cplx(:,:,:,N),u_cplx(:,:,:,NP1)  values of u at time t,t+dt  etc
!              amp(:,:,:,3)   exp(-1/Re kappa^2 dt ) and 
!                             exp(-1/Pr_1*Re kappa^2 dt )  exp(-1/Pr_2*Re kappa^2 dt )
!                             where kappa^2 = kx^2 + ky^2 + kz^2 
!              dt             time increment
!              step_flag      "euler","AB2","AB3" 
!                             character flag indicating method to apply


       implicit none
       integer             :: NP1
       double precision    :: dt2,dt12
       
       call tic(clock_start(step_id),rate)
       NP1=NM1    ! current values in locations N
                     
       if( step_flag .eq. 'euler' ) then
                      
           u_cplx(:,:,:,NP1) = amp(:,:,:,1)*( u_cplx(:,:,:,N) + dt*Tr1(:,:,:,M) )
           v_cplx(:,:,:,NP1) = amp(:,:,:,1)*( v_cplx(:,:,:,N) + dt*Tr2(:,:,:,M) )
           w_cplx(:,:,:,NP1) = amp(:,:,:,1)*( w_cplx(:,:,:,N) + dt*Tr3(:,:,:,M) )
           s1_cplx(:,:,:,NP1)= amp(:,:,:,2)*( s1_cplx(:,:,:,N)+ dt*Tr4(:,:,:,M) )
           s2_cplx(:,:,:,NP1)= amp(:,:,:,3)*( s2_cplx(:,:,:,N)+ dt*Tr5(:,:,:,M) )
           step_flag = 'AB2'    ! next time use AB2
        
       elseif( step_flag .eq. 'AB2' ) then
        dt2=dt/2.
           u_cplx(:,:,:,NP1)=  amp(:,:,:,1)*(u_cplx(:,:,:,N) + dt2*(3*Tr1(:,:,:,M)-amp(:,:,:,1)*Tr1(:,:,:,MM1)))
           v_cplx(:,:,:,NP1)=  amp(:,:,:,1)*(v_cplx(:,:,:,N) + dt2*(3*Tr2(:,:,:,M)-amp(:,:,:,1)*Tr2(:,:,:,MM1)))
           w_cplx(:,:,:,NP1)=  amp(:,:,:,1)*(w_cplx(:,:,:,N) + dt2*(3*Tr3(:,:,:,M)-amp(:,:,:,1)*Tr3(:,:,:,MM1)))
           s1_cplx(:,:,:,NP1)=amp(:,:,:,2)*(s1_cplx(:,:,:,N) + dt2*(3*Tr4(:,:,:,M)-amp(:,:,:,2)*Tr4(:,:,:,MM1)))
           s2_cplx(:,:,:,NP1)=amp(:,:,:,3)*(s2_cplx(:,:,:,N) + dt2*(3*Tr5(:,:,:,M)-amp(:,:,:,3)*Tr5(:,:,:,MM1)))
           step_flag = 'AB3'    ! next time use AB3
        
       elseif( step_flag .eq. 'AB3' ) then
        dt12=dt/12.
           u_cplx(:,:,:,NP1) = amp(:,:,:,1)*( u_cplx(:,:,:,N) +  &
              dt12*(23*Tr1(:,:,:,M)-16*amp(:,:,:,1)*Tr1(:,:,:,MM1)+5*(amp(:,:,:,1)**2)*Tr1(:,:,:,MM2) )  )
           v_cplx(:,:,:,NP1) = amp(:,:,:,1)*( v_cplx(:,:,:,N) +  &
              dt12*(23*Tr2(:,:,:,M)-16*amp(:,:,:,1)*Tr2(:,:,:,MM1)+5*(amp(:,:,:,1)**2)*Tr2(:,:,:,MM2) )  )
           w_cplx(:,:,:,NP1) = amp(:,:,:,1)*( w_cplx(:,:,:,N) +  &
              dt12*(23*Tr3(:,:,:,M)-16*amp(:,:,:,1)*Tr3(:,:,:,MM1)+5*(amp(:,:,:,1)**2)*Tr3(:,:,:,MM2) )  )
           s1_cplx(:,:,:,NP1)= amp(:,:,:,2)*( s1_cplx(:,:,:,N)+  &
              dt12*(23*Tr4(:,:,:,M)-16*amp(:,:,:,2)*Tr4(:,:,:,MM1)+5*(amp(:,:,:,2)**2)*Tr4(:,:,:,MM2) )  )
           s2_cplx(:,:,:,NP1)= amp(:,:,:,3)*( s2_cplx(:,:,:,N)+  &
              dt12*(23*Tr5(:,:,:,M)-16*amp(:,:,:,3)*Tr5(:,:,:,MM1)+5*(amp(:,:,:,3)**2)*Tr5(:,:,:,MM2) )  )
       endif

       call toc(clock_start(step_id),rate,secs(step_id))         
       cumsecs(step_id)=cumsecs(step_id) + secs(step_id)
       numcalls(step_id)=numcalls(step_id)+1
     
       return 
       end subroutine step
!=============================================================== 
!===============================================================



!=============================================================== 
!===============================================================
  subroutine advance_particles
   use flags_counters_etc
   use independent_variables
   use dimensional_params
   use particles
   use mpi_params
   implicit none
   real(kind=8)              :: dt2,dt12
   integer                   :: mypts
   integer, allocatable      :: my_labels(:)
   real(kind=8), allocatable :: xvals(:)
   real(kind=8), allocatable :: yvals(:)
   real(kind=8), allocatable :: zvals(:)
   
   if(ntrajectories<=0) return
   
   call tic(clock_start(advance_particles_id),rate)
   
   call mpi_barrier(comm,ierr)
   call get_ghosted_velocity   !! e.g. gvelocity(0:nx+1,0:ny+1,0:locnz+1,3)  
   
   allocate( my_labels(ntrajectories) )
   allocate( xvals(ntrajectories) )
   allocate( yvals(ntrajectories) )
   allocate( zvals(ntrajectories) )
   my_labels=0  ;  xvals=0.d0  ;  yvals=0.d0    ;   zvals=0.d0
      
   call update_owners !! match current positions w/ processors
   
   call find_my_particles(my_labels,xvals,yvals,zvals,mypts) !! get position/vel for myid particles
    
   call spline_interp(xvals,yvals,zvals,mypts,my_labels) !! get u,v,w at the positions owned by myid
    
   if( particle_step_flag .eq. 'euler' ) then
   
    do i=1,mypts  !! loop through the particles on processor myid
     j=my_labels(i)     
     positions(j,1)=positions(j,1) + dt*uvels(j,M)    
     positions(j,2)=positions(j,2) + dt*vvels(j,M)
     positions(j,3)=positions(j,3) + dt*wvels(j,M)
    enddo
    particle_step_flag='AB2'    
    
   elseif( particle_step_flag .eq. 'AB2' ) then
    dt2=dt/2.
    do i=1,mypts  !! loop through the particles on processor myid
     j=my_labels(i)
     positions(j,1)=positions(j,1) + dt2*( 3.*uvels(j,M) - uvels(j,MM1) )
     positions(j,2)=positions(j,2) + dt2*( 3.*vvels(j,M) - vvels(j,MM1) )
     positions(j,3)=positions(j,3) + dt2*( 3.*wvels(j,M) - wvels(j,MM1) )
    enddo
    particle_step_flag='AB3'
    
   elseif( particle_step_flag .eq. 'AB3' ) then   
    dt12=dt/12.
    do i=1,mypts  !! loop through the particles on processor myid
     j=my_labels(i)
     positions(j,1)=positions(j,1) + dt12*(23.*uvels(j,M) - 16.*uvels(j,MM1) + 5.*uvels(j,MM2) )
     positions(j,2)=positions(j,2) + dt12*(23.*vvels(j,M) - 16.*vvels(j,MM1) + 5.*vvels(j,MM2) )
     positions(j,3)=positions(j,3) + dt12*(23.*wvels(j,M) - 16.*wvels(j,MM1) + 5.*wvels(j,MM2) )
    enddo    
   endif

   if(numprocs>1)  call globally_update_particle_data
    
   deallocate( my_labels,xvals,yvals,zvals )
    
   call toc(clock_start(advance_particles_id),rate,secs(advance_particles_id)) 
   cumsecs(advance_particles_id)=cumsecs(advance_particles_id) + secs(advance_particles_id)
   numcalls(advance_particles_id)=numcalls(advance_particles_id)+1
   
   if(myid==0) then
    open(1,file='output/track',position='append')
      write(1,*) t_secs,positions(1,1:3)*Lz
    close(1)
   endif
   
 end subroutine advance_particles
 



