!!==========================================================
!!==========================================================
module user_routines
!! module for user-customized subroutines and functions
use user_data               
implicit none
public   :: user_ics                 !!routine defining initial conditions
public   :: user_forcing             !!routine defining any external forcing
public   :: user_particle_positions  !!routine defining initial locations
public   :: UserBdryData             !!routine for specifying boundary-condition data
public   :: eval_s1_bar              !!routine defining ambient s1 profile & z derivs
public   :: eval_s2_bar              !!routine defining ambient s2 profile & z derivs
public   :: map_s2x_at_gridpoints    !!routine to define mapping for distributing gridpoints
public   :: point_map_x2s            !!routine to evaluate map at given point x
public   :: point_map_s2x            !!routine to evaluate map at given point s

!!==========================================================    
Contains
!!==========================================================

!!==========================================================
subroutine user_ics(xval,yval,zval,Fvals,id,nx,ny,nz)
!!===========================================================
!!  User defined subroutine to specify initial conditions.
!!  Inputs and outputs are all in dimensional units.
!!  
!!
!!  inputs:
!!   x,y,z   coordinate arrays (in meters) that define 
!!           the region over which forcing functions
!!           are to be evaluated
!! 
!!   id      field id [u,v,w,s1',s2'] <---[1,2,3,4,5]
!!             where s1 = s1bar(z) + s1'(x,y,z,t)
!!             and   s2 = s2bar(z) + s2'(x,y,z,t)
!!
!!  outputs:
!!          F(:,:,:) the initial conditions for variable id
!!          F is to be returned in DIMENSIONAL units 
!!          (i.e. m/s or deg C etc )
!!   
!!===========================================================
 use mpi_params
 use user_data, only: Lx,Ly,Lz,f,N2,dgrad,rho_0,      &
                      g,beta_dimensional,runlabel  
 implicit none
 real(kind=8)           ::  xval(nx)
 real(kind=8)           ::  yval(ny)
 real(kind=8)           ::  zval(nz)
 real(kind=8)           ::  Fvals(nx,ny,nz)
 integer                ::  id,ierr
 integer                ::  nx,ny,nz 
 integer                ::  i,j,k,kk
 
 !! rigid_lid_IW
 real(kind=8)     :: omega,omega2,neg_dgrad,A,f2,pi
 real(kind=8)     :: kx,kx2,ky,ky2,kz,kz2,C(2,2)
 pi=4.*atan(1.)
 
 
  f2=f*f 
  kx=2.*pi/Lx
  if(ny==1) then
   ky=0.*pi/Ly
  else
   ky=2.*pi/Ly
  endif
  kz=pi/Lz
 
  kx2=kx*kx
  ky2=ky*ky
  kz2=kz*kz
 
  omega2=((kx2+ky2)*N2 + kz2*f2)  &
         / (kx2 + ky2 + kz2)
  omega=sqrt(omega2)
 
  neg_dgrad=-(dgrad)   !! NOTE ======>> - <<======== sign !!!!
 
  A = 0.01
 
  C(1,1)=A*kx*kz*omega/(kx2+ky2)
  C(2,1)=A*ky*kz*omega/(kx2+ky2)
  C(1,2)=A*ky*kz*f/(kx2+ky2)
  C(2,2)=A*kx*kz*f/(kx2+ky2)
  
  write(0,*) ' inertial freq  ',f,'   1/s'
  write(0,*) ' BV freq        ',sqrt(N2),'   1/s'
  write(0,*) ' IW freq        ',omega,'   1/s'
  write(0,*) ' IW cph         ',omega*3600.d0/(2.*pi)
  write(0,*) ' IW per         ',2.*pi/omega,'  secs'
  write(0,*) ' IW per         ',(2.*pi/omega)/60,'  mins'
  write(0,*) ' IW per         ',(2.*pi/omega)/3600,'  hrs'
  write(0,*) ' BV per         ',(2.*pi/sqrt(N2))/60,'  mins'
  write(0,*) ' BV per         ',(2.*pi/sqrt(N2)),'  secs'
  write(0,*) ' Inertial per   ',(2.*pi/f)/3600,'  hrs'
  write(0,*) ' Inertial per   ',(2.*pi/f),'  secs'
  write(0,*) ' horizontal wavelength   ',(2.*pi/sqrt(kx2+ky2)),'  m'
  write(0,*) ' vertical wavelength     ',(2.*pi/sqrt(kz2)),'  m'
    

 select case(id)
 
  case(1)        !! ics for u [m/s]
  
   do i=1,nx
    do j=1,ny
     do k=1,nz
      Fvals(i,j,k) =  C(1,1)*cos(kz*zval(k))*cos(kx*xval(i) + ky*yval(j))     &
                    - C(1,2)*cos(kz*zval(k))*sin(kx*xval(i) + ky*yval(j))
     enddo
    enddo
   enddo
  
  case(2)        !! ics for v [m/s]
  
   do i=1,nx
    do j=1,ny
     do k=1,nz
      Fvals(i,j,k) =  C(2,1)*cos(kz*zval(k))*cos(kx*xval(i) + ky*yval(j))     &
                   +  C(2,2)*cos(kz*zval(k))*sin(kx*xval(i) + ky*yval(j))  
      enddo
    enddo
   enddo
  
  case(3)        !! ics for w [m/s]
  
   do i=1,nx
    do j=1,ny
     do k=1,nz
      Fvals(i,j,k) = A*omega*sin(kz*zval(k))*sin(kx*xval(i) + ky*yval(j))
     enddo
    enddo
   enddo
   
  case(4)        !! ics for s1' 
  
   do i=1,nx
    do j=1,ny
     do k=1,nz
      Fvals(i,j,k) = -A*neg_dgrad*sin(kz*zval(k))*cos(kx*xval(i) + ky*yval(j))
     enddo
    enddo
   enddo
  
  case(5)        !! ics for s2' 
  
   do i=1,nx
    do j=1,ny
     do k=1,nz
       Fvals(i,j,k) = exp(-((zval(k)-Lz)/(Lz/8.))**2 )
     enddo
    enddo
   enddo

  case default
   stop 'error in calling user_ics: id out of range'
 end select

return
end subroutine user_ics
!!========================================================== 
!!==========================================================
      
      
      


!!========================================================== 
!!==========================================================
subroutine user_forcing(x,y,z,t,FF,id,m1,m2,m3)
!!===========================================================
!!  User defined subroutine to add time/space dependent
!!  forcing to the equations of motion. Inputs and outputs
!!  for this routine are all in dimensional units.
!!  
!!
!!  inputs:
!!   x,y,z   coordinate arrays [m] that define 
!!           the region over which forcing
!!           functions are to be evaluated
!!   t       current time in seconds  
!!   id      field id [u,v,w,s1',s2'] <---[1,2,3,4,5]
!!
!!  outputs:
!!          FF(:,:,:) the rhs term for equation id
!!          FF is to be returned in DIMENSIONAL units 
!!          (i.e. m/s2 or deg C/s etc )
!!   
!!===========================================================
 use user_data
 use dependent_variables
 use intermediate_variables
 use etc,  only: MM0
 implicit none
 integer           ::  m1,m2,m3
 real(kind=8)      ::  x(m1)
 real(kind=8)      ::  y(m2)
 real(kind=8)      ::  z(m3)
 real(kind=8)      ::  t
 real(kind=8)      ::  FF(m1,m2,m3)
 integer           ::  id
  
 real(kind=8)  :: xpos,ypos,zpos
 integer       :: i,j,k
 

!!==================================================================
 select case(id)
 
  case(1)   ! u 
   FF=0.d0
  case(2)   ! v
   FF=0.d0
  case(3)   ! w 
   FF=0.d0
  case(4)   ! s1
   FF=0.d0
  case(5)   ! s2 
   FF=0.d0
  
  case default
   stop 'error in calling user_forcing: id out of range'
  
 end select 
return
end subroutine user_forcing
!!==================================================================== 


!!==================================================================== 
subroutine UserBdryData(id,dim,endpt,bdry_type,m1,m2,x1,x2,vals)
!!====================================================================
!!
!!   A boundary condition "type" has been specified for each variable
!!   at each of the six faces [x=0,Lx] [y=0,Ly] [z=0,Lz]
!!
!!   The boundary conditions have the form  
!!   Var(id)(:,:) = vals(:,:) 
!!                      at coordinate(idim)'s BEGINNING/END (Dirichlet)
!!         or
!!   d/dn Var(id)(:,:)=vals(:,:) 
!!                        at coordinate(idim)'s BEGINNING/END (Neumann)
!!
!!   My convention here is that d/dn is equal to d/d(coordinate)
!!   at both ends it is NOT the OUTWARD facing normal.
!!   e.g for dim=1, d/dn=d/dx at both x=0,Lx
!!
!!   The boundary data "vals" must be specified in this routine.
!!
!!   In the example below, all boundary conditions are homogeneous 
!!   and therefore vals(:,:) is set to 0.0 except at the upper rigid lid,
!!   where a stress (neumann) condition is imposed (du/dz=vals(:,:)).
!!        ( upper rigid lid: dim=3, endpt=ENDING, x1<==x, x2<==y )
!!
!!   ==================================================================
!!   INPUT
!!   ==================================================================
!!   id            (1,2,3,4,5) ==> (u,v,w,s1',s2')
!!   dim                    (x,y,z)==>(1,2,3)
!!   endpt              (start,end)==>(1,2)=(BEGINNING,ENDING)
!!   m1,m2              array dims of bdry face
!!   bdry_type           dirichlet, neumann
!!   x1(:),x2(:)         y&z, x&z or x&y    
!!                       for dim=(1,2,3) respectively
!!                       positions given in METERS
!!   ==================================================================
!!   OUTPUT
!!   ==================================================================
!!   vals(_n1,_n2)  user defined boundary values in DIMENSIONAL UNITS
!!   ==================================================================
 use user_data
 implicit none
 integer          :: id  
 integer          :: dim      
 integer          :: endpt 
 integer          :: m1,m2 
 integer          :: i,j
 integer,parameter:: BEGINNING=1,ENDING=2
 character(len=80):: bdry_type       
 real(kind=8)     :: x1(m1),x2(m2) 
 real(kind=8)     :: vals(m1,m2)                                        
 real(kind=8)     :: x,y,z

 
 if(dim==1 .and. xdim_periodic) then
  vals=0.
  return
 endif
 
 if(dim==2 .and. ydim_periodic) then
  vals=0.
  return
 endif
 
 select case(id)  !! value of id determines which block to execute
 
  case(1)               !! bcs for u or du/dn
   vals(:,:) = 0.d0

  case(2)               !! bcs for v or dv/dn
   vals(:,:) = 0.d0
  
  case(3)               !! bcs for w or dw/dn
   vals(:,:) = 0.d0
  
  case(4)               !! bcs for s1 or ds1/dn
   vals(:,:) = 0.d0
  
  case(5)               !! bcs for s2 or ds2/dn
   vals(:,:) = 0.d0
  
 end select
 
return 
end subroutine UserBdryData
      
!!================================================================ 
!!================================================================ 
 subroutine eval_s1_bar(s1_bar,s1_bar_z,s1_bar_zz)  
  use user_data, only: Lz
  use independent_variables,  only: Grid
  use dependent_variables , only : velocity,scalars

!!================================================================ 
!!  Code is written to solve the eqn of motion for s1'
!!   where s1 = s1_bar(z) + s1'(x,y,z,t).
!!   Supply s1_bar and its z derivatives here.
!!   Setting s1_bar=0 is fine if there is no sensible 
!!   ambient distribution of s1 in your problem.
!!
!! z:           vertical z spatial position  [m]
!! Lz:          vertical size of computational domain [m]
!! s1_scale:    characteristic range of s1 values, used to scale s1
!!              units are dimensional, 
!!              e.g. kg/m3 for s1--> perturbation density
!!
!! output 
!! s1_bar, s1_bar_z, s1_bar_zz
!! scalar concentration and z derivs   [dimensional units]
 implicit none
 real(kind=8),dimension(:),intent(inout) :: s1_bar,s1_bar_z,s1_bar_zz
!!=================================================================
!!         USER CUSTOMIZED CODE SEGMENT STARTS  HERE  
!!                    S1_BAR DEPENDs ONLY ON Z  
!!         The meaning of s1 is defined by scalar_kind(1) 
!!         in problem_params (temp,salinity,density,tracer)
!!=================================================================
 integer :: k,my_k
 real(kind=8) :: zpos 
 real(kind=8), parameter    :: dRdz = 40.243
  
  !!==============================
  !! s1 ===> dens in kg/m3
  !!==============================
  do k=1,scalars(1)%DataMap%locnz
     my_k = velocity(1)%DataMap%my_global_kvals(k)
     zpos = Lz*Grid(0)%SpaceDims(3)%x(my_k) ! [m] 
     s1_bar(k) = dRdz*(Lz-zpos)
     s1_bar_z(k) = -dRdz
     s1_bar_zz(k) = 0.0
  enddo

 return
 end subroutine eval_s1_bar
!!================================================================ 
!!================================================================ 






!!================================================================ 
!!================================================================ 
 subroutine eval_s2_bar(s2_bar,s2_bar_z,s2_bar_zz)  
  use user_data
!!================================================================ 
!!  Code is written to solve the eqn of motion for s2'
!!   where s2 = s2_bar(z) + s2'(x,y,z,t).
!!   Supply s2_bar and its z derivatives here.
!!   Setting s2_bar=0 is fine if there is no sensible 
!!   ambient distribution of s2 in your problem.
!!
!! z:           vertical z spatial position  [m]
!! Lz:          vertical size of computational domain [m]
!! s2_scale:    characteristic range of s2 values, used to scale s2
!!              units are dimensional, 
!!              e.g. kg/m3 for s2--> perturbation density
!!
!! output 
!! s2_bar, s2_bar_z, s2_bar_zz
!! scalar concentration and z derivs   [dimensional units]
 implicit none
 real(kind=8),dimension(:),intent(inout) :: s2_bar,s2_bar_z,s2_bar_zz
!!=================================================================
!!         USER CUSTOMIZED CODE SEGMENT STARTS  HERE  
!!                    s2_BAR DEPENDs ONLY ON Z  
!!         The meaning of s2 is defined by scalar_kind(1) 
!!         in problem_params (temp,salinity,density,tracer)
!!=================================================================

  !!==============================
  !! s2 ===> passive tracer
  !!==============================
   s2_bar = 0.d0
   s2_bar_z = 0.d0      
   s2_bar_zz = 0.0
      
 return
end subroutine eval_s2_bar
!!================================================================ 
!!================================================================ 



!!================================================================ 
!!================================================================ 
 subroutine map_s2x_at_gridpoints(idim,periodic,s,Lx,         &
                                  x,dxds,d2xds2,maplabel)
                                  
  use user_data, only: MapParams,runlabel
  implicit none
  integer            :: idim,N
  real(kind=8)       :: a,b,h,Lx
  real(kind=8)       :: s(:),ds,dt,t
  real(kind=8)       :: x(:),pi
  real(kind=8)       :: dxds(:)
  real(kind=8)       :: d2xds2(:)
  character(len=80)  :: maplabel
  logical            :: periodic

!!==================================================
!! user can declare new local variables here 
!!==================================================
  real(kind=8)  :: beta,s0,alpha,gamma,num,denom
  integer       :: i,p,mapid
!!==================================================

  N = ubound(s,1)  !!  # of discrete gridpoints
  pi=4.*atan(1.)
  
  if( periodic ) then
   ds=1./N
  else
   ds=1./(N-1.)
  endif

  !! x endpoints
  a = 0.d0  !! by convention
  b = Lx
  
  !! by convention, parameterize w/ s in [0,1]
  s(:) = (/ (ds*i, i = 0,N-1) /)
  
  !!MAPPING FOR X DIRECTION
  if( idim == 1) then
    mapid=1
        
     if( mapid==1) then
      p=-999
      beta=-999.d0
      s0=-999.d0
     elseif( mapid==2 ) then        
      p = 2        !! controls width of clustered region
                   !! p=3 gives wider region than p=2
      beta = .25*b !! controls severity of clustering  .85
                   !! 0<= beta <=b , as beta->b, --> unclustered grid
      s0 = 0.5     !! sets center of clustered region
     endif
    
    
!!MAPPING FOR Y DIRECTION
  elseif( idim == 2) then

    mapid=1
       
    if( mapid==1) then
     p=-999
     beta=-999.d0
     s0=-999.d0
    elseif( mapid==2 ) then        
     p = 1        !! controls width of clustered region
                  !! p=3 gives wider region than p=2
     beta = .95*b !! controls severity of clustering
                  !! 0<= beta <=b , as beta->b, --> unclustered grid
     s0 = 0.5     !! sets center of clustered region
    endif
    
    

!!MAPPING FOR Z DIRECTION
  elseif( idim == 3) then

    mapid=1   !! simple linear s in [0,1] <--> x in [0,b]
    
    if( mapid==1) then
     p=-999
     beta=-999.d0
     s0=-999.d0
    elseif( mapid==2 ) then
     p = 1        !! controls width of clustered region
                  !! p=3 gives wider region than p=2
     beta = .25*b !! controls severity of clustering
                  !! 0<= beta <=b , as beta->b, --> unclustered grid
     s0 = 1.0     !! sets center of clustered region
    endif
    
    
  elseif( idim == -998 ) then   !! standard chebychev approach
    !! s in [-1,1] <--> x in [0,b]
    !! cheby differentiation will give df/ds, then df/dx=df/ds * ds/dx
    !! the t<-->s transformation is buried in the chebyshev transform/
    !! recursion/transform procedure.
    maplabel = 'cheby'  !! not well tested yet
    dt=pi/(N-1.)
    do i=1,N
     t=(i-1.)*dt          !! parameter t in [0,pi]
     s(i)=-cos(t)         !! s in [-1,1]
    enddo
    
  endif
 !!==================================================================
  
  
  
  if( mapid==1 ) then
   
   !! simple linear mapping [0,1] ->[0,b]
    x(:) =  (b-a) * s(:) + a
    dxds(:) =  (b-a)
    d2xds2(:) = 0.d0
    maplabel='uniform_spacing'
  
  elseif(mapid==2) then
  
    !! some derived quantities    
    num=(2*p+1)*(b-beta);
    denom=b*( ((1-s0)/b)**(2*p+1) + (s0/b)**(2*p+1) )
    alpha = num/denom;
    gamma = (alpha*b/(2*p+1) )*(s0/b)**(2*p+1);
    maplabel='cluster'
    
    !! evaluate the mapping at gridpoints
    do i=1,N
     x(i)=(alpha*b/(2*p+1))*((s(i)-s0)/b)**(2*p+1) + beta*s(i) + gamma
     dxds(i)= alpha*((s(i)-s0)/b)**(2*p) + beta
     d2xds2(i)=(2.*p*alpha/b)*((s(i)-s0)/b)**(2*p-1)
    enddo
            
  elseif(mapid==3) then
      
    !! map x in [0,b] to s in [-1 1]
    x(:) = (b/2.d0)*(s(:)+1.d0)
    dxds(:) = b/2.d0
    d2xds2(:) = 0.d0
    
  endif
  
  !! save the mapping parameters
  MapParams(idim,1) = p   ! save as real even though its an integer
  MapParams(idim,2) = beta
  MapParams(idim,3) = s0
  MapParams(idim,4) = mapid

   
 end subroutine map_s2x_at_gridpoints
!!================================================================ 
!!================================================================ 


!!================================================================ 
!! You don't need to edit below here unless you are defining your 
!! own custom maps, in which case, use  mapid > 3 and use the
!! code below as a guide for doing something similar.
!!================================================================


!!================================================================ 
!!================================================================ 
 subroutine point_map_x2s(x,idim,periodic,Lx,N,s)
  use user_data,     only: MapParams
  implicit none
  real(kind=8)          :: x,s,Lx
  real(kind=8)          :: a,b,h
  integer               :: idim,N
  logical               :: periodic
  
  !!==================================================
  !! user can declare new local variables here 
  !!==================================================
  real(kind=8)  :: beta,s0,tol
  integer       :: p,mapid
  !!==================================================
  
  a = 0.d0  !! by convention
  b=Lx
  
  !! user params for coord. map
  p = MapParams(idim,1)  !! convert back to integer
  beta = MapParams(idim,2)
  s0 = MapParams(idim,3)
  mapid = MapParams(idim,4)
  
  tol=1.e-5
  
!! this is the part the user needs to specify
!! must be consistent with mapping functions used
!! in map_s2x_at_gridpoints etc

    if( mapid == 1 ) then       !! simple linear
     s = (x-a)/(b-a)
    elseif( mapid == 2 ) then   !! cluster w/ (p,beta,s0)
     call NewtonCluster(s,p,beta,s0,b,tol,x)
    endif
   
 end subroutine point_map_x2s
!!================================================================ 
!!================================================================ 


!!================================================================ 
!!================================================================ 
 subroutine point_map_s2x(s,idim,periodic,Lx,N,x,x_s)
  use user_data,     only: MapParams
  implicit none
  real(kind=8)          :: x,s,x_s,Lx
  real(kind=8)          :: a,b,h
  integer               :: idim,N
  logical               :: periodic
  
  !!==================================================
  !! user can declare new local variables here 
  !!==================================================
  real(kind=8)  :: beta,s0,num,denom,alpha,gamma
  integer       :: p,mapid
  !!==================================================

  a = 0.d0  !! by convention
  b=Lx
  mapid = MapParams(idim,4)

!! this is the part the user needs to specify
!! must be consistent with mapping functions used
!! in map_s2x_at_gridpoints etc

   if( mapid==1 ) then   !! simple linear map
    x =  (b-a) * s + a
    x_s = b-a
   elseif( mapid==2 ) then
    !! user params for cluster map
    p = MapParams(idim,1)  !! convert back to integer
    beta = MapParams(idim,2)
    s0 = MapParams(idim,3)
    !! derived params for cluster map
    num=(2*p+1)*(b-beta);
    denom=b*( ((1-s0)/b)**(2*p+1) + (s0/b)**(2*p+1) )
    alpha = num/denom;
    gamma = (alpha*b/(2*p+1) )*(s0/b)**(2*p+1);
    x = (alpha*b/(2*p+1))*((s-s0)/b)**(2*p+1) + beta*s + gamma
    x_s = (alpha)*((s-s0)/b)**(2*p) + beta
   endif


 end subroutine point_map_s2x
!!================================================================ 
!!================================================================ 

!!================================================================ 
!!================================================================
 subroutine NewtonCluster(s,p,beta,s0,b,tol,x) 
  implicit none
  real(kind=8)      :: s,beta,s0,b
  real(kind=8)      :: alpha,gamma,tol
  real(kind=8)      :: F,df,x,ds,num,denom
  integer           :: k,kmax=500,p

  if( x==0 ) then
   s=0; return
  else
   s=b/(x)  !! initial guess
  endif 
  
  num=(2*p+1)*(b-beta);
  denom=b*( ((1-s0)/b)**(2*p+1) + (s0/b)**(2*p+1) )
  alpha = num/denom;
  gamma = (alpha*b/(2*p+1) )*(s0/b)**(2*p+1)
    
  do k=1,kmax 
   F = x - (alpha*b/(2*p+1))*((s-s0)/b)**(2*p+1) + beta*s + gamma
   df = -( alpha*((s-s0)/b)**(2*p) + beta )
   ds = - F/dF
   s = s + ds
   !write(0,*) F,df,x
   if( abs(F) < tol .or. abs(ds) < tol ) exit
  enddo
  if( abs(F) > tol .and. abs(ds) > tol )  &
    stop 'No convergence NewtonCluster'
 end subroutine NewtonCluster
!!================================================================ 
!!================================================================


 
!!================================================================ 
!!================================================================ 
subroutine user_particle_positions(locations,np,Lx,Ly,Lz)
 !!   locations(i,j)  i=1,np
 !!                   j=1,2,3  
 !!                   for x,y,z in 
 !!                   (0,Lx),(0,Ly),(0,Lz) [m]
 !!  urv <---> uniformly distributed random variable in [0,1]
 implicit none
 integer           :: i, j, np
 real(kind=8)      :: Lx, Ly, Lz
 real(kind=8)      :: locations(np,3)
 real(kind=8)      :: urv,sides(3)
 
 call RANDOM_SEED  !! I'm using the F90 intrinsic functions
                   !! RANDOM_SEED, RANDOM_NUMBER here
 sides(:) = (/Lx,Ly,Lz/)
  do i=1,np       
   do j=1,3  
     call RANDOM_NUMBER( urv )
     locations(i,1) = urv*sides(j)  
   enddo      
  enddo 
 return
 
end subroutine user_particle_positions
!!================================================================ 
!!================================================================ 

!======================================================================
!======================================================================
subroutine eval_surface(x,h,h_x,h_y)
 use user_data,  only: Lx,Ly,Lz
 use mpi_params, only: myid,numprocs,comm
 implicit none
 integer           :: i,pid,ierr
 real(kind=8)      :: x(2),h,h_x,h_y,junk
 real(kind=8)      :: xval,yval
 real(kind=8),save :: sigmax,x0,h0,floor
 logical,save      :: first_entry=.TRUE.
!======================================================================
!     inputs are DIMENSIONLESS x,y pts --> x(2)
!     output is  corresponding DIMENSIONLESS surface height
!     and, if requested, the partials h_x and h_y
!
!     (x,y,h(x,y)) is the point on surface z=h(x,y)
       xval=Lz*x(1)
       yval=Lz*x(2)
!======================================================================
!
 if( first_entry ) then
  !! read input file one processor at a time
  do pid=0,numprocs-1   
   if(pid==myid) then
     open(1,file='input/forcing_params.nliwi',position='rewind')
       !! skip over 1st 6 vals (ampl,freq & sponge params)
       do i=1,6
        read(1,*) junk
       enddo
       !! read the topo params
       read(1,*) x0
       read(1,*) floor
       read(1,*) h0
       read(1,*) sigmax
     close(1)
   endif   
   call mpi_barrier(comm,ierr)       
  enddo
  first_entry=.FALSE.
 endif
                    
      h = ( floor + (h0)*exp(-((xval-x0)/(sigmax) )**2) )/Lz   !! [dless]
      
      h_x = -(2.*h0/(sigmax**2))*((xval-x0))  &
             *exp(-((xval-x0)/(sigmax) )**2)                   !! [dless]
             
      h_y = 0.d0                                               !! [dless]
      
      
 return
end subroutine eval_surface
!======================================================================
!======================================================================


 
end module user_routines
