Module class_BoundaryOperator  
  use class_DataDecomposition
  use class_TensorProductGrid
  use class_CoordinateMap
  use class_ScalarVariable
  
  implicit none
  public  :: NewBoundaryOperator
  public  :: UpdateBoundaryOperator
  public  :: DeleteBoundaryOperator
  public  :: ApplyB
  public  :: SwitchType

 
 !!A class containing data and algorithms for applying
 !!boundary conditions to 3d data "cube" at a specified boundary.
 
 type BoundaryOperator   
  character(len=80)        :: label             !! Dirichlet,Neumann,Mixed,Periodic
  integer                  :: order             !! order of 1-sided deriv operator
  integer                  :: DimId             !! Which coordinate boundary operator B acts on (1,2,3)
  integer                  :: EndId             !! Whether B acts at initial(0) or final(1)  point of coord. range
  real(kind=8),pointer     :: alpha(:,:)        !! alpha --> B(phi) = alpha*phi * beta * phi'  
  real(kind=8),pointer     :: beta(:,:)         !! beta  --> B(phi) = alpha*phi * beta * phi'  
  real(kind=8),pointer     :: BoundaryData(:,:) !! data such that the boundary condition is B(phi) = data
  integer                  :: n1,n2             !! (n1,n2) --> array size for alpha,beta,BoundaryData
  type( CoordinateMap )    :: CoordMap          !! CoordMap associated with dimension DimId  
  logical                  :: initialized
 end type BoundaryOperator
  
Contains                      ! member functionality

 subroutine DeleteBoundaryOperator(B)
  type( BoundaryOperator ),intent(inout)  :: B
  deallocate(B%alpha)
  deallocate(B%beta)
  deallocate(B%BoundaryData)
  call DeleteCoordinateMap( B%CoordMap )
 end subroutine DeleteBoundaryOperator
 
 subroutine NewBoundaryOperator  &
           (label,order,DimId,EndId,Grid,DataMap,Val,B)
  use class_DataDecomposition
  use class_TensorProductGrid
  use class_CoordinateMap
  implicit none
  
  type( BoundaryOperator ),intent(inout)  :: B
  type( TensorProductGrid ),intent(in)    :: Grid
  type( DataDecomposition ),intent(in)    :: DataMap
  real(kind=8),intent(in)                 :: Val
  integer,intent(in)                      :: order
  integer,intent(in)                      :: DimId,EndId
  character( len=80 ),intent(inout)       :: label
  integer                                 :: n1,n2
  integer                                 :: myid,numprocs
  integer                                 :: ierr,rcode(3)=0
  
  include 'mpif.h'
  
  call mpi_comm_rank(MPI_COMM_WORLD,myid,ierr)      
  call mpi_comm_size(MPI_COMM_WORLD,numprocs,ierr)
  
  if(trim(label)=='Dirichlet') label='dirichlet'
  if(trim(label)=='Dirchlet') label='dirichlet'
  if(trim(label)=='dirchlet') label='dirichlet'
  if(trim(label)=='Neumann') label='neumann'
  if(trim(label)=='Neuman') label='neumann'
  if(trim(label)=='neuman') label='neumann'
  if(trim(label)=='Periodic') label='periodic'
  if(trim(label)=='Mixed') label='mixed'
  
  if( B%initialized ) then
   !if(myid==0) write(0,*) ' Boundary operator exists -- immediate return'
   !if(myid==0) write(0,*) ' dim, end ===>',B%DimId,B%EndId
   !if(myid==0) write(0,*) ' label ===> ',B%label
   return !! operator already exixts
  endif
  
  if( EndId==2 ) then
   write(0,*) ' Convention for dimension endids is 0/1 for start/end '
   write(0,*) ' NewBoundaryOperator called w/ EndId=2,  STOPPING NOW '
   stop
  endif
    
  B%label=label
  B%order=order
  B%EndId=EndId 
  B%DimId=DimId
  B%CoordMap  =     &
     NewCoordinateMap(Grid%SpaceDims(DimId)%npts, &
                      Grid%SpaceDims(DimId)%label,&
                      Grid%SpaceDims(DimId)%x,    &
                      Grid%SpaceDims(DimId)%s,    &
                      Grid%SpaceDims(DimId)%dxds, &
                      Grid%SpaceDims(DimId)%d2xds2,&
                      Grid%SpaceDims(DimId)%L)
     
  select case(DimId)
   case(1)
    n1=DataMap%SlabDims(2,myid)
    n2=DataMap%SlabDims(3,myid)
   case(2)
    n1=DataMap%SlabDims(1,myid)
    n2=DataMap%SlabDims(3,myid)
   case(3)
    n1=DataMap%SlabDims(1,myid)
    n2=DataMap%SlabDims(2,myid)
  end select
  
  B%n1=n1
  B%n2=n2
  
  allocate( B%alpha(n1,n2), stat=rcode(1) )
  allocate( B%beta(n1,n2), stat=rcode(2) )
  allocate( B%BoundaryData(n1,n2), stat=rcode(3) )
  if( any( rcode /= 0 ) ) Stop 'Allocation failed: NewBoundaryOperator'
  
   if(label == 'dirichlet') then
    B%alpha(:,:)=1.
    B%beta(:,:)=0.
    B%BoundaryData(:,:)=Val !! simple init, set data to constant
     
   elseif(label == 'neumann') then
    B%alpha(:,:)=0.
    B%beta(:,:)=1.
    B%BoundaryData(:,:)=Val !! simple init, set data to constant
        
   elseif(label == 'mixed') then
    B%alpha(:,:)=1.
    B%beta(:,:)=1.
    B%BoundaryData(:,:)=Val !! simple init, set data to constant
     
   elseif(label == 'periodic') then
     !! no values
   else
    write(0,*) 'Boundary operator incorrectly specified'
    stop    
   endif  
  B%initialized=.TRUE.
 end subroutine NewBoundaryOperator
 
 subroutine UpdateBoundaryOperator(B,ArrayName,vals,n1,n2)
 !!Use this function to change values associated with an existing
 !!boundary operator. Legal values of label are 'alpha','beta','data'.
 !!This function can be called to enable nontrivial, inhomogeneous conditions.
 !!The BoundaryData could be changed at each time step if, for example, a boundary 
 !!condition were to be time dependent.
  implicit none
  
  integer                  :: n1,n2
  type( BoundaryOperator ) :: B
  character( len=80 )      :: ArrayName
  real(kind=8)             :: vals(n1,n2)
    
  if( B%n1 .ne. n1 .or. B%n2 .ne. n2 ) then
   write(0,*) 'UpdateBoundaryOperator: input array inconsistent with input operator'
   stop
  endif
  
  select case(ArrayName)
   case('alpha')
    B%alpha=vals
   case('beta')
    B%beta=vals
   case('data')
    B%BoundaryData=vals
  end select  
    
 end subroutine UpdateBoundaryOperator
 
 subroutine SwitchType(B,newtype)
 !!Use this function to change values associated with an existing
 !!boundary operator. Legal values of label are 'alpha','beta','data'.
 !!This function can be called to enable nontrivial, inhomogeneous conditions.
 !!The BoundaryData could be changed at each time step if, for example, a boundary 
 !!condition were to be time dependent.
  implicit none
  character(len=80)   :: newtype
  
  type( BoundaryOperator ) :: B
      
  select case(trim(newtype))
   case('neumann')      !! switch to neumann
    B%alpha=0.d0
    B%beta =1.d0
    B%label='neumann'
   case('dirichlet')    !! switch to dirichlet
    B%alpha=1.d0
    B%beta =0.d0
    B%label='dirichlet'
   case('periodic')    !! switch to periodic 
    B%label='periodic'
  end select  
    
 end subroutine SwitchType
 
 
 
 
 subroutine ApplyB(B,phi,B_of_Phi)  
  use class_ScalarVariable
  implicit none 
  type( BoundaryOperator ),intent(in) :: B
  type( ScalarVariable ),intent(inout):: phi
  type( ScalarVariable ),intent(inout):: B_of_phi
  real(kind=8)              :: ds,h,xx
  real(kind=8), allocatable :: Vals(:,:) 
  real(kind=8), allocatable :: A(:)
  integer                   :: dim
  integer                   :: endpt
  integer                   :: i,j,k
  integer                   :: n1,n2
  integer                   :: N
  integer                   :: idx(0:5),jdx
  integer                   :: order,adj_order    
  integer                   :: myid
  integer                   :: ierr
  integer                   :: numprocs
  
  include 'mpif.h'
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)     
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
  
  n1=B%n1
  n2=B%n2
  dim=B%DimId
  endpt=B%EndId  !!convention: 0 / 1 for start / end  
  order=B%order
  N=phi%DataMap%SlabDims(dim,myid)     
  
  !! Reduce order as needed for small grid sizes
  adj_order=minval( (/order,N-2/) )   !!  adjusted order
    
  allocate( A(0:adj_order) )
  allocate( Vals(n1,n2) )   
  
  !! See Abramowitz & Stegun p. 914 
   select case(adj_order)
    case(1) 
     A(0)=-1.0 ; A(1)=1.0
    case(2)
     A(0)=-3.; A(1)=4.; A(2)=-1.
    case(3)
     A(0)=-11.; A(1)=18.; A(2)=-9.; A(3)=2.          
    case(4)
     A(0)=-50.; A(1)=96.; A(2)=-72.; A(3)=32.; A(4)=-6.   
    case(5)
     A(0)=-274.; A(1)=600.; A(2)=-600.; A(3)=400.; A(4)=-150.; A(5)=24.
   end select

  !! Set up coefficients for general, mixed bc representation
  !! Same discretizations used at each end, except forward/backward
  !! property which introduces a reversal of order & a 
  !! sign change to the coefficients
  if( endpt == 0 ) then  
   jdx=1     !! *START* of coord. range 
   idx(0:5)=(/1,2,3,4,5,6/)
   if( dim==3 .and. myid .ne. 0 ) stop 'wrong processor applying z=0 bdry operator'
  elseif( endpt == 1 ) then
   jdx=N     !! *END* of coord. range
   idx(0:5)=(/N,N-1,N-2,N-3,N-4,N-5/)
   A=-A
   if( dim==3 .and. myid .ne. numprocs-1 ) stop 'wrong processor applying z=Lz bdry operator'
  else
   stop 'Error in designating endpoint in ApplyB (B%endpt)'
  endif
    
  !! Discrete derivatives are expressed w/ respect to the uniformly
  !! spaced parameter s. What's really needed (for Neumann) are derivatives
  !! with respect to x. The stretching factor dx/ds is used to convert the
  !! s derivative to an x derivative. I need to think a bit about endpt
  !! singularities in mappings.
     
  if( phi%Grid%SpaceDims(dim)%label == 'StandardCheby' ) then
  !! use unevenly spaced formula in x, bypass x<->s altogether 
  !! as the transformation dx/ds is singular at the ends
       
    h=phi%Grid%SpaceDims(dim)%ds
    stop 'bdry formulas for case cheby not yet implemented in ClassBoundaryOperator'
      
  else
     
    !! h = ds*dx/ds  ---> results are Neumann BCs expressed in x not parameter s       
    h=phi%Grid%SpaceDims(dim)%ds * phi%Grid%SpaceDims(dim)%dxds(jdx)
      
  endif

  if(adj_order==1) xx=1./(h)
  if(adj_order==2) xx=1./(2.*h)
  if(adj_order==3) xx=1./(3.*2.*h)
  if(adj_order==4) xx=1./(4.*3.*2.*h)
  if(adj_order==5) xx=1./(5.*4.*3.*2.*h)

  !! Set up the discretization coefficients for the Neumann
  !! part of the boundary operator: B[f] = alpha*f + beta*df/dx
  !! Convention: the Neumann part is the derivative with respect
  !! to the coordinate, say x,y,z, not the parameter s and not the
  !! "outward facing unit normal".
  !! Mixed form B_of_Phi = alpha f + beta df/dx 
  !! dfdx(0) = (1/h)(-f0 + f1) + O(h2) 
     
  Vals(:,:)=0.0
  if( dim==1 ) then
   Vals=0.0  
   do i=1,adj_order+1
    do j=1,n1
     do k=1,n2
      Vals(j,k) = Vals(j,k) + A(i-1)*Phi%SlabVals(idx(i-1),j,k)
     enddo
    enddo
   enddo

   
   do j=1,n1
    do k=1,n2
      B_of_Phi%SlabVals(jdx,j,k) = B%alpha(j,k) * Phi%SlabVals(jdx,j,k)  &
                                 + B%beta(j,k) * Vals(j,k) * xx
    enddo
   enddo
   
  elseif( dim==2 ) then
   Vals=0.0  
   do j=1,adj_order+1
    do i=1,n1
     do k=1,n2
      Vals(i,k) = Vals(i,k) + A(j-1)*Phi%SlabVals(i,idx(j-1),k)
     enddo
    enddo
   enddo
   
   do i=1,n1
    do k=1,n2
      B_of_Phi%SlabVals(i,jdx,k) = B%alpha(i,k) * Phi%SlabVals(i,jdx,k)  &
                                 + B%beta(i,k) * Vals(i,k) * xx
    enddo
   enddo
   
                                     
  elseif( dim==3 ) then
  
   Vals=0.0  
   do k=1,adj_order+1
    do i=1,n1
     do j=1,n2
      Vals(i,j) = Vals(i,j) + A(k-1)*Phi%SlabVals(i,j,idx(k-1))
     enddo
    enddo
   enddo
   
   do i=1,n1
    do j=1,n2
      B_of_Phi%SlabVals(i,j,jdx) = B%alpha(i,j) * Phi%SlabVals(i,j,jdx)  &
                                 + B%beta(i,j) * Vals(i,j) * xx
    enddo
   enddo

  endif
  deallocate( Vals )
  deallocate( A )
  
  end subroutine ApplyB
 
end Module class_BoundaryOperator


