Module class_SAOperator  !! Self-Adjoint Operator
use class_TensorProductGrid
use class_DataDecomposition
use class_BoundaryOperator
use class_Derivative
use class_ScalarVariable

!! L[f] = {d/dx( A(x,y,z) df/dx ) + d/dy( B(x,y,z) df/dy ) +  d/dz( C(x,y,z) df/dz )}
!!        in interior
!! B[f] = g on boundaries
implicit none
public   ::    NewSAOperator      !! constructor
public   ::    DeleteSAOperator   !! destructor
public   ::    ApplySAOperator
public   ::    SAGetSizes

type SAOperator 
  type( BoundaryOperator  )   :: BdryOps(3,2)
  integer                     :: level    !! 0 is finest level
  integer                     :: nlevels  
  integer                     :: nrows    !! global number of grdpoints/rows/eqns
  type( TensorProductGrid )   :: Grid
  type( DataDecomposition )   :: DataMap         !! decomposition map
  character (len=80)          :: DerivTypes(6)   !! labels defining differentiation methods
  type( Derivative )          :: Derivs(6)       !! initialized derivatives for each dim
end type SAOperator


Contains            

 subroutine NewSAOperator(f,BdryOps,L,mglevels)
  use class_ScalarVariable
  use class_BoundaryOperator
  use class_TensorProductGrid
  use class_Derivative
  use class_DataDecomposition
  
  implicit none
  type( ScalarVariable ),intent(in)      :: f
  type( TensorProductGrid )              :: Grid
  type( BoundaryOperator  ),intent(in)   :: BdryOps(3,2)
  type( SAOperator ),intent(inout)       :: L
  
  integer                                :: n1,n2,n3
  integer                                :: i,j,k
  integer                                :: mglevels
  integer                                :: myid,ierr
  
  include 'mpif.h'
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr) 
  
  
  n1 = f%DataMap%SlabDims(1,myid)
  n2 = f%DataMap%SlabDims(2,myid)
  n3 = f%DataMap%SlabDims(3,myid)
    
  L%level = f%DataMap%level
  L%nlevels = mglevels
  L%nrows = n1*n2*n3       !! ====> save myid's allocated rows
  do i=1,3
   do j=1,2
    L%BdryOps(i,j) = BdryOps(i,j)
   enddo
  enddo
  L%DataMap = f%DataMap
  L%DerivTypes =  f%DerivTypes
  L%Derivs =  f%Derivs
  L%Grid =  f%Grid
  
 end subroutine NewSAOperator


subroutine DeleteSAOperator(L)
  use class_BoundaryOperator
  use class_TensorProductGrid
  use class_Derivative
  use class_DataDecomposition
  
  type( SAOperator ),intent(inout)       :: L
  integer                                :: i,j
    
  do i=1,3
   do j=1,2
    call DeleteBoundaryOperator( L%BdryOps(i,j) )
   enddo
  enddo
  
  call DeleteDataDecomposition(L%DataMap)
  call DeleteDerivative(L%Derivs(1))
  call DeleteDerivative(L%Derivs(2))
  call DeleteDerivative(L%Derivs(3))
  call DeleteTensorProductGrid(L%Grid)
    
 end subroutine DeleteSAOperator
 
 
 subroutine SAGetSizes(L,nx_local,ny_local,nz_local,  &
                         nx_global,ny_global,nz_global) 
  type( SAOperator ),intent(in)       :: L
  integer     :: nx_local,ny_local,nz_local
  integer     :: nx_global,ny_global,nz_global
  integer     :: myid,ierr
  
  include 'mpif.h'
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)
  
  nx_local = L%DataMap%SlabDims(1,myid)
  ny_local = L%DataMap%SlabDims(2,myid) 
  nz_local = L%DataMap%SlabDims(3,myid) 
  nx_global= L%Grid%SpaceDims(1)%npts
  ny_global= L%Grid%SpaceDims(2)%npts
  nz_global= L%Grid%SpaceDims(3)%npts  
 end subroutine SAGetSizes
  
  
 subroutine ApplySAOperator(L,f,L_of_f,temp1,temp2,c1,c2,c3,grad)
  use class_ScalarVariable
  
  implicit none
  type( SAOperator )                   :: L
  type( ScalarVariable )               :: f,L_of_f
  type( ScalarVariable )               :: temp1,temp2
  type( ScalarVariable ),optional      :: grad(3)
  integer                     :: dim,ierr,i,j,k
  real(kind=8)                :: c1,c2,c3
  integer,SAVE                :: myid,numprocs,nx,ny,locnz
  logical                     :: forward,inverse
  logical,SAVE                :: first_entry=.TRUE.  
  include 'mpif.h'
  
  !!=====================================================
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   nx = L%DataMap%SlabDims(1,myid) 
   ny = L%DataMap%SlabDims(2,myid) 
   locnz = L%DataMap%SlabDims(3,myid)
   first_entry=.FALSE.
  endif
  !!=====================================================
  
  !!=====================================================
  !! If the input is given as expansion coefficients, 
  !! do an inverse transform3d of f to get grdpt values.
  !! Always begin the differentiations from physical
  !! space. Scale and add helmholtz term as necessary.
  !! Return gridpoint values of L[f].
  !!=====================================================
  if( f%wavenumber_values .and. .NOT. f%gridpoint_values ) then
   forward=.FALSE.
   inverse=.TRUE.
   call Transform3D(f,f,temp1,temp2,forward,inverse)
   !! now reset flags before differentiating
   f%gridpoint_values=.TRUE.
   f%wavenumber_values=.FALSE.
   f%forward_transform_only=.FALSE.
   f%inverse_transform_only=.FALSE.
  endif
      
    
  !!==============================================
  !! keep separate copies of boundary values
  !! this step requires that f be in phys. space
  !!==============================================
  f%DirichletWest=f%SlabVals(1,:,:)
  f%DirichletEast=f%SlabVals(nx,:,:)
  if(ny>1) f%DirichletFront=f%SlabVals(:,1,:)
  if(ny>1) f%DirichletBack=f%SlabVals(:,ny,:)
  if(myid==numprocs-1) f%DirichletTop=f%SlabVals(:,:,locnz)
  if(myid==0) f%DirichletBottom=f%SlabVals(:,:,1)
  
  !!===============================================================
  !! Set the flags & derivative types for the 2nd derivatives of f
  !!===============================================================
  temp1%Derivs(1)=f%Derivs(4)              ! method for d/dx (df/dx)
  !temp1%DerivTypes(1)=f%DerivTypes(4)      ! type for d/dx (df/dx)
  
  temp1%Derivs(2)=f%Derivs(5)              ! method for d/dy (df/dy)
  !temp1%DerivTypes(2)=f%DerivTypes(5)      ! type for d/dy (df/dy)
  
  temp1%Derivs(3)=f%Derivs(6)              ! method for d/dz (df/dz)
  !temp1%DerivTypes(3)=f%DerivTypes(6)      ! type for d/dz (df/dz)
    
  temp1%gridpoint_values=.TRUE.
  temp1%wavenumber_values=.FALSE.
  temp1%forward_transform_only=.FALSE.
  temp1%inverse_transform_only=.FALSE.
    
  !!=======================================================
  !! Now take first derivs of f wrt x,y,z storing results
  !! in tmp1. Then take second derivs of temp1 wrt x,y,z,
  !! accumulating the results in L_of_f.
  !!=======================================================
  dim=1
  !!=======================================================
  call Differentiate(f,dim,temp1)
   f%NeumannWest=temp1%SlabVals(1,:,:)
   f%NeumannEast=temp1%SlabVals(nx,:,:)
  call Differentiate(temp1,dim,L_of_f)
  if( present(grad) ) call CopySlabData_a2b(temp1,grad(1)) 
  
  
  !!=======================================================
  dim=2
  !!=======================================================
  if(ny>1) then
   call Differentiate(f,dim,temp1) 
    f%NeumannFront=temp1%SlabVals(:,1,:)
    f%NeumannBack =temp1%SlabVals(:,ny,:)
   call Differentiate(temp1,dim,temp2) 
   L_of_f%SlabVals = L_of_f%SlabVals + temp2%SlabVals
   if( present(grad) ) call CopySlabData_a2b(temp1,grad(2))
  else
   if( present(grad) ) grad(2)%SlabVals=0.d0
  endif
    
  !!=======================================================
  dim=3
  !!=======================================================
  call Differentiate(f,dim,temp1)
   if(myid==numprocs-1) f%NeumannTop=temp1%SlabVals(:,:,locnz)
   if(myid==0)          f%NeumannBottom=temp1%SlabVals(:,:,1)   
  call Differentiate(temp1,dim,temp2)  
  L_of_f%SlabVals = L_of_f%SlabVals + temp2%SlabVals
  if( present(grad) ) call CopySlabData_a2b(temp1,grad(3))
  
    
  !!=================================================================
  !! At this point, we have (f_xx + f_yy + f_zz)
  !! stored in L_of_f  (s<->x transformations completed).
  !! The actual operator L_of_f, however is defined as
  !!
  !!      L_of_f = c3*c1*(f_xx + f_yy + f_zz) + c3*c2*f
  !!
  !!=================================================================
  if( c3*c1 .ne. 1.d0 ) then
   L_of_f%SlabVals = (c3*c1) * L_of_f%SlabVals
  endif
  if( c3*c2 .ne. 0.d0 ) then
   L_of_f%SlabVals = L_of_f%SlabVals + c3*c2*f%SlabVals
  endif
  !!=================================================================
  !! Return the result as gridpoint values in L_of_f.
  !!=================================================================
    
  return      
 end subroutine ApplySAOperator

   
 
end Module class_SAOperator
