Module class_ScalarVariable         !============================================
use class_TensorProductGrid
use class_Derivative
use class_DataDecomposition
  
implicit none
public   ::    NewScalar             !! constructor, only allocates array for Slab
public   ::    DeleteScalarVariable  !! destructor
public   ::    NewVector             !! merge 3 scalars into a ScalarVariable array
public   ::    DeleteVector          !! destructor
public   ::    Norm2                 !! rms
public   ::    NormInf               !! inf norm
public   ::    Differentiate         !! differentiate ScalarVariable
public   ::    Gradient              !! gradient of ScalarVariable
public   ::    Divergence            !! divergence of vector { ScalarVariable(3) }
public   ::    Curl                  !! curl of vector { ScalarVariable(3) }
public   ::    Laplacian             !! gradient of ScalarVariable
public   ::    Transform3D           !! do trigonometric transforms w/ derivs

public   ::    AddSlabs              !! c = a + b 
public   ::    SubtractSlabs         !! c = a - b 
public   ::    MultSlabs             !! c = a * b
public   ::    MultSlabsByConstR     !! c = a * b,  b=scalar constant
public   ::    MultSlabsByConstL     !! c = a * b,  a=scalar constant
public   ::    SlabDotProduct        !! dp = [a,b,c] . [d,e,f]   (Slab arrays)
public   ::    ScalarDotProduct      !! dp = [a,b,c] . [d,e,f]   (ScalarVariables)
public   ::    MyAxpy                !! y <-- alpha*x + y w/ ScalarVariable inputs
public   ::    MyScal                !! x <-- alpha*x     w/ ScalarVariable inputs
public   ::    AddRConst              !! S.V. = S.V. + scalar_constant (real,kind=8)
public   ::    AddIConst              !! S.V. = S.V. + scalar_constant (integer)

private  ::    NewScalarSC           !! allocates arrays for both Slabs and Cols
private  ::    ApplyDerivative       !! low level utility routine
public   ::    CopySlabData_a2b      !! low level copy using blas
public   ::    AllocateCol           !! allocate array to hold vals in Col storage mode
public   ::    DeallocateCol         !! deallocate Col storage array
private  ::    ChebyRecursion        !! recursion to get Cheby coeffs of derivative given coeffs of function
 
type ScalarVariable 
  character (len=80)          :: Name            !! Descriptive name
  type( TensorProductGrid )   :: Grid            !! Grid over which scalar is defined
  type( DataDecomposition )   :: DataMap         !! decomposition map
  character (len=80)          :: DerivTypes(6)   !! labels defining differentiation methods
  type( Derivative )          :: Derivs(6)       !! initialized derivatives for each dim
  real(kind=8),pointer        :: SlabVals(:,:,:) !! data values for scalar in "slab" storage
  real(kind=8),pointer        :: ColVals(:,:,:)  !! data values for scalar in "col" storage
  real(kind=8),pointer        :: Ghost_x(:,:,:)  !! ny*nz*2 for left/right planes of ghost values 
  real(kind=8),pointer        :: Ghost_y(:,:,:)  !! etc
  real(kind=8),pointer        :: Ghost_z(:,:,:)  !!
  real(kind=8),pointer        :: DirichletEast(:,:)  !! array of data values
  real(kind=8),pointer        :: DirichletWest(:,:)
  real(kind=8),pointer        :: DirichletFront(:,:)
  real(kind=8),pointer        :: DirichletBack(:,:)
  real(kind=8),pointer        :: DirichletTop(:,:)
  real(kind=8),pointer        :: DirichletBottom(:,:)
  real(kind=8),pointer        :: NeumannEast(:,:)  !! array of deriv values
  real(kind=8),pointer        :: NeumannWest(:,:)
  real(kind=8),pointer        :: NeumannFront(:,:)
  real(kind=8),pointer        :: NeumannBack(:,:)
  real(kind=8),pointer        :: NeumannTop(:,:)
  real(kind=8),pointer        :: NeumannBottom(:,:)
  
  real(kind=8)                :: ScaleFactor             !! scale used to nondimensionalize scalar
  logical                     :: initialized             !! to tell whether data arrays are allocated
  logical                     :: gridpoint_values        !! these can be useful under some circumstances
  logical                     :: wavenumber_values       !! to control behavior of "differentiate" method
  logical                     :: forward_transform_only  !!  ""
  logical                     :: inverse_transform_only  !!  ""
end type ScalarVariable


!! overload assignment operator =
interface assignment (=)
 module procedure EqualScalars
end interface

interface operator(+)
 module procedure AddSlabs
end interface

interface operator(-)
 module procedure SubtractSlabs
end interface

interface operator(*)
 module procedure MultSlabs, MultSlabsByConstR, MultSlabsByConstL
end interface

interface operator(/)
 module procedure DivideSlabs
end interface

Contains    
 !!======================================
 !
 ! ASSIGNMENT OPERATOR: new = old
 !
 !!======================================
 subroutine EqualScalars(new,old)
  use etc
  implicit none
  type( ScalarVariable ),intent(in)    :: old
  type( ScalarVariable ),intent(out)   :: new
  integer                              :: npts(3)
  integer                              :: myid
  include 'mpif.h'
  
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)     
      
  if( new%initialized ) then 
   new%SlabVals=old%SlabVals     
  else 
   npts(1)  = old%DataMap%SlabDims(1,myid)
   npts(2)  = old%DataMap%SlabDims(2,myid)
   npts(3)  = old%DataMap%SlabDims(3,myid)
   allocate( new%SlabVals(npts(1),npts(2),npts(3)), stat=ierr )
   if( ierr /= 0  ) Stop 'Allocation failed: EqualScalars'
   if(myid==0) then
    NewAllocationGB=npts(1)*npts(2)*npts(3)*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    open(2,file='output/MemoryLog')
    write(2,*) '  equal scalars old-->new        ...........',old%name,'--->',new%name
    write(2,*) '  New allocation (Column storage)  .........',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
    close(2)
  endif
   call CopySlabData_a2b(old,new)
  endif
  
  new%Name = old%Name
  new%Grid = old%Grid 
  new%DataMap = old%DataMap
  new%DerivTypes = old%DerivTypes
  new%ScaleFactor = old%ScaleFactor
  new%Ghost_x = old%Ghost_x
  new%Ghost_y = old%Ghost_y
  new%Ghost_z = old%Ghost_z
  new%DirichletEast = old%DirichletEast
  new%DirichletWest = old%DirichletWest
  new%DirichletFront = old%DirichletFront
  new%DirichletBack = old%DirichletBack
  new%DirichletTop = old%DirichletTop
  new%DirichletBottom = old%DirichletBottom
  new%NeumannEast = old%NeumannEast
  new%NeumannWest = old%NeumannWest
  new%NeumannFront = old%NeumannFront
  new%NeumannBack = old%NeumannBack
  new%NeumannTop = old%NeumannTop
  new%NeumannBottom = old%NeumannBottom
  new%gridpoint_values = old%gridpoint_values
  new%wavenumber_values = old%wavenumber_values
  new%forward_transform_only=old%forward_transform_only
  new%inverse_transform_only=old%inverse_transform_only
  
  !!Duplicate deriv objects, dont compute new ones
  do i=1,6 
   call EqualDerivs(new%Derivs(i),old%Derivs(i))
  enddo 
  
  new%initialized=.TRUE.
 end subroutine EqualScalars


 !!======================================
 !
 ! ADDITION OPERATOR: c=a+b
 !
 !!======================================
 function AddSlabs(a,b) result(c)
 !! c = a+b
  implicit none
  type( ScalarVariable),intent(in)     :: a,b
  type( ScalarVariable)                :: c  
  if( c%initialized ) then
   c%SlabVals = a%SlabVals + b%SlabVals
  else
   c=a  !! use overloaded assignment
   c%SlabVals = c%SlabVals + b%SlabVals
  endif
 end function AddSlabs
 
 
 !!======================================
 !
 ! SUBTRACTION OPERATOR: c=a-b
 !
 !!======================================
 function SubtractSlabs(a,b) result(c)
  !! c = a-b
  implicit none
  type( ScalarVariable),intent(in)     :: a,b
  type( ScalarVariable)                :: c  
  if( c%initialized ) then
   c%SlabVals = a%SlabVals - b%SlabVals
  else
   c=a  !! use overloaded assignment
   c%SlabVals = c%SlabVals - b%SlabVals
  endif
 end function SubtractSlabs
  
  
 !!======================================
 !
 ! ELEMENTWISE MULTIPLICATION OPERATORs: c=a*b
 !
 !!======================================
 function MultSlabs(a,b) result(c)
  !! c = a*b
  implicit none
  type( ScalarVariable),intent(in)     :: a,b
  type( ScalarVariable)                :: c  
  if( c%initialized ) then
   c%SlabVals = a%SlabVals * b%SlabVals
  else
   c=a  !! use overloaded assignment
   c%SlabVals = c%SlabVals * b%SlabVals
  endif    
 end function MultSlabs

 function MultSlabsByConstR(a,b) result(c) 
 !! c = a*b, b=scalar constant
 implicit none
  type( ScalarVariable),intent(in)      :: a
  real(kind=8),intent(in)               :: b
  type( ScalarVariable)                 :: c  
  if( c%initialized ) then
   c%SlabVals = a%SlabVals * b
  else
   c=a  !! use overloaded assignment
   c%SlabVals = c%SlabVals * b
  endif    
end function MultSlabsByConstR

function MultSlabsByConstL(a,b) result(c) 
 !! c = a*b, a=scalar constant
 implicit none  
  real(kind=8),intent(in)              :: a
  type( ScalarVariable),intent(in)     :: b
  type( ScalarVariable)                :: c  
  if( c%initialized ) then
   c%SlabVals = a * b%SlabVals
  else
   c=b  !! use overloaded assignment
   c%SlabVals = c%SlabVals * a
  endif    
end function MultSlabsByConstL


 !!======================================
 !
 ! ELEMENTWISE DIVISION OPERATOR: c=a/b
 !
 !!======================================
 function DivideSlabs(a,b) result(c) 
 !! c = a/b
 implicit none
  type( ScalarVariable),intent(in)     :: a,b
  type( ScalarVariable)                :: c  
  if( c%initialized ) then
   c%SlabVals = a%SlabVals / b%SlabVals
  else
   c=a  !! use overloaded assignment
   c%SlabVals = c%SlabVals / b%SlabVals
  endif    
 end function DivideSlabs


 !!======================================
 !
 ! CONSTRUCTORS
 !
 !!======================================
 subroutine NewScalar(Name,Grid,DataMap,ScaleFactor,var) 
  use class_TensorProductGrid
  use class_DataDecomposition
  use class_Derivative
  use etc
  
  implicit none
  type( ScalarVariable )                    :: var
  type( TensorProductGrid )                 :: Grid
  type( DataDecomposition )                 :: DataMap
  real(kind=8)                              :: ScaleFactor
  integer                                   :: npts(3)
  character (len=80 )                       :: Name
  integer                                   :: myid,numprocs
  logical                                   :: second_deriv
  include 'mpif.h'
  
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)     
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr) 
  
  if(myid==0) then
   open(2,file=memoryfile,position='append')
  endif
  
  if( var%initialized ) call DeleteScalarVariable(var) 
  
  var%Name = Name
  var%Grid = Grid 
  var%DataMap = DataMap
  var%DerivTypes = DataMap%DerivTypes
  !var%DerivTypes = var%DataMap%DerivTypes 
  
  
  !!Create the new derivatives objects given details in DataMap
  !!Second_deriv is a logical arg telling NewDeriv to
  !!look at DataMap%DerivType(i) or at DataMap%DerivType(i+3)
  !!to determine what type of derivative object to construct.
  do i=1,3 
   second_deriv=.FALSE.
   call NewDerivative(i,DataMap,Grid,var%Derivs(i),second_deriv)
   second_deriv=.TRUE.
   call NewDerivative(i,DataMap,Grid,var%Derivs(i+3),second_deriv)
  enddo  
  
  
  var%ScaleFactor = ScaleFactor 
  !!Allocate and zero out an array for slab storage. 
  npts(1)  = DataMap%SlabDims(1,myid)
  npts(2)  = DataMap%SlabDims(2,myid)
  npts(3)  = DataMap%SlabDims(3,myid)
      
  allocate( var%SlabVals(npts(1),npts(2),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'Allocation failed: NewScalar'
  var%SlabVals=0.d0 
  
  if(myid==0) then
   NewAllocationGB=npts(1)*npts(2)*npts(3)*8*GBperByte
   AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
   write(2,*) '  Variable        ..........................',var%name
   write(2,*) '  New allocation  ..........................',NewAllocationGB
   write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
   write(2,*) '   '
  endif
     
  !! Allow for ghost data in periodic directions
  if( trim(var%DerivTypes(1)) == 'fourier'   &
     .or. trim(var%DerivTypes(1)) == 'zeroderiv' ) then
     
   allocate( var%Ghost_x(npts(2),npts(3),2), stat=ierr )
   if( ierr /= 0 ) Stop 'Ghost_x allocation failed: NewScalar'
   var%Ghost_x=0.d0
   
   if(myid==0) then
    NewAllocationGB=2*npts(2)*npts(3)*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    write(2,*) '  x ghost data    ..........................'
    write(2,*) '  New allocation  ..........................',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
   endif
   
  endif
  
  if( trim(var%DerivTypes(2)) == 'fourier'    &
     .or. trim(var%DerivTypes(1)) == 'zeroderiv' ) then
   allocate( var%Ghost_y(npts(1),npts(3),2), stat=ierr )
   if( ierr /= 0 ) Stop 'Ghost_y allocation failed: NewScalar'
   var%Ghost_y=0.d0
      
   if(myid==0) then
    NewAllocationGB=2*npts(1)*npts(3)*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    write(2,*) '  y ghost data    ..........................'
    write(2,*) '  New allocation  ..........................',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
   endif

  endif
  
  !! Allow for ghost data in vertical direction
  allocate( var%Ghost_z(npts(1),npts(2),2), stat=ierr )
  if( ierr /= 0 ) Stop 'Ghost_z allocation failed: NewScalar'
  var%Ghost_z=0.d0
     
   if(myid==0) then
    NewAllocationGB=2*npts(2)*npts(1)*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    write(2,*) '  z ghost data    ..........................'
    write(2,*) '  New allocation  ..........................',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
   endif
   

  
  allocate( var%DirichletEast(npts(2),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'DirichletEast allocation failed: NewScalar'
  var%DirichletEast=0.d0
  allocate( var%DirichletWest(npts(2),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'DirichletWest allocation failed: NewScalar'
  var%DirichletWest=0.d0
  allocate( var%DirichletFront(npts(1),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'DirichletFront allocation failed: NewScalar'
  var%DirichletFront=0.d0
  allocate( var%DirichletBack(npts(1),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'DirichletBack allocation failed: NewScalar'
  var%DirichletBack=0.d0
  allocate( var%DirichletTop(npts(1),npts(2)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'DirichletTop allocation failed: NewScalar'
  var%DirichletTop=0.d0
  allocate( var%DirichletBottom(npts(1),npts(2)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'DirichletBottom allocation failed: NewScalar'
  var%DirichletBottom=0.d0
  
  allocate( var%NeumannEast(npts(2),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'NeumannEast allocation failed: NewScalar'
  var%NeumannEast=0.d0
  allocate( var%NeumannWest(npts(2),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'NeumannWest allocation failed: NewScalar'
  var%NeumannWest=0.d0
  allocate( var%NeumannFront(npts(1),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'NeumannFront allocation failed: NewScalar'
  var%NeumannFront=0.d0
  allocate( var%NeumannBack(npts(1),npts(3)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'NeumannBack allocation failed: NewScalar'
  var%NeumannBack=0.d0
  allocate( var%NeumannTop(npts(1),npts(2)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'NeumannTop allocation failed: NewScalar'
  var%NeumannTop=0.d0
  allocate( var%NeumannBottom(npts(1),npts(2)), stat=ierr )   !! this is the actual data array
  if( ierr /= 0 ) Stop 'NeumannBottom allocation failed: NewScalar'
  var%NeumannBottom=0.d0
  
    
   if(myid==0) then
    open(1,file=docfile,position='append')
     write(1,*) ' '
     write(1,*) 'ScalarVariable ----> ',trim(var%Name),' <---- created. '
     write(1,*) 'Data storage array has (slab) size             ',(npts(i),i=1,3)
     write(1,*) 'Derivative types :  ',var%DerivTypes
     write(1,*) 'Array globally initialized to zero. '
     write(1,*) ' '
    close(1)
   endif
   
      
   if(myid==0) then
    NewAllocationGB=2*2*(npts(1)*npts(2)+npts(2)*npts(3)+npts(1)*npts(3))*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    write(2,*) '  Dirichlet & Neumann data    ..............'
    write(2,*) '  New allocation  ..........................',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
    close(2)
   endif

  var%gridpoint_values = .TRUE.
  var%wavenumber_values = .FALSE.
  var%forward_transform_only = .FALSE.
  var%inverse_transform_only = .FALSE.
  var%initialized=.TRUE.
  
 end subroutine NewScalar
 
 subroutine NewVector(s1,s2,s3,v)  
  type( ScalarVariable ),intent(inout)  :: v(3)
  type( ScalarVariable ),intent(in)     :: s1,s2,s3  
  call EqualScalars(v(1),s1)            !! v(1) <-- s1
  call EqualScalars(v(2),s2)            !! v(2) <-- s2
  call EqualScalars(v(3),s3)            !! v(3) <-- s3
 end subroutine NewVector
  
 subroutine NewScalarSC(Name,Grid,DataMap,ScaleFactor,var) 
  use class_TensorProductGrid
  use class_DataDecomposition
  use class_Derivative
  use etc
  
  implicit none
  type( ScalarVariable ),intent(inout)      :: var
  type( TensorProductGrid ),intent(in)      :: Grid
  type( DataDecomposition ),intent(in)      :: DataMap
  real(kind=8),intent(in)                   :: ScaleFactor
  character (len=80 ),intent(in)            :: Name
  integer                                   :: npts(3)
  integer                                   :: myid,numprocs
  include 'mpif.h'
   
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      !!  ====> myid
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)  !!  ====> numprocs
  
  call NewScalar(Name,Grid,DataMap,ScaleFactor,var)
  
  !! Now allocate and zero out an additional array for col storage. 
  npts(:) = DataMap%ColDims(:,myid)
  
  allocate( var%ColVals(npts(1),npts(2),npts(3)) )   !! this is the actual data array
  var%ColVals=0.d0 
  
  if(myid==0) then
   NewAllocationGB=npts(1)*npts(2)*npts(3)*8*GBperByte
   AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
   open(2,file='output/MemoryLog')
   write(2,*) '  Variable        ..........................',Name
   write(2,*) '  New allocation (Column storage)  .........',NewAllocationGB
   write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
   write(2,*) '   '
   close(2)
  endif
  
  var%initialized=.TRUE.
 end subroutine NewScalarSC
 
 !!======================================
 !
 ! DESTRUCTORS
 !
 !!======================================
 subroutine DeleteScalarVariable(var)
  use etc
  use class_TensorProductGrid
  use class_Derivative
  use class_DataDecomposition  
  type( ScalarVariable )   :: var
  integer                              :: nx 
  integer                              :: ny
  integer                              :: nz  
  integer, SAVE                        :: myid,numprocs
  logical, SAVE                        :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr) 
   first_entry=.FALSE.
  endif
  
  nx=var%DataMap%SlabDims(1,myid)
  ny=var%DataMap%SlabDims(2,myid)
  nz=var%DataMap%SlabDims(3,myid)
    
  !!Delete Slab & Col storage arrays
  if( associated(var%SlabVals) ) deallocate( var%SlabVals )
  if( associated(var%ColVals) )  deallocate( var%ColVals ) 
  
  if(myid==0) then
    NewAllocationGB = -nx*ny*nz*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    open(2,file='output/MemoryLog')
    write(2,*) '  Variable        ..........................',var%name
    write(2,*) '  New allocation (Column storage)  .........',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
    close(2)
   endif
      
  !!Delete derived type members
  call DeleteDataDecomposition(var%DataMap)
     !! --> deletes arrays containing the sizes of the data arrays
  call DeleteDerivative(var%Derivs(1))
     !! --> deallocates  {kx} or {A_diff,ipiv} for dim 1, depending on type
  call DeleteDerivative(var%Derivs(2))
     !! --> deallocates  {kx} or {A_diff,ipiv} for dim 2, depending on type
  call DeleteDerivative(var%Derivs(3))
     !! --> deallocates  {kx} or {A_diff,ipiv} for dim 3, depending on type
  call DeleteDerivative(var%Derivs(4))
     !! --> deallocates  {kx} or {A_diff,ipiv} for dim 1, depending on type
  call DeleteDerivative(var%Derivs(5))
     !! --> deallocates  {kx} or {A_diff,ipiv} for dim 2, depending on type
  call DeleteDerivative(var%Derivs(6))
     !! --> deallocates  {kx} or {A_diff,ipiv} for dim 3, depending on type
  call DeleteTensorProductGrid(var%Grid) !!!this is the bad one
  
 end subroutine DeleteScalarVariable
 
 subroutine DeleteVector(v)
  implicit none
  type( ScalarVariable ), intent(inout)   :: v(3)  
  integer                                 :: i  
   do i=1,3
    if( v(i)%initialized ) call DeleteScalarVariable( v(i) )
   enddo
 end subroutine DeleteVector
 
 !!======================================
 !
 ! UTILITY
 !
 !!======================================
 subroutine AllocateCol(var)
  use etc
  implicit none
  type( ScalarVariable ),intent(inout) :: var
  integer                              :: nx 
  integer                              :: ny
  integer                              :: nz  
  integer, SAVE                        :: myid,numprocs
  logical, SAVE                        :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr) 
   first_entry=.FALSE.
  endif
  
  call DeAllocateCol(var)
  nx=var%DataMap%ColDims(1,myid)
  ny=var%DataMap%ColDims(2,myid)
  nz=var%DataMap%ColDims(3,myid) 
  allocate(var%ColVals(nx,ny,nz), stat=ierr)
  if( ierr /= 0  ) Stop 'Allocation failed: AllocateCol'
  
  if(myid==0) then
   NewAllocationGB=nx*ny*nz*8*GBperByte
   AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
   open(2,file='output/MemoryLog')
   write(2,*) '  Variable        ..........................',var%name
   write(2,*) '  New allocation (Column storage)  .........',NewAllocationGB
   write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
   write(2,*) '   '
   close(2)
  endif
  
 end subroutine AllocateCol
 
 subroutine DeAllocateCol(var)
  use etc
  implicit none
  type( ScalarVariable ),intent(inout) :: var
  integer                              :: nx 
  integer                              :: ny
  integer                              :: nz  
  integer, SAVE                        :: myid,numprocs
  logical, SAVE                        :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr) 
   first_entry=.FALSE.
  endif
  
  nx=var%DataMap%ColDims(1,myid)
  ny=var%DataMap%ColDims(2,myid)
  nz=var%DataMap%ColDims(3,myid)
  
   deallocate( var%ColVals )
   if(myid==0) then
    NewAllocationGB = -nx*ny*nz*8*GBperByte
    AllocatedMemoryGB = AllocatedMemoryGB + NewAllocationGB
    open(2,file='output/MemoryLog')
    write(2,*) '  Variable        ..........................',var%name
    write(2,*) '  New allocation (Column storage)  .........',NewAllocationGB
    write(2,*) '  Total Allocated Memory in GB .............',AllocatedMemoryGB
    write(2,*) '   '
    close(2)
   endif
 end subroutine DeAllocateCol
 
 subroutine CopySlabData_a2b(a,b)
 implicit none
 type( ScalarVariable ), intent(in)     :: a
 type( ScalarVariable ), intent(inout)  :: b
 integer                                :: N
 integer                                :: ierr,inc=1
 integer, SAVE                          :: myid
 logical, SAVE                          :: first_entry=.TRUE.
    
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   first_entry=.FALSE.
  endif
  
 N = a%DataMap%SlabDims(1,myid)   &
   * a%DataMap%SlabDims(2,myid)   &
   * a%DataMap%SlabDims(3,myid)
     call DCOPY( N, a%SlabVals, inc, b%SlabVals, inc )  !! b<--a 
 end subroutine CopySlabData_a2b
 
 subroutine GetLocalArraySize(f,n1,n2,n3)
 implicit none
 type( ScalarVariable )     :: f 
 integer                    :: n1,n2,n3
!!Get extents of local part of array
 n1= ubound( f%SlabVals,1 )
 n2= ubound( f%SlabVals,2 )
 n3= ubound( f%SlabVals,3 )
end subroutine GetLocalArraySize

subroutine GetCoordVal(f,idim,i,x)
 implicit none
 integer                    :: idim,i,i0
 real(kind=8)               :: x
 type( ScalarVariable )     :: f 
 i0=0
 if(idim==3) call GetOffset(f,i0) !!Need offset for x3 dimension
 x = f%Grid%SpaceDims(idim)%x(i0+i)  
end subroutine GetCoordVal

subroutine GetDimlessCoordVal(f,idim,i,x,L0)
 implicit none
 integer                    :: idim,i,i0
 real(kind=8)               :: x,L0
 type( ScalarVariable )     :: f 
 i0=0
 if(idim==3) call GetOffset(f,i0) !!Need offset for x3 dimension
 x = ( f%Grid%SpaceDims(idim)%x(i0+i) )/L0 
end subroutine GetDimlessCoordVal

subroutine GetOffset(f,k0) 
 implicit none
 type( ScalarVariable )     :: f
 integer                    :: k0,ierr 
 integer, SAVE              :: myid,numprocs
 logical, SAVE              :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr) 
   first_entry=.FALSE.
  endif
 
  select case ( mod(f%DataMap%nz,2) )  
   case(0)              ! nz is even
    k0 = myid*(f%DataMap%nz/numprocs) 
   case(1)              ! nz is odd
    k0 = myid*((f%DataMap%nz-1)/numprocs) 
  end select 
  
end subroutine GetOffset

 
 
 
 !!======================================
 !
 ! OPERATORS
 !
 !!======================================
 
 subroutine Norm2(phi,rms)
  implicit none  
  integer                   :: n1,n2,n3,i,j,k,ierr
  integer                   :: nlocal,nglobal,one=1
  type( ScalarVariable )    :: phi
  real(kind=8)              :: rms,local_sum,global_sum
  integer, SAVE             :: myid
  logical, SAVE             :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   first_entry=.FALSE.
  endif
  
  n1=phi%DataMap%SlabDims(1,myid)
  n2=phi%DataMap%SlabDims(2,myid)
  n3=phi%DataMap%SlabDims(3,myid)
  local_sum=0.d0
  nlocal=n1*n2*n3
  
  ! get total number of elements in vector ==> nglobal
  call mpi_allreduce(nlocal,nglobal,one,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,ierr)
  
  do k=1,n3
   do j=1,n2
    do i=1,n1
     local_sum = local_sum + phi%SlabVals(i,j,k)**2
    enddo
   enddo
  enddo
  
  call mpi_allreduce(local_sum,global_sum,one,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierr)
  rms = sqrt( global_sum/nglobal )
 end subroutine Norm2
 
  subroutine NormInf(phi,ans)
  implicit none  
  integer                   :: ierr
  integer                   :: one=1
  type( ScalarVariable )    :: phi
  real(kind=8)              :: ans,local_max,global_max
  include 'mpif.h'
   
  local_max = MAXVAL( abs(phi%SlabVals) )
  call mpi_allreduce(local_max,global_max,one,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,ierr)
  ans = global_max
  
 end subroutine NormInf

 
 subroutine Differentiate(phi,idim,d_phi,call_id)
  use class_Derivative
  implicit none  
  logical, SAVE                         :: first_entry = .TRUE.
  integer,intent(in)                    :: idim
  type( ScalarVariable ),intent(inout)  :: phi
  type( ScalarVariable ),intent(inout)  :: d_phi
  real(kind=8),allocatable,SAVE         :: col_array(:,:,:,:)
  integer,SAVE                          :: n1,n2,n3,myid,ny
  integer                               :: ierr,k
  character(len=80),optional            :: call_id
  character(len=80)                     :: caller
  include 'mpif.h'
  
  !!==================================================
  !!==================================================
  !! I'm assuming here w/ SAVE that each processor 
  !! will only be differentiating
  !! one array shape repeatedly. If not, 
  !! return to allocate-deallocate logic.
  !!==================================================
  !!==================================================
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)
   n1=phi%DataMap%ColDims(1,myid)
   n2=phi%DataMap%ColDims(2,myid)
   n3=phi%DataMap%ColDims(3,myid)
   ny=phi%DataMap%SlabDims(2,myid)
   allocate( col_array(n1,n2,n3,2),stat=ierr )
   if(ierr .ne. 0)  &
    stop 'Problem allocating col_array: Differentiate'
   if( ubound(phi%SlabVals,2) /= n2 )  &
    stop 'ny problem btwn slabs & cols: Differentiate'
   first_entry=.FALSE.
  endif
  !!==================================================
  !!==================================================
  if( .not. present(call_id) ) then
   caller='unknown'
  else
   caller=call_id
  endif
  
  !!==================================================
  !!  if 2d in y, then ddy=0 
  !!==================================================
  if(idim==2 .and. ny==1) then
   d_phi%SlabVals=0.d0
   return
  endif
  !!==================================================
  !!==================================================
  
  !!==================================================
  !!  if derivs in some direction defined to be zero ...
  !!==================================================
  if( trim( phi%Derivs(idim)%MethodLabel ) == 'zeroderiv') then
   d_phi%SlabVals=0.d0
   return
  endif
  !!==================================================
  !!==================================================
  
  
  !!==================================================
  !!  for z derivs, shuffle data then differentiate
  !!  and unshuffle....
  !!==================================================
  if(idim==3) then
     call slabs_to_columns(phi,col_array(1,1,1,1),n1,n2,n3)
     
     call ApplyDerivativeToArray(phi,  &
            col_array(1,1,1,1),        &
            col_array(1,1,1,2),        &
            phi%Derivs(idim),n1,n2,n3,caller)
                                  
     call columns_to_slabs(d_phi,col_array(1,1,1,2),n1,n2,n3)
     
  elseif(idim<3) then 
  !!==================================================
  !!  no need for data swaps...
  !!==================================================
   call ApplyDerivative(phi,d_phi,phi%Derivs(idim),caller) 
   
  endif 
  
  return
 end subroutine Differentiate


 subroutine Gradient(scalar,grad)
  implicit none  
  type( ScalarVariable )                :: scalar
  type( ScalarVariable ),intent(inout)  :: grad(3) 
  
   call Differentiate(scalar,1,grad(1))
   call Differentiate(scalar,2,grad(2))
   call Differentiate(scalar,3,grad(3))   
 end subroutine Gradient 
 
 
 
 subroutine Transform3D(in,out,tmp1,tmp2,forward,inverse)
  implicit none  
  type( ScalarVariable ),intent(inout)  ::  in,out
  type( ScalarVariable ),intent(inout)  ::  tmp1,tmp2
  logical                               ::  forward,inverse
  integer                               ::  i,nx,ny,nz
  character(len=80)                     ::  name1,name2
  
  nx=ubound( in%SlabVals,1 )
  ny=ubound( in%SlabVals,2 )
  nz=ubound( in%SlabVals,3 )
  
  if( forward .and. .not. inverse ) then
   in%gridpoint_values = .TRUE.
   in%wavenumber_values = .FALSE.
   in%forward_transform_only = .TRUE.
   in%inverse_transform_only = .FALSE.
   
   tmp1%gridpoint_values = .TRUE.
   tmp1%wavenumber_values = .FALSE.
   tmp1%forward_transform_only = .TRUE.
   tmp1%inverse_transform_only = .FALSE.
   
   tmp2%gridpoint_values = .TRUE.
   tmp2%wavenumber_values = .FALSE.
   tmp2%forward_transform_only = .TRUE.
   tmp2%inverse_transform_only = .FALSE.
   
  elseif( inverse .and. .not. forward ) then
   in%gridpoint_values = .FALSE.
   in%wavenumber_values = .TRUE.
   in%forward_transform_only = .FALSE.
   in%inverse_transform_only = .TRUE.
   
   tmp1%gridpoint_values = .FALSE.
   tmp1%wavenumber_values = .TRUE.
   tmp1%forward_transform_only = .FALSE.
   tmp1%inverse_transform_only = .TRUE.
   
   tmp2%gridpoint_values = .FALSE.
   tmp2%wavenumber_values = .TRUE.
   tmp2%forward_transform_only = .FALSE.
   tmp2%inverse_transform_only = .TRUE.
  else
   stop 'problem in Transform3D, forward and inverse cant both be true!!'
  endif
  
  
  name1=tmp1%Name
  name2=tmp2%Name
  tmp1%Name=in%Name
  tmp2%Name=in%Name
  
  !! Duplicate deriv objects, dont compute new ones
  do i=1,3
   tmp1%Derivs(i)=in%Derivs(i)
   if(ny>1) tmp2%Derivs(i)=in%Derivs(i)
  enddo
  
  
  if(forward) then
   call Differentiate(in,1,tmp1)      !! transform only in x
   if( ny > 1 ) then
    call Differentiate(tmp1,2,tmp2)   !! transform only in y
    call Differentiate(tmp2,3,out)    !! transform only in z
   elseif( ny==1 ) then
    call Differentiate(tmp1,3,out)    !! transform only in z
   endif
  elseif(inverse) then
   call Differentiate(in,3,tmp1)      !! transform only in z
   if( ny > 1 ) then
    call Differentiate(tmp1,2,tmp2)   !! transform only in y
    call Differentiate(tmp2,1,out)    !! transform only in x
   elseif( ny==1 ) then
    call Differentiate(tmp1,1,out)    !! transform only in x
   endif
  endif
  
                                     
  !! cleanup
  in%forward_transform_only = .FALSE.
  in%inverse_transform_only = .FALSE.
  out%forward_transform_only = .FALSE.
  out%inverse_transform_only = .FALSE.
  tmp1%forward_transform_only = .FALSE.
  tmp1%inverse_transform_only = .FALSE.
  tmp2%forward_transform_only = .FALSE.
  tmp2%inverse_transform_only = .FALSE.
  tmp1%Name=name1
  tmp2%Name=name2
  
  if( forward .and. .not. inverse ) then
   out%gridpoint_values = .FALSE.
   out%wavenumber_values = .TRUE.
  elseif( inverse .and. .not. forward ) then
   out%gridpoint_values = .TRUE.
   out%wavenumber_values = .FALSE.
  endif
   
 end subroutine Transform3D
!!====================================================================== 
!!====================================================================== 



 
!!====================================================================== 
!!====================================================================== 
 subroutine Divergence(v,div,temp,components) 
  use class_Derivative
  implicit none  
  integer                                        :: idim(3),ny,i
  character(len=80)                              :: text_string
  type( ScalarVariable ),intent(inout)           :: v(3) 
  type( ScalarVariable ),intent(inout)           :: div
  type( ScalarVariable )                         :: temp
  type( ScalarVariable ),optional,intent(inout)  :: components(3)
  
  idim(:)=(/1,2,3/)
  ny = ubound(v(1)%SlabVals,2)
  text_string='Divergence'    !! id the calling routine
  
  if( .not. div%initialized ) then
   write(0,*) 'the result of a Vector divergence must be'
   write(0,*) 'a previously existing ScalarVariable'
   stop
  endif
  
  !!=======================================================
  !! always start divergence calc from physical space ...
  !!=======================================================
  do i=1,3
   v(i)%gridpoint_values      =.TRUE. 
   v(i)%wavenumber_values     =.FALSE.
   v(i)%inverse_transform_only=.FALSE. 
   v(i)%forward_transform_only=.FALSE. 
  enddo
  
  !!==================================================
  !! if we are storing the individual derivatives....
  !!==================================================
  if( present(components) ) then
  
   call Differentiate(v(1),idim(1),components(1),text_string)     !! d/dx   
   call Differentiate(v(3),idim(3),components(3),text_string)     !! d/dz
   div%SlabVals = components(1)%SlabVals  &
                + components(3)%SlabVals  
                
   if( ny > 1 ) then
    call Differentiate(v(2),idim(2),components(2),text_string)    !! d/dy
    div%SlabVals = div%SlabVals + components(2)%SlabVals 
   endif
   
  else 
  !!=====================================================
  !! if we don't need to keep the individual derivatives
  !!=====================================================
  
   call Differentiate(v(1),idim(1),div,text_string)      !! d/dx
   call Differentiate(v(3),idim(3),temp,text_string)     !! d/dz
   div%SlabVals = div%SlabVals + temp%SlabVals
   
   if( ny > 1 ) then
    call Differentiate(v(2),idim(2),temp,text_string)    !! d/dy
    div%SlabVals = div%SlabVals  + temp%SlabVals
   endif
      
  endif

 end subroutine Divergence
!!====================================================================== 
!!====================================================================== 



!!====================================================================== 
!!====================================================================== 
 subroutine Laplacian(phi,grad,Lphi,temp)
 !!Computes and returns both the gradient and the laplacian of phi
  implicit none 
  type( ScalarVariable )     :: phi
  type( ScalarVariable )     :: grad(3) 
  type( ScalarVariable )     :: Lphi
  type( ScalarVariable )     :: temp
  
  !!===============================================================
  !! Set the flags & derivative types for the 2nd derivatives of phi
  !! i.e. assign the correct derivative objects to the vector SV grad
  !!===============================================================
  grad(1)%Derivs(1)=phi%Derivs(4)              ! method for d/dx (dphi/dx)  
  grad(2)%Derivs(2)=phi%Derivs(5)              ! method for d/dy (dphi/dy)  
  grad(3)%Derivs(3)=phi%Derivs(6)              ! method for d/dz (dphi/dz)
    
  call Gradient(phi,grad)
  call Divergence(grad,Lphi,temp) 
 end subroutine Laplacian
!!====================================================================== 
!!====================================================================== 
 
  
 subroutine Curl(s,curlv,temp)
  use class_Derivative
  implicit none  
  type( ScalarVariable ),intent(inout) :: s(3)
  type( ScalarVariable ),intent(inout) :: curlv(3)
  type( ScalarVariable )               :: temp(2)
  character( len=80 )                  :: Name
  character( len=4 )                   :: flag  
   !! needs to be redone     
 end subroutine Curl 
 
 subroutine MyAxpy(alpha,x,y) !!  y <== alpha*x + y
  implicit none
  type( ScalarVariable ), intent(in)     :: x
  type( ScalarVariable ), intent(inout)  :: y
  real(kind=8)                           :: alpha
  integer                                :: N,n1,n2,n3
  integer                                :: inc=1
  call GetLocalArraySize(y,n1,n2,n3)
  N=n1*n2*n3
  call daxpy(N, alpha, x%SlabVals, inc, y%SlabVals, inc)
 end subroutine MyAxpy
 
 subroutine MyScal(alpha,x) !!  x <== alpha*x 
  implicit none
  type( ScalarVariable ), intent(inout)  :: x
  real(kind=8)                           :: alpha
  integer                                :: N,n1,n2,n3
  integer                                :: inc=1
  call GetLocalArraySize(x,n1,n2,n3)
  N=n1*n2*n3
  call dscal(N, alpha, x%SlabVals, inc)
 end subroutine MyScal
 
 subroutine AddRConst(alpha,x) !!  x <== x + alpha
  implicit none
  type( ScalarVariable ), intent(inout)  :: x
  real(kind=8)                           :: alpha
  x%SlabVals = x%SlabVals + alpha
 end subroutine AddRConst
 
 subroutine AddIConst(alpha,x) !!  x <== x + alpha
  implicit none
  type( ScalarVariable ), intent(inout)  :: x
  integer                                :: alpha
  x%SlabVals = x%SlabVals + alpha
 end subroutine AddIConst
 
 
 
 
subroutine SlabDotProduct(dp,sign,a,b,c,d,e,f,m1,m2,m3)
 implicit none
 integer                                         :: sign
 integer                                         :: m1,m2,m3
 real(kind=8), dimension(m1,m2,m3), intent(out)  :: dp
 real(kind=8), dimension(m1,m2,m3), intent(in)   :: a,b,c
 real(kind=8), dimension(m1,m2,m3), intent(in)   :: d,e,f
 dp = sign*(a*d + b*e + c*f)
end subroutine SlabDotProduct

subroutine SlabDotSaxpy(dp,sign,a,b,c,d,e,f)
 implicit none
 integer                                      :: sign
 real(kind=8), dimension(:,:,:), intent(out)  :: dp
 real(kind=8), dimension(:,:,:), intent(in)   :: a,b,c
 real(kind=8), dimension(:,:,:), intent(in)   :: d,e,f
 dp = dp + sign*(a*d + b*e + c*f)
end subroutine SlabDotSaxpy

subroutine ScalarDotProduct(dp,sign,a,b,c,d,e,f)
 implicit none
 integer                 :: sign
 integer                 :: m1,m2,m3
 type( ScalarVariable )  :: dp
 type( ScalarVariable )  :: a,b,c
 type( ScalarVariable )  :: d,e,f
 call GetLocalArraySize(dp,m1,m2,m3)
 call SlabDotProduct( dp%SlabVals,   &
                             sign,   &      
                       a%SlabVals,   &
                       b%SlabVals,   &
                       c%SlabVals,   &
                       d%SlabVals,   &
                       e%SlabVals,   &
                       f%SlabVals,   &
                       m1,m2,m3)
end subroutine ScalarDotProduct

subroutine ScalarDotSaxpy(dp,sign,a,b,c,d,e,f)
 implicit none
 integer                             :: sign
 type( ScalarVariable ), intent(out) :: dp
 type( ScalarVariable ), intent(in)  :: a,b,c
 type( ScalarVariable ), intent(in)  :: d,e,f
 call SlabDotSaxpy(   dp%SlabVals,   &
                             sign,   &
                       a%SlabVals,   &
                       b%SlabVals,   &
                       c%SlabVals,   &
                       d%SlabVals,   &
                       e%SlabVals,   &
                       f%SlabVals )
end subroutine ScalarDotSaxpy
 
 subroutine ApplyDerivative(phi,d_phi,deriv,call_id) 
 !! all derivatives are taken wrt equally spaced parameter s
 !! each result is then mult by the appropriate 1/dx/ds vector
  use class_Derivative
  use class_DataDecomposition
  
  implicit none  
  include 'fftw3.f'
  
  character(len=80),optional         :: call_id
  type(ScalarVariable),intent(in)    :: phi
  type(ScalarVariable),intent(inout) :: d_phi
  type(Derivative),intent(in)        :: deriv
  integer*8                          :: plan
  integer, dimension(3)              :: SlabDims
  integer, dimension(3)              :: SlabVals
  integer, dimension(3)              :: ColDims
  integer, dimension(3)              :: ColVals
  integer                            :: idim,ierr
  integer                            :: n,nb2,nsmall
  integer                            :: i,j,k,kend
  complex(kind=8),save               :: cmplxi,xx
  real(kind=8),allocatable,save      :: temp(:,:,:)
  real(kind=8),allocatable,save      :: small_x(:,:,:)
  real(kind=8),allocatable,save      :: small2_x(:,:,:) 
  real(kind=8),allocatable,save      :: small_y(:,:,:)
  real(kind=8),allocatable,save      :: small2_y(:,:,:) 
  real(kind=8),allocatable,save      :: dsdx(:)
  real(kind=8),allocatable,save      :: dsdy(:)
  real(kind=8)                       :: NormFactor
  integer, save                      :: myid,numprocs,nx,ny,nz
  logical, save                      :: first_entry=.TRUE.
  
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   cmplxi=cmplx(0.d0,1.d0)
   nx=phi%DataMap%SlabVals(1,myid)
   ny=phi%DataMap%SlabVals(2,myid)
   nz=phi%DataMap%SlabVals(3,myid)

   allocate( dsdx(nx), dsdy(ny) ) 
   allocate(temp(nx,ny,nz))
   allocate( small_x(nx-2,ny,nz),small2_x(nx-2,ny,nz) )
   allocate( small_y(nx,ny-2,nz),small2_y(nx,ny-2,nz) )

   dsdx(:) = 1.d0/phi%Grid%SpaceDims(1)%dxds(:)
   dsdy(:) = 1.d0/phi%Grid%SpaceDims(2)%dxds(:)

   first_entry=.FALSE.
  endif 
      
  idim=deriv%DimIndex
  n = deriv%DataDim
  nb2 = n/2

  
  if( deriv%MethodLabel=='fourier' ) then
   
   NormFactor=1.d0/n

   if(idim==1) then 
     !!=======================================================
     !! If we are starting out in physical space, do the
     !! forward transform in x use fftw3 real to halfcomplex
     !!=======================================================
     if( phi%gridpoint_values ) then
      
      if( phi%forward_transform_only ) then
       call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals,d_phi%SlabVals)
       return
      else
       call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals,temp )
      endif
            
     endif
     
     !!=======================================================
     !! provided we are not doing the inv transform only,
     !! do wavenumber multiplication and phase shift
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then
      
      !!================================================
      !! Multiply by (-) wavenunmbers and normalize
      !!===========================================
      if( phi%gridpoint_values ) then  !! transforms done here, data is in temp 
      
      !! we do a  "-ik_x" multiply:
      !! here is the "i":
      !! phase (location)  shift necessary for real--> 1/2 complex transform
       do k=1,nz
        do j=1,ny
         do i=2,nb2
          xx = temp(i,j,k)* deriv%kfilter(i)
          temp(i,j,k) = temp(n-i+2,j,k)* deriv%kfilter(i)
          temp(n-i+2,j,k)=xx
         enddo
        enddo
       enddo
       
       !! here is the "-k_x":
       do k=1,nz
        do j=1,ny                        
         temp(:,j,k) = - deriv%kx(:)*NormFactor    &
                       * temp(:,j,k)
        enddo
       enddo
       temp(nb2+1,:,:) = 0.d0    !! nyquist
       
      elseif(phi%wavenumber_values) then  !! data not in temp
      !! same procedure but w/ data starting in phi rather than in temp
      
      !! we do a  "-ik_x" multiply:
      !! here is the "i":
      !! phase (location)  shift necessary for real--> 1/2 complex transform
       do k=1,nz
        do j=1,ny
         do i=2,nb2
          xx = phi%SlabVals(i,j,k) * deriv%kfilter(i)
          phi%SlabVals(i,j,k) = phi%SlabVals(n-i+2,j,k) * deriv%kfilter(i)
          phi%SlabVals(n-i+2,j,k)=xx
         enddo
        enddo
       enddo
       
       !! here is the "-k_x":
       do k=1,nz
        do j=1,ny                        
         temp(:,j,k) = deriv%kx(:)*NormFactor    &
                       * phi%SlabVals(:,j,k)
        enddo
       enddo
       temp(nb2+1,:,:) = 0.d0    !! nyquist
       
      endif
            
     endif
     
     !!==================================================================
     !! Now do the inverse transform, fftw3 halfcomplex to real
     !! if inverse_transform only, we need to normalize 
     !! and return before applying ds/dz
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then
      call dfftw_execute_r2r( deriv%inverse_plan,temp,d_phi%SlabVals)
     elseif(phi%inverse_transform_only ) then
      call dfftw_execute_r2r( deriv%inverse_plan,phi%SlabVals,d_phi%SlabVals)
      d_phi%SlabVals=d_phi%SlabVals*NormFactor
      return
     endif
     
     
     !!======================================
     !! deal w/ x<-->s transformation
     !! dfdx = (df/ds * ds/dx) 
     !!======================================
      do k=1,nz
       do j=1,ny 
        d_phi%SlabVals(:,j,k) = d_phi%SlabVals(:,j,k) * dsdx
       enddo
      enddo
           
     return
   elseif(idim==2) then !! must loop over k because stride is not constant   
     !!===================================================================
     !! If we are starting out in physical space, do the
     !! forward transforms, this time organized as complex to complex dft
     !! regarding the data along the i dimension as nx/2 complex elements
     !!===================================================================
     
     if( phi%gridpoint_values ) then
        
       if( phi%forward_transform_only ) then  !! result===> d_phi
        do k=1,nz
         call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals(1,1,k),d_phi%SlabVals(1,1,k))
        enddo
        return
       else           !! result===> temp
        do k=1,nz
         call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals(1,1,k),temp(1,1,k))
        enddo
       endif
       
     endif
     
     !!=======================================================
     !! provided we are not doing the inv transform only,
     !! multiply by data by (-ik) and normalize
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then
     
      if( phi%gridpoint_values ) then  !! we did transform, data in temp

      !! we do a  "-ik_y" multiply:
      !! here is the "i":
      !! phase (location)  shift necessary for real--> 1/2 complex transform
       do k=1,nz
        do i=1,nx
         do j=2,nb2
          xx = temp(i,j,k) * deriv%kfilter(j)
          temp(i,j,k) = temp(i,n-j+2,k) * deriv%kfilter(j)
          temp(i,n-j+2,k)=xx
         enddo
        enddo
       enddo
       
       !! and here is the "-k_y":
       do k=1,nz
        do j=1,ny 
         do i=1,nx    !!/2
          temp(i,j,k) = -deriv%kx(j)*temp(i,j,k)*NormFactor
         enddo
        enddo
       enddo
       temp(:,nb2+1,:) = 0.d0    !! nyquist
       
      elseif( phi%wavenumber_values ) then  !! we didnt transform, data not in temp
      !! same procedure but with data in phi
      
      !! we do a  "-ik_y" multiply:
      !! here is the "i":
      !! phase (location)  shift necessary for real--> 1/2 complex transform
       do k=1,nz
        do i=1,nx
         do j=2,nb2
          xx = phi%SlabVals(i,j,k) * deriv%kfilter(j)
          phi%SlabVals(i,j,k) = phi%SlabVals(i,n-j+2,k) * deriv%kfilter(j)
          phi%SlabVals(i,n-j+2,k)=xx
         enddo
        enddo
       enddo
       
       !! and here is the "-k_y":
       do k=1,nz
        do j=1,ny 
         do i=1,nx 
           temp(i,j,k) = -deriv%kx(j)*phi%SlabVals(i,j,k)*NormFactor
         enddo
        enddo
       enddo
       temp(:,nb2+1,:) = 0.d0    !! nyquist
       
      endif
     endif
     
     !!=========================================================
     !! do inverse transforms, again as complex to complex dft
     !! if inverse_transform only, we need to normalize 
     !! and return before applying ds/dz
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then  !! data in ctemp
      do k=1,nz
        call dfftw_execute_r2r(deriv%inverse_plan,temp(1,1,k),d_phi%SlabVals(1,1,k))
      enddo
     elseif(phi%inverse_transform_only ) then   !! data in phi
      do k=1,nz
        call dfftw_execute_r2r(deriv%inverse_plan,phi%SlabVals(1,1,k),d_phi%SlabVals(1,1,k))
      enddo
      d_phi%SlabVals=d_phi%SlabVals*NormFactor
      return
     endif
     
     !!======================================
     !! deal w/ y<-->s transformation
     !! dfdy = (df/ds * ds/dy) 
     !!======================================
     do k=1,nz
      do j=1,ny
       do i=1,nx 
        d_phi%SlabVals(i,j,k) = d_phi%SlabVals(i,j,k) * dsdy(j)
       enddo
      enddo
     enddo
  
    return
   endif
  endif   !! end Fourier block
   
  
  if( deriv%MethodLabel=='cos' )  then 
    
   NormFactor = 1./(2.*(n-1.))
   
   if(idim==1) then
                   
     nsmall = nx-2

     !!=======================================================
     !! If we are starting out in physical space, or if we
     !! only need to do an inverse transform, then cos transform
     !! the input data
     !!=======================================================
     if( phi%gridpoint_values .or. phi%inverse_transform_only ) then

       if( phi%forward_transform_only ) then
        call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals,d_phi%SlabVals)
        return
       elseif(phi%inverse_transform_only) then
        call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals,d_phi%SlabVals)
        d_phi%SlabVals=d_phi%SlabVals*NormFactor
        return
       else
        call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals,temp)
       endif 
       
     endif
    
     !!=======================================================
     !! multiply by (-1)* wavenumbers and by normalization factor
     !!=======================================================
     
      !! if we skipped the transform so data NOT in temp...
      if(phi%wavenumber_values) then 
       do k=1,nz
        do j=1,ny
         do i=1,nsmall
          small_x(i,j,k) = - deriv%kx(i+1)*deriv%kfilter(i+1)   &
                           * phi%SlabVals(i+1,j,k)*NormFactor
         enddo
        enddo
       enddo
      !! else we did the transform so data IS in temp...
      elseif(phi%gridpoint_values) then
       do k=1,nz
        do j=1,ny
         do i=1,nsmall
          small_x(i,j,k) = - deriv%kx(i+1)*deriv%kfilter(i+1)  &
                           * temp(i+1,j,k)*NormFactor
         enddo
        enddo
       enddo
      endif
      
     
     !!=======================================================
     !! do the inverse transform (sin) on the smaller array
     !! zero out endpts (mean & nyquist) for sin expansions
     !!=======================================================
      call dfftw_execute_r2r( deriv%inverse_plan,small_x,small2_x) 
      do k=1,nz
       do j=1,ny
        d_phi%SlabVals(2:nx-1,j,k)=small2_x(1:nsmall,j,k)        
        d_phi%SlabVals(1,j,k)=0.d0
        d_phi%SlabVals(nx,j,k)=0.d0
       enddo
      enddo
      

     !!======================================
     !! deal w/ x<-->s transformation
     !! dfdx = (df/ds * ds/dx) 
     !!======================================
     do k=1,nz
      do j=1,ny 
       d_phi%SlabVals(:,j,k) = d_phi%SlabVals(:,j,k) * dsdx
      enddo
     enddo
     
    return
   elseif(idim==2) then !! must loop over k because stride is not constant
                   
     nsmall = ny-2
     
     !!=======================================================
     !! If we are starting out in physical space, or if we
     !! only need to do an inverse transform, then cos transform
     !! the input data
     !!=======================================================
     if( phi%gridpoint_values .or. phi%inverse_transform_only ) then
        
       if( phi%forward_transform_only ) then
        do k=1,nz
         call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals(1,1,k),d_phi%SlabVals(1,1,k))
        enddo
        return
       elseif(phi%inverse_transform_only) then
        do k=1,nz
         call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals(1,1,k),d_phi%SlabVals(1,1,k))
        enddo
        d_phi%SlabVals=d_phi%SlabVals*NormFactor
        return
       else
        do k=1,nz
         call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals(1,1,k),temp(1,1,k))
        enddo
       endif
       
     endif
               
     !!=======================================================
     !! multiply by (-1)* wavenumbers and by normalization factor
     !!=======================================================
     
      !! if we didn't do transform, data is NOT in temp
      if(phi%wavenumber_values) then
       do k=1,nz
        do j=1,nsmall
         do i=1,nx      
          small_y(i,j,k) = - deriv%kx(j+1)*deriv%kfilter(j+1)   &
                           * phi%SlabVals(i,j+1,k) * NormFactor
         enddo
        enddo
       enddo
      !! else if we did do transform, data IS in temp
      elseif(phi%gridpoint_values) then
       do k=1,nz
        do j=1,nsmall
         do i=1,nx      
          small_y(i,j,k) = - deriv%kx(j+1)*deriv%kfilter(j+1)   &
                           * temp(i,j+1,k) * NormFactor
         enddo
        enddo
       enddo
      endif
      
     
     !!=======================================================
     !! do the inverse transforms
     !!=======================================================
      do k=1,nz
       call dfftw_execute_r2r(deriv%inverse_plan,small_y(1,1,k),d_phi%SlabVals(1,2,k))
      enddo
      d_phi%SlabVals(:,1,:)=0.d0
      d_phi%SlabVals(:,nsmall+2,:)=0.d0 !! endpts always zero for sin expansions
          
          
     !!======================================
     !! deal w/ y<-->s transformation
     !! dfdy = (df/ds * ds/dy) 
     !!======================================
     do k=1,nz
      do j=1,ny
       do i=1,nx 
        d_phi%SlabVals(i,j,k) = d_phi%SlabVals(i,j,k) * dsdy(j)
       enddo
      enddo
     enddo
          
    return
   endif     
  endif
  
  if( deriv%MethodLabel=='sin' ) then 
   
   NormFactor = 1./(2.*(n-1.))
   
   if(idim==1) then 
     nsmall = nx-2
     
     !!=======================================================
     !! If we are starting out in physical space, or if we
     !! only need to do an inverse transform, then sin transform
     !! the input data
     !!=======================================================
     if( phi%gridpoint_values .or. phi%inverse_transform_only ) then

      !! extract the data w/o the zero endpts
      do i=1,nsmall
       small_x(i,:,:) = phi%SlabVals(i+1,:,:)
      enddo
     
      if( phi%forward_transform_only ) then
       call dfftw_execute_r2r(deriv%forward_plan,small_x,small2_x)
       do i=1,nsmall
        d_phi%SlabVals(i+1,:,:)=small2_x(i,:,:)
       enddo
       d_phi%SlabVals(1,:,:)=0.d0
       d_phi%SlabVals(nx,:,:)=0.d0
       return
      elseif(phi%inverse_transform_only) then
       call dfftw_execute_r2r(deriv%forward_plan,small_x,small2_x)
       do i=1,nsmall
        d_phi%SlabVals(i+1,:,:)=small2_x(i,:,:)*NormFactor
       enddo
       d_phi%SlabVals(1,:,:)=0.d0
       d_phi%SlabVals(nx,:,:)=0.d0
       return
      else
       call dfftw_execute_r2r(deriv%forward_plan,small_x,small2_x)
      endif
      
     endif
     
     
     !!=======================================================
     !! multiply by (+1)* wavenumbers and by normalization factor
     !!=======================================================
     
      !! if we didn't do transform, data is NOT in small2_x
      if(phi%wavenumber_values) then
       do k=1,nz
        do j=1,ny
         do i=1,nx
          temp(i,j,k) = deriv%kx(i)*deriv%kfilter(i)   &
                       * phi%SlabVals(i,j,k) * NormFactor
         enddo
        enddo
       enddo
      !! if we did do transform, data IS in small2_x
      elseif(phi%gridpoint_values) then
       do k=1,nz
        do j=1,ny
         do i=1,nsmall
          temp(i+1,j,k) = deriv%kx(i+1)*deriv%kfilter(i+1)   &
                        * small2_x(i,j,k) * NormFactor
         enddo
        enddo
       enddo
      endif
      temp(1,:,:)=0.d0        !! derivative of a sin expansion has no mean
      temp(nsmall+2,:,:)=0.d0 !! nyquist
    
     !!=======================================================
     !! do the inverse transforms
     !!=======================================================
     call dfftw_execute_r2r( deriv%inverse_plan,temp,d_phi%SlabVals ) 
     
     !!======================================
     !! deal w/ x<-->s transformation
     !! dfdx = (df/ds * ds/dx) 
     !!======================================
     do k=1,nz
      do j=1,ny
       d_phi%SlabVals(:,j,k) = d_phi%SlabVals(:,j,k) * dsdx
      enddo
     enddo
     
    return
    
   elseif(idim==2) then !! must loop over k because stride is not constant
     nsmall = ny-2
     
     !!=======================================================
     !! If we are starting out in physical space, or if we
     !! only need to do an inverse transform, then sin transform
     !! the input data
     !!=======================================================
     if( phi%gridpoint_values .or. phi%inverse_transform_only ) then
         
       !! extract the data w/o the zero endpts
       do j=1,nsmall
        small_y(:,j,:) = phi%SlabVals(:,j+1,:)
       enddo
       !! do the forward transform (requires loop due to stride issues)
       
       if( phi%forward_transform_only ) then
        do k=1,nz
         call dfftw_execute_r2r( deriv%forward_plan, small_y(1,1,k), small2_y(1,1,k) )
        enddo
        do j=1,nsmall
         d_phi%SlabVals(:,j+1,:)=small2_y(:,j,:)
        enddo
        d_phi%SlabVals(:,1,:)=0.d0
        d_phi%SlabVals(:,ny,:)=0.d0
        return
       elseif(phi%inverse_transform_only) then
        do k=1,nz
         call dfftw_execute_r2r( deriv%forward_plan, small_y(1,1,k), small2_y(1,1,k) )
        enddo
        do j=1,nsmall
         d_phi%SlabVals(:,j+1,:)=small2_y(:,j,:)*NormFactor
        enddo
        d_phi%SlabVals(:,1,:)=0.d0
        d_phi%SlabVals(:,ny,:)=0.d0
        return
       else
        do k=1,nz
         call dfftw_execute_r2r( deriv%forward_plan, small_y(1,1,k), small2_y(1,1,k) )
        enddo
       endif
       
     endif
      
    
     !!=======================================================
     !! multiply by (+1)* wavenumbers and by normalization factor
     !!=======================================================
     
      !! if we didn't do transform, data is NOT in small2_y
      if(phi%wavenumber_values) then
       do k=1,nz
        do j=1,nsmall
         do i=1,nx
          temp(i,j+1,k) = deriv%kx(j+1)*deriv%kfilter(j+1)   &
                        * phi%SlabVals(i,j+1,k) * NormFactor
         enddo
        enddo
       enddo
      !! if we did do transform, data IS in small2_y
      elseif(phi%gridpoint_values) then
       do k=1,nz
        do j=1,nsmall
         do i=1,nx
          temp(i,j+1,k) = deriv%kx(j+1)*deriv%kfilter(j+1)   &
                        * small2_y(i,j,k) * NormFactor
         enddo
        enddo
       enddo
       temp(:,1,:)=0.d0           !! derivative of a sin expansion has no mean
       temp(:,nsmall+2,:)=0.d0    !! nyquist
      endif
      
     
     !!=======================================================
     !! do the inverse transforms
     !!=======================================================
      do k=1,phi%DataMap%SlabDims(3,myid)
       call dfftw_execute_r2r(deriv%inverse_plan,temp(1,1,k),d_phi%SlabVals(1,1,k))
      enddo
     
     !!======================================
     !! deal w/ x<-->s transformation
     !! dfdy = (df/ds * ds/dy) 
     !!======================================
     do k=1,nz
      do j=1,ny
       do i=1,nx 
        d_phi%SlabVals(i,j,k) = d_phi%SlabVals(i,j,k) * dsdy(j)
       enddo
      enddo
     enddo
          
    return
   endif
  endif
  
  if( deriv%MethodLabel=='cheby' )  then      
     if( idim==1 ) then
      !!=======================================================
      !! If we are starting out in physical space, do the
      !! forward transform (requires loop due to stride issues)
      !!=======================================================
     if( phi%gridpoint_values        .and.   &
         .NOT. phi%wavenumber_values .and.   &
         .NOT. phi%inverse_transform_only)   then
         
      !! constant strides, can do all the transforms in a single call
      !! do cosine transform(s)
      call dfftw_execute_r2r(deriv%forward_plan,  &
                             phi%SlabVals, &
                             d_phi%SlabVals )
      if( phi%forward_transform_only ) return
     endif
     
     
     !!=======================================================
     !! provided we are not doing the inv transform only,
     !! get normalized expansion coefficients for deriv from
     !! coeffs of the function recursively
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then
      if(phi%wavenumber_values) then
       call ChebyRecursion( idim,phi%DataMap%SlabDims,phi%SlabVals,temp)
      elseif(phi%gridpoint_values) then
       call ChebyRecursion( idim,phi%DataMap%SlabDims,d_phi%SlabVals,temp)
      endif
     endif
     
     !!=======================================================
     !! inverse transform, i.e. another round of cosine transform(s)
     !! if inverse_transform only, ***we need to normalize???***
     !! and return before applying ds/dz
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then 
      call dfftw_execute_r2r(deriv%forward_plan,temp,d_phi%SlabVals)
     elseif( phi%inverse_transform_only ) then
      call dfftw_execute_r2r(deriv%forward_plan,phi%SlabVals,d_phi%SlabVals)
      return
     endif
    
     !!======================================
     !! deal w/ x<-->s transformation
     !! dfdz = (df/ds * ds/dz) 
     !!======================================
     do k=1,nz
      do j=1,ny 
       d_phi%SlabVals(:,j,k) = d_phi%SlabVals(:,j,k) * dsdx
      enddo
     enddo
    return
    
   elseif( idim==2 ) then 
      !!=======================================================
      !! If we are starting out in physical space, do the
      !! forward transform (requires loop due to stride issues)
      !!=======================================================
      if( phi%gridpoint_values        .and.   &
         .NOT. phi%wavenumber_values .and.   &
         .NOT. phi%inverse_transform_only)   then
         
       !!here we must loop over k planes 
       !!because memory stride is not constant
       kend=phi%DataMap%ColVals(3,myid)
       !! do cosine transform(s)
       do k=1,kend
        call dfftw_execute_r2r( deriv%forward_plan,  &
                                phi%SlabVals(1,1,k), &
                                d_phi%SlabVals(1,1,k) )
       enddo
       if( phi%forward_transform_only ) return
      endif
      
      !!=======================================================
      !! provided we are not doing the inv transform only,
      !! get normalized expansion coefficients for deriv from
      !! coeffs of the function recursively
      !!=======================================================
      if( .NOT. phi%inverse_transform_only ) then
       if(phi%wavenumber_values) then
        call ChebyRecursion( idim,phi%DataMap%SlabDims,phi%SlabVals,temp)
       elseif(phi%gridpoint_values) then
        call ChebyRecursion( idim,phi%DataMap%SlabDims,d_phi%SlabVals,temp)
       endif
      endif
     
     !!=======================================================
     !! inverse transform, i.e. another round of cosine transform(s)
     !! if inverse_transform only, ***we need to normalize???***
     !! and return before applying ds/dz
     !!=======================================================
     if( .NOT. phi%inverse_transform_only ) then
      do k=1,kend
       call dfftw_execute_r2r(deriv%forward_plan,  &
                             temp(1,1,k),d_phi%SlabVals(1,1,k))
      enddo
     elseif(phi%inverse_transform_only) then
      do k=1,kend
       call dfftw_execute_r2r(deriv%forward_plan,  &
                             phi%SlabVals(1,1,k),d_phi%SlabVals(1,1,k))
      enddo
      return
     endif
        
    !!======================================
    !! deal w/ x<-->s transformation
    !! dfdz = (df/ds * ds/dz) 
    !!======================================
    do k=1,nz
     do j=1,ny
      do i=1,nx 
       d_phi%SlabVals(i,j,k) = d_phi%SlabVals(i,j,k) * dsdy(j)
      enddo
     enddo
    enddo
    
    return
   endif
  endif
  
  if( deriv%MethodLabel=='compact' )  then
  
   if( phi%wavenumber_values        .or.   &
       .NOT. phi%gridpoint_values ) then
      if(myid==0)   &
      write(0,*) 'method = compact but flag says data in wavenumber space'
      stop ' stopping in ApplyDerivative '
   endif
         
    if(idim==1) then
     call compact_ddx(phi,d_phi)
    elseif(idim==2) then
     call compact_ddy(phi,d_phi)
    elseif(idim==3) then
     call compact_ddz(phi,d_phi)
    endif
  endif  

 end subroutine ApplyDerivative
 
 
 
 subroutine ApplyDerivativeToArray(SV,phi,d_phi,deriv,nx,ny,nz,call_id) 
 !! all derivatives are taken wrt equally spaced parameter s
 !! each result needs to be mult by the appropriate 1/dx/ds vector
  use class_Derivative
  use class_DataDecomposition
  
  implicit none  
  include 'fftw3.f'
  
  type(ScalarVariable),intent(in):: SV
  character(len=80),optional     :: call_id
  integer                        :: nx,ny,nz
  real(kind=8)                   :: phi(nx,ny,nz)
  real(kind=8)                   :: d_phi(nx,ny,nz)
  
  type( Derivative ), intent(in):: deriv
  integer                       :: ierr
  integer                       :: i,j,k
  real(kind=8)                  :: NormFactor,alpha
  integer, SAVE                 :: myid,numprocs,idim,n
  integer, SAVE                 :: nb2,npts,nsmall
  logical, SAVE                 :: first_entry=.TRUE.
  complex(kind=8), SAVE         :: cmplxi
  complex(kind=8),allocatable,SAVE    :: ctemp(:,:,:)
  real(kind=8),allocatable, SAVE      :: temp(:,:,:)
  real(kind=8),allocatable, SAVE      :: small(:,:,:)
  real(kind=8),allocatable, SAVE      :: small2(:,:,:) 
  real(kind=8),allocatable, SAVE      :: dsdz(:)
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   cmplxi=cmplx(0.d0,1.d0)  
   idim=deriv%DimIndex
   n = deriv%DataDim
   nb2 = n/2
   npts = nx*ny
   nsmall = nz-2
   allocate( temp(nx,ny,nz) )
   allocate( ctemp(nx,ny,nz) )
   allocate( small(nx,ny,nz-2) )
   allocate( small2(nx,ny,nz-2) )
   allocate( dsdz(nz) )
   dsdz(:) = 1.d0/SV%Grid%SpaceDims(idim)%dxds(:)
   first_entry = .FALSE.
  endif 
  
  
  if( deriv%MethodLabel=='fourier' ) then
   
    NormFactor=1.d0/n

     !!=======================================================
     !! If we are starting out in physical space, do the
     !! forward transforms, organized as complex to complex dfts
     !!=======================================================
     if( SV%gridpoint_values )   then
       if(SV%forward_transform_only) then
        call dfftw_execute_dft( deriv%forward_plan, phi, d_phi )
        return
       else
        call dfftw_execute_dft( deriv%forward_plan, phi, ctemp )
       endif
     endif
     
     if(.NOT. SV%inverse_transform_only)   then
      !!=========================================
      !! Multiply by data by (-ik) and normalize
      !!=========================================
      if( SV%wavenumber_values ) then !!no forward transform in this call
       do k=1,nz
        !!ctemp(:,:,k) = -cmplxi * deriv%kx(k)*phi(:,:,k)*NormFactor
        alpha = -cmplxi * deriv%kx(k) * deriv%kfilter(k) * NormFactor
        call DSCAL(npts,alpha,phi(1,1,k),1)  !! input data in phi
       enddo
      elseif(SV%gridpoint_values) then !!I did the forward transform in this call
       do k=1,nz
        !!ctemp(:,:,k) = -cmplxi * deriv%kx(k)*ctemp(:,:,k)*NormFactor
        alpha = -cmplxi * deriv%kx(k) * deriv%kfilter(k) * NormFactor
        call DSCAL(npts,alpha,ctemp(1,1,k),1)   !! data in ctemp
       enddo
      endif
      
     endif
     
     !!=======================================================
     !! do inverse transforms, again as complex to complex dft
     !! if inverse_transform only, we need to normalize 
     !! and return before applying ds/dz
     !!=======================================================
     if(.NOT. SV%inverse_transform_only)   then      
      call dfftw_execute_dft( deriv%inverse_plan, ctemp, d_phi )
     elseif(SV%inverse_transform_only)   then
      alpha = NormFactor
      call DSCAL(nz*npts,alpha,phi(1,1,1),1)
      call dfftw_execute_dft( deriv%inverse_plan, phi, d_phi )
      return
     endif
     
     !!======================================
     !! dfdz = (df/ds * ds/dz) 
     !!======================================
     do k=1,nz  
       !!d_phi(:,:,k) = d_phi(:,:,k) * dsdz(k)
       alpha = dsdz(k)
       call DSCAL(npts,alpha,d_phi(1,1,k),1)
     enddo
      
    return
    
  endif   !! end Fourier block
   
  
  if( deriv%MethodLabel=='cos' )  then 
             
    NormFactor = 1./(2.*(n-1.))
      
     !!=======================================================
     !! If we are starting out in physical space, or if we
     !! only need to do an inverse transform, then cos transform
     !! the input data
     !!=======================================================
     if( SV%gridpoint_values .or. SV%inverse_transform_only ) then
      
      if(SV%forward_transform_only) then
       call dfftw_execute_r2r(deriv%forward_plan,phi,d_phi)
       return
      elseif( SV%inverse_transform_only ) then 
       !! cos transform is its own inverse
       call dfftw_execute_r2r(deriv%forward_plan,phi,d_phi)
       d_phi=d_phi*NormFactor
       return
      else
       call dfftw_execute_r2r(deriv%forward_plan,phi,temp)
      endif
      
     endif
           
      !!=========================================
      !! Multiply by data by (-ik) and normalize
      !!=========================================
      if( SV%wavenumber_values ) then !! input data in phi
       do k=1,nsmall
        small(:,:,k) = - deriv%kx(k+1)*deriv%kfilter(k+1)   &
                       * phi(:,:,k+1) * NormFactor
       enddo
      elseif(SV%gridpoint_values) then !! data now in temp
       do k=1,nsmall
        small(:,:,k) = - deriv%kx(k+1)*deriv%kfilter(k+1)   &
                       * temp(:,:,k+1) * NormFactor
       enddo
      endif
      
          
     !!=======================================================
     !! do inverse transforms
     !!=======================================================
     
      call dfftw_execute_r2r(deriv%inverse_plan,small,d_phi(1,1,2))
      !! set top & bottom planes to zero
      alpha = 0.d0
      call DSCAL(npts,alpha,d_phi(1,1,1),1)
      call DSCAL(npts,alpha,d_phi(1,1,nsmall+2),1)
     
     
     !!======================================
     !! dfdz = (df/ds * ds/dz) 
     !!======================================
     do k=2,nz-1  !!d_phi(:,:,k) = d_phi(:,:,k) * dsdz(k)
       alpha = dsdz(k)
       call DSCAL(npts,alpha,d_phi(1,1,k),1)
     enddo
          
   return
  endif


  
  if( deriv%MethodLabel=='sin' ) then 
   
    NormFactor = 1./(2.*(n-1.))
 
     !!=======================================================
     !! If we are starting out in physical space, or if we
     !! only need to do an inverse transform, then sin transform
     !! the input data
     !!=======================================================
     if( SV%gridpoint_values .or. SV%inverse_transform_only ) then
         
      !! extract the data w/o the zero endpts
      do k=1,nsmall
       !!small(:,:,k) = phi(:,:,k+1)
       call DCOPY(npts,phi(1,1,k+1),1,small(1,1,k),1)
      enddo
      
      call dfftw_execute_r2r( deriv%forward_plan, small, small2 )
      
      if( SV%forward_transform_only) then
       do k=1,nsmall
        d_phi(:,:,k+1)=small2(:,:,k)
       enddo
       d_phi(:,:,1)=0.d0    !! mean
       d_phi(:,:,nz)=0.d0   !! nyquist
       return
      elseif(SV%inverse_transform_only) then
       do k=1,nsmall
        d_phi(:,:,k+1)=small2(:,:,k)*NormFactor
       enddo
       d_phi(:,:,1)=0.d0    !! bottom
       d_phi(:,:,nz)=0.d0   !! top
       return
      endif      
     endif
     
      !!=========================================
      !! Multiply by data by (kz) and normalize
      !!=========================================
      if( SV%wavenumber_values ) then !! data is in phi
       do k=2,nz-1
        temp(:,:,k) = deriv%kx(k)*deriv%kfilter(k)   &
                    * phi(:,:,k) * NormFactor
       enddo
       alpha = 0.d0
       call DSCAL(npts,alpha,temp(1,1,1),1)
       call DSCAL(npts,alpha,temp(1,1,nz),1)

      elseif(SV%gridpoint_values) then  !! data is in small2
      
       do k=1,nsmall
        temp(:,:,k+1) = deriv%kx(k+1)*deriv%kfilter(k+1)   &
                      * small2(:,:,k) * NormFactor
       enddo
       
       
       alpha = 0.d0
       call DSCAL(npts,alpha,temp(1,1,1),1)
       call DSCAL(npts,alpha,temp(1,1,nz),1)
      endif
            
     !!=======================================================
     !! do inverse transforms
     !!=======================================================
     call dfftw_execute_r2r(deriv%inverse_plan,temp,d_phi)
     
     !!======================================
     !! dfdz = (df/ds * ds/dz) 
     !!======================================
     do k=1,nz  
       !!d_phi(:,:,k) = d_phi(:,:,k) * dsdz(k)
       alpha = dsdz(k)
       call DSCAL(npts,alpha,d_phi(1,1,k),1)
     enddo  
          
   return
  endif
  
  if( deriv%MethodLabel=='cheby' )  then      
   
     !!=======================================================
     !! If we are starting out in physical space, do the
     !! forward transforms
     !!=======================================================
     if( SV%gridpoint_values        .and.   &
         .NOT. SV%wavenumber_values .and.   &
         .NOT. SV%inverse_transform_only)   then
      !! constant strides, can do all the transforms in a single call
      !! do cosine transform(s)
      call dfftw_execute_r2r( deriv%forward_plan,phi,d_phi )
      if( SV%forward_transform_only) return
     endif
    
    if(.NOT. SV%inverse_transform_only)   then
     !! get normalized expansion coefficients for deriv 
     !! from coeffs of the function recursively
     if( SV%wavenumber_values ) then
      call ChebyRecursion( idim, SV%DataMap%ColDims, phi,temp)
     elseif(SV%gridpoint_values) then
      call ChebyRecursion( idim, SV%DataMap%ColDims, d_phi,temp)
     endif
    endif
    
    !!=======================================================
    !! inverse transform, i.e. another round of cosine transform(s)
    !! if inverse_transform only, ***we need to normalize???***
    !! and return before applying ds/dz
    !!=======================================================
    if(.NOT. SV%inverse_transform_only)   then
     call dfftw_execute_r2r(deriv%forward_plan,temp,d_phi)
    elseif(SV%inverse_transform_only)   then
     call dfftw_execute_r2r(deriv%forward_plan,phi,d_phi)
     return
    endif
     
    !!======================================
    !! dfdz = (df/ds * ds/dz) 
    !!======================================
    do k=1,nz  
      !!d_phi(:,:,k) = d_phi(:,:,k) * dsdz(k)
      alpha = dsdz(k)
      call DSCAL(npts,alpha,d_phi(1,1,k),1)
    enddo
   
   return        
  endif
  
  if( deriv%MethodLabel=='compact' )  then
  
   if( SV%wavenumber_values        .or.   &
       .NOT. SV%gridpoint_values ) then
      if(myid==0)   &
      write(0,*) 'method = compact but flag says data in wavenumber space'
      stop ' stopping in ApplyDerivativeToArray '
   endif
   
   call compact_ddz_array(SV,phi,d_phi,nx,ny,nz)
  endif
  
 end subroutine ApplyDerivativeToArray
 
 
 subroutine ChebyRecursion( idim, dims, a, b )
  implicit none
  
  integer                    :: idim, dims(3),i,k,N
  real(kind=8),intent(inout) :: a(dims(1),dims(2),dims(3))    ! cheby coeffs for function f
  real(kind=8),intent(inout) :: b(dims(1),dims(2),dims(3))    ! cheby coeffs for function f' in idim direction
  
  N=dims(idim)  !! length in requested direction
  
  !!The recursion relation here assumes a map from s in [-1,1] --> theta in [0,pi]
  
  if( idim==1 ) then
    
    a(1,:,:)=0.5*a(1,:,:)
    a(N,:,:)=0.5*a(N,:,:)
    a=a/(N-1.)
    b(N,:,:)=a(N,:,:)

    b(N-1,:,:) = 2.d0*(N-1)*a(N-1+1,:,:)
    do k=N-1-2,1,-1
     b(k+1,:,:) = b(k+3,:,:) + 2.*(k+1)*a(k+2,:,:)
    enddo
    b(1,:,:)=a(2,:,:)+0.5*b(3,:,:)
    
    b(1,:,:) = 2.*b(1,:,:)
    b(N,:,:) = 2.*b(N,:,:)
 
  elseif( idim==2 ) then
  
    a(:,1,:)=0.5*a(:,1,:)
    a(:,N,:)=0.5*a(:,N,:)
    a=a/(N-1.)
    b(:,N,:)=a(:,N,:)

    b(:,N-1,:) = 2.d0*(N-1)*a(:,N-1+1,:)
    do k=N-1-2,1,-1
     b(:,k+1,:) = b(:,k+3,:) + 2.*(k+1)*a(:,k+2,:)
    enddo
    b(:,1,:)=a(:,2,:)+0.5*b(:,3,:)
    
    b(:,1,:) = 2.*b(:,1,:)
    b(:,N,:) = 2.*b(:,N,:)
    
  elseif( idim==3 ) then
  
    a(:,:,1)=0.5*a(:,:,1)
    a(:,:,N)=0.5*a(:,:,N)
    a=a/(N-1.)
    b(:,:,N)=a(:,:,N)

    b(:,:,N-1) = 2.d0*(N-1)*a(:,:,N-1+1)
    do k=N-1-2,1,-1
     b(:,:,k+1) = b(:,:,k+3) + 2.*(k+1)*a(:,:,k+2)
    enddo
    b(:,:,1)=a(:,:,2)+0.5*b(:,:,3)
    
    b(:,:,1) = 2.*b(:,:,1)
    b(:,:,N) = 2.*b(:,:,N)
        
  endif
  
 end subroutine ChebyRecursion
 

subroutine compact_ddx(F,dF)
 use class_CoordinateMap
 implicit none
 type( ScalarVariable ), intent(inout)   :: dF
 type( ScalarVariable ), intent(in)      :: F
 type( CoordinateMap )                   :: CoordMap
 integer                                 :: idim=1
 integer                                 :: i,j,k,ierr
 integer, SAVE                           :: myid,numprocs
 integer,SAVE                            :: m1,m2,m3
 real(kind=8),SAVE                       :: h,a1,b1,c1
 real(kind=8),allocatable,SAVE           :: work(:,:)
 real(kind=8),allocatable,SAVE           :: dsdx(:)
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   !!===================================================================
   !! m1,m2,m3    dimensions of array with data to be differentiated
   !! h           x grid spacing
   !! a1,b1,c1    coefficients of method, assigned to Derivative object
   !!===================================================================
   m1=F%DataMap%SlabDims(1,myid)
   m2=F%DataMap%SlabDims(2,myid)
   m3=F%DataMap%SlabDims(3,myid)
   allocate(work(m1,m2))
   allocate(dsdx(m1))
   CoordMap = F%Grid%SpaceDims(idim)
   h = CoordMap%ds
   dsdx = 1.d0/CoordMap%dxds  
   a1 = F%Derivs(idim)%a1
   b1 = F%Derivs(idim)%b1
   c1 = F%Derivs(idim)%c1
   first_entry = .FALSE.
  endif 
 

  
!!loop through each xy horizontal plane       
 do k=1,m3
 
 !Create rhs vectors, store in work(:,:)
 work(1,:)=( -(5./2.)*F%SlabVals(1,:,k)+(2.)*F%SlabVals(2,:,k)   &
                           +(1./2.)*F%SlabVals(3,:,k) )/h
                           
 work(2,:)=( -(5./9.)*F%SlabVals(1,:,k)-(1./2.)*F%SlabVals(2,:,k)   &
                           +(1.)*F%SlabVals(3,:,k)+(1./18.)*F%SlabVals(4,:,k) )/h

 work(3,:)= ( (1./36.)*(F%SlabVals(5,:,k)-F%SlabVals(1,:,k))   &
             + (7./9.)*(F%SlabVals(4,:,k)-F%SlabVals(2,:,k)) )/h  

 do i=4,m1-3
  work(i,:)= ( (c1/6.)*(F%SlabVals(i+3,:,k)-F%SlabVals(i-3,:,k))   &
             + (b1/4.)*(F%SlabVals(i+2,:,k)-F%SlabVals(i-2,:,k))   & 
             + (a1/2.)*(F%SlabVals(i+1,:,k)-F%SlabVals(i-1,:,k)) )/h  
 enddo

 work(m1-2,:)= (-(1./36.)*(F%SlabVals(m1-4,:,k)-F%SlabVals(m1,:,k))   &
                - (7./9.)*(F%SlabVals(m1-3,:,k)-F%SlabVals(m1-1,:,k)) )/h  
                
 work(m1-1,:)=( (5./9.)*F%SlabVals(m1,:,k)+(1./2.)*F%SlabVals(m1-1,:,k)  &
                         -(1)*F%SlabVals(m1-2,:,k)-(1./18.)*F%SlabVals(m1-3,:,k) )/h
                         
 work(m1,:) = ( (5./2.)*F%SlabVals(m1,:,k)-(2.)*F%SlabVals(m1-1,:,k)  &
                         -(1./2.)*F%SlabVals(m1-2,:,k) )/h

!!Compute the x derivatives for plane k, overwriting the array work
  call DGBTRS('N',m1,2,2,m2,F%Derivs(idim)%A_Diff,7,F%Derivs(idim)%ipiv,work,m1,ierr)

      
!!store the results in deriv(:,:,k) and apply metric from s <--> x
!! i.e. dF/dx = dF/ds *ds/dx
   do j=1,m2
    dF%SlabVals(:,j,k)=work(:,j) * dsdx(:)
   enddo

!!Go on to the next k plane
 enddo

 return
end subroutine compact_ddx


subroutine compact_ddy(F,dF)
 use class_CoordinateMap
 implicit none
 type( ScalarVariable ), intent(inout)   :: dF
 type( ScalarVariable ), intent(in)      :: F
 type( CoordinateMap )                   :: CoordMap
 integer                                 :: idim=2
 integer                                 :: i,j,k,ierr
 integer, SAVE                           :: myid,numprocs
 integer,SAVE                            :: m1,m2,m3
 real(kind=8),SAVE                       :: h,a1,b1,c1
 real(kind=8),allocatable,SAVE           :: work(:,:)
 real(kind=8),allocatable,SAVE           :: dsdy(:) 
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   !!===================================================================
   !! m1,m2,m3    dimensions of array with data to be differentiated
   !! h           x grid spacing
   !! a1,b1,c1    coefficients of method, assigned to Derivative object
   !!=================================================================== 
   m1=F%DataMap%SlabDims(1,myid)
   m2=F%DataMap%SlabDims(2,myid)
   m3=F%DataMap%SlabDims(3,myid)
   allocate(work(m2,m1))
   allocate(dsdy(m2))
   CoordMap = F%Grid%SpaceDims(idim)
   h = CoordMap%ds
   dsdy = 1.d0/CoordMap%dxds  !! remember, all low level notation uses "x"
   a1 = F%Derivs(idim)%a1
   b1 = F%Derivs(idim)%b1
   c1 = F%Derivs(idim)%c1
   first_entry=.FALSE.
  endif 
 
 

!!loop through each xy horizontal plane       
 do k=1,m3

!! These are high order one sided formula, to use, 
!! the setup matrix need needs to be adjusted
! work(1,:)=( -(274.)*F%SlabVals(:,1,k)+(600.)*F%SlabVals(:,2,k)   &
!             -(600.)*F%SlabVals(:,3,k)+(400.)*F%SlabVals(:,4,k)   &
!             -(150.)*F%SlabVals(:,5,k)+ (24.)*F%SlabVals(:,6,k) )/(120.*h)  
             
!work(2,:)=( -(24.)*F%SlabVals(:,1,k)-(130.)*F%SlabVals(:,2,k)   &
!             +(240.)*F%SlabVals(:,3,k)-(120.)*F%SlabVals(:,4,k)   &
!             +(40.)*F%SlabVals(:,5,k)- (6.)*F%SlabVals(:,6,k) )/(120.*h)  

!      create rhs vectors, store in work(:,:,1)
 work(1,:)=( -(5./2.)*F%SlabVals(:,1,k)+(2.)*F%SlabVals(:,2,k)   &
             +(1./2.)*F%SlabVals(:,3,k) )/h
             
 work(2,:)=( -(5./9.)*F%SlabVals(:,1,k)-(1./2.)*F%SlabVals(:,2,k)   &
             +(1.)*F%SlabVals(:,3,k)+(1./18.)*F%SlabVals(:,4,k) )/h


 work(3,:) = ( (1./36.)*( F%SlabVals(:,5,k) - F%SlabVals(:,1,k) )   &
              + (7./9.)*( F%SlabVals(:,4,k) - F%SlabVals(:,2,k) ) )/h  
              
 do j=4,m2-3
  work(j,:) = (  (c1/6.)*( F%SlabVals(:,j+3,k) - F%SlabVals(:,j-3,k) )   &
               + (b1/4.)*( F%SlabVals(:,j+2,k) - F%SlabVals(:,j-2,k) )   &
               + (a1/2.)*( F%SlabVals(:,j+1,k) - F%SlabVals(:,j-1,k) ) )/h  
 enddo

 work(m2-2,:) = (-(1./36.)*( F%SlabVals(:,m2-4,k) - F%SlabVals(:,m2,k) )   &
              - (7./9.)*( F%SlabVals(:,m2-3,k) - F%SlabVals(:,m2-1,k) ) )/h  
              
 work(m2-1,:)= ( (5./9.)*F%SlabVals(:,m2,k)+(1./2.)*F%SlabVals(:,m2-1,k)   &
               -(1.)*F%SlabVals(:,m2-2,k)-(1./18.)*F%SlabVals(:,m2-3,k) )/h
               
 work(m2,:)= ( (5./2.)*F%SlabVals(:,m2,k)-(2.)*F%SlabVals(:,m2-1,k)   &
               -(1./2.)*F%SlabVals(:,m2-2,k) )/h

!! These are high order one sided formula, to use, 
!! the setup matrix need needs to be adjusted
 !work(m2-1,:)=( (24.)*F%SlabVals(:,m2,k)+(130.)*F%SlabVals(:,m2-1,k)   &
 !            -(240.)*F%SlabVals(:,m2-2,k)+(120.)*F%SlabVals(:,m2-3,k)   &
 !            -(40.)*F%SlabVals(:,m2-4,k)+ (6.)*F%SlabVals(:,m2-5,k) )/(120.*h) 
             
 !work(m2,:)=(  (274.)*F%SlabVals(:,m2,k)-(600.)*F%SlabVals(:,m2-1,k)   &
 !            +(600.)*F%SlabVals(:,m2-2,k)-(400.)*F%SlabVals(:,m2-3,k)   &
 !            +(150.)*F%SlabVals(:,m2-4,k)- (24.)*F%SlabVals(:,m2-5,k) )/(120.*h)  
             
             
!Compute the y derivatives for plane k, overwriting the array work
  call DGBTRS('N',m2,2,2,m1,F%Derivs(idim)%A_Diff,7,F%Derivs(idim)%ipiv,work,m2,ierr)
        
!!store the results in deriv(:,:,k) and apply metric from s <--> y
!! i.e. dF/dy = dF/ds *ds/dy
 dF%SlabVals(1:m1,1:m2,k)=transpose( work(1:m2,1:m1)  )
 do i=1,m1
  dF%SlabVals(i,:,k) = dF%SlabVals(i,:,k) * dsdy(:) 
 enddo                                               

!Go on to the next k plane
 enddo

 return
end subroutine compact_ddy


subroutine compact_ddz(F,dF)
 use class_CoordinateMap
 implicit none
 type( ScalarVariable ), intent(inout)   :: dF
 type( ScalarVariable ), intent(in)      :: F
 integer                                 :: idim=3
 integer                                 :: i,j,k
 integer,save                                 :: m1,m2,m3
 real(kind=8),save                            :: h,a1,b1,c1
 real(kind=8),allocatable,save                :: work(:,:)
 real(kind=8),allocatable,save                :: dsdz(:)
 integer                                 :: ierr
 integer, SAVE                           :: myid,numprocs
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
 !!=================================================================== 
 !! m1,m2,m3    dimensions of array with data to be differentiated
 !! h           x grid spacing
 !! a1,b1,c1    coefficients of method, assigned to Derivative object
 !!===================================================================
 
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   !! in idim=3 case, we work w/ data stord in "Column" format
   m1=F%DataMap%ColDims(1,myid)
   m2=F%DataMap%ColDims(2,myid)
   m3=F%DataMap%ColDims(3,myid)
   allocate(work(m3,m1))
   allocate(dsdz(m3))
   h = F%Grid%SpaceDims(idim)%ds
   dsdz = 1.d0/F%Grid%SpaceDims(idim)%dxds
   a1 = F%Derivs(idim)%a1
   b1 = F%Derivs(idim)%b1
   c1 = F%Derivs(idim)%c1
   first_entry=.FALSE.
  endif 
 
 
 
!loop through each xz plane       
 do j=1,m2

!! These are high order one sided formula, to use, 
!! the setup matrix need needs to be adjusted
 !work(1,:)=( -(274.)*F%ColVals(:,j,1)+(600.)*F%ColVals(:,j,2)   &
 !            -(600.)*F%ColVals(:,j,3)+(400.)*F%ColVals(:,j,4)   &
 !            -(150.)*F%ColVals(:,j,5)+(24.)*F%ColVals(:,j,6) )/(120.*h)  
 !work(2,:)=( -(24.)*F%ColVals(:,j,1)-(130.)*F%ColVals(:,j,2)   &
 !            +(240.)*F%ColVals(:,j,3)-(120.)*F%ColVals(:,j,4)   &
 !            +(40.)*F%ColVals(:,j,5)-(6.)*F%ColVals(:,j,6) )/(120.*h)   
 
 !create rhs vectors, store in work(:,:,1)
 work(1,:)=(-(5./2.)*F%ColVals(:,j,1)+(2.)*F%ColVals(:,j,2)   &
          +(1./2.)*F%ColVals(:,j,3) )/h
          
 work(2,:)=(-(5./9.)*F%ColVals(:,j,1)-(1./2.)*F%ColVals(:,j,2)   &
          +(1.)*F%ColVals(:,j,3)+(1./18.)*F%ColVals(:,j,4) )/h
             
 work(3,:)=( (1./36.)* ( F%ColVals(:,j,5) - F%ColVals(:,j,1) )  &
            + (7./9.)* ( F%ColVals(:,j,4) - F%ColVals(:,j,2) ) )/h
            
 do k=4,m3-3
  work(k,:)=( (c1/6.)* ( F%ColVals(:,j,k+3) - F%ColVals(:,j,k-3) )  &
             + (b1/4.)* ( F%ColVals(:,j,k+2) - F%ColVals(:,j,k-2) )  &
             + (a1/2.)* ( F%ColVals(:,j,k+1) - F%ColVals(:,j,k-1) ) )/h
 enddo
 
 work(m3-2,:)=(-(1./36.)* ( F%ColVals(:,j,m3-4) - F%ColVals(:,j,m3) )  &
            - (7./9.)* ( F%ColVals(:,j,m3-3) - F%ColVals(:,j,m3-1) ) )/h
 work(m3-1,:)= ( (5./9.)*F%ColVals(:,j,m3)+(1./2.)*F%ColVals(:,j,m3-1)   &
                   -(1.)*F%ColVals(:,j,m3-2)-(1./18.)*F%ColVals(:,j,m3-3) )/h                   
 work(m3 ,:)= ( (5./2.)*F%ColVals(:,j,m3)-(2.)*F%ColVals(:,j,m3-1)   &
                   -(1./2.)*F%ColVals(:,j,m3-2) )/h
!! These are high order one sided formula, to use, 
!! the setup matrix need needs to be adjusted
 !work(m3-1,:)=( (24.)*F%ColVals(:,j,m3)+(130.)*F%ColVals(:,j,m3-1)   &
 !             -(240.)*F%ColVals(:,j,m3-2)+(120.)*F%ColVals(:,j,m3-3)   &
 !              -(40.)*F%ColVals(:,j,m3-4)+(6.)*F%ColVals(:,j,m3-5) )/(120.*h)   
 !work(m3,:)=( (274.)*F%ColVals(:,j,m3)-(600.)*F%ColVals(:,j,m3-1)   &
 !            +(600.)*F%ColVals(:,j,m3-2)-(400.)*F%ColVals(:,j,m3-3)   &
 !            +(150.)*F%ColVals(:,j,m3-4)-(24.)*F%ColVals(:,j,m3-5) )/(120.*h)  
             
             
!Compute the z derivatives for plane j, overwriting the array work
  call DGBTRS('N',m3,2,2,m1,F%Derivs(idim)%A_Diff,7,F%Derivs(idim)%ipiv,work(1,1),m3,ierr)

!!store the results in deriv(:,j,:) and apply metric from s <--> z
!! i.e. dF/dz = dF/ds *ds/dz
 do i=1,m1
  dF%ColVals(i,j,:)=work(:,i) * dsdz(:)
 enddo
      
!Go on to the next j plane
 enddo

end subroutine compact_ddz

subroutine compact_ddz_array(SV,F,dF,n1,n2,n3)
 use class_CoordinateMap
 implicit none
 type(ScalarVariable),intent(in)         :: SV
 integer                                 :: n1,n2,n3
 real(kind=8), intent(inout)             :: dF(n1,n2,n3)
 real(kind=8), intent(inout)             :: F(n1,n2,n3)
 integer                                 :: idim=3
 integer                                 :: i,j,k,ierr
 integer, SAVE                           :: myid,numprocs
 integer,SAVE                            :: m1,m2,m3
 real(kind=8),SAVE                       :: h,a1,b1,c1
 real(kind=8),allocatable,SAVE           :: work(:,:)
 real(kind=8),allocatable,SAVE           :: dsdz(:)
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   !!=================================================================== 
   !! m1,m2,m3    dimensions of array with data to be differentiated
   !! h           x grid spacing
   !! a1,b1,c1    coefficients of method, assigned to Derivative object
   !!=================================================================== 
   !! in idim=3 case, we work w/ data stored in "Column" format
   m1=SV%DataMap%ColDims(1,myid)
   m2=SV%DataMap%ColDims(2,myid)
   m3=SV%DataMap%ColDims(3,myid)
   allocate(work(m3,m1))
   allocate(dsdz(m3))
   h = SV%Grid%SpaceDims(idim)%ds
   dsdz = 1.d0/SV%Grid%SpaceDims(idim)%dxds
   a1 = SV%Derivs(idim)%a1
   b1 = SV%Derivs(idim)%b1
   c1 = SV%Derivs(idim)%c1
   first_entry=.FALSE.
  endif  
 
!loop through each xz plane       
 do j=1,m2
 
 !create rhs vectors, store in work(:,:,1)
 work(1,:)=(-(5./2.)*F(:,j,1)+(2.)*F(:,j,2)   &
          +(1./2.)*F(:,j,3) )/h
          
 work(2,:)=(-(5./9.)*F(:,j,1)-(1./2.)*F(:,j,2)   &
          +(1.)*F(:,j,3)+(1./18.)*F(:,j,4) )/h
             
 work(3,:)=( (1./36.)* ( F(:,j,5) - F(:,j,1) )  &
            + (7./9.)* ( F(:,j,4) - F(:,j,2) ) )/h
            
 do k=4,m3-3
  work(k,:)=( (c1/6.)* ( F(:,j,k+3) - F(:,j,k-3) )  &
             + (b1/4.)* ( F(:,j,k+2) - F(:,j,k-2) )  &
             + (a1/2.)* ( F(:,j,k+1) - F(:,j,k-1) ) )/h
 enddo
 
 work(m3-2,:)=(-(1./36.)* ( F(:,j,m3-4) - F(:,j,m3) )  &
            - (7./9.)* ( F(:,j,m3-3) - F(:,j,m3-1) ) )/h
 work(m3-1,:)= ( (5./9.)*F(:,j,m3)+(1./2.)*F(:,j,m3-1)   &
                   -(1.)*F(:,j,m3-2)-(1./18.)*F(:,j,m3-3) )/h                   
 work(m3 ,:)= ( (5./2.)*F(:,j,m3)-(2.)*F(:,j,m3-1)   &
                   -(1./2.)*F(:,j,m3-2) )/h             
             
!Compute the z derivatives for plane j, overwriting the array work
 call DGBTRS('N',m3,2,2,m1,SV%Derivs(idim)%A_Diff,7,SV%Derivs(idim)%ipiv,work(1,1),m3,ierr)

!!store the results in deriv(:,j,:) and apply metric from s <--> z
!! i.e. dF/dz = dF/ds *ds/dz
 do i=1,m1
  dF(i,j,:)=work(:,i) * dsdz(:)
 enddo
      
!Go on to the next j plane
 enddo

end subroutine compact_ddz_array

subroutine filter_x(F,FF)
 implicit none
 type( ScalarVariable ), intent(inout)   :: FF
 type( ScalarVariable ), intent(inout)   :: F
 integer                                 :: idim=1
 integer                                 :: m1,m2,m3
 integer                                 :: i,j,k
 real(kind=8)                            :: a,b,c,d
 real(kind=8),allocatable                :: work(:,:)
 integer                                 :: ierr
 integer, SAVE                           :: myid,numprocs
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   first_entry=.FALSE.
  endif 
 
!!===================================================================
 !! m1,m2,m3    dimensions of array with data to be differentiated
 !! a,b,c,d     coefficients of method, assigned to deriv object
!!===================================================================
 
 m1=F%DataMap%SlabDims(1,myid)
 m2=F%DataMap%SlabDims(2,myid)
 m3=F%DataMap%SlabDims(3,myid)
 allocate(work(m1,m2))
 
 a = F%Derivs(idim)%a
 b = F%Derivs(idim)%b
 c = F%Derivs(idim)%c
 d = F%Derivs(idim)%d

!!loop through each xy horizontal plane       
 do k=1,m3
 
 !Create rhs vectors, store in work(:,:)
 work(1,:)=  (15./16.)*F%SlabVals(1,:,k)                             & 
            +(1./16)*( 4.*F%SlabVals(2,:,k) - 6.*F%SlabVals(3,:,k)   &
                      +4.*F%SlabVals(4,:,k) - F%SlabVals(5,:,k) )   
                           
 work(2,:)=  (3./4.)*F%SlabVals(2,:,k)                             & 
            +(1./16)*( F%SlabVals(1,:,k) + 6.*F%SlabVals(3,:,k)   &
                      -4.*F%SlabVals(4,:,k) + F%SlabVals(5,:,k) )

 work(3,:)=  (5./8.)*F%SlabVals(3,:,k)                             & 
            +(1./16)*(-F%SlabVals(1,:,k) + 4.*F%SlabVals(2,:,k)   &
                      +4.*F%SlabVals(4,:,k) - F%SlabVals(5,:,k) )

 do i=4,m1-3
  work(i,:)= ( a*(F%SlabVals(i,:,k))   & 
             + b*(F%SlabVals(i+1,:,k)+F%SlabVals(i-1,:,k))   & 
             + c*(F%SlabVals(i+2,:,k)+F%SlabVals(i-2,:,k))   &
             + d*(F%SlabVals(i+3,:,k)+F%SlabVals(i-3,:,k))  )
 enddo

 work(m1,:)=  (15./16.)*F%SlabVals(m1,:,k)                             & 
            +(1./16)*( 4.*F%SlabVals(m1-1,:,k) - 6.*F%SlabVals(m1-2,:,k)   &
                      +4.*F%SlabVals(m1-3,:,k) - F%SlabVals(m1-4,:,k) )   
                           
 work(m1-1,:)=  (3./4.)*F%SlabVals(m1-1,:,k)                             & 
            +(1./16)*( F%SlabVals(m1,:,k) + 6.*F%SlabVals(m1-2,:,k)   &
                      -4.*F%SlabVals(m1-3,:,k) + F%SlabVals(m1-4,:,k) )

 work(m1-2,:)=  (5./8.)*F%SlabVals(m1-2,:,k)                             & 
            +(1./16)*(-F%SlabVals(m1,:,k) + 4.*F%SlabVals(m1-1,:,k)   &
                      +4.*F%SlabVals(m1-3,:,k) - F%SlabVals(m1-4,:,k) )

!!Compute the filtered for plane k, overwriting the array work
  call DGBTRS('N',m1,2,2,m2,F%Derivs(idim)%A_filt,7,F%Derivs(idim)%jpiv,work,m1,ierr)

  do j=1,m2
   FF%SlabVals(:,j,k)=work(:,j)
  enddo

!!Go on to the next k plane
 enddo

 if(allocated(work)) deallocate(work)
 return
end subroutine filter_x

subroutine filter_y(F,FF)
 implicit none
 type( ScalarVariable ), intent(inout)   :: FF
 type( ScalarVariable ), intent(inout)   :: F
 integer                                 :: idim=2
 integer                                 :: m1,m2,m3
 integer                                 :: i,j,k
 real(kind=8)                            :: a,b,c,d
 real(kind=8),allocatable                :: work(:,:)
 integer                                 :: ierr
 integer, SAVE                           :: myid,numprocs
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   first_entry=.FALSE.
  endif 
 
!!===================================================================
 !! m1,m2,m3    dimensions of array with data to be differentiated
 !! a,b,c,d     coefficients of method, assigned to deriv object
!!===================================================================
 
 m1=F%DataMap%SlabDims(1,myid)
 m2=F%DataMap%SlabDims(2,myid)
 m3=F%DataMap%SlabDims(3,myid)
 allocate(work(m2,m1))
 
 a = F%Derivs(idim)%a
 b = F%Derivs(idim)%b
 c = F%Derivs(idim)%c
 d = F%Derivs(idim)%d

!!loop through each xy horizontal plane       
 do k=1,m3

 work(1,:)=  (15./16.)*F%SlabVals(:,1,k)                             & 
            +(1./16)*( 4.*F%SlabVals(:,2,k) - 6.*F%SlabVals(:,3,k)   &
                      +4.*F%SlabVals(:,4,k) - F%SlabVals(:,5,k) )   
                           
 work(2,:)=  (3./4.)*F%SlabVals(:,2,k)                             & 
            +(1./16)*( F%SlabVals(:,1,k) + 6.*F%SlabVals(:,3,k)   &
                      -4.*F%SlabVals(:,4,k) + F%SlabVals(:,5,k) )

 work(3,:)=  (5./8.)*F%SlabVals(:,3,k)                             & 
            +(1./16)*(-F%SlabVals(:,1,k) + 4.*F%SlabVals(:,2,k)   &
                      +4.*F%SlabVals(:,4,k) - F%SlabVals(:,5,k) ) 
              
 do j=4,m2-3
  work(j,:)= ( a*(F%SlabVals(:,j,k))   & 
             + b*(F%SlabVals(:,j+1,k)+F%SlabVals(:,j-1,k))   & 
             + c*(F%SlabVals(:,j+2,k)+F%SlabVals(:,j-2,k))   &
             + d*(F%SlabVals(:,j+3,k)+F%SlabVals(:,j-3,k))  ) 
 enddo

 work(m2,:)=  (15./16.)*F%SlabVals(:,m2,k)                             & 
             +(1./16)*( 4.*F%SlabVals(:,m2-1,k) - 6.*F%SlabVals(:,m2-2,k)   &
                      +4.*F%SlabVals(:,m2-3,k) - F%SlabVals(:,m2-4,k) )   
                           
 work(m2-1,:)=  (3./4.)*F%SlabVals(:,m2-1,k)                             & 
             +(1./16)*( F%SlabVals(:,m2,k) + 6.*F%SlabVals(:,m2-2,k)   &
                      -4.*F%SlabVals(:,m2-3,k) + F%SlabVals(:,m2-4,k) )

 work(m2-2,:)=  (5./8.)*F%SlabVals(:,m2-2,k)                             & 
             +(1./16)*(-F%SlabVals(:,m2,k) + 4.*F%SlabVals(:,m2-1,k)   &
                      +4.*F%SlabVals(:,m2-3,k) - F%SlabVals(:,m2-4,k) ) 

             
  call DGBTRS('N',m2,2,2,m1,F%Derivs(idim)%A_filt,7,F%Derivs(idim)%jpiv,work,m2,ierr)
          
!!store the results in deriv(:,:,k) 
 FF%SlabVals(1:m1,1:m2,k)=transpose( work(1:m2,1:m1)  )
                                              
!Go on to the next k plane
 enddo

 if(allocated(work)) deallocate(work)
 return
end subroutine filter_y

subroutine filter_z(F,FF)
 implicit none
 type( ScalarVariable ), intent(inout)   :: FF
 type( ScalarVariable ), intent(inout)   :: F
 integer                                 :: idim=3
 integer                                 :: m1,m2,m3
 integer                                 :: i,j,k
 real(kind=8)                            :: a,b,c,d
 real(kind=8),allocatable                :: work(:,:)
 integer                                 :: ierr
 integer, SAVE                           :: myid,numprocs
 logical, SAVE                           :: first_entry=.TRUE.
   
  include 'mpif.h'
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   first_entry=.FALSE.
  endif 
 
 !!===================================================================
 !! m1,m2,m3    dimensions of array with data to be differentiated
 !! a,b,c,d     coefficients of method, assigned to deriv object
!!===================================================================
 
 !! in idim=3 case, we work w/ data stord in "Column" format
 m1=F%DataMap%ColDims(1,myid)
 m2=F%DataMap%ColDims(2,myid)
 m3=F%DataMap%ColDims(3,myid)
 allocate(work(m3,m1))
  
 a = F%Derivs(idim)%a
 b = F%Derivs(idim)%b
 c = F%Derivs(idim)%c
 d = F%Derivs(idim)%d
 
!loop through each xz plane       
 do j=1,m2

 !create rhs vectors, store in work(:,:,1)
 work(1,:)=  (15./16.)*F%SlabVals(:,j,1)                             & 
            +(1./16)*( 4.*F%SlabVals(:,j,2) - 6.*F%SlabVals(:,j,3)   &
                      +4.*F%SlabVals(:,j,4) - F%SlabVals(:,j,5) )   
                           
 work(2,:)=  (3./4.)*F%SlabVals(:,j,2)                             & 
            +(1./16)*( F%SlabVals(:,j,1) + 6.*F%SlabVals(:,j,3)   &
                      -4.*F%SlabVals(:,j,4) + F%SlabVals(:,j,5) )

 work(3,:)=  (5./8.)*F%SlabVals(:,j,3)                             & 
            +(1./16)*(-F%SlabVals(:,j,1) + 4.*F%SlabVals(:,j,2)   &
                      +4.*F%SlabVals(:,j,4) - F%SlabVals(:,j,5) ) 
            
 do k=4,m3-3
  work(k,:)= ( a*(F%SlabVals(:,j,k))   & 
             + b*(F%SlabVals(:,j,k+1)+F%SlabVals(:,j,k-1))   & 
             + c*(F%SlabVals(:,j,k+2)+F%SlabVals(:,j,k-2))   &
             + d*(F%SlabVals(:,j,k+3)+F%SlabVals(:,j,k-3))  ) 
 enddo
 
 work(m3,:)=  (15./16.)*F%SlabVals(:,j,m3)                             & 
            +(1./16)*( 4.*F%SlabVals(:,j,m3-1) - 6.*F%SlabVals(:,j,m3-2)   &
                      +4.*F%SlabVals(:,j,m3-3) - F%SlabVals(:,j,m3-4) )   
                           
 work(m3-1,:)=  (3./4.)*F%SlabVals(:,j,m3-1)                             & 
            +(1./16)*( F%SlabVals(:,j,m3) + 6.*F%SlabVals(:,j,m3-2)   &
                      -4.*F%SlabVals(:,j,m3-3) + F%SlabVals(:,j,m3-4) )

 work(m3-2,:)=  (5./8.)*F%SlabVals(:,j,m3-2)                             & 
            +(1./16)*(-F%SlabVals(:,j,m3) + 4.*F%SlabVals(:,j,m3-1)   &
                      +4.*F%SlabVals(:,j,m3-3) - F%SlabVals(:,j,m3-4) )
                      
  call DGBTRS('N',m3,2,2,m1,F%Derivs(idim)%A_filt,7,F%Derivs(idim)%jpiv,work(1,1),m3,ierr)

!!store the results in FF(:,j,:)
 do i=1,m1
  FF%ColVals(i,j,:)=work(:,i)
 enddo
      
!Go on to the next j plane
 enddo

 if(allocated(work)) deallocate(work)
end subroutine filter_z

!=============================================================== 
!===============================================================
 subroutine get_ghost_data(xdim_periodic,ydim_periodic,F)
  
  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: F
  logical                :: xdim_periodic,ydim_periodic
  integer                :: status_array(mpi_status_size,1024)
  integer                :: nreqs,reqs(1024),ierr
  integer                :: nx,ny,locnz
  integer, SAVE          :: myid,numprocs
  logical, SAVE          :: first_entry=.TRUE.
   
  if( first_entry ) then
   call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      
   call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)
   first_entry=.FALSE.
  endif 

 
  
  !!============================================================================
  !! for x periodic case
  !!============================================================================
  if( xdim_periodic ) then
   nx = F%DataMap%SlabDims(1,myid)
   F%Ghost_x(:,:,1) = F%SlabVals(nx,:,:)  !! left
   F%Ghost_x(:,:,2) = F%SlabVals(1,:,:)   !! right
  endif
  
  !!============================================================================
  !! for y periodic case
  !!============================================================================
  if( ydim_periodic ) then
   ny = F%DataMap%SlabDims(2,myid)
   F%Ghost_y(:,:,1) = F%SlabVals(:,ny,:)  !! front
   F%Ghost_y(:,:,2) = F%SlabVals(:,1,:)   !! back
  endif
  !!============================================================================
  !!============================================================================
   
  if(numprocs==1) return 
  !!============================================================================
  !!============================================================================
  
  
  !!============================================================================
  !!MULTI-PROCESSOR CASE: INTERIOR NODES  ======================================
  !!============================================================================
  if( numprocs>1 .and. myid > 0 .and. myid < numprocs-1 ) then  
  
   nreqs=0
   !get data from north neighbor
   call get_gvals_from_north(nreqs,reqs,F)
   nreqs=nreqs+1
   
   !get data from south neighbor
   call get_gvals_from_south(nreqs,reqs,F)
   nreqs=nreqs+1
   
   !send data to north
   call send_gvals_north(nreqs,reqs,F)
   nreqs=nreqs+1
   
   !send data to south
   call send_gvals_south(nreqs,reqs,F)
   nreqs=nreqs+1
   
   call mpi_waitall(nreqs,reqs,status_array,ierr)
  
  !!============================================================================
  !!MULTI-PROCESSOR CASE:  TOP NODE  ===========================================
  !!============================================================================
  elseif( numprocs>1 .and. myid==numprocs-1 ) then 
   nreqs=0
   
   !get data from south neighbor
   call get_gvals_from_south(nreqs,reqs,F)
   nreqs=nreqs+1
   
   call send_gvals_south(nreqs,reqs,F)   !!
   nreqs=nreqs+1
   
   call mpi_waitall(nreqs,reqs,status_array,ierr)
  
  !!============================================================================
  !!MULTI-PROCESSOR CASE:  BOTTOM NODE  ========================================
  !!============================================================================
  elseif( numprocs>1 .and. myid==0 ) then  !! bottom node
   
   nreqs=0
   call get_gvals_from_north(nreqs,reqs,F)  !!get data from north
   nreqs=nreqs+1
   
   !send data to north
   call send_gvals_north(nreqs,reqs,F)
   nreqs=nreqs+1
      
   call mpi_waitall(nreqs,reqs,status_array,ierr)
      
  endif  
  
 end subroutine get_ghost_data
!=============================================================== 
!===============================================================


!!===========================================================
subroutine get_gvals_from_north(numreqs,reqs,F)
  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: F
  integer                :: comm,status_array(mpi_status_size,1024)
  integer                :: numreqs,reqs(1024),myid,ierr,numprocs
  integer                :: nx,ny,locnz,source,dest,tag,numwords,j
  
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      !!  ====> myid
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)  !!  ====> numprocs
       
  if(numprocs==1) then
   write(0,*) 'myid  numprocs ',myid,numprocs
   write(0,*) 'I dont have a north neighbor!!'
   stop ' stopping in get_vals_from_north'
  endif
  
  nx = F%DataMap%SlabDims(1,myid)
  ny = F%DataMap%SlabDims(2,myid)
  locnz = F%DataMap%SlabDims(3,myid)
       
  source=myid+1!! NORTH neighbor
  dest=myid   !! me
  numwords=(nx)*(ny)  
  tag=100
    
  j=numreqs+1
  call MPI_IRECV(F%Ghost_z(1,1,2),numwords,    &
                 MPI_DOUBLE_PRECISION,source,tag,comm,  &
                 reqs(j),ierr )
                  
   return
  end subroutine get_gvals_from_north
  !!===========================================================
  
!!===========================================================
subroutine get_gvals_from_south(numreqs,reqs,F)
  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: F
  integer                :: comm,status_array(mpi_status_size,1024)
  integer                :: numreqs,reqs(1024),myid,ierr,numprocs
  integer                :: nx,ny,locnz,source,dest,tag,numwords,j
  
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      !!  ====> myid
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)  !!  ====> numprocs
       
  if(numprocs==1) then
   write(0,*) 'myid  numprocs ',myid,numprocs
   write(0,*) 'I dont have a south neighbor!!'
   stop ' stopping in get_vals_from_south'
  endif
  
  nx = F%DataMap%SlabDims(1,myid)
  ny = F%DataMap%SlabDims(2,myid)
  locnz = F%DataMap%SlabDims(3,myid)
       
  source=myid-1!! SOUTH neighbor
  dest=myid   !! me
  numwords=(nx)*(ny)  
  tag=100
    
  j=numreqs+1
  call MPI_IRECV(F%Ghost_z(1,1,1),numwords,    &
                 MPI_DOUBLE_PRECISION,source,tag,comm,  &
                 reqs(j),ierr )
                  
   return
  end subroutine get_gvals_from_south
  !!===========================================================
  
  
!!=========================================================== 
!!===========================================================
 subroutine send_gvals_south(numreqs,reqs,F)
  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: F
  integer                :: comm,status_array(mpi_status_size,1024)
  integer                :: numreqs,reqs(1024),myid,ierr,numprocs
  integer                :: nx,ny,locnz,source,dest,tag,numwords,j
  
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      !!  ====> myid
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)  !!  ====> numprocs
       
  if(numprocs==1) then
   write(0,*) 'myid  numprocs ',myid,numprocs
   write(0,*) 'I dont have a south neighbor!!'
   stop ' stopping in send_vals_south'
  endif
  
  nx = F%DataMap%SlabDims(1,myid)
  ny = F%DataMap%SlabDims(2,myid)
  locnz = F%DataMap%SlabDims(3,myid)
               
  source=myid  !! me
  dest=myid-1  !! SOUTH neighbor
  numwords=(nx)*(ny) 
  tag=100
   
  !! send lowest data to south
  j=numreqs+1   !! nreqs is the number of previous requests for myid
  call MPI_ISEND(F%SlabVals(1,1,1),numwords,   &
                 MPI_DOUBLE_PRECISION,dest,tag,comm, &
                 reqs(j),ierr)
          
  return
 end subroutine send_gvals_south
  !!===========================================================
  
!!=========================================================== 
!!===========================================================
 subroutine send_gvals_north(numreqs,reqs,F)
  implicit none
  include 'mpif.h'
  type( ScalarVariable ) :: F
  integer                :: comm,status_array(mpi_status_size,1024)
  integer                :: numreqs,reqs(1024),myid,ierr,numprocs
  integer                :: nx,ny,locnz,source,dest,tag,numwords,j
  
  call mpi_comm_rank (MPI_COMM_WORLD,myid,ierr)      !!  ====> myid
  call mpi_comm_size (MPI_COMM_WORLD,numprocs,ierr)  !!  ====> numprocs
       
  if(numprocs==1) then
   write(0,*) 'myid  numprocs ',myid,numprocs
   write(0,*) 'I dont have a south neighbor!!'
   stop ' stopping in send_vals_south'
  endif
  
  nx = F%DataMap%SlabDims(1,myid)
  ny = F%DataMap%SlabDims(2,myid)
  locnz = F%DataMap%SlabDims(3,myid)
               
  source=myid  !! me
  dest=myid+1  !! NORTH neighbor
  numwords=(nx)*(ny) 
  tag=100
   
  !! send top data to north
  j=numreqs+1   !! nreqs is the number of previous requests for myid
  call MPI_ISEND(F%SlabVals(1,1,locnz),numwords,   &
                 MPI_DOUBLE_PRECISION,dest,tag,comm, &
                 reqs(j),ierr)
          
  return
 end subroutine send_gvals_north
  !!===========================================================



end Module class_ScalarVariable
