!!================================================================
!!================================================================
subroutine pressure_solve
!!================================================================
!!================================================================
 use class_ScalarVariable
 use user_data,               only: psolve_exp_coeffs,    &
                                    psolve_at_gridpoints, &
                                    nx,ny
 use mpi_params,              only: myid
 use intermediate_variables,  only: Phi,scratch,rhs_b,    &
                                    implicit_rhs,         &
                                    explicit_rhs
 use etc,                     only: field_id,             &
                                    numzplanes_below,     &
                                    istep,N,NM1,MM0,M_oldest
 use dependent_variables,     only: velocity                                    
 implicit none
 real(kind=8),save               :: dsdx,dsdy,dsdz
 real(kind=8)                    :: xx,ff
 integer                         :: i,j,k,kg
 integer,save                    :: locnz
 logical                         :: forward
 logical                         :: inverse 
 logical                         :: nonzero_rhs
 logical,save                    :: uniform_spacing
 logical,save                    :: first_entry=.TRUE.
 
 !!================================================================
 if( first_entry ) then
  locnz=Phi(0)%DataMap%SlabVals(3,myid)
  dsdx = 1.d0/Phi(0)%Grid%SpaceDims(1)%dxds(1)
  if(ny>1) then
   dsdy = 1.d0/Phi(0)%Grid%SpaceDims(2)%dxds(1)
  else
   dsdy = 0.d0
  endif
  dsdz = 1.d0/Phi(0)%Grid%SpaceDims(3)%dxds(1)
  uniform_spacing=.FALSE.
  if( Phi(0)%Grid%SpaceDims(1)%uniform_dx  .and.      &
      Phi(0)%Grid%SpaceDims(2)%uniform_dx  .and.      &
      Phi(0)%Grid%SpaceDims(3)%uniform_dx  ) then
   uniform_spacing=.TRUE.
  endif
  first_entry=.FALSE.
 endif
 !!================================================================ 
 field_id=6

 !!================================================================
 !! Two approaches: solve for expansion coefficients, implicit BCs
 !!                 solve for gridpoint values, explicit BCs
 !! In either case, on entry we the have rhs_b gridpoint values.
 !!================================================================
 if( psolve_exp_coeffs ) then
 
  if(myid==0)   &
   write(0,*)'... solving for pressure expansion coefficients'
    
  !!============================================================
  !! transform rhs_b in all three directions, store result back
  !! into rhs_b, use scratch(1:2) as scratch arrays. Make sure the
  !! deriv operators for rhs_b match those of phi.
  !!============================================================
   forward=.TRUE.
   inverse=.FALSE.
   do i=1,3
    rhs_b%Derivs(i)=Phi(0)%Derivs(i) !!overloaded operator (=)
   enddo
   call Transform3D(rhs_b,rhs_b,scratch(1),scratch(2),forward,inverse)
  
  
  !!============================================================
  !! Now, we solve for the expansion coefficients in one of two
  !! ways. If ALL grid mappings are uniform, we have a straight-
  !! forward, analytical solution in wavenumber space. If, any 
  !! of the "x,y,z" <===> s mappings are nonuniform, then we
  !! can't explicitly form the soln in wavenumber space but can
  !! construct a reasonable approximation assuming the grid 
  !! stretchings don't dominate the transformations of dxx etc.
  !! Either way, construct the "uniform grid" solution here.
  !!============================================================
        
   if(uniform_spacing) then 
    if(myid==0) then 
     write(0,*)   &
     '...     uniform grid ==> analytical solve in wavenumber space'
    endif
   elseif(.NOT. uniform_spacing) then
    if(myid==0) then 
     write(0,*)   &
     '...     nonuniform grid ==> approximate solve in wavenumber space'
    endif
   endif
    
    do k=1,locnz
     kg = numzplanes_below(myid) + k
      do j=1,ny
       do i=1,nx
        
         xx = -( Phi(0)%Derivs(1)%kx(i)  * dsdx )**2  &
              -( Phi(0)%Derivs(2)%kx(j)  * dsdy )**2  &
              -( Phi(0)%Derivs(3)%kx(kg) * dsdz )**2 
              
         ff = Phi(0)%Derivs(1)%kfilter(i)     &
             *Phi(0)%Derivs(2)%kfilter(j)     &
             *Phi(0)%Derivs(3)%kfilter(kg)
         
         Phi(0)%SlabVals(i,j,k) =  (rhs_b%SlabVals(i,j,k)/xx) * ff
               
                  
        enddo
       enddo
      enddo
      Phi(0)%SlabVals(1,1,1)=0.d0
    
        
    !!===============================================================
    !! If Phi(0) is only a first approximation, call the iterative 
    !! solver with rhs_b in wavenumber space and the approximation 
    !! of the expansion coefficients used as the initial guess.
    !!===============================================================
    if(.NOT. uniform_spacing) then  !! No analytical soln because
                                    !!(s<-->x) must be done in p space.
                                    !! But still do iterations in w-n space.    
     write(0,*)    &
     '...     nonuniform grid ==> iterative solve in wavenumber space'
     
     Phi(0)%wavenumber_values = .TRUE.
     Phi(0)%gridpoint_values  = .FALSE.     
     Phi(0)%forward_transform_only=.FALSE.
     Phi(0)%inverse_transform_only=.FALSE.
     call DoSolve(Phi(0))
    
    endif   
    
    !!=====================================================
    !! Compute the gradient of Phi(0), storing
    !! the results in the arrays diff(3,4,5) which are
    !! free prior to the diffusion solves. The values
    !! of Phi(0)%gridpoint_values & Phi(0)%wavenumber_values
    !! will be checked in the differentiation routines
    !! and used to determine whether to do a transform
    !! before the 1st wavenumber multiply or not.
    !!=====================================================
    !Phi(0)%wavenumber_values = .TRUE.
    !Phi(0)%gridpoint_values  = .FALSE.     
    !Phi(0)%forward_transform_only=.FALSE.
    !Phi(0)%inverse_transform_only=.FALSE.
    !call Gradient(Phi(0),diff(3))    
        
    
    !!========================================================
    !! 3D inverse transform the pressure. This step could be
    !! skipped, only the gradient is needed to do the time step.
    !!========================================================    
    forward=.FALSE.
    inverse=.TRUE. 
    Phi(0)%inverse_transform_only=.TRUE.
    call Transform3D(Phi(0),Phi(0),scratch(1),scratch(2),forward,inverse)
    
    Phi(0)%wavenumber_values = .FALSE.
    Phi(0)%gridpoint_values  = .TRUE.      
    Phi(0)%forward_transform_only=.FALSE.
    Phi(0)%inverse_transform_only=.FALSE.
    call Gradient(Phi(0),explicit_rhs(1,M_oldest)) 
    return
        
    
  elseif(  psolve_at_gridpoints ) then
  !!========================================================
  !! If solving for the Phi(0) gridpoint values, go straight  
  !! to the iterative solver, leaving rhs_b in its physical  
  !! space representation.
  !!========================================================
  
    write(0,*)'...   iterative solve for gridpoint values'
  
    Phi(0)%gridpoint_values  = .TRUE.
    Phi(0)%wavenumber_values = .FALSE.
    Phi(0)%forward_transform_only=.FALSE.
    Phi(0)%inverse_transform_only=.FALSE.
    rhs_b%gridpoint_values  = .TRUE.
    rhs_b%wavenumber_values = .FALSE.
    rhs_b%forward_transform_only=.FALSE.
    rhs_b%inverse_transform_only=.FALSE.
   
    call DoSolve(Phi(0))
    
    !!===================================================== 
    !! Store the gradients in explicit_rhs(1,M_oldest) which are
    !! free until the the next time step. The values
    !! of Phi%gridpoint_values & Phi%wavenumber_values
    !! will be checked in the differentiation routines
    !! and used to determine whether to do a transform
    !! before the 1st wavenumber multiply or not.
    !!=====================================================
 
    call Gradient(Phi(0),explicit_rhs(1,M_oldest))
  
  endif  
  if(myid==0) write(0,*)  &
  '...     pressure and pressure gradient computed'
        
end subroutine pressure_solve
!!=====================================================
!!=====================================================




!!=====================================================
!!=====================================================
subroutine diffusion_solve
 use class_ScalarVariable
 use operators,               only: alpha,beta
 use mpi_params,              only: myid
 use intermediate_variables,  only: rhs_b,scratch
 use dependent_variables,     only: velocity,scalars
 use etc,                     only: field_id,numzplanes_below,N,NM1
 use user_data,               only: nx,ny,do_immersed_boundary, &
                                    usolve_exp_coeffs,          &
                                    vsolve_exp_coeffs,          &
                                    wsolve_exp_coeffs,          &
                                    s1solve_exp_coeffs,         &
                                    s2solve_exp_coeffs,         &
                                    runlabel,dt,LES,DNS,        &
                                    alpha_les,p_les
                                    
 implicit none
 integer,save                    :: locnz,i,j,k,id,kg,myexp
 real(kind=8),save               :: dsdx,dsdy,dsdz,xx_dns,xx_les,ff,pi
 logical                         :: p_space,w_space
 logical                         :: forward,inverse
 logical, save                   :: uniform_spacing
 logical, save                   :: first_entry=.TRUE.
 
 !!========================================================
 !!========================================================


 if( first_entry ) then
  locnz=velocity(1)%DataMap%SlabVals(3,myid)
  dsdx = 1.d0/velocity(1)%Grid%SpaceDims(1)%dxds(1)
  dsdy = 1.d0/velocity(1)%Grid%SpaceDims(2)%dxds(1)
  dsdz = 1.d0/velocity(1)%Grid%SpaceDims(3)%dxds(1)
  uniform_spacing=.TRUE.
  if(.NOT. velocity(1)%Grid%SpaceDims(1)%uniform_dx) uniform_spacing=.FALSE.
  if(.NOT. velocity(1)%Grid%SpaceDims(2)%uniform_dx) uniform_spacing=.FALSE.
  if(.NOT. velocity(1)%Grid%SpaceDims(3)%uniform_dx) uniform_spacing=.FALSE.
  myexp=2
  if( trim(runlabel)=='longs_IW' ) myexp=6
  first_entry=.FALSE.
 endif
 !!========================================================
 !!========================================================
 id=field_id
 
 !!========================================================
 !!========================================================
 !! determine whether we solve in physical or wavenumber space
 !!========================================================
 !!========================================================
 p_space=.TRUE.
 w_space=.FALSE.
 if(id==1 .and. usolve_exp_coeffs) w_space=.TRUE.
 if(id==2 .and. vsolve_exp_coeffs) w_space=.TRUE.
 if(id==3 .and. wsolve_exp_coeffs) w_space=.TRUE.
 if(id==4 .and. s1solve_exp_coeffs) w_space=.TRUE.
 if(id==5 .and. s2solve_exp_coeffs) w_space=.TRUE.
 if( w_space ) p_space=.FALSE.
  
 
 if( p_space ) then

 !!=========================================================
 !! for soln in physical space, make sure flags are correct
 !! and then do an iterative solve for gridpoint values
 !!=========================================================
 
  rhs_b%gridpoint_values  = .TRUE.
  rhs_b%wavenumber_values = .FALSE.
  rhs_b%forward_transform_only=.FALSE.
  rhs_b%inverse_transform_only=.FALSE.
 
  if( id <= 3 ) then
   if(myid==0) write(0,*)'... diffusion update of gridpoint values for velocity: ',id
   velocity(id)%gridpoint_values      =.TRUE.
   velocity(id)%wavenumber_values     =.FALSE.
   velocity(id)%forward_transform_only=.FALSE.
   velocity(id)%inverse_transform_only=.FALSE.
   
   call DoSolve(velocity(id))
   
  elseif( id > 3) then
   if(myid==0) write(0,*)'... diffusion update of gridpoint values for scalar: ',id-3
   scalars(id-3)%gridpoint_values      =.TRUE.
   scalars(id-3)%wavenumber_values     =.FALSE.
   scalars(id-3)%forward_transform_only=.FALSE.
   scalars(id-3)%inverse_transform_only=.FALSE.
   
   call DoSolve(scalars(id-3))
  
  else
   stop 'id problem in DiffusionSolve, physical space'
  endif 
 
 
 elseif( w_space ) then

 if( id <= 3 ) then
   if(myid==0)   &
     write(0,*)'... diffusion update of expansion coeffs for velocity: ',id
 elseif( id > 3 .and. id < 6) then
   if(myid==0)   &
     write(0,*)'... diffusion update of expansion coeffs for scalar:   ',id-3
 else
   stop 'id problem in DiffusionSolve, wavenumber space'
 endif
 
 !!============================================================
 !! make sure rhs_b derivs are set correctly
 !! so that transforms are done properly
 !!============================================================
  do i=1,3   !!overloaded operator (=)
   if( id <= 3 ) then
    rhs_b%Derivs(i)=velocity(id)%Derivs(i)
   elseif( id > 3 .and. id < 6) then
    rhs_b%Derivs(i)=scalars(id-3)%Derivs(i)
   endif
  enddo
 
 !!============================================================
 !! transform rhs_b in all three directions, store result back
 !! into rhs_b, use scratch(1:2) as scratch arrays.
 !!============================================================
  forward=.TRUE.
  inverse=.FALSE. 
  call Transform3D(rhs_b,rhs_b,scratch(1),scratch(2),forward,inverse) 
       
    
  !!============================================================
  !!============================================================
  !! if we have constant grid spacing in each dimension, 
  !! (and no IB to worry about) then
  !! we can do the transformed diffusion update analytically
  !!
  !! transformed eqns have the form
  !!       dy/dt = ( xx_dns + xx_les )y
  !!                xx_dns = alpha*(k^2 + l^2 + m^2)
  !!                xx_les = alpha_les(1)*k^(2*p_les) 
  !!                       + alpha_les(2)*l^(2*p_les)
  !!                       + alpha_les(3)*m^(2*p_les)
  !!
  !!      analytic soln is
  !!      
  !!        y = y_n-1 * exp{ -( xx_dns + xx_les )*dt }
  !!
  !! where
  !! alpha = 1/Re  or 1/PrRe
  !!
  !!============================================================
  !!============================================================
  
  if(uniform_spacing .and. .not. do_immersed_boundary) then
   if(myid==0 ) write(0,*)'...  No IB, constant grid spacing, solve is analytical'
   
     if( id <= 3 ) then ! velocity components
          
      do k=1,locnz
       kg = numzplanes_below(myid) + k
       do j=1,ny
        do i=1,nx
        
         xx_dns = alpha(id)*(                                         &
                    ( velocity(id)%Derivs(1)%kx(i)  * dsdx )**2       &
                   +( velocity(id)%Derivs(2)%kx(j)  * dsdy )**2       &
                   +( velocity(id)%Derivs(3)%kx(kg) * dsdz )**2   )    
         
         xx_dns = exp(-xx_dns * dt )
         
         xx_les = (                                                            &
           alpha_les(1)*( velocity(id)%Derivs(1)%kx(i) * dsdx )**(2*p_les(1))  &
          +alpha_les(2)*( velocity(id)%Derivs(2)%kx(j) * dsdy )**(2*p_les(2))  &
          +alpha_les(3)*( velocity(id)%Derivs(3)%kx(kg) * dsdz )**(2*p_les(3)) )
                           
          
         xx_les = exp(-xx_les * dt )
         
         ff = velocity(id)%Derivs(1)%kfilter(i)     &
             *velocity(id)%Derivs(2)%kfilter(j)     &
             *velocity(id)%Derivs(3)%kfilter(kg)
         
         velocity(id)%SlabVals(i,j,k) = ff * rhs_b%SlabVals(i,j,k)*xx_dns*xx_les   
                   
        enddo
       enddo
      enddo
      
     elseif( id==4 .or. id==5 ) then  ! scalars components
      do k=1,locnz
       kg = numzplanes_below(myid) + k
       do j=1,ny
        do i=1,nx
        
         xx_dns = alpha(id)*(                                         &
                    ( scalars(id-3)%Derivs(1)%kx(i)  * dsdx )**2       &
                   +( scalars(id-3)%Derivs(2)%kx(j)  * dsdy )**2       &
                   +( scalars(id-3)%Derivs(3)%kx(kg) * dsdz )**2   )    
         
         xx_dns = exp(-xx_dns * dt )
         
         xx_les = (                                                          &
           alpha_les(1)*( scalars(id-3)%Derivs(1)%kx(i) * dsdx )**(2*p_les(1)) &
          +alpha_les(2)*( scalars(id-3)%Derivs(2)%kx(j) * dsdy )**(2*p_les(2)) &
          +alpha_les(3)*( scalars(id-3)%Derivs(3)%kx(kg) * dsdz )**(2*p_les(3)) )
                           
          
         xx_les = exp(-xx_les * dt )
         
         ff = scalars(id-3)%Derivs(1)%kfilter(i)     &
             *scalars(id-3)%Derivs(2)%kfilter(j)     &
             *scalars(id-3)%Derivs(3)%kfilter(kg)
         
         scalars(id-3)%SlabVals(i,j,k) = ff * rhs_b%SlabVals(i,j,k)*xx_dns*xx_les
         
         
        enddo
       enddo
      enddo
     endif  !! end velocity or scalar
    
    else  !! either we have a stretched grid or an IB ==> iterative solve
    
     !! No analytical soln because
     !!(s<-->x) must be done in p space
     !! or there is an additional IB term.
     !! We still do iterations in w-n space however.    
     if(myid==0) write(0,*)'... IB and/or non-uniform grid spacing, solve is iterative'
          
     if( id <= 3 ) then
      velocity(id)%gridpoint_values      =.FALSE.
      velocity(id)%wavenumber_values     =.TRUE.
      velocity(id)%forward_transform_only=.FALSE.
      velocity(id)%inverse_transform_only=.FALSE.
      
          call DoSolve(velocity(id))
      
     else
      scalars(id-3)%gridpoint_values      =.FALSE.
      scalars(id-3)%wavenumber_values     =.TRUE.
      scalars(id-3)%forward_transform_only=.FALSE.
      scalars(id-3)%inverse_transform_only=.FALSE.
      
          call DoSolve(scalars(id-3))
      
     endif
    
    endif
      
  !!=========================================================
  !!=========================================================
  !! In either case,
  !! transform the wavenumber soln back to physical space,
  !! overwriting the input arrays.
  !!=========================================================
  !!========================================================= 
  forward=.FALSE.
  inverse=.TRUE.  
  if(id<=3) then
   call Transform3D(velocity(id),velocity(id),scratch(1),scratch(2),forward,inverse)
  elseif(id==4 .or. id==5) then
   call Transform3D(scalars(id-3),scalars(id-3),scratch(1),scratch(2),forward,inverse)
  endif
                  
 endif   !! end p_space or w_space block
  
  
end subroutine diffusion_solve
!!=====================================================
!!=====================================================



!!=====================================================
!!=====================================================
subroutine  EllipticOp(Amat,petsc_x,petsc_y,jerr) 
!! High Order Approximation to Differential Operator
!!===============================================
!! This defines the matrix multiply operation
!! for the shell matrix A which exists only
!! on the finest grid level i=0.
!! Args: compute  petsc_y <-- A petsc_x 
!!===============================================
 use class_ScalarVariable
 use class_SAOperator
 
 use InterfacePetsc        
 use operators,               only: L,alpha,beta,h2_scale_factor
 use dependent_variables
 use mpi_params
 use immersed_boundary
 use intermediate_variables,  only: Phi,scratch,rhs_b
 use etc,                     only: field_id,numzplanes_below
 use user_data,               only: nx,ny,nz,                &
                                    dt,Re,Pr,BCtype,         &
                                    xdim_implicit_bcs,       &
                                    ydim_implicit_bcs,       &
                                    zdim_implicit_bcs,       &
                                    psolve_exp_coeffs,       &
                                    usolve_exp_coeffs,       &
                                    vsolve_exp_coeffs,       &
                                    wsolve_exp_coeffs,       &
                                    s1solve_exp_coeffs,      &
                                    s2solve_exp_coeffs,      &
                                    DNS,LES,alpha_les,p_les, &
                                    do_immersed_boundary
                                    
 implicit none
#include "finclude/petsc.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"
#include "finclude/petscpc.h"
#include "finclude/petscda.h"
#include "finclude/petscksp.h"
#include "finclude/petscmg.h"
 Mat    Amat  
 Vec    petsc_x,petsc_y

  
 integer                :: jerr,idim,i,j,k,kg,id
 integer                :: endid
 real(kind=8)           :: h

 integer, SAVE          :: locnx,locny,locnz,level
 real(kind=8), SAVE     :: h2,xx,dsdx,dsdy,dsdz
 logical, SAVE          :: first_entry=.TRUE.
 logical                :: forward,inverse
 logical                :: p_solve,w_solve

 
 if( first_entry ) then
  level=0  !! fine grid only 
  locnx=scratch(1)%DataMap%SlabVals(1,myid)
  locny=scratch(1)%DataMap%SlabVals(2,myid)
  locnz=scratch(1)%DataMap%SlabVals(3,myid)
  dsdx = 1.d0/velocity(1)%Grid%SpaceDims(1)%dxds(1)
  dsdy = 1.d0/velocity(1)%Grid%SpaceDims(2)%dxds(1)
  dsdz = 1.d0/velocity(1)%Grid%SpaceDims(3)%dxds(1)
  first_entry = .FALSE.
 endif
 id=field_id
 
 xx=h2_scale_factor
 p_solve=.TRUE.
 w_solve=.FALSE.
 if( psolve_exp_coeffs  .and. field_id==6 ) xx=1.d0
 if( usolve_exp_coeffs  .and. field_id==1 ) xx=1.d0
 if( vsolve_exp_coeffs  .and. field_id==2 ) xx=1.d0
 if( wsolve_exp_coeffs  .and. field_id==3 ) xx=1.d0
 if( s1solve_exp_coeffs .and. field_id==4 ) xx=1.d0
 if( s2solve_exp_coeffs .and. field_id==5 ) xx=1.d0
 if( xx==1.d0 ) then
  w_solve=.TRUE.
  p_solve=.FALSE.
 endif

!!============================================================= 
!! Make sure scratch(1) has the right derivative types
!! and labels for both first and second derivatives
!! scratch(1) should inherit flag values from the variable
!! being solved for
!!============================================================= 
 do i=1,6
 
  if(id<=3) then
   scratch(1)%Derivs(i)=velocity(id)%Derivs(i)
   !scratch(1)%DerivTypes(i)=velocity(id)%DerivTypes(i)
   
   scratch(1)%gridpoint_values=velocity(id)%gridpoint_values
   scratch(1)%wavenumber_values=velocity(id)%wavenumber_values
   scratch(1)%forward_transform_only=velocity(id)%forward_transform_only
   scratch(1)%inverse_transform_only=velocity(id)%inverse_transform_only
   
  elseif(id==4 .or. id==5) then
  
   scratch(1)%Derivs(i)=scalars(id-3)%Derivs(i)
   !scratch(1)%DerivTypes(i)=scalars(id-3)%DerivTypes(i)
      
   scratch(1)%gridpoint_values=scalars(id-3)%gridpoint_values
   scratch(1)%wavenumber_values=scalars(id-3)%wavenumber_values
   scratch(1)%forward_transform_only=scalars(id-3)%forward_transform_only
   scratch(1)%inverse_transform_only=scalars(id-3)%inverse_transform_only

  elseif(id==6) then
  
   scratch(1)%Derivs(i)=Phi(0)%Derivs(i)
   !scratch(1)%DerivTypes(i)=Phi(0)%DerivTypes(i)
   
   scratch(1)%gridpoint_values=Phi(0)%gridpoint_values
   scratch(1)%wavenumber_values=Phi(0)%wavenumber_values
   scratch(1)%forward_transform_only=Phi(0)%forward_transform_only
   scratch(1)%inverse_transform_only=Phi(0)%inverse_transform_only
  endif
  
 enddo
    

!!============================================================= 
!! Store input vector as a ScalarVariable
!!=============================================================
 call PetscVec2Scalar(petsc_x,scratch(1))

 
 !!=============================================================
 !! Apply the self-adjoint operator L to f,
 !! storing the result in the ScalarVariable scratch(2)
 !!    L_of_f = scratch(2) <====  xx*alpha*(f_xx + f_yy + f_zz)
 !!                             + xx*beta*f
 !! (for AM 3 we have beta = -2/dt for velocity & scalars)
 !! where alpha and beta depend on which variable we're working
 !! with and xx is simply a scale factor. e.g. beta=0 for pressure.
 !! An optional SV(3) array can be also filled w/ Grad f values if
 !! it is added as the last argument to the call to ApplySAOperator.
 !! Regardless of whether the input is in physical or wavenumber space, 
 !! the ApplySAOperator routine ALWAYS returns its results in 
 !! physical space (had to be in pspace for ds/dx metric)
 !!=============================================================  
  
  call ApplySAOperator(L(level),scratch(1),scratch(2),   &
                                scratch(3),scratch(4),   &
                                alpha(id),beta(id),xx)
  
                            
 !!=============================================================
 !!=============================================================
 !! SAOperator returns result in physical space. 
 !! (had to apply the final dsdx metric in phys. space)
 !! ===> need to transform back to wavenumber space before 
 !! returning control of the iterations to petsc. Make sure
 !! the transform types for s2 match those of variable "id"
 !!=============================================================
 !!=============================================================
 if( w_solve ) then
  forward=.TRUE.
  inverse=.FALSE.
  do i=1,3
    scratch(2)%Derivs(i)=scratch(1)%Derivs(i)
  enddo
  call Transform3D(scratch(2),scratch(2),scratch(3),scratch(4),forward,inverse)
  
  
  if(LES) then
   !!=============================================================
   !! get transform of input field (petsc_x=scratch(1))
   !!=============================================================
   call Transform3D(scratch(1),scratch(5),scratch(3),scratch(4),forward,inverse)
   
   !!=============================================================   
   !! loop through wavenumber space & update scratch(2)
   !! adding in the les terms:  alpha_les(i)*k(i)**(2*p_les(i)) * f^
   !!=============================================================
   do k=1,locnz
    kg = numzplanes_below(myid) + k
     do j=1,ny
      do i=1,nx
          scratch(2)%SlabVals(i,j,k) = scratch(2)%SlabVals(i,j,k)              &
          +(alpha_les(1)*( scratch(1)%Derivs(1)%kx(i) * dsdx )**(2*p_les(1))   &
           +alpha_les(2)*( scratch(1)%Derivs(2)%kx(j) * dsdx )**(2*p_les(2))   &
           +alpha_les(3)*( scratch(1)%Derivs(3)%kx(kg)* dsdx )**(2*p_les(3)) ) &
           * scratch(5)%SlabVals(i,j,k)
      enddo
     enddo
   enddo
  endif
  
  goto 102  !! don't apply any explicit bdry operators
 endif
  
  
 !!=========================
 !!  EAST AND WEST FACES
 !!=========================
 idim=1
 if(xdim_implicit_bcs) goto 100
 
 
  endid=1
  i=1
  if(BCtype(field_id,idim,endid)=='dirichlet') then
   scratch(2)%SlabVals(i,:,:) = scratch(1)%DirichletWest(:,:)
  elseif(BCtype(field_id,idim,endid)=='neumann') then
   scratch(2)%SlabVals(i,:,:) = scratch(1)%NeumannWest(:,:)
  endif
  
  endid=2
  i=locnx
  if(BCtype(field_id,idim,endid)=='dirichlet') then
   scratch(2)%SlabVals(i,:,:) = scratch(1)%DirichletEast(:,:)
  elseif(BCtype(field_id,idim,endid)=='neumann') then
   scratch(2)%SlabVals(i,:,:) = scratch(1)%NeumannEast(:,:)
  endif
   
 
 
   
 !!=========================
 !!  FRONT AND BACK FACES
 100 idim=2
 if(ydim_implicit_bcs) goto 101
 !!=========================
  
  endid=1
  j=1
  if(BCtype(field_id,idim,endid)=='dirichlet') then
   scratch(2)%SlabVals(:,j,:) = scratch(1)%DirichletFront(:,:)
  elseif(BCtype(field_id,idim,endid)=='neumann') then
   scratch(2)%SlabVals(:,j,:) = scratch(1)%NeumannFront(:,:)
  endif
  
  endid=2
  j=locny
  if(BCtype(field_id,idim,endid)=='dirichlet') then
   scratch(2)%SlabVals(:,j,:) = scratch(1)%DirichletBack(:,:)
  elseif(BCtype(field_id,idim,endid)=='neumann') then
   scratch(2)%SlabVals(:,j,:) = scratch(1)%NeumannBack(:,:)
  endif
     
  
 !!=========================
 !!  BOTTOM 
 !!=========================
101 idim=3
  endid=1
  k=1
  if(zdim_implicit_bcs) goto 102
   
   if(BCtype(field_id,idim,endid)=='dirichlet') then
    scratch(2)%SlabVals(:,:,k) = scratch(1)%DirichletBottom(:,:)
   elseif(BCtype(field_id,idim,endid)=='neumann') then
    scratch(2)%SlabVals(:,:,k) = scratch(1)%NeumannBottom(:,:)
   endif
       
 !!=========================
 !!  TOP
 idim=3
 endid=2
 k=locnz
 !!=========================
 
  if(BCtype(field_id,idim,endid)=='dirichlet') then
   scratch(2)%SlabVals(:,:,k) = scratch(1)%DirichletTop(:,:)
  elseif(BCtype(field_id,idim,endid)=='neumann') then
   scratch(2)%SlabVals(:,:,k) = scratch(1)%NeumannTop(:,:)
  endif
     
!! Store final result in petsc vector
102 continue
call ScalarVariable2PetscVec(scratch(2),petsc_y)



end subroutine EllipticOp
!!=====================================================
!!=====================================================




!!=====================================================
!!=====================================================
 subroutine Create_L_Solver(ksp_L)
!!=====================================================
!!=====================================================
 use class_ScalarVariable
 use class_SAOperator
 
 use operators, only: L,A,A_PC
 use user_data, only: xdim_periodic,      &
                      ydim_periodic,      &
                      zdim_periodic,      &
                      usolve_exp_coeffs,  &
                      vsolve_exp_coeffs,  &
                      wsolve_exp_coeffs,  &
                      s1solve_exp_coeffs, &
                      s2solve_exp_coeffs, &
                      psolve_exp_coeffs
                      
 use etc,       only: field_id
 use mpi_params
 implicit none
#include "finclude/petsc.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"
#include "finclude/petscpc.h"
#include "finclude/petscda.h"
#include "finclude/petscksp.h"
#include "finclude/petscmg.h"

 KSP                   ksp_L
 PC                    pc_L
 MatNullSpace          nullsp    
 Vec                   RHS,X
 PetscTruth            flag_pcasm
 PetscTruth            flag_hypre
 PetscTruth            flag_diag
 PetscTruth            flag_fgmres
 PetscTruth            flag_bcgs
 PetscTruth            flag_guess
 PetscTruth            flag_PC_pressure
 PetscTruth            flag_PC_othervars
  
 integer,allocatable :: idx(:)
 integer             :: nx_local,ny_local,nz_local
 integer             :: nx_global,ny_global,nz_global
 integer             :: localrows,localcols
 integer             :: globalrows,globalcols
 integer             :: i,id,nz,nnz,jerr
 integer             :: izero=0,ione=1,level=0
 real(kind=8)        :: spai_tol=0.9d0
 logical             :: PC_A
 logical             :: PC_A_PHYSICAL
 logical             :: PC_A_WAVENUMBER
 logical,save        :: first_entry=.TRUE.
 external  EllipticOp
 
 id=field_id
 if(myid==0 .and. id==1 ) write(0,*) '... Creating KSP Solver: L[f], u '
 if(myid==0 .and. id==2 ) write(0,*) '... Creating KSP Solver: L[f], v '
 if(myid==0 .and. id==3 ) write(0,*) '... Creating KSP Solver: L[f], w '
 if(myid==0 .and. id==4 ) write(0,*) '... Creating KSP Solver: L[f], s1 '
 if(myid==0 .and. id==5 ) write(0,*) '... Creating KSP Solver: L[f], s2 '
 if(myid==0 .and. id==6 ) write(0,*) '... Creating KSP Solver: L[f], phi '
 
 !!==========================================================
 !! Find out whether to create a preconditioner matrix A_PC
 !!==========================================================
 PC_A=.FALSE.
 if(field_id==6) then
  call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-PC_pressure', &
                           flag_PC_pressure,jerr)
  if(flag_PC_pressure == 1) PC_A=.TRUE.
 else
  call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-PC_othervars', &
                           flag_PC_othervars,jerr)
  if(flag_PC_othervars == 1) PC_A=.TRUE.
 endif
 
 !!==========================================================
 !! If preconditioning, do in physical or wavenumber space?
 !!==========================================================
 PC_A_WAVENUMBER=.FALSE.
 PC_A_PHYSICAL=.FALSE.
 if( PC_A ) then
  if(id==1 .and. usolve_exp_coeffs)  PC_A_WAVENUMBER=.TRUE.
  if(id==2 .and. vsolve_exp_coeffs)  PC_A_WAVENUMBER=.TRUE.
  if(id==3 .and. wsolve_exp_coeffs)  PC_A_WAVENUMBER=.TRUE.
  if(id==4 .and. s1solve_exp_coeffs) PC_A_WAVENUMBER=.TRUE.
  if(id==5 .and. s2solve_exp_coeffs) PC_A_WAVENUMBER=.TRUE.
  if(id==6 .and. psolve_exp_coeffs)  PC_A_WAVENUMBER=.TRUE.
  if(.NOT. PC_A_WAVENUMBER) PC_A_PHYSICAL=.TRUE.
 endif
      
 
!!======================================================================
!! Create Shell Matrix A & set Operator Routines
!!======================================================================
!! determine layout from SA operator at level 0
 call SAGetSizes(L(level),nx_local,ny_local,nz_local,  &
                       nx_global,ny_global,nz_global)
                       
 localrows=nx_local*ny_local*nz_local   
 globalrows=nx_global*ny_global*nz_global 
 localcols=localrows
 globalcols=globalrows
 
 
 !!=============================================================================
 !!=============================================================================
 !! shell matrix for higher order self adjoint operator on fine grid
 !!=============================================================================
 !!=============================================================================
 if(myid==0) write(0,*) '... Creating shell matrix A.....',field_id
 if(myid==0) write(0,*) '... PC_A             .....',PC_A
 if(myid==0) write(0,*) '... PC_A_PHYSICAL    .....',PC_A_PHYSICAL
 if(myid==0) write(0,*) '... PC_A_WAVENUMBER  .....',PC_A_WAVENUMBER
 call MatCreateShell(comm,localrows,localcols,globalrows, &
                     globalcols,PETSC_NULL_INTEGER,A,jerr) 
 if( jerr .ne. 0 ) stop 'error in call to MatCreateShell'
 call MatShellSetOperation(A,MATOP_MULT,EllipticOp,jerr)
 if( jerr .ne. 0 ) stop 'error in call to MatShellSetOperation' 
  
 !!=============================================================================
 !!=============================================================================
 !! Explicit matrix for preconditioning. 
 !! Results stored in petsc Matrix A_PC 
 !! in module "operators"
 !!=============================================================================
 !!=============================================================================
 if(PC_A) then
 
  if(first_entry) then
   if(PC_A_PHYSICAL) nnz=5
   if(PC_A_WAVENUMBER) nnz=1
   call MatCreateMPIAIJ(comm,localrows,localcols,globalrows,globalcols,   &
                        nnz,PETSC_NULL_INTEGER,ione,PETSC_NULL_INTEGER,   &
                        A_PC,jerr)
   first_entry=.FALSE.
  endif
  call MatZeroEntries(A_PC,jerr)
  if( jerr .ne. 0 ) stop 'error in call to MatZeroEntries'
  
  if( PC_A_PHYSICAL ) then
   if(myid==0) write(0,*) '...  calling FillA_FD (finite difference)'
   call FillA_FD(myid,localrows,globalrows)
   
  elseif(PC_A_WAVENUMBER) then
   if(myid==0) write(0,*) '...  calling FillA_WS (wavenumber space)'
   call FillA_WS(myid,localrows,globalrows)
  endif
   
 endif
 
 !!=============================================================================
 !!=============================================================================
 !! Create KSP solver object 
 !!=============================================================================
 !!=============================================================================
 call KSPCreate(comm,ksp_L,jerr)
                          if( jerr .ne. 0 ) stop 'error in call to KSPCreate'
 if(PC_A) then   
  call KSPSetOperators(ksp_L,A,A_PC,SAME_PRECONDITIONER,jerr)
 else
  call KSPSetOperators(ksp_L,A,A,SAME_PRECONDITIONER,jerr)
 endif
 
 flag_fgmres=1
 flag_bcgs=0
 call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-petsc_fgmres',flag_fgmres,jerr)
 call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-petsc_bcgs',flag_bcgs,jerr)
 
 
 if(myid==0)  &
 write(0,*) '... Setting KSP type.....(fgmres,bcgs)...',flag_fgmres,flag_bcgs
 if( flag_fgmres /= 0 ) then
     call KSPSetType(ksp_L,KSPFGMRES,jerr) 
     !!call KSPGMRESSetPreAllocateVectors(ksp_L,jerr)
 endif
 if( flag_bcgs.ne. 0 )   call KSPSetType(ksp_L,KSPBCGS,jerr) 
       if( jerr /= 0 ) stop 'error in call to KSPSetType'

 call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-nonzero_guess',flag_guess,jerr)
 if( flag_guess /= 0 ) call KSPSetInitialGuessNonzero(ksp_L,PETSC_TRUE,jerr)
       if( jerr /= 0 ) stop 'error in call to KSPSetInitialGuessNonzero'
       
 call KSPSetFromOptions(ksp_L,jerr)
      if( jerr /= 0 ) stop 'error in call to KSPSetFromOptions'
                          
 call KSPGetPC(ksp_L,pc_L,jerr)
      if( jerr /= 0 ) stop 'error in call to KSPGetPC' 

 !!=============================================================================
 !!=============================================================================
 !! Create & Attach a PC object 
 !!=============================================================================
 !!=============================================================================
 
 flag_pcasm=1
 flag_hypre=0
 flag_diag=0
 call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-petsc_pcasm',flag_pcasm,jerr)
 call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-petsc_hypre',flag_hypre,jerr)
 call PetscOptionsHasName(PETSC_NULL_CHARACTER,'-petsc_diag',flag_diag,jerr)
 
 if(PC_A) then  
  if(myid==0) &
   write(0,*) '... Setting KSP PC type.....(pcasm,hypre,diag)...',  &
                   flag_pcasm,flag_hypre,flag_diag
  if (flag_diag /= 0 )  call PCSetType(pc_L,PCJACOBI,jerr)
  if (flag_pcasm /= 0)  call PCSetType(pc_L,PCASM,jerr) !
  if (flag_hypre /= 0)  call PCSetType(pc_L,PCHYPRE,jerr)
  if( jerr .ne. 0 ) stop 'error in call to PCSetType'
 elseif(.NOT. PC_A ) then
  call PCSetType(pc_L,PCNONE,jerr)
   if( jerr .ne. 0 ) stop 'error in call to PCSetType'
 endif
 
 call PCSetFromOptions(pc_L,jerr)
end subroutine Create_L_Solver
!!=========================================================
!!=========================================================



!!=========================================================
!!=========================================================
subroutine DoSolve(f)
!! Solve system Ax=b <===> L[f]=b
!! The rhs ScalarVariable b is available
!! via "use intermediate_variables"
!! ScalarVariable f is the fine grid solution 
 use InterfacePetsc 
 use intermediate_variables,  only: rhs_b
 use etc,                     only: field_id,istep,istart                                    
 use operators
 use mpi_params
 
 implicit none
 PetscTruth               flg
 type( ScalarVariable )   :: f
 integer                  :: id,i,k,jerr,maxits(6)
 integer,save             :: nx_local,ny_local,nz_local
 integer,save             :: nx_global,ny_global,nz_global
 integer,save             :: localrows,localcols,locnz
 integer,save             :: globalrows,globalcols
 integer                  :: idim,num_its(6)
 real(kind=8)             :: rtol(6),atol(6),dtol(6),res_norm(6)
 real(kind=8)             :: xmem(6)
 logical,save             :: first_entry=.TRUE.
 
  id=field_id
  if(myid==0)   &
  write(0,*) ' .......... begining iterative solve for field ......',id
 
 if(first_entry) then
  !!=====================================================================
  !! determine layout from my SA operator L
  !! and create petsc vectors to use during solves
  !!=====================================================================
  call SAGetSizes(L(0),nx_local,ny_local,nz_local,  &
                  nx_global,ny_global,nz_global )
                       
  localrows = nx_local*ny_local*nz_local   
  globalrows= nx_global*ny_global*nz_global
  localcols=localrows       
  globalcols=globalrows
  locnz=f%DataMap%SlabVals(3,myid)
  first_entry=.FALSE.
 endif
  
  call VecCreateMPI(comm,localrows,globalrows,petsc_soln,jerr)
  if( jerr .ne. 0 ) stop 'error in call to VecCreateMPI: petsc_soln'
  call VecDuplicate(petsc_soln,petsc_b,jerr)
  if( jerr .ne. 0 ) stop 'error in call to VecDuplicate'
 
 !!==========================================================================
 !! put scalarVariable rhs --> petsc vector
 !!==========================================================================
  call ScalarVariable2PetscVec(rhs_b,petsc_b)
  
 
 if(id==6 ) then
 !!==========                         ===========
 !!              pressure  solve 
 !!==========                         ===========
 
   
  !! initial guess
  call SetGuess(petsc_soln)   
    
  !! (re)set tolerances etc
  rtol(id)=1.e-5     !! 1.e-4
  atol(id)=1.e-6     !! 1.e-7
  dtol(id)=1.e15
  maxits(id) = 1e4
  !! over-ride with value from PetscOptions
  call PetscOptionsGetInt(PETSC_NULL_CHARACTER,    &
            '-pressure_its',maxits(id),flg,jerr)
  call KSPSetTolerances(     &
         ksp(id),rtol(id),   &
         atol(id),dtol(id),  &
         maxits(id),jerr )
  if( jerr .ne. 0 ) stop 'error in call to KSPSetTolerances'
  
      
   if(myid==0) write(0,*) '...   calling KSPSolve ksp'
    call KSPSolve(ksp(id),petsc_b,petsc_soln,jerr)
    if( jerr .ne. 0 ) stop 'error in call to KSPSolve'
   if(myid==0) write(0,*) '...   done with KSPSolve ksp'
   
   call KSPGetIterationNumber(ksp(id),num_its(id),jerr)
   call KSPGetResidualNorm(ksp(id),res_norm(id),jerr)
   !call PetscMemoryGetCurrentUsage(xmem(id),jerr)
      
   if(myid==0) then
    if(istep==istart)open(1,file='output/pressure_iterations',position='rewind')
    if(istep>istart) open(1,file='output/pressure_iterations',position='append')
     write(1,*) istep,num_its(id),res_norm(id)   !,xmem(i)
    close(1)
   endif
        
 
 !!==========                  =====================
 else !!  all other variables (u,v,w,s1,s2)
 !!==========                  =====================
 

  !! set tolerances for the various Helmholtz eqns
  if(id==1) then      !!!   ============ tolerances for u   ======================
   rtol(id)=1.e-10
   atol(id)=1.e-12
   dtol(id)=1.e15
   maxits(id)=10
  elseif(id==2) then  !!!   ============ tolerances for v   ======================
   rtol(id)=1.e-10
   atol(id)=1.e-12
   dtol(id)=1.e15
   maxits(id)=10
  elseif(id==3) then  !!!   ============ tolerances for w   ======================
   rtol(id)=1.e-10
   atol(id)=1.e-12
   dtol(id)=1.e15
   maxits(id)=10
  elseif(id==4) then  !!!   ============ tolerances for s1  ======================
   rtol(id)=1.e-10
   atol(id)=1.e-12
   dtol(id)=1.e15
   maxits(id)=10
  elseif(id==5) then  !!!   ============ tolerances for s2  ======================
   rtol(id)=1.e-10
   atol(id)=1.e-12
   dtol(id)=1.e15
   maxits(id)=10
  endif
  
  !! over-ride with value from PetscOptions
  call PetscOptionsGetInt(PETSC_NULL_CHARACTER,    &
            '-othervars_its',maxits(id),flg,jerr)
  call KSPSetTolerances(     &
         ksp(id),rtol(id),   &
         atol(id),dtol(id),  &
         maxits(id),jerr )
  if( jerr .ne. 0 )          &
  stop 'error in call to KSPSetTolerances'
  
  !! set nonzero initial guess
  call SetGuess(petsc_soln)
  
  !! do the solve
  call KSPSolve(ksp(id),petsc_b,petsc_soln,jerr)
  if( jerr .ne. 0 ) stop 'error in call to KSPSolve'
  
  !! get the iteration count etc
  call KSPGetIterationNumber(ksp(id),num_its(id),jerr)
  call KSPGetResidualNorm(ksp(id),res_norm(id),jerr)
  !call PetscMemoryGetCurrentUsage(xmem(id),jerr)
   
  !! write iteration count to text file
  if(myid==0) then
   if(istep==istart .and. id== 1) open(1,file='output/u_iterations',position='rewind')
   if(istep>istart .and. id== 1)  open(1,file='output/u_iterations',position='append')
   if(istep==istart .and. id== 2) open(1,file='output/v_iterations',position='rewind')
   if(istep>istart .and. id== 2)  open(1,file='output/v_iterations',position='append')
   if(istep==istart .and. id== 3) open(1,file='output/w_iterations',position='rewind')
   if(istep>istart .and. id== 3)  open(1,file='output/w_iterations',position='append')
   if(istep==istart .and. id== 4) open(1,file='output/s1_iterations',position='rewind')
   if(istep>istart .and. id== 4)  open(1,file='output/s1_iterations',position='append')
   if(istep==istart .and. id== 5) open(1,file='output/s2_iterations',position='rewind')
   if(istep>istart .and. id== 5)  open(1,file='output/s2_iterations',position='append')
    write(1,*) istep,num_its(id),res_norm(id)   !,xmem(i)
   close(1)
  endif
  
 endif
 
 
!!==================================================================
!! Store petsc_soln  =======>  ScalarVariable f 
!!==================================================================
  call PetscVec2Scalar(petsc_soln,f)
 
!!==================================================================
!! Release memory used by petsc 
!!==================================================================
 call VecDestroy(petsc_b,jerr)
 call VecDestroy(petsc_soln,jerr) 
   
 return
end subroutine DoSolve 
!!=========================================================
!!=========================================================




!===================================================    
!===================================================
integer function node_index(i,j,k)
 use user_data,  only: nx,ny
 implicit none
 integer::i,j,k
 !! 3d natural indexing of nodes/eqns
 !! i,j,k zero based indices
 node_index = k*(nx*ny) + j*(nx) + i 
 return
end function node_index
!===================================================    
!===================================================


!!=====================================================
!!=====================================================
subroutine FillA_FD(myid,localrows,globalrows)
 use operators,              only: L,A_PC,alpha,beta,h2_scale_factor
 use intermediate_variables, only: scratch
 use user_data
 use etc,         only:docfile,           &
                       field_id,          &
                       numzplanes_below,  &
                       numzplanes
 implicit none

#include "finclude/petsc.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"

 PetscInt                    mx,my,mz
 PetscInt                    xs,ys,zs
 PetscInt                    xm,ym,zm
 
 integer                     :: i,j,k,myid,localrows,globalrows
 integer                     :: idim,jerr,one=1,MM,level,locnz
 real(kind=8)                :: h1,h2,h3
 real(kind=8)                :: dxds(3),ds(3),dx(3),dxdx
 real(kind=8)                :: scale_factor(2)
 real(kind=8)                :: diag_val,scaled_diag,x_s(3),x_ss(3)
 integer,allocatable         :: idx(:)
 real(kind=8),allocatable    :: vals(:)
 character(len=80)           :: btype(3,2)
  
 !!===============================================
 !!  size parameters
 !!===============================================
 
 !! zero based start indices for myid in 3d representation
 xs=0
 ys=0
 zs=numzplanes_below(myid)
 
 !! zero based end indices for myid in 3d representation
 xm=xs + nx-1
 ym=ys + ny-1
 locnz=numzplanes(myid)
 zm=zs + locnz - 1
 
 !! global number of nodes in each direction
 mx=nx
 my=ny
 mz=nz
 
 level=0
 if(myid==0) then
  open(1,file=docfile,position='append')
  write(1,*) ' '
  write(1,*) '... Assembling A_FD at level ',level
 endif
  
 do i=1,3
  ds(i)=L(level)%Grid%SpaceDims(i)%ds
 enddo
       
 !!===============================================
 !!  scale factors
 !!===============================================
  
 dxdx = h2_scale_factor
 scale_factor(1) = alpha(field_id)*dxdx  !! for Grad^2 terms
 scale_factor(2) = dxdx                  !! for neumann bdry terms (inverse)
 scaled_diag = beta(field_id)*dxdx       !! for xtra helmholtz terms
 
 
 !!===============================================
 !!  BC labels
 !!===============================================
 do idim=1,3
  do j=1,2
   btype(idim,j) = BCtype(field_id,idim,j)
  enddo
 enddo
 
 !!===============================================
 !!  Fill the main portion of A_FD
 !!===============================================
 do k=zs,zm    !! loop through my assigned data using global indexing
 
    if( k > 0 .and. k < mz-1 ) then
     ds(3)= (L(0)%Grid%SpaceDims(3)%s(k+2) - L(0)%Grid%SpaceDims(3)%s(k))/(2.)
    elseif( k==0 ) then
     ds(3)= (L(0)%Grid%SpaceDims(3)%s(2) - L(0)%Grid%SpaceDims(3)%s(1))
    elseif( k==mz-1 ) then
     ds(3)= (L(0)%Grid%SpaceDims(3)%s(k+1) - L(0)%Grid%SpaceDims(3)%s(k))/(2.)
    endif
    
    h3=ds(3)*L(0)%Grid%SpaceDims(3)%dxds(k+1)  !! ds*dx/ds
    x_s(3)=L(0)%Grid%SpaceDims(3)%dxds(k+1)
    x_ss(3)=L(0)%Grid%SpaceDims(3)%d2xds2(k+1)
 
 
   do j=ys,ym
     
     if( j > 0 .and. j < my-1 ) then
      ds(2)= (L(0)%Grid%SpaceDims(2)%s(j+2) - L(0)%Grid%SpaceDims(2)%s(j))/(2.)
     elseif( j==0 ) then
      ds(2)=1.d0
     elseif( j==my-1 ) then
      ds(2)= (L(0)%Grid%SpaceDims(2)%s(j+1) - L(0)%Grid%SpaceDims(2)%s(j))/(2.)
     endif
     
      h2=ds(2)*L(0)%Grid%SpaceDims(2)%dxds(j+1)
      x_s(2)=L(0)%Grid%SpaceDims(2)%dxds(j+1)
      x_ss(2)=L(0)%Grid%SpaceDims(2)%d2xds2(j+1)
      
      
     do i=xs,xm 
       
        if( i > 0 .and. i < mx-1 ) then
         ds(1)= (L(0)%Grid%SpaceDims(1)%s(i+2) - L(0)%Grid%SpaceDims(1)%s(i))/(2.)
        elseif( i==0 ) then
         ds(1)= (L(0)%Grid%SpaceDims(1)%s(2) - L(0)%Grid%SpaceDims(1)%s(1))
        elseif( i==mx-1 ) then
         ds(1)= (L(0)%Grid%SpaceDims(1)%s(i+1) - L(0)%Grid%SpaceDims(1)%s(i))/(2.)
        endif
     
        h1=ds(1)*L(0)%Grid%SpaceDims(1)%dxds(i+1)
        x_s(1)=L(0)%Grid%SpaceDims(1)%dxds(i+1)
        x_ss(1)=L(0)%Grid%SpaceDims(1)%d2xds2(i+1)  
          
        if(ny>1) then  
           call SetMatrixElements  &
             ( i,j,k,h1,h2,h3,      &
               x_s,x_ss,btype,      &
               scale_factor,        &
               mx,my,mz,            &
               scaled_diag )
         elseif(ny==1) then               
           call SetMatrixElements2D  &
             ( i,j,k,h1,h3,          &
               x_s,x_ss,btype,       &
               scale_factor,         &
               mx,mz,                &
               scaled_diag )
         endif
               
              
                             
       enddo
     enddo
 enddo
 
 !!============================================================
 !!  Clear cache & leave matrix ready to use
 !!============================================================
 call MatAssemblyBegin(A_PC,MAT_FINAL_ASSEMBLY,jerr)
      if(jerr .ne.0 ) then 
       write(0,*) 'Error calling MatAssemblyBegin '
       stop
      endif
 call MatAssemblyEnd(A_PC,MAT_FINAL_ASSEMBLY,jerr)
      if(jerr .ne.0 ) then 
       write(0,*) 'Error calling MatAssemblyEnd '
       stop
      endif
 
  if(myid==0) write(0,*) 'A_FD elements set and matrix object assembled assembled.'
 return
end
!!=====================================================
!!=====================================================



!!=====================================================
!!=====================================================
subroutine SetMatrixElements(i,j,k,dx,dy,dz,x_s,x_ss,btype, &
                             scale_factor,mx,my,mz,scaled_diag)
 use operators,    only: A_PC
 use user_data,    only: xdim_periodic,  &
                         ydim_periodic,  &
                         zdim_periodic,  &
                         do_immersed_boundary
 use etc,          only: field_id
 use immersed_boundary
 use mpi_params
 
 implicit none
 integer              :: i,j,k,nnz
 integer              :: idim,mx,my,mz
 integer              :: jerr,one=1
 integer              :: row,col(12),ifor,ibac,jfor,jbac
 real(kind=8)         :: dx,dy,dz,x_s(3),x_ss(3),hmin,v(12)
 real(kind=8)         :: scale_factor(2),Gscale,Dscale
 real(kind=8)         :: scaled_diag,zero=0.d0
 character(len=80)    :: btype(3,2)
    
#include "finclude/petsc.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"
    
 integer, external   :: node_index
  
 Gscale = scale_factor(1)  !! scale for Grad^2 terms
 Dscale = scale_factor(2)  !! scale for Dirichlet terms
 
 !!  Should never be defining eqns for nodes outside these ranges
  if( i<0 ) STOP 'i<0 in SetMatrixElements'
  if( i>mx-1 ) STOP 'i>mx-1 in SetMatrixElements'
  if( j<0 ) STOP 'j<0 in SetMatrixElements'
  if( j>my-1 ) STOP 'j>my-1 in SetMatrixElements'
  if( k<0 ) STOP 'k<0 in SetMatrixElements'
  if( k>mz-1 ) STOP 'k>mz-1 in SetMatrixElements'
 
  v(:)=0. ; row=0.; col=0.
 
 
 !! 3d natural indexing of nodes/eqns
  row = node_index(i,j,k) 
    
 !!==========================================================
 !!       top and bottom faces, including edges
 !!========================================================== 
 
 if( k==0 .and. trim(btype(3,1))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i,j,k+1)
  col(3) = node_index(i,j,k+2)
  v(1) = -(3.d0/(2.*dz)) / x_s(3) 
  v(2) =  (4.d0/(2.*dz)) / x_s(3)
  v(3) = -(1.d0/(2.*dz)) / x_s(3)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 
 elseif( k==mz-1 .and. trim(btype(3,2))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i,j,k-1)
  col(3) = node_index(i,j,k-2)
  v(1) =  (3.d0/(2.*dz)) / x_s(3) 
  v(2) = -(4.d0/(2.*dz)) / x_s(3)
  v(3) =  (1.d0/(2.*dz)) / x_s(3)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( k==0 .and. trim(btype(3,1))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( k==mz-1 .and. trim(btype(3,2))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 endif
 
 !!==========================================================
 !!       interior points
 !!==========================================================
 
 if( i>0 .and. i<mx-1  .and. j>0 .and. j<my-1 .and. k>0 .and. k<mz-1 ) then
  col(1) = node_index(i-1,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i,j-1,k)
  col(4) = node_index(i,j+1,k)
  col(5) = node_index(i,j,k-1)
  col(6) = node_index(i,j,k+1)
  col(7) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=2
  v(3) =  1./(dy**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  v(4) =  1./(dy**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  idim=3
  v(5) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(6) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(7) = -2./(dx**2)  -2./(dy**2) -2./(dz**2)
  
  
  nnz=7
  v=Gscale*v
  v(7)=v(7) + scaled_diag
  
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
   goto 99
    
 endif
 
 
 
 !!==========================================================
 !!      nonperiodic  left & right faces, including edges
 !!========================================================== 
 
 if( i==0 .and. trim(btype(1,1))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i+2,j,k)
  v(1) = -3.d0/(2.*dx) 
  v(2) =  4.d0/(2.*dx)
  v(3) = -1.d0/(2.*dx)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 
 elseif( i==mx-1 .and. trim(btype(1,2))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i-1,j,k)
  col(3) = node_index(i-2,j,k)
  v(1) =  3.d0/(2.*dx) 
  v(2) = -4.d0/(2.*dx)
  v(3) =  1.d0/(2.*dx)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( i==0 .and. trim(btype(1,1))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( i==mx-1 .and. trim(btype(1,2))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 endif
 
 
 
 !!==========================================================
 !!      nonperiodic  front & back faces, including edges
 !!========================================================== 
 
 
 if( j==0 .and. trim(btype(2,1))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i,j+1,k)
  col(3) = node_index(i,j+2,k)
  v(1) = -3.d0/(2.*dy) 
  v(2) =  4.d0/(2.*dy)
  v(3) = -1.d0/(2.*dy)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 
 elseif( j==my-1 .and. trim(btype(2,2))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i,j-1,k)
  col(3) = node_index(i,j-2,k)
  v(1) =  3.d0/(2.*dy) 
  v(2) = -4.d0/(2.*dy)
  v(3) =  1.d0/(2.*dy)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( j==0 .and. trim(btype(2,1))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( j==my-1 .and. trim(btype(2,2))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 endif
 
 
 !!==========================================================
 !!      x dir periodic, y nonperiodic
 !!========================================================== 
 
 if( i==0 .and. xdim_periodic .and.      &
     trim(btype(2,1))=='neumann' .or.    &
     trim(btype(2,1))=='dirichlet') then
     
  col(1) = node_index(mx-1,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i,j-1,k)
  col(4) = node_index(i,j+1,k)
  col(5) = node_index(i,j,k-1)
  col(6) = node_index(i,j,k+1)
  col(7) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=2
  v(3) =  1./(dy**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  v(4) =  1./(dy**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  idim=3
  v(5) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(6) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(7) = -2./(dx**2)  -2./(dy**2) -2./(dz**2)
  
  nnz=7
  v=Gscale*v
  v(7)=v(7) + scaled_diag
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( i==mx-1 .and. xdim_periodic .and.      &
     trim(btype(2,2))=='neumann' .or.    &
     trim(btype(2,2))=='dirichlet') then
     
  col(1) = node_index(i-1,j,k)
  col(2) = node_index(0,j,k)
  col(3) = node_index(i,j-1,k)
  col(4) = node_index(i,j+1,k)
  col(5) = node_index(i,j,k-1)
  col(6) = node_index(i,j,k+1)
  col(7) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=2
  v(3) =  1./(dy**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  v(4) =  1./(dy**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  idim=3
  v(5) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(6) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(7) = -2./(dx**2)  -2./(dy**2) -2./(dz**2)
  
  nnz=7
  v=Gscale*v
  v(7)=v(7) + scaled_diag
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 endif
 
 
 !!==========================================================
 !!      y dir periodic, x nonperiodic
 !!========================================================== 
 
 if( j==0 .and. ydim_periodic .and.      &
     trim(btype(1,1))=='neumann' .or.    &
     trim(btype(1,1))=='dirichlet') then
     
  col(1) = node_index(i-1,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i,my-1,k)
  col(4) = node_index(i,j+1,k)
  col(5) = node_index(i,j,k-1)
  col(6) = node_index(i,j,k+1)
  col(7) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=2
  v(3) =  1./(dy**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  v(4) =  1./(dy**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  idim=3
  v(5) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(6) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(7) = -2./(dx**2)  -2./(dy**2) -2./(dz**2)
  
  nnz=7
  v=Gscale*v
  v(7)=v(7) + scaled_diag
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( j==my-1 .and. ydim_periodic .and.      &
     trim(btype(1,2))=='neumann' .or.    &
     trim(btype(1,2))=='dirichlet') then
     
  col(1) = node_index(i-1,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i,j-1,k)
  col(4) = node_index(i,0,k)
  col(5) = node_index(i,j,k-1)
  col(6) = node_index(i,j,k+1)
  col(7) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=2
  v(3) =  1./(dy**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  v(4) =  1./(dy**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  idim=3
  v(5) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(6) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(7) = -2./(dx**2)  -2./(dy**2) -2./(dz**2)
  
  nnz=7
  v=Gscale*v
  v(7)=v(7) + scaled_diag
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 endif
 
 
 !!==========================================================
 !!      periodic in both x and y directions
 !!==========================================================
 
 if(  xdim_periodic .and.  ydim_periodic ) then
  !! ok
 else
  write(0,*) 'logic descending to xy periodic, but problem detected for field: ',field_id
  write(0,*) 'x bdry types: ',trim(btype(1,1)),'  ',trim(btype(1,2))
  write(0,*) 'y bdry types: ',trim(btype(2,1)),'  ',trim(btype(2,2))
  stop 'logic problem filling A_FD'
 endif
 
 if( i .ne. 0 .and. i .ne. mx-1 .and. j .ne. 0 .and. j .ne. my-1 ) then
  write(0,*) 'logic problem filling A_FD',i,j,k
  stop
 endif
 ifor=i+1
 ibac=i-1
 jfor=j+1
 jbac=j-1
 
 if(i==0) ibac=mx-1
 if(i==mx-1) ifor=0
 if(j==0) jbac=my-1
 if(j==my-1) jfor=0
 
  col(1) = node_index(ibac,j,k)
  col(2) = node_index(ifor,j,k)
  col(3) = node_index(i,jbac,k)
  col(4) = node_index(i,jfor,k)
  col(5) = node_index(i,j,k-1)
  col(6) = node_index(i,j,k+1)
  col(7) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=2
  v(3) =  1./(dy**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  v(4) =  1./(dy**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dy)
  idim=3
  v(5) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(6) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
       
  v(7) = -2./(dx**2)  -2./(dy**2) -2./(dz**2)
  
  nnz=7
  v=Gscale*v
  v(7)=v(7) + scaled_diag
  
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
      

99 return

end subroutine SetMatrixElements







!!=====================================================
!!=====================================================
subroutine SetMatrixElements2D(i,j,k,dx,dz,x_s,x_ss,btype, &
                             scale_factor,mx,mz,scaled_diag)
 use operators,    only: A_PC
 use user_data,    only: xdim_periodic,  &
                         zdim_periodic,  &
                         do_immersed_boundary
 use etc,          only: field_id
 use immersed_boundary
 use mpi_params
 
 implicit none
 integer              :: i,j,k,nnz
 integer              :: idim,mx,mz
 integer              :: jerr,one=1
 integer              :: row,col(12),ifor,ibac,jfor,jbac
 real(kind=8)         :: dx,dz,x_s(3),x_ss(3),hmin,v(12)
 real(kind=8)         :: scale_factor(2),Gscale,Dscale
 real(kind=8)         :: scaled_diag,zero=0.d0
 character(len=80)    :: btype(3,2)
    
#include "finclude/petsc.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"
    
 integer, external   :: node_index
  
 
 Gscale = scale_factor(1)  !! scale for Grad^2 terms
 Dscale = scale_factor(2)  !! scale for Dirichlet terms
 
 !!  Should never be defining eqns for nodes outside these ranges
  if( i<0 ) STOP 'i<0 in SetMatrixElements'
  if( i>mx-1 ) STOP 'i>mx-1 in SetMatrixElements'
  if( k<0 ) STOP 'k<0 in SetMatrixElements'
  if( k>mz-1 ) STOP 'k>mz-1 in SetMatrixElements'
 
  v(:)=0. ; row=0.; col=0.
 
 
 !! 3d natural indexing of nodes/eqns
  row = node_index(i,j,k) 
    
 !!==========================================================
 !!       top and bottom faces, including edges
 !!========================================================== 
 
 
 if( k==0 .and. trim(btype(3,1))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i,j,k+1)
  col(3) = node_index(i,j,k+2)
  v(1) = -(3.d0/(2.*dz)) / x_s(3) 
  v(2) =  (4.d0/(2.*dz)) / x_s(3)
  v(3) = -(1.d0/(2.*dz)) / x_s(3)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 
 elseif( k==mz-1 .and. trim(btype(3,2))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i,j,k-1)
  col(3) = node_index(i,j,k-2)
  v(1) =  (3.d0/(2.*dz)) / x_s(3) 
  v(2) = -(4.d0/(2.*dz)) / x_s(3)
  v(3) =  (1.d0/(2.*dz)) / x_s(3)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( k==0 .and. trim(btype(3,1))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( k==mz-1 .and. trim(btype(3,2))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 endif
 
 !!==========================================================
 !!       interior points
 !!==========================================================
 
 if( i>0 .and. i<mx-1  .and. k>0 .and. k<mz-1 ) then
  col(1) = node_index(i-1,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i,j,k-1)
  col(4) = node_index(i,j,k+1)
  col(5) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=3
  v(3) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(4) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(5) = -2./(dx**2) -2./(dz**2)
  
  
  nnz=5
  v=Gscale*v
  v(nnz)=v(nnz) + scaled_diag
  
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
   goto 99
    
 endif
 
 
 
 !!==========================================================
 !!      nonperiodic (x) left & right faces, including edges
 !!========================================================== 
 
 if( i==0 .and. trim(btype(1,1))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i+2,j,k)
  v(1) = -3.d0/(2.*dx) 
  v(2) =  4.d0/(2.*dx)
  v(3) = -1.d0/(2.*dx)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 
 elseif( i==mx-1 .and. trim(btype(1,2))=='neumann' ) then
  col(1) = node_index(i,j,k)
  col(2) = node_index(i-1,j,k)
  col(3) = node_index(i-2,j,k)
  v(1) =  3.d0/(2.*dx) 
  v(2) = -4.d0/(2.*dx)
  v(3) =  1.d0/(2.*dx)
  nnz=3
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( i==0 .and. trim(btype(1,1))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( i==mx-1 .and. trim(btype(1,2))=='dirichlet' ) then
  col(1) = node_index(i,j,k)
  v(1) = 1.d0 
  nnz=1
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
 endif
 
  
 !!==========================================================
 !!      x dir periodic
 !!========================================================== 
 
 if( i==0 .and. xdim_periodic ) then
     
  col(1) = node_index(mx-1,j,k)
  col(2) = node_index(i+1,j,k)
  col(3) = node_index(i,j,k-1)
  col(4) = node_index(i,j,k+1)
  col(5) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=3
  v(3) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(4) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(5) = -2./(dx**2) -2./(dz**2)
  
  nnz=5
  v=Gscale*v
  v(nnz)=v(nnz) + scaled_diag
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 elseif( i==mx-1 .and. xdim_periodic ) then
     
  col(1) = node_index(i-1,j,k)
  col(2) = node_index(0,j,k)
  col(3) = node_index(i,j,k-1)
  col(4) = node_index(i,j,k+1)
  col(5) = node_index(i,j,k)
  
  idim=1
  v(1) =  1./(dx**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  v(2) =  1./(dx**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dx)
  idim=3
  v(3) =  1./(dz**2) + (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  v(4) =  1./(dz**2) - (x_ss(idim)/x_s(idim)**2)*1.d0/(2.*dz)
  
  v(5) = -2./(dx**2)-2./(dz**2)
  
  nnz=5
  v=Gscale*v
  v(nnz)=v(nnz) + scaled_diag
          
  call MatSetValues(A_PC,one,row,nnz,col,v,INSERT_VALUES,jerr)
    if( jerr .ne.0 ) then
     write(0,*) 'Error calling MatSetValues ',i,j,k
     write(0,*) 'myid ',myid
     stop
    endif
  goto 99
  
 endif
       

99 return

end subroutine SetMatrixElements2D


!!=====================================================
!!=====================================================
subroutine GetLabel(id,idim,endid,label)
 use operators
 implicit none
 integer            :: id,idim,endid
 character(len=80)  :: label
 if(id==6) then
   label=PhiBdryOps(idim,endid)%label
  elseif(id==1) then
   label=u_BdryOps(idim,endid)%label
  elseif(id==2) then
   label=v_BdryOps(idim,endid)%label
  elseif(id==3) then
   label=w_BdryOps(idim,endid)%label
  elseif(id==4) then
   label=s1_BdryOps(idim,endid)%label
  elseif(id==5) then
   label=s2_BdryOps(idim,endid)%label
  endif
end subroutine GetLabel

!!=====================================================
!!=====================================================
subroutine GetSval(id,idim,endid,i,j,S)
 use operators
 implicit none
 integer            :: id,idim,endid,i,j
 real(kind=8)       :: S
 if(id==6) then
   S=PhiBdryOps(idim,endid)%BoundaryData(i,j)
  elseif(id==1) then
   S=u_BdryOps(idim,endid)%BoundaryData(i,j)
  elseif(id==2) then
   S=v_BdryOps(idim,endid)%BoundaryData(i,j)
  elseif(id==3) then
   S=w_BdryOps(idim,endid)%BoundaryData(i,j)
  elseif(id==4) then
   S=s1_BdryOps(idim,endid)%BoundaryData(i,j)
  elseif(id==5) then
   S=s2_BdryOps(idim,endid)%BoundaryData(i,j)
  endif
end subroutine GetSval



!!=====================================================
!!=====================================================
subroutine SetGuess(petsc_vec)
 use dependent_variables,    only: velocity,scalars
 use intermediate_variables, only: Phi,rhs_b  
 use etc,                    only: field_id
 use InterfacePetsc
 
 implicit none
#include "finclude/petsc.h"
#include "finclude/petscvec.h"  
 Vec   petsc_vec
    
 if( field_id < 4 )    &
     call ScalarVariable2PetscVec(velocity(field_id),petsc_vec)
 if( field_id == 4 )   &
     call ScalarVariable2PetscVec(scalars(1),petsc_vec)
 if( field_id == 5 )   &
     call ScalarVariable2PetscVec(scalars(2),petsc_vec)
 if( field_id == 6 )   &
     call ScalarVariable2PetscVec(Phi(0),petsc_vec)
    
end subroutine SetGuess
!!=====================================================
!!=====================================================






!!============================================================
!!============================================================
subroutine FillA_WS(myid,localrows,globalrows)
!!   fill explicit A for preconditioning in wavenumber space
!!============================================================
 use operators,              only: L,A_PC,alpha,beta
 use dependent_variables,    only: velocity
 use intermediate_variables, only: scratch
 use user_data
 use etc,                    only:docfile,           &
                                  field_id,          &
                                  numzplanes_below,  &
                                  numzplanes
 implicit none

#include "finclude/petsc.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"

 PetscInt                    mx,my,mz
 PetscInt                    xs,ys,zs
 PetscInt                    xm,ym,zm
 
 integer                     :: i,j,k,myid
 integer                     :: localrows,globalrows
 integer                     :: level,locnz
 integer                     :: row,col
 integer                     :: jerr,one=1,nnz,id
 real(kind=8)                :: dsdx,dsdy,dsdz,mean_value
 real(kind=8)                :: xx,kx,ky,kz
 integer, external           :: node_index
  
 id=field_id
 !!===============================================
 !!  size parameters
 !!=============================================== 
 !! zero based start indices for myid in 3d representation
 xs=0
 ys=0
 zs=numzplanes_below(myid)
 
 !! zero based end indices for myid in 3d representation
 xm=xs + nx-1
 ym=ys + ny-1
 locnz=numzplanes(myid)
 zm=zs + locnz - 1
 
 !! global number of nodes in each direction
 mx=nx
 my=ny
 mz=nz
 
 level=0
 if(myid==0) then
  open(1,file=docfile,position='append')
  write(1,*) ' '
  write(1,*) '... Assembling A_PC in wavenumber space at level ',level
 endif
  
       
 !!===============================================
 !!  scale factors for x<<==>>s, use mean values
 !!===============================================
 mean_value=SUM(L(0)%Grid%SpaceDims(1)%dxds(:))/nx
 dsdx=1.d0/mean_value
 
 mean_value=SUM(L(0)%Grid%SpaceDims(2)%dxds(:))/ny
 dsdy=1.d0/mean_value
 
 mean_value=SUM(L(0)%Grid%SpaceDims(3)%dxds(:))/nz
 dsdz=1.d0/mean_value
 
 
 !!===============================================
 !!  Fill the matrix A_PC: the approximation to
 !!  L = alpha*grad^2 + beta
 !!  obtained by assuming the s<-->x transformations
 !!  are constant, using mean values.
 !!  A_PC is a simple diagonal matrix.
 !!===============================================
 do k=zs,zm    !! loop through my assigned data using global indexing
 
    kz = velocity(1)%Derivs(3)%kx(k+1) 
 
   do j=ys,ym
     
     ky = velocity(1)%Derivs(2)%kx(j+1)
            
     do i=xs,xm 
       
        !! 3d natural indexing of nodes/eqns
        row = node_index(i,j,k)
        col = row
        nnz = 1
  
        kx = velocity(1)%Derivs(1)%kx(i+1) 
        
        xx = -alpha(id)*( (dsdx*kx)**2 + (dsdy*ky)**2 + (dsdz*kz)**2 ) &
             + beta(id)
             
        call MatSetValues(A_PC,one,row,nnz,col,xx,INSERT_VALUES,jerr)
        if( jerr .ne.0 ) then
         write(0,*) 'Error calling MatSetValues wavenumber space ',i,j,k
         write(0,*) 'myid ',myid
         stop
        endif
                             
     enddo
   enddo
 enddo
 
 !!============================================================
 !!  Clear cache & leave matrix ready to use
 !!============================================================
 call MatAssemblyBegin(A_PC,MAT_FINAL_ASSEMBLY,jerr)
      if(jerr .ne.0 ) then 
       write(0,*) 'Error calling MatAssemblyBegin '
       stop
      endif
 call MatAssemblyEnd(A_PC,MAT_FINAL_ASSEMBLY,jerr)
      if(jerr .ne.0 ) then 
       write(0,*) 'Error calling MatAssemblyEnd '
       stop
      endif
 
  if(myid==0) &
  write(0,*) 'A_PC elements set and matrix object assembled assembled wavenumber space'
 return
end
!!=====================================================
!!=====================================================
