!!==========================================================
!!==========================================================
module user_routines
!! module for user-customized subroutines and functions

implicit none

public   :: user_ics                 !!routine defining initial conditions
public   :: user_forcing             !!routine defining any external forcing
public   :: user_bdry_forcing        !!routine for specifying boundary-condition data


!!==========================================================    
Contains
!!==========================================================


!!==========================================================
  subroutine user_ics(xval,yval,zval,id,m1,m2,m3,FF)
    !!===========================================================
    !!  User defined subroutine to specify initial conditions.
    !!  Inputs and outputs are all in dimensional units.
    !!  
    !!
    !!  inputs:
    !!   x,y,z   coordinate arrays (in meters) that define 
    !!           the region over which forcing functions
    !!           are to be evaluated
    !! 
    !!   id      field id [u,v,w,s1',s2'] <---[1,2,3,4,5]
    !!             where s1 = s1bar(z) + s1'(x,y,z,t)
    !!             and   s2 = s2bar(z) + s2'(x,y,z,t)
    !!
    !!  outputs:
    !!          F(:,:,:) the initial conditions for variable id
    !!          F is to be returned in DIMENSIONAL units 
    !!          (i.e. m/s or deg C etc )
    !!   
    !!===========================================================

    use dimensional_scales ,only: f,DEL,Nsquared,g,rho0,pi,L_scale,density_scale
    use dependent_variables,only: rho_bar,s1_bar_zeta,s2_bar_zeta
    use mpi_parameters     ,only: myid,comm,nprocs
    use counters_flags_etc ,only: ierr
    use grid_info          ,only: grid,Lx,Ly,Lz

    implicit none

    include '../input/problem_size.h' 
    include 'mpif.h'

    integer           ::  m1,m2,m3         
    integer           ::  id
    real              ::  xval(m1)
    real              ::  yval(m2)
    real              ::  zval(m3)
    real              ::  FF(m1,m2,m3)

    integer           :: i,j,k
    real              :: xpos,ypos,zpos,arg
    real              :: A_ics,k_ics,l_ics,m_ics,omega2,omega_ics
    real,dimension(2,2) :: c 

    integer           :: myproc
    real              :: ambient_density_gradient,N2_ics
    real,dimension(m1,m3) :: drhobardz
    real,dimension(:),allocatable :: drhobardz_tot

    !! Definition du processeur pour l'echange
    if ( myid == 0 ) then
            myproc=0
    else
            myproc=1
    endif
    !! allocation de "rho_bar_tot" pour le processeur "myproc" seulement
    if ( myid == myproc ) allocate ( drhobardz_tot(m3*nprocs) )
    !! calculde la derive de rho_bar(-,z) (dimensionne) selon z
    call compact_ddz_mpi( rho_bar*density_scale+rho0 , drhobardz , 1 , -1e32 , m1 , 1 , m3)
    !! collecte par le processeur "myproc" de toutes les valeurs de drhobardz
    call mpi_gather( drhobardz(1,:)/L_scale , m3 , MPI_REAL , drhobardz_tot , m3 , MPI_REAL , myproc , comm , ierr )
    !! moyenne sur toutes les valeurs
    if ( myid == myproc ) ambient_density_gradient = sum( drhobardz_tot ) / ( m3*nprocs)
    !! Diffusion  tous les processeurs la valeur "ambient_density_gradient"
    call mpi_bcast( ambient_density_gradient , 1 , MPI_REAL , myproc , comm , ierr ) 
    !! deallocation de "drhobardz_tot"
    if ( myid == myproc ) deallocate ( drhobardz_tot )

    N2_ics = (-g/rho0) * ambient_density_gradient

    !  Internal wave parameters
    A_ics=1.                                     ! displacement [m]
    k_ics=((1)*(2.*pi/Lx))                       ! [1/m]
    l_ics=((0)*(2.*pi/Ly))                       ! [1/m]
    m_ics=((1)*(pi/Lz))                          ! [1/m]

    !  Internal wave dispersion relation
    omega2 = ( ((f)**2)*m_ics**2 + N2_ics*(k_ics**2+l_ics**2) )/ &
         ( k_ics**2 + l_ics**2 + m_ics**2 )
    omega_ics=sqrt(omega2)                       ! [1/s]


    !  Constants in front of u & v expressions  [m/s]
    c(1,1)=A_ics*k_ics*m_ics*omega_ics/(k_ics**2+l_ics**2)
    c(1,2)=A_ics*l_ics*m_ics*f/(k_ics**2+l_ics**2)
    c(2,1)=A_ics*l_ics*m_ics*omega_ics/(k_ics**2+l_ics**2)
    c(2,2)=A_ics*k_ics*m_ics*f/(k_ics**2+l_ics**2)
    
    if ( id == 1 .and. myid == 0 ) then
           print*,'m1,m2,m3=',m1,m2,m3
           print*,'Lx,Ly,Lz=',Lx,Ly,Lz
           print*,'f,N,ambient_density_gradient=',f,sqrt(N2_ics),ambient_density_gradient
           print*,'k_ics,l_ics,m_ics=',k_ics,l_ics,m_ics
           print*,'c(1,1),c(1,2),c(2,1),c(2,2)=',c(1,1),c(1,2),c(2,1),c(2,2)
           print*,' '
           write(6,*) ' TEST WAVE FREQUENCY (1/s) :  ',omega_ics
           write(6,*) ' TEST WAVE PERIOD    (min) : ', (2*pi/omega_ics)/60.
   endif       

    !     arg=k_ics*x + l_ics*y   ! generally -omega_ics*t but set t=0

    select case(id)

    case(1)        !! ics for u [m/s]
       do i=1,m1
          do j=1,m2
             do k=1,m3        
                xpos = xval(i)     ! [m]
                ypos = yval(j)     ! [m]
                zpos = zval(k)     ! [m] 
                arg  = k_ics*xpos + l_ics*ypos 

                FF(i,j,k) = c(1,1) * cos(m_ics*zpos)*cos(arg) - c(1,2) * cos(m_ics*zpos)*sin(arg)

             enddo
          enddo
       enddo

    case(2)        !! ics for v [m/s]
       do i=1,m1         
          do j=1,m2
             do k=1,m3
                xpos = xval(i)     ! [m]
                ypos = yval(j)     ! [m]
                zpos = zval(k)     ! [m]  
                arg  = k_ics*xpos + l_ics*ypos

                FF(i,j,k) = c(2,1) * cos(m_ics*zpos)*cos(arg) + c(2,2) * cos(m_ics*zpos)*sin(arg) 

             enddo
          enddo
       enddo

    case(3)        !! ics for w [m/s]
       do i=1,m1
          do j=1,m2
             do k=1,m3 
                xpos = xval(i)     ! [m]
                ypos = yval(j)     ! [m]
                zpos = zval(k)     ! [m] 
                arg  = k_ics*xpos + l_ics*ypos

                FF(i,j,k) = A_ics*omega_ics*sin(m_ics*zpos)*sin(arg)

             enddo
          enddo
       enddo

    case(4)        !! ics for s1' [in this case deg C]
       do i=1,m1
          do j=1,m2
             do k=1,m3
                xpos = xval(i)     ! [m]
                ypos = yval(j)     ! [m]
                zpos = zval(k)     ! [m] 
                arg  = k_ics*xpos + l_ics*ypos

                FF(i,j,k) = -A_ics*s1_bar_zeta(i,k)*sin(m_ics*zpos)*cos(arg)

             enddo
          enddo
       enddo


    case(5)        !! ics for s2' [concentration units]
       do i=1,m1
          do j=1,m2
             do k=1,m3
                xpos = xval(i)     ! [m]
                ypos = yval(j)     ! [m]
                zpos = zval(k)     ! [m] 
                arg  = k_ics*xpos + l_ics*ypos

                FF(i,j,k) = -A_ics*s2_bar_zeta(i,k)*sin(m_ics*zpos)*cos(arg)

             enddo
          enddo
       enddo


    case default
       stop 'error in calling user_ics: id out of range'
    end select

    return
  end subroutine user_ics
!!========================================================== 
!!==========================================================
      

      
!!==========================================================
!!==========================================================
subroutine user_forcing(x,y,z,t,FF,id,m1,m2,m3)
  !!===========================================================
  !!  User defined subroutine to add time/space dependent
  !!  forcing to the equations of motion. Inputs and outputs
  !!  for this routine are all in dimensional units.
  !!
  !!
  !!  inputs:
  !!   x,y,z   coordinate arrays [m] that define
  !!           the region over which forcing
  !!           functions are to be evaluated
  !!   t       current time in seconds
  !!   id      field id [u,v,w,s1',s2'] <---[1,2,3,4,5]
  !!
  !!  outputs:
  !!          FF(:,:,:) the rhs term for equation id
  !!          FF is to be returned in DIMENSIONAL units
  !!          (i.e. m/s2 or deg C/s etc )
  !!
  !!===========================================================
  use dependent_variables, only: s1_bar_zeta,s2_bar_zeta
  use dimensional_scales,  only: f,pi,t_scale
  use mpi_parameters,      only: myid        
  use counters_flags_etc,  only: istep
  use grid_info,           only: Lx,Ly,Lz,dt

  implicit none
  integer   ::  m1,m2,m3
  real      ::  x(m1)
  real      ::  y(m2)
  real      ::  z(m3)
  real      ::  t
  integer   ::  id
  real,dimension(:,:,:),intent(out)  ::  FF

  real      :: xpos,ypos,zpos
  integer   :: i,j,k
  real      :: A_ics,xfreq,arg,G1,G2,sigma1,sigma2,sigma3
  integer   :: iend_storm
  real      :: Vt,lt,xt,H,zscale,init,gauss,window
  
  iend_storm= floor((0.5*(3600*24))/(dt*t_scale))

  select case(id)

  case(1)   ! u

  if ( istep .le. iend_storm ) then          

!=========FORCAGE TYPE ASARO========
      Vt=1.e-6
      lt=400.e3
      xt=Ly/2
      H=55.
      zscale=Lz/50
      init=atan( (Lz-H) / zscale )
      do i=1,m1
      do j=1,m2
      do k=1,m3
          gauss=exp(-(y(j)-xt)**2/lt**2)
          window=init-atan( (Lz-H-z(k)) / zscale )
          FF(i,j,k)=Vt*gauss*window*sin(y(j)*pi/Ly)
      enddo
      enddo
      enddo
      if ( Vt .ne. 0. ) then
        print*,'========='
        print*,' Forcage, istep=',istep,'  field=',id
        print*,'========='
      endif
!===================================
 
  else
  
     FF=0.d0

  endif   

  case(2)   ! v
     FF=0.d0

  case(3) ! w
     FF=0.d0

  case(4) ! s1
     FF=0.d0

  case(5) ! s2
     FF=0.d0

  case default
     stop 'error in calling user_forcing: id out of range'

  end select

  return

end subroutine user_forcing
!!===============================================================
!!===============================================================




!!==========================================================
subroutine user_bdry_forcing(x,y,z,t,FF,id,m1,m2,m3)
!!===========================================================
!!  User defined subroutine to add time/space dependent
!!  forcing to the equations of motion. Inputs and outputs
!!  for this routine are all in dimensional units.
!!
!!
!!  inputs:
!!   x,y,z   coordinate arrays [m] that define
!!           the region over which forcing
!!           functions are to be evaluated
!!   t       current time in seconds
!!   id      field id [u,v,w,s1',s2'] <---[1,2,3,4,5]
!!
!!  outputs:
!!          FF(:,:,:) the rhs term for equation id
!!          FF is to be returned in DIMENSIONAL units
!!          (i.e. m/s2 or deg C/s etc )
!!
!!===========================================================
 use stress_parameters    !! tau_0, delta, Patch_Radius
 use dependent_variables, only: s1_bar_zeta,s2_bar_zeta
 use dimensional_scales , only: pi,f,t_scale, rho0
 use user_parameters    , only: Km_of_z
 use counters_flags_etc , only: istep
 use grid_info          , only: dt,Lx,Ly,Lz


implicit none
  integer   ::  m1,m2,m3
  real      ::  x(m1)
  real      ::  y(m2)
  real      ::  z(m3)
  real      ::  t
  real      ::  FF(m1,m2)
  integer   ::  id
  real      ::  xpos,ypos,zpos
  integer   ::  i,j,k,istep_Ti



  !! surface stress non_zero value just for 1/2 Ti. 
  !! 0.5 Ti
  istep_Ti= floor((0.5*(2*pi/f))/(dt*t_scale)) 
  !! 1 Ti
  !istep_Ti= floor((1.0*(2*pi/f))/(dt*t_scale))

  select case(id)

    case(1)   ! u

!! xlv
!========surface stress imposed==========
 !if ( istep < istep_Ti ) then
 !        tau_0 = 0.2
 !elseif ( istep >= istep_Ti .and. istep <= 2*istep_Ti ) then        
 !        tau_0 = 0.4 * ( 1 - istep / (2*istep_Ti) )
 !elseif ( istep >= istep_Ti .and. tau_0 /= 0. ) then  
         tau_0 = 0.
 !        print*,' '
 !        print*,' Until now, surface stress is set to zero !'
 !        print*,' '
 !else
 !   print*,' ... '         
 !endif

 !! stress = tau_0 = rho0*nu*Shear(z=Lz)
 !! ===> bc for dUdz:  Shear = tau_0/(rho0*nu)
 !!Shear = tau_0/(rho0*nu) * t_scale
 Shear = tau_0/(rho0*Km_of_z(m3)) * t_scale
 print*,'     ... Shear=',tau_0/(rho0*Km_of_z(m3)),'  tau_0=',tau_0
 print*,'         m3=',m3,'  Km_of_z(locnz)=',Km_of_z(m3)
 print*,'         Patch_Radius=',Patch_Radius,'  delta=',delta

 do i=1,m1
 do j=1,m2
      xpos= x(i)
      ypos= y(j)
!!!      FF(i,j)= 1 - tanh( delta*( (xpos-Lx/2)**2 + (ypos-Ly/2)**2 - Patch_Radius**2 ) )

       FF(i,j)= cos(ypos*pi/Ly)*exp( -( (ypos - Ly/2)/20e3)**2 )
 enddo
 enddo
 FF=Shear * FF / maxval(abs(FF))

 open(933,file='../output/stress.data',status='unknown')
     do i=1,m1
     do j=1,m2
        write(933,*) x(i),y(j),FF(i,j)
     enddo
     enddo
 close(933)

 if (tau_0 .ne. 0. ) then
   print*,'========='
   print*,' Wind stress, istep=',istep,'  field=',id
   print*,'========='
 endif
!========================================

 case(2) ! v
   FF=0.d0
   
  case(3) ! w
   FF=0.d0
  
  case(4) ! s1

    do j=1,m2
      FF(:,j)= - s1_bar_zeta(:,m3)
    enddo
  
  case(5) ! s2
  
    do j=1,m2
      FF(:,j)= - s1_bar_zeta(:,m3)
    enddo

  case default
   stop 'error in calling user_bdry_forcing: id out of range'
  
 end select

return

end subroutine user_bdry_forcing
!!===============================================================
!!===============================================================








 
end module user_routines
