subroutine XY_FFT(in_data,out_data,m1,m2,m3,isign)
! Routine to compute the x&y (normalized) FFT (1st 2 dimensions)
! of a 3d field stored in in_data(1:m1,1:m2,1:m3) assuming
! a uniformly spaced grid mesh in x&y.

!  Normalization:   computing a forward transform and then an
!                   inverse transform of the result in FFTW
!                   results in the original data scaled by N.
!                   Here we will normalizs transforms in BOTH
!                   directions by sqrt(N).
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       inputs
!              in_data     m1,m2,m3 array of input values
!              m1,m2,m3    dimensions of field to be transformed
!              isign       +1=forward=real_to_complex, -=inverse=complex_to_real
!
!       output
!              out_data    m1,m2,m3 array containing the result
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 use boundary_information
 use mpi_parameters
 use counters_flags_etc
 implicit none

 include  '../input/problem_size.h'

 integer, intent(in)                              :: isign,m1,m2,m3
 real, dimension(m1,m2,m3)                        :: in_data
 real, dimension(m1,m2,m3)                        :: out_data
!!! integer                                          :: i,j,k
!!! integer                                          :: ierr

 real, allocatable                                :: in(:),out(:)
!!! real, allocatable                                :: tmp(:,:,:)
 real, dimension(nx,ny,locnz)                     :: tmp
 real                                             :: pi,xnorm
 integer*8                                        :: plan !size must match pointer size, 8 for DEC alphas
 integer                                          :: N
 include 'fftw_kw.inc'                            !  predefined FFTW constants

 if( boundary_type2(1,1) .ne. 'periodic' .or. boundary_type4(1,1) .ne. 'periodic' ) then 
  write(0,*) '**********   BOTH X&Y DIRECTIONS SHOULD BE PERIODIC  **********'
  write(0,*) '**********   IF XY_FFT IS TO BE CALLED !!           ***********'
  STOP
 endif

! X preliminaries, setup fftw "plan"
  pi=4.*atan(1.)
  N=m1-1                                  ! m1 includes both boundary points, N only 1
  xnorm = 1./sqrt( float(N) )
!!!  allocate( in(N),out(N),tmp(m1,m2,m3) )
  allocate( in(N),out(N) )

  if( isign == 1 ) then
   call rfftw_f77_create_plan(plan,N,FFTW_REAL_TO_COMPLEX,FFTW_ESTIMATE)
  elseif( isign == -1 ) then
   call rfftw_f77_create_plan(plan,N,FFTW_COMPLEX_TO_REAL,FFTW_ESTIMATE)
  else
   write(0,*) ' **** XY_FFT:   SIGN OF TRANSFORM MUST BE +1/-1  FOR FORWARD/INVERSE DIRECTIONS '
  endif

! do the x transforms; store result in tmp
  do j=1,m2
   do k=1,m3
    in(1:N)=in_data(1:N,j,k)
    call rfftw_f77_one(plan,in,out)  
    tmp(1:N,j,k)=out(1:N)*xnorm
   enddo
  enddo
  if( isign==-1 ) tmp(m1,:,:)=tmp(1,:,:)  ! extra periodic endpoint
  if( isign== 1 ) tmp(N/2+1,:,:)=0.0      ! zero out nyquist component
  call rfftw_f77_destroy_plan(plan)
  deallocate( in,out )

!!!  call MPI_BARRIER(comm,ierr)
 
! Y preliminaries, setup fftw "plan"
  N=m2-1                                  ! m2 includes both boundary points, N only 1
  xnorm = 1./sqrt( float(N) )
  allocate( in(N),out(N) )

  if( isign == 1 ) then
   call rfftw_f77_create_plan(plan,N,FFTW_REAL_TO_COMPLEX,FFTW_ESTIMATE)
  elseif( isign == -1 ) then
   call rfftw_f77_create_plan(plan,N,FFTW_COMPLEX_TO_REAL,FFTW_ESTIMATE)
  else
   write(0,*) ' **** XY_FFT:   SIGN OF TRANSFORM MUST BE +1/-1  FOR FORWARD/INVERSE DIRECTIONS '
  endif

! do the y transforms; store result in tmp
!!!   print *,'START y-trans loop, m1,m3,N,myid,istep,isign = ',m1,m3,N,myid,istep,isign
  do i=1,m1
   do k=1,m3
    in(1:N)=tmp(i,1:N,k)
     call rfftw_f77_one(plan,in,out)  
     out_data(i,1:N,k)=out(1:N)*xnorm
   enddo
  enddo
!!!   print *,'END y-trans loop, myid,istep,isign = ',myid,istep,isign
  if( isign==-1 ) out_data(:,m2,:)=out_data(:,1,:)  ! extra periodic endpoint
  if( isign== 1 ) out_data(:,N/2+1,:)=0.  ! zero out near nyquist components
  call rfftw_f77_destroy_plan(plan)
!!!  deallocate( in,out,tmp )
  deallocate( in,out )

end subroutine XY_FFT
