#include "config.h"

! ****************************************************************************
program NS3Dstrat
!
! ***************************************************************************
!
!       Code NS3Dstrat - LadHyX
!
! ***************************************************************************
!       PRESENTATION
! ***************************************************************************
!
!       Ce code resout les equations de Navier-Stokes 2D ou 3D par une methode pseudo spectrale.
!       La solution calcule est donc considre priodique sur le domaine de rsolution.
!       Le schema temporel utilise est d'ordre deux (Adams Bashforth).
!       Le code peut eventuellement etre mis en oeuvre dans le cas stratifie (approximations de Boussinesq).
!       Les equations peuvent etre resolues en perturbation (autour d'un etat de base 2D) ou non et dans un cas lineaire ou non.
!       Dans le cas d'un run en perturbation, l'tat de base vbx, vby, vbz, wbx, wby, wbz est considere comme invariant suivant z et
!       ne dpend pas du temps.
!
!       L'tat de base 2D tout comme le champ de vitesse initial peuvent soit tre definis en appelant une sous-fonction
!       interne ou alors en lisant un fichier d'entre.
!
!
! ***************************************************************************
!       UTILISATION DU CODE
! ***************************************************************************
!
!       Pour utiliser ce code, il faut :
!           - editer le fichier "config.h"
!           - editer le fichier "data.in"
!           - compiler (dans l'ordre indiqu) et lier les fichiers source "parametres.F90", "fft.F90", "sous_routines.F90"
!             et "NS3D.F90".
!             Attention lors de l'tape de compilation, il faut activer (si ce n'est pas dja le cas par dfaut)
!             le passage pralable du prprocesseur C/C++.
!             Il faut egalement penser a passer au linker l'emplacement des librairies de la FFT utilisee (FFTW, ASL...)
!           - l'executable obtenu doit ensuite tre execute dans le dossier qui contient le fichier "data.in" et
!             eventuellement les diffrents fichiers de donnes d'entre ("velocity.init", "base2D.init", "velo2D.init").
!
!       Toutes ces tapes peuvent tre automatises par l'utilisation d'un script shell (c'est mme fortement conseill).
!
!
! ***************************************************************************
!       ENTREES / SORTIES
! ***************************************************************************
!
!  Fichiers d'entre :
!
! - config.h : ce fichier contient tous les parametres ncessaires  l'tape de compilation du code. Il est utilise par le
!              preprocesseur C/C++.
!
! - data.in : ce fichier contient l'ensemble des parametres utiles  l'tape d'excution du code. Il est lu par le code en
!             debut d'execution.
!
! - velocity.init : ce fichier est utilis lorsque l'on souhaite reprendre un run.
!
!                   Contient en binaire ("unformatted" pour Fortran) les champs de l'etat initial :
!                   vox(x,y,z), voy(x,y,z), voz(x,y,z), rho(x,y,z), wox(x,y,z), woy(x,y,z), woz(x,y,z)
!                   qui sont de taille nxmax*nymax*nzmax, et les champs de l'etat de base vbx(x,y), vby(x,y), vbz(x,y),
!                   wbx(x,y), wby(x,y),wbz(x,y) de taille nxmax*nymax.
!
!                   De faon plus prcise le format du fichier est le suivant:
!
!                   i,1,nxmax,nymax,nzmax,lx,ly,lz,truncate,type_trunc,rtrunc_x,rtrunc_y,rtrunc_z,nu,stratification,
!                   xns,schm,omega2,perturbe,lin,j,i,vox,j,i,voy,j,i,voz,j,i,rho,j,i,fx,j,i,fy,j,i,fz,j,
!                   i,vbx,vby,vbz,j,i,wbx,wby,wbz,j
!
!                   1 est un flag permet d'eviter de verifier le format du fichier (big/little endian par exemple)
!                   nxmax, nymax, nzmax sont les dimensions suivant chaque direction (nombre de points de colocation)
!                   lx,ly,lz correspondent aux dimensions physiques de la bote selon les directions x,y et z
!                   truncate est un logical qui indique si le desaliasing est utilise ou non
!                   time est l'instant ou commence la simulation
!                   vox, voy, voz sont les pertubations de vitesse
!                   rho est la perturbation de densite
!                   wox, woy, woz sont la vorticite de la perturbation de vitesse
!                   vbx, vby, vbz, wbx, wby, wbz sont respectivement la vitesse et la vorticite de l'tat de base.
!
!                   i et j sont des entiers qui sont ajouts par Fortran lors de l'enregistrement de donnes binaires.
!                   Ils correspondent respectivement  la taille en octets de l'enregistrement qui suit ou qui prcde.
!
!                   Dans les mtadonnes, le flag "1" est utile pour verifier que l'on ne fait pas d'erreur lors des
!                   entrees sorties en unformatted(little-endian, big-endian). Les autres mtadonnes (dimensions etc)
!                   permettent d'viter galement pas mal d'erreurs.
!
!
! - base2D.init : ce fichier permet de definir un etat de base ne dpendant que de x et y.
!
!                 Contient vbx(x,y), vby(x,y), vbz(x,y), wbx(x,y), wby(x,y), wbz(x,y) en "unformatted".
!
!                 Les champs sont de taille nxmax*nymax.
!                 lx,ly correspondent aux dimensions physiques de la bote selon x et y
!
!                 Le format prcis du fichier est :
!                 i,1,nxmax,nymax,lx,ly,j,i,vbx,vby,vbz,j,i,wbx,wby,wbz,j
!
!
! - velo2D.init : ce fichier permet de definir un champs de vitesse initial ne dpendant que de x et y.
!
!                 Contient vix(x,y), viy(x,y), viz(x,y) en "unformatted". Les champs sont de taille nxmax*nymax.
!
!                 De faon plus prcise le format du fichier est le suivant:
!                 i,1,nxmax,nymax,lx,ly,j,i,vix,viy,viz,j
!
!
!  Fichiers de sortie :
!
!  - velo_rho_vort.t=####.### : contient vox(x,y,z), voy(x,y,z), voz(x,y,z), rho(x,y,z), wox(x,y,z), wox(x,y,z), wox(x,y,z),
!                               vbx(x,y), vby(x,y), vbz(x,y), wbx(x,y), wby(x,y),wbz(x,y) en "binary".
!                               Ces fichiers contiennent la vitesse, la densite, la vorticite et l'tat de base
!
!                               Ils sont au exactement format qui peut tre utilis pour reprendre un run
!                               (fichier "velocity.init") cd avec des metadonnees au debut etc.
!
!  - energie_quadratique_mode : contient l'energie de chaque mode kx et kz en moyennant suivant ky
!                               Pour le format prcis du fichier voir directement dans le code...
!
!
!
! ***************************************************************************
!       DIVERS
! ***************************************************************************
!
!      - Au niveau du temps de calcul, le temps cpu varie approximativement comme Tcpu = Const*nb_steps*N*ln(N)
!        avec nb_steps le nombre de pas de temps, N la dimension totale du probleme N=nxmax*nymax*nzmax et Const une constante
!        qui dpend de l'environnement de calcul utilis.
!
!        Attention, il s'agit ici d'une estimation du temps cpu (temps de calcul effectivement realise) et non
!        pas du temps de restitution (= temps que l'utilisateur doit attendre en vrai). Le temps de restitution
!        peut etre superieur au temps cpu (par exemple lorsqu'on execute le code sur un processeur utilise pour d'autres processus)
!        ou inferieur (lorsqu'on execute le code sur un environnement parallele o plusieurs processeurs travaillent ensemble).
!
!        Pour info la constante Const vaut environ :
!        6.5E-9 sec sur un NEC SX-5 (fft de la bibliotheque ASL)
!        6.E-8 sec sur un Xeon 3.06GHz (fft de la bibliotheque FFTW 3.01)
!        2.5E-7 sec sur un Pentium4 800 MHz (fft de la bibliotheque FFTW 3.01)
!        8.E-8 sec sur un Power4 1.3 GHz (fft de la bibiliotheque ESSL)
!
!        Ces valeurs sont valables pour une simulation en fluide non-stratifie. Dans le cas d'un fluide stratifie, il faut ajouter
!        de 20  40 % de temps en plus parce qu'il y a (notamment) des fft supplementaires a effectuer.
!
!        Ces infos permettent d'avoir une ide tres approximative (a un facteur 1.5 pres disons...) du temps de calcul necessaire et
!        donc, en fonction de l'environnement de calcul, du temps de restitution.
!
!
!      - Avec la librairie FFTW. Il n'y a pas de restrictions particulires sur les dimensions du problme.
!        Toutes les valeurs entires peuvent tre utilises mme si certaines donnent de meilleurs rsultats en terme de
!        vitesse de calcul (puissance de 2,3,5 etc).
!
!        Avec la librairie ASL, il faut que nxmax, nymax, nzmax >= 2 mais aprs toutes les valeurs sont acceptables.
!
!        Pour la librairie ESSL, il faut que nxmax, nymax, nzmax appartiennent  la liste suivante
!        (on se limite ici aux entiers infrieurs  1024; pour les autres cf la doc ESSL)
!          2         4         6         8        10        12        14        16        18
!         20        22        24        28        30        32        36        40        42
!         44        48        56        60        64        66        70        72        80
!         84        88        90        96       110       112       120       126       128
!        132       140       144       154       160       168       176       180       192
!        198       210       220       224       240       252       256       264       280
!        288       308       320       330       336       352       360       384       396
!        420       440       448       462       480       504       512       528       560
!        576       616       630       640       660       672       704       720       768
!        770       792       840       880       896       924       960       990      1008
!       1024
!
!      - Le meme code peut etre utilise pour simuler un seul mode suivant la direction z.
!        Pour realiser une simulation dans ce mode, il faut que le fichier config.h contienne "#define mode_kz" et la longueur d'
!        onde du mode simule vaut alors lz. Pour une longueur d'onde infini (kz=0), on peut indiquer lz=0 dans le fichier data.in.
!        De faon plus precise, le code simule l'volution de champs de la forme (u,v,w,rho)(x,y)exp(i.kz.z)*g(t) avec
!        u et v reels, w et rho imaginaires pur, g reel (et donc en particulier il faut que le taux de croissance sigma soit
!        purement reel).
!        Attention les resultats obtenus ne sont valables que pour une simulation lineaire autout d'un etat de base ne dependant
!        que de x et y et n'ayant pas de vitesse suivant la direction z.
!        Numriquement on travaille alors vox = Re(u) = u, voy = Re(v) = v, voz = Im(w) = -i*w, rho_du_code = Im(rho) = -i*rho
!
!      - On peut galement simuler un mode purement 2D. On n'est pas restreint alors a une simulation lineaire. Il suffit d'indiquer
!        nz = 1
!
!
!      - Remarques sur l'utilisation memoire du code :
!        La majorite (environ 90-95%) des besoins memoire du code est representee par 12 champs de rels double precision
!        et 1 champ de logical qui sont tous environ de taille nxmax*nymax*nzmax (pour la dimension exacte,
!        voir la remarque suivante). Ces champs sont places en memoire data.
!
!        Si on utilise la libraire FFT ASL, il faut en plus ajouter un champ de travail supplmentaire de rels double
!        precision environ de taille nxmax*nymax*nzmax. Dans le cas d'une utilisation en parallele, il y a un champs
!        de travail par processus parallele. Ces champs de travail sont places par defaut dans la memoire stack
!        (je crois...).
!
!      - toujours pour les besoins memoires, la sous fonction "out_champs" peut necessiter d'importants besoins en
!        memoire stack parce que chaque enregistrement ecrit une partie de tableau et non pas la totalite (Exemple:
!        "write(76) vox(0:nxmax-1,0:nymax-1,0:nzmax-1)" et non pas "write(76) vox"). En fait la gestion de la memoire
!        dans ce cas (utilisation de memoire stack ou non pour stocker temporairement les parties de tableau qui vont
!        etre ecrites) depend pas mal du compilateur. Tout cela n'est pas tres clair mais la conclusion est que si
!        le code plante en debut d'execution, il y a de bonnes chances que cela soit lie a ce besoin en memoire stack.
!        Il faut alors veiller a ne pas limiter la quantite de memoire stack utilisable sur le systeme
!        (ulimit -s unlimited sur Linux par exemple).
!
!
!      - Numeriquement, dans le code, les champs de grandes dimensions (vox,voy,voz,rho,vox2,voy2,voz2,rho2,fx,fy,fz,frho,trunctab)
!        ne sont pas de taille nxmax*nymax*nzmax mais de taille (2*(int(nxmax/2)+1)+padx) * (nymax+pady) * (nzmax+padz).
!        Cela est du au fait que pour calculer des fft reel/complexe, il faut un tableau pour les coefficients spectraux legerement
!        plus grand que le tableau stockant les valeurs physiques. On ajoute de plus des valeurs "inutiles" dans le calcul avec les
!        coefficients padx, pady, padz de facon a optimiser les acces memoire (c'est surtout important sur le NEC de l'IDRIS;
!        les valeurs adaptees de padx, pady, padz dependent des dimensions nxmax, nymax, nzmax ainsi que de l'architecture
!        de l'ordinateur utilise).
!
!      - il est conseill de compiler le code en forant le codage des rels simple precision en double precision. Cela
!        amliore la prcision des calculs (argument "-r8" pour le compilateur Intel Fortran par exemple).
!
!
! ***************************************************************************
!       CHANGELOG
! ***************************************************************************
!
!       v1.36b (09/10/2006 - A. Deloncle) :
!        - changement des librairies de FFT utilisees. Cette modification reprend en fait les modifs
!          apportees au code depuis juin 2005 et present dans la version principale de developpement 2.*
!          Elle n'a ete faite que pour permettre d'utiliser l'ancienne version du code 1.36  l'IDRIS
!          avec le nouvel ordinateur vectoriel Brodie.
!          La nouvelle version developpee 2.* comprend d'autres amliorations importantes notamment la
!          parallelisation MPI
!
!
!       v1.36 (16/06/2005 - A. Deloncle) :
!        - possibilite de simuler un unique mode kz
!        - utilisation de transformees de Fourier 2D des lors que DIMZ=1
!        - affichage du temps de calcul ecoule et d'une estimation du temps restant
!
!       v1.35 (31/03/2005 - A. Deloncle) :
!        - Modification du format des fichiers d'entres/sorties (tous les champs sont dans un mme fichier,
!          ajout de mtadonnes)
!        - modification de la fft ASL pour viter un bug ( confirmer...) existant dans la librairie (problme avec le champ
!          de travail de la librairie ASL)
!        - ajout d'une fonction pour ecrire l'energie quadratique moyenne suivant ky pour chaque mode kx et kz
!
!       v1.33 (15/02/2005 - A. Deloncle) :
!        - Modification de l'en tte des fichiers d'entres/sorties : "1, nxmax, nymax, nzmax, lx, ly, lz"
!
!       v1.31 (10/02/2005 - A. Deloncle) :
!        - changement du format des entres-sorties pour gagner en besoin memoire stack et faire moins d'erreurs
!          (ajout des flags "1, nxmax, nymax, nzmax")
!        - ecriture de commentaires plus nombreux sur le code (description des entrees-sorties etc.)
!
!       v1.3 (10/11/2004 - A. Deloncle) :
!               - passage du code a la norme Fortran 90/95 (modules, format libre, fonctions intrinseques sur les tableaux...)
!               - davantage de choix pour la librairie FFT utilisee (ajout de FFTW v3.01 en version mono-processeur et parallele,
!                 ajout de la librairie ASL en version parallele, ajout de la librairie ESSL version mono-processeur)
!               - correction de quelques "bugs" (troncature, nombre de Schmidt, termes preexponentiels pour valeurs extremes de k)
!               - economie de memoire (reecriture des fonctions "white_noise", "gen_base2D", "gen_velo2D", tableau de troncature
!                 code en logical plutt qu'en double precision)
!               - possibilite de faire une troncature carree
!               - optimisation du temps de calcul (reecriture des fonctions de la boucle principale, parallelisation openmp)
!               - possibilite de tester la presence de NaN et d'interrompre le calcul le cas echeant (uniquement si la fonction
!                 "isnan" est disponible sur le compilateur utilise)
!               - changement de la mthode de projection de la vitesse vo dans le plan orthogonal  k
!               - suppression de variables/fonctions inutiles
!               - plus d'utilisation du fait que div(u)=0 pour gagner un champ en stockage (gain en temps de calcul
!                 mais utilisation d'un champ supplementaire en espace memoire).
!               - d'autres choses dont je ne me souviens plus...
!
!       v1.2 (P. Billant) :
!               - possibilite de desactiver la stratification pour ne pas passer par les FFT necessaires
!                 a l'equation de la densite
!               - routine separee pour le calcul de u*grad(rho)
!               - attention : les coordonnees x et z ont ete inversees pour que l'etat de base soit
!                 dans le plan (x,y) et la stratification et la rotation planetaire sur z
!
!       v1.1 (20/04/04 - J.M. Chomaz) :
!               - cas stratifie
!               - version portable
!
!       v1.0 (Vincent, Meneguzzi, Brancher, Delbende, Gallaire) :
!           Code original :
!
!           Three dimensional Navier-Stokes Homogeneous Turbulence using Fourier spectral method.
!
!           Version NEC-SX5 par F. Gallaire avoid memory bank conflicts
!           Resolution de l'equation de Navier-Stokes 3d  288***3 sur Cray2 (252 megamots)
!           Ecrit par A. Vincent et M. Meneguzzi avec des FFT de C. Temperton.
!           Version avec huit champs: un champ de stokage (vkx0,vy,vz) dans lequel on ne travaille jamais
!           et 6 champs de travail (vox,voy,voz) et (fx,fy,fz)
!           Dans cette version, on impose la divergence nulle ce qui conduit au recalcul de la troisieme
!           composante de la vitesse pour laquelle seule l'harmonique kz=0 est conservee.
!           L'absence d'un neuvieme champ impose d'utiliser des permutations de type:
!           (B,A,C)=(B,A,A)o(B,B,A)o(A,B,A)o(A,B,C)
!           Version avec extraction de structures pour etude en temps.
!
! ***************************************************************************

    ! Modules externes utilises par le programme
    use parametres
    use fft
    use sous_routines

    implicit none

    ! format d'entrees-sorties utilises
    character(len=100), parameter :: aff_tps = '(A22,TR1,F9.2,TR1,A3,TR3,F6.2,TR1,A1)'

#ifdef test_isnan
    integer :: ix, iy, iz
    logical :: detect_nan
#endif

    ! Lecture des parametres de la simulation dans le fichier "data.in"
    call enter_param

    ! Si le nombre de points suivant z est 1, on evite de tronquer cet unique mode en mettant un rayon de troncature tres eleve
    if (nzmax==1) then
        rtrunc_z = 1.E30
    end if

    ! Quelques initialisations supplementaires utiles par la suite
    pi = acos(-1.)

    ! On affiche les parametres du run
    call out_param

    write(*,*) 'INITIALISATION'

    ! Calcul des coordonnees des points de colocation
    call init_coord(xx,yy,zz)

    ! Initialisation des nombres d'ondes
    call init_k(kx,k2x,ky,k2y,kz,k2z,cax,ca2x)

    ! Initialisation du tableau de troncature
    if (truncate) then
        call init_trunctab(trunctab)
    end if

    ! initialization of the Fast Fourier Transform (FFT) functions
    write(*,*)
    write(*,*) 'Initialization of the Fast Fourier Transformation (FFT)'
    call init_fft(vox,nxmax,nymax,nzmax,dimx_tab,dimy_tab,dimz_tab,fft_work)
    write(*,*) '  -> '//fft_lib_name
    write(*,*)

    ! ******************************
    ! Definition des conditions initiales
    ! Deux possibilites :
    !     - soit il s'agit d'une reprise de run : on lit alors le fichier de conditions initiales
    !     - soit il s'agit d'un nouveau run, il faut alors definir un etat de base et/ou une vitesse initiale
    ! ******************************

    if (dRho_read) then
  
     write(*,*) "  -> lecture du fichier 'dRho_dz.init'"
     call read_dRho(dRho_dz)
    
    end if


    if (reprise_run) then
        write(*,*) "Reprise d'un run"
        write(*,*) "  -> lecture du fichier 'velocity.init'"
        call read_run(vox,voy,voz,rho,fx,fy,fz,vbx,vby,vbz,wbx,wby,wbz,begin)
    else
        write(*,*) "Etat de base"

        if (base2D_read) then
            write(*,*) "  -> lecture du fichier 'base2D.init'"
            call read_base2D(vbx,vby,vbz,wbx,wby,wbz)
        else if (base2D_sub) then
            write(*,*)  "  -> generation de l'etat de base2D"
            call gen_base2D(vbx,vby,vbz,wbx,wby,wbz,vox2,voy2,voz2)
        else
            write(*,*) "  -> etat de base2D nul"
            vbx=0
            vby=0
            vbz=0
            wbx=0
            wby=0
            wbz=0
        end if

        write(*,*)
        write(*,*) "Vitesse initiale"

        if (velo2D_read) then
            write(*,*) "  -> lecture du fichier 'velo2D.init'"
            call read_velo2D(vox,voy,voz)
        else if (velo2D_sub) then
            write(*,*)  "  -> generation de la vitesse initiale"
            call gen_velo2D(vox,voy,voz)
        else
            write(*,*) "  -> vitesse initiale nulle"
            vox = 0.
            voy = 0.
            voz = 0.
            rho = 0.
        end if

        ! Ajout d'un bruit blanc  la vitesse initiale
        if (lnoise) then
            write(*,*) "  -> ajout d'un bruit blanc a la vitesse initiale"
            call white_noise(vox,voy,voz)
        end if

    end if

    ! on sauvegarde tous les champs initiaux
    vox2 = vox
    voy2 = voy
    voz2 = voz

    call fwd_fft(vox2,fft_work)
    call fwd_fft(voy2,fft_work)
    call fwd_fft(voz2,fft_work)
    call rotv(vox2,voy2,voz2,fx,fy,fz)
    call bck_fft(fx,fft_work)
    call bck_fft(fy,fft_work)
    call bck_fft(fz,fft_work)

    call out_champs(vox,voy,voz,rho,fx,fy,fz)

    ! ******************************
    ! Premier pas de calcul
    ! ******************************
    write(*,*)
    write(*,*) 'PREMIER PAS DE TEMPS'

    call fwd_fft(vox,fft_work)
    call fwd_fft(voy,fft_work)
    call fwd_fft(voz,fft_work)

    if (stratification) then
         call fwd_fft(rho,fft_work)
    end if

    ! On projette dans le plan orthogonal  k et on tronque
    call projection(vox,voy,voz)
    if(truncate) then
        call trunc0(vox,voy,voz,rho)
    endif

    ! Le premier pas est calcule  partir d'un schema de Runge-Kutta d'ordre 2
    call raccro(vox,voy,voz,rho,fx,fy,fz,frho,vox2,voy2,voz2,rho2)

    ! on projette et on tronque
    call projection(vox,voy,voz)
    if(truncate) then
        call trunc0(vox,voy,voz,rho)
    endif

    write(*,*) '  -> OK'

    ! ******************************
    ! Pas suivants de calcul
    ! ******************************
    write(*,*)
    write(*,*) 'PAS DE TEMPS SUIVANTS'

    ! on initialise les compteurs du temps passe dans chaque sous-routine
    tps_adams_step1 = 0
    tps_adams_step2 = 0
    tps_rotv = 0
    tps_fwd_fft = 0
    tps_bck_fft = 0
    tps_out = 0
    tps_vecpro2 = 0
    tps_urho = 0
    tps_div_speciale = 0
    tps_projection = 0
    tps_trunc0 = 0
    tps_detect_nan = 0
    tps_total = 0

    call cpu_time(tps_temp)
    tps_total = tps_total - tps_temp

    ! On rentre dans la boucle avec les valeurs des champs a l'instant it-1 et on en sort avec les valeurs a l'instant it
    ! Globalement ici, on rentre donc avec les valeurs  l'instant 1 et on en sort avec les valeurs  l'instant itmax
    do it=2,itmax
        !write(*,*) 'boucle it2, itmax' 
        ! Shma d'ordre deux d'Adams-Bashforth
        call adams_bashforth(vox,voy,voz,rho,fx,fy,fz,frho,vox2,voy2,voz2,rho2)

        ! on projette et on tronque
        call cpu_time(tps_temp)
        tps_projection = tps_projection - tps_temp
        call projection(vox,voy,voz)
        call cpu_time(tps_temp)
        tps_projection = tps_projection + tps_temp

        if(truncate) then
            call cpu_time(tps_temp)
            tps_trunc0 = tps_trunc0 - tps_temp
            call trunc0(vox,voy,voz,rho)
            call cpu_time(tps_temp)
            tps_trunc0 = tps_trunc0 + tps_temp
        endif

! On teste la prsence ou non de NaN dans les rsultats obtenus. Le cas chant, on stoppe le calcul.
! Attention, la fonction "isnan" ne fait pas partie de la norme Fortran 90/95 et il faut donc tre sr
! que le compilateur utilise l'accepte. C'est le cas par exemple de l'Intel Fortran Compiler 8.
! On ne cherche des NaN que dans le champ vox pour gagner en temps de calcul. De toute faon si un NaN apparat
! dans un autre champ, il devrait se propager  vox en quelques pas de temps.
#ifdef test_isnan
        call cpu_time(tps_temp)
        tps_detect_nan = tps_detect_nan - tps_temp

        detect_nan = .false.
        do iz=0, nzmax-1
            do iy=0, nymax-1
                do ix=0, nxmax+1
                    detect_nan = detect_nan .or. isnan(vox(ix,iy,iz))
                end do
            end do
        end do

        call cpu_time(tps_temp)
        tps_detect_nan = tps_detect_nan + tps_temp

        if (detect_nan) then
            write(*,*) "ERREUR : il y a un Nan dans les resultats obtenus. Arret du calcul."
            exit
        end if
#endif

    end do

    call cpu_time(tps_temp)
    tps_total = tps_total + tps_temp

    write(*,*)
    write(*,*) 'SIMULATION TERMINEE'

    ! on sauvegarde tous les champs finaux (instant itmax)
    call rotv(vox,voy,voz,fx,fy,fz)
    call bck_fft(vox,fft_work)
    call bck_fft(voy,fft_work)
    call bck_fft(voz,fft_work)

    if (stratification) then
        call bck_fft(rho,fft_work)
    end if

    call bck_fft(fx,fft_work)
    call bck_fft(fy,fft_work)
    call bck_fft(fz,fft_work)

    it = itmax
    call out_champs(vox,voy,voz,rho,fx,fy,fz)


    ! on affiche toutes les infos sur le temps de calcul des differentes sous-routines de la boucle principale
    write(*,*)
    write(*,*)
    write(*,*) '---------------------------------------------------'
    write(*,*) 'TEMPS DE CALCUL DANS LA BOUCLE PRINCIPALE'
    write(*,*) '---------------------------------------------------'
    write(*,*)
    write(*,aff_tps) 'Temps total :', tps_total, 'sec', 100*tps_total/tps_total,'%'
    write(*,*)
    write(*,aff_tps) 'fwd_fft :', tps_fwd_fft, 'sec', 100*tps_fwd_fft/tps_total,'%'
    write(*,aff_tps) 'bck_fft :', tps_bck_fft, 'sec', 100*tps_bck_fft/tps_total,'%'
    write(*,aff_tps) 'vecpro2 :', tps_vecpro2, 'sec', 100*tps_vecpro2/tps_total,'%'
    write(*,aff_tps) 'projection :', tps_projection, 'sec', 100*tps_projection/tps_total,'%'
    write(*,aff_tps) 'adams_step1 :', tps_adams_step1, 'sec', 100*tps_adams_step1/tps_total,'%'
    write(*,aff_tps) 'adams_step2 :', tps_adams_step2, 'sec', 100*tps_adams_step2/tps_total,'%'
    write(*,aff_tps) 'copie :', tps_copie, 'sec', 100*tps_copie/tps_total,'%'
    write(*,aff_tps) 'div_speciale :', tps_div_speciale, 'sec', 100*tps_div_speciale/tps_total,'%'
    write(*,aff_tps) 'rotv :', tps_rotv, 'sec', 100*tps_rotv/tps_total,'%'
    write(*,aff_tps) 'out :', tps_out, 'sec', 100*tps_out/tps_total,'%'
    write(*,aff_tps) 'urho :', tps_urho, 'sec', 100*tps_urho/tps_total,'%'
    write(*,aff_tps) 'trunc0 :', tps_trunc0, 'sec', 100*tps_trunc0/tps_total,'%'
    write(*,aff_tps) 'detect_nan :', tps_detect_nan, 'sec', 100*tps_detect_nan/tps_total,'%'
    write(*,*)
    write(*,aff_tps) 'Total sous-routines :', tps_adams_step1+tps_adams_step2+tps_rotv+tps_fwd_fft+tps_bck_fft+ &
                                              tps_out+tps_vecpro2+tps_urho+tps_div_speciale+tps_projection+ &
                                              tps_trunc0+tps_copie+tps_detect_nan, 'sec', &
                                              100*(tps_adams_step1+tps_adams_step2+tps_rotv+tps_fwd_fft+tps_bck_fft+ &
                                              tps_out+tps_vecpro2+tps_urho+tps_div_speciale+tps_projection+ &
                                              tps_trunc0+tps_copie+tps_detect_nan)/tps_total,'%'
    write(*,*)


end program
