!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organization : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************


!***************************************************************************
!  FFT Module
!***************************************************************************
!****h* ns3d/fft
!  NAME
!    fft
!
!  DESCRIPTION
!    Implement common drivers for different FFT libraries: JMFFT-8.0, FFTW 3.01, ASL, ESSL
!    Every FFT library must be Initialized with init_fft() subroutine before performing
!    forward FFT with fwd_fft() or backward FFT with bck_fft()
!***



#if defined(JMFFT)
! ############################################################
!
!    JMFFT-8.0: a Fortran portable library by Jean-Marie Teuler, CNRS, teuler@lcp.u-psud.fr
!               for 3D/2D FFT real -> complex
! ############################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2

    contains

! We include all the JMFFT library files
#include "JMFFT-8.0/src/jmccfft.f90"
#include "JMFFT-8.0/src/jmccfft2d.f90"
#include "JMFFT-8.0/src/jmccfft3d.f90"
#include "JMFFT-8.0/src/jmccfftm.f90"
#include "JMFFT-8.0/src/jmccm1d.f90"
#include "JMFFT-8.0/src/jmccm1d2.f90"
#include "JMFFT-8.0/src/jmccm1d3.f90"
#include "JMFFT-8.0/src/jmccm1d4.f90"
#include "JMFFT-8.0/src/jmccm1d5.f90"
#include "JMFFT-8.0/src/jmccm1dp.f90"
#include "JMFFT-8.0/src/jmcctranspcs.f90"
#include "JMFFT-8.0/src/jmcfftmlt.f90"
#include "JMFFT-8.0/src/jmcftfax.f90"
#include "JMFFT-8.0/src/jmcsfft.f90"
#include "JMFFT-8.0/src/jmcsfft2d.f90"
#include "JMFFT-8.0/src/jmcsfft3d.f90"
#include "JMFFT-8.0/src/jmcsfftm.f90"
#include "JMFFT-8.0/src/jmcsm1d.f90"
#include "JMFFT-8.0/src/jmcsm1dxy.f90"
#include "JMFFT-8.0/src/jmdecoup.f90"
#include "JMFFT-8.0/src/jmdecoup3.f90"
#include "JMFFT-8.0/src/jmerreur1.f90"
#include "JMFFT-8.0/src/jmerreur2.f90"
#include "JMFFT-8.0/src/jmerreur3.f90"
#include "JMFFT-8.0/src/jmerreur4.f90"
#include "JMFFT-8.0/src/jmfact.f90"
#include "JMFFT-8.0/src/jmfftfax.f90"
#include "JMFFT-8.0/src/jmgetcode.f90"
#include "JMFFT-8.0/src/jmgeterreur.f90"
#include "JMFFT-8.0/src/jmgeterror.f90"
#include "JMFFT-8.0/src/jmgetmessage.f90"
#include "JMFFT-8.0/src/jmgetnwork.f90"
#include "JMFFT-8.0/src/jmgetsetcode.f90"
#include "JMFFT-8.0/src/jmgetseterreur.f90"
#include "JMFFT-8.0/src/jmgetseterror.f90"
#include "JMFFT-8.0/src/jmgetsetnwork.f90"
#include "JMFFT-8.0/src/jmrfftmlt.f90"
#include "JMFFT-8.0/src/jmscfft.f90"
#include "JMFFT-8.0/src/jmscfft2d.f90"
#include "JMFFT-8.0/src/jmscfft3d.f90"
#include "JMFFT-8.0/src/jmscfftm.f90"
#include "JMFFT-8.0/src/jmscm1d.f90"
#include "JMFFT-8.0/src/jmscm1dxy.f90"
#include "JMFFT-8.0/src/jmsetcode.f90"
#include "JMFFT-8.0/src/jmseterreur.f90"
#include "JMFFT-8.0/src/jmseterror.f90"
#include "JMFFT-8.0/src/jmsetnwork.f90"
#include "JMFFT-8.0/src/jmtable.f90"
#include "JMFFT-8.0/src/jmtransp.f90"
#include "JMFFT-8.0/src/jmsinfftmlt.f90"
#include "JMFFT-8.0/src/jmcosfftmlt.f90"
#include "JMFFT-8.0/src/jmsinfftfax.f90"
#include "JMFFT-8.0/src/jmcosfftfax.f90"
#include "JMFFT-8.0/src/jmtablesin.f90"
#include "JMFFT-8.0/src/jmtablesincos.f90"

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        ! n3 \= 1; we use a 3D FFT
        if (n3 /= 1) then
             fft_lib_name = 'JMFFT-8.0 3D (Author: Jean-Marie Teuler, CNRS)'

            scaling = 1./(n1*n2*n3)

            ! the library JMFFT requires two working arrays while only one is given in argument of the subroutine init_fft.
            ! We create artificially two arrays into our single array "work" by determining lower and upper bounds of inner arrays.
            lbnd1 = 1
            ubnd1= 100 + 2*(n1+n2+n3)
            lbnd2 = ubnd1+1
            ubnd2 = ubnd1 + 512*max(n1,n2,n3)

            call SCFFT3D(0,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)

        ! n3=1; we use a 2D FFT
        else
            fft_lib_name = 'JMFFT-8.0 2D (Author: Jean-Marie Teuler, CNRS)'

            scaling = 1./(n1*n2)

            lbnd1 = 1
            ubnd1= 100 + 2*(n1+n2)
            lbnd2 = ubnd1+1
            ubnd2 = ubnd1 + 512*max(n1,n2)

            call SCFFT2D(0,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if


    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call SCFFT3D(-1,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        else
            call SCFFT2D(-1,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call CSFFT3D(1,n1,n2,n3,1.,b,d1/2,d2,b,d1,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        else
            call CSFFT2D(1,n1,n2,1.,b,d1/2,b,d1,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if

    end subroutine

end module



#elif defined(FFTW) || defined(FFTW_OPENMP)
! #######################################################
!
!     Library FFTW - mono-threaded version for 3D (and 2D) FFT real -> complex
!
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    ! plans and parameters used by FFTW
    integer(kind=8), save :: plan, plan_i
    integer, parameter :: FFTW_FORWARD=-1
    integer, parameter :: FFTW_BACKWARD=+1
    integer, parameter :: FFTW_MEASURE=0
    integer, parameter :: FFTW_EXHAUSTIVE=8
    integer, parameter :: FFTW_PATIENT=32
    integer, parameter :: FFTW_ESTIMATE=64

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    contains

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !    n1,n2,n3: logical dimensions of the FFT along the first, second and third directions.
    !              The FFT real -> complex which is the 1st of the 3 required 1D FFT is performed along the 1st direction.
    !              Because of conjugate relations, only half of the modes along this 1st direction are stored in spectral space.
    !    d1,d2,d3: physical dimensions of the array. d1 >= 2*(n1/2+1), d2 >= n2, d3 >= n3
    !    work: this argument is optional and not used here. It is kept only for compatibilty with other FFT interfaces.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(in), optional :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3
        integer, dimension(0:2) :: ta, nembed, nembed_out
        logical :: trash

#ifdef FFTW_OPENMP
        integer return_thread

        call dfftw_init_threads(return_thread)
        call dfftw_plan_with_nthreads(FFTW_OPENMP)

        write(fft_lib_name,'(A38,I3)') 'FFTW 3.x OpenMP | Nb threads: ', FFTW_OPENMP
#else
        fft_lib_name = 'FFTW 3.x'
#endif

        ! Only to avoid the warning "The variable 'work' was not used" that is issued by some compilers.
        trash = present(work)

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3
        scaling = 1./(n1*n2*n3)

        ta(0) = n1
        ta(1) = n2
        ta(2) = n3
        nembed(0) = d1
        nembed(1) = d2
        nembed(2) = d3
        nembed_out(0) = d1/2
        nembed_out(1) = d2
        nembed_out(2) = d3

        call dfftw_plan_many_dft_r2c(plan,3,ta,1,b,nembed,1,1,b,nembed_out,1,1,FFTW_PATIENT)
        call dfftw_plan_many_dft_c2r(plan_i,3,ta,1,b,nembed_out,1,1,b,nembed,1,1,FFTW_PATIENT)

    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b(x1,x2,x3): field in physical space
    !    work: this argument is optional and not used here. It is kept only for compatibilty with other FFT interfaces.
    !
    !  OUTPUT
    !    b(2,xk1,xk2,xk3): field in spectral space. Real and imaginary parts are stores contiguously.
    !***

        implicit none
        double precision, dimension(0:d1-1,0:d2-1,0:d3-1), intent(inout) :: b
        double precision, dimension(*), intent(in), optional :: work
        logical :: trash

        ! Only to avoid the warning "The variable 'work' was not used" that is issued by come compilers.
        trash = present(work)

        call dfftw_execute_dft_r2c(plan, b, b)

        ! we normalize the Fourier coefficients
        b = scaling*b

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b(2,xk1,xk2,xk3): field in spectral space. Real and imaginary parts are stored contiguously.
    !    work: this argument is optional and not used here. It is kept only for compatibilty with other FFT interfaces.
    !
    !  OUTPUT
    !    b(x1,x2,x3): field in physical space
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(in), optional :: work
        logical :: trash

        ! Only to avoid the warning "The variable 'work' was not used" that is issued by some compilers.
        trash = present(work)

        call dfftw_execute_dft_c2r(plan_i,b,b)

    end subroutine

end module



#elif defined(MATHKEISAN)
! ############################################################
!
!    MathKeisan: fft library for NEC-SX8 for 3D/2D FFT real -> complex
!
! ############################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2, lbnd3, ubnd3

    contains

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        ! n3 \= 1; we use a 3D FFT
        if (n3 /= 1) then
            fft_lib_name = 'MathKeisan 3D'

            scaling = 1./(n1*n2*n3)

            ! the library MathKeisan requires two working arrays while only one is given in argument of the subroutine init_fft.
            ! We create artificially two arrays into our single array "work" by determining lower and upper bounds of inner arrays.
            lbnd1 = 1
            ubnd1 = 2*(n1+n2+n3)+195
            lbnd2 = ubnd1+1
            ubnd2 = ubnd1 + 4*256*max(n1,n2,n3)

            call DZFFT3D(0,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)

        ! n3=1; we use a 2D FFT
        else
            fft_lib_name = 'MathKeisan 2D'

            scaling = 1./(n1*n2)

            lbnd1 = 1
            ubnd1 = 2*(n1+n2)+130
            lbnd2 = ubnd1+1
            ubnd2 = ubnd1 + 6*n1*n2

            call DZFFT2D(0,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if


    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !
    !  REM
    !    The sign in the exponential term for the forward and backward FFT
    !    is the inverse than usual
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call DZFFT3D(-1,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        else
            call DZFFT2D(-1,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !
    !  REM
    !    The sign in the exponential term for the forward and backward FFT
    !    is the inverse than usual
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call ZDFFT3D(1,n1,n2,n3,1.,b,d1/2,d2,b,d1,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        else
            call ZDFFT2D(1,n1,n2,1.,b,d1/2,b,d1,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if

    end subroutine

end module



#elif defined(ESSL)
! #######################################################
!
!       Library IBM ESSL - FFT real -> complex
!
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    ! useful for the 2D FFT to build working arrays
    integer, save :: naux1
    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2

    contains


! *************************************************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *************************************************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(out), optional :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3
        double precision :: aux

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        if (n3 /= 1) then
            fft_lib_name = 'ESSL 3D'

            scaling = 1./(n1*n2*n3)
        else
            fft_lib_name = 'ESSL 2D'

            scaling = 1./(n1*n2)

            ! we determine the required size for the working array
            ! The ESSL documentation indicates that for a 2D real to complex FFT, the following storage is required:
            ! If n <= 2048, use naux1 = 42000.
            ! If n > 2048, use naux1 = 40000+1.64n1+2.28n2,
            ! where n = max(n1/2, n2).
            if (max(n1/2,n2) <= 2048) then
                naux1 = 42000
            else
                naux1 = 40000+1.64*n1+2.28*n2
            end if

            lbnd1 = 1
            ubnd1 = naux1
            lbnd2 = naux1+1
            ubnd2 = 2*naux1

            call DRCFT2(1,b,d1,b,d1/2,n1,n2,1,scaling,work(lbnd1:ubnd1),naux1,aux,0)
            call DCRFT2(1,b,d1/2,b,d1,n1,n2,-1,1.0,work(lbnd2:ubnd2),naux1,aux,0)
        endif

    end subroutine


! *************************************************************************
    subroutine fwd_fft(b,work)
! *************************************************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        double precision :: aux

        if (n3 /= 1) then
            call DRCFT3(b,d1,d1*d2,b,d1/2,d1/2*d2,n1,n2,n3,1,scaling,aux,0)
        else
            call DRCFT2(0,b,d1,b,d1/2,n1,n2,1,scaling,work(lbnd1:ubnd1),naux1,aux,0)
        end if

    end subroutine


! *************************************************************************
    subroutine bck_fft(b, work)
! *************************************************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !***

        implicit none
        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        double precision :: aux

        if (n3 /= 1) then
            call DCRFT3(b,d1/2,d1/2*d2,b,d1,d1*d2,n1,n2,n3,-1,1.0,aux,0)
        else
            call DCRFT2(0,b,d1/2,b,d1,n1,n2,-1,1.0,work(lbnd2:ubnd2),naux1,aux,0)
        end if

    end subroutine

end module


#endif
