
        subroutine fill_wave_vectors

!===========================================================================
!   
!  fill_wave_vectors - Routine to compute discrete wavenumber vectors
!                      appropriate for mixed Fourier(x,y) and sin/cos(z)
!                      or triply Fourier transforms, depending on bc_flag. 
!                      An integrating factor which is useful
!                      for the explicit time stepping is also initialized.
!
!===========================================================================

      use mpi_params
      use dimensional_params
      use dimensionless_params
      use flags_counters_etc
      use independent_variables
      
      implicit none
      real(kind=8)            :: f1,f2,f3,maxk,maxl,maxm,r2
      real(kind=8)            :: r2_scaled,taper,dk
      integer                 :: ii
    
!
!  Compute the local portion of the k wavevector (2*pi*k/Lx) for k's(myid))
!  Negative k wavenumbers not needed due to conjugate symmetry
!
      dk = 2.0*pi/(Lx/Lz)      ! all lengths nondimensionalized by Lz
      ii = myid*locnx
      do i = 1,locnx
       wnx(i) = ii*dk
       ii=ii+1
      enddo

!
!   Compute the l wavevector (2*pi*l/ly for whole table)
!
      if( ny > 0 ) then
       do j = 2, ny/2
        wny(j) = 2.0*pi*(j-1)/(Ly/Lz)
        wny(ny-j+2) = -wny(j)
       enddo
       wny(1) = 0.0
       wny(ny/2+1) = 2.0*pi*(ny/2)/(Ly/Lz)
      elseif( ny==0 ) then
       wny(1) = 0.0   ! the only y wavenumber needed
      endif

!
!  Compute the m (vertical, z) wavevector
!

      if( trim(bc_flag) == 'zslip' ) then   ! sin and cos expansions in z
       do k = 1,n3
        wnz(k) = pi*(k-1.d0)/(Lz/Lz)
       enddo
       maxm=wnz(n3)
      elseif( trim(bc_flag) == 'zperiodic') then  ! Fourier expansions in z
       do k = 2, nz/2
        wnz(k) = 2.d0*pi*(k-1.d0)/(Lz/Lz)
        wnz(nz-k+2) = -wnz(k)
       enddo
       wnz(1) = 0.d0
       wnz(nz/2+1) = 2.d0*pi*(nz/2.d0)/(Lz/Lz)
       maxm=wnz(nz/2+1)
      endif

!
!   Compute the integrating factor for modes within truncation sphere
!   Store only the local portion of the arrays.
!
       f1 = 1.d0/Re            ! factor for velocity
       f2 = 1.d0/(Re*Pr_1)     ! factor for scalar 1
       f3 = 1.d0/(Re*Pr_2)     ! factor for scalar 2

!      Cutoff values of wavenumbers (max values representable on grid)
       maxk=(nx/2.d0)*2.d0*pi/(Lx/Lz)
       maxl=(ny/2.d0)*2.d0*pi/(Ly/Lz)
       

!      for 2d problems, set lmax very large to avoid division by zero: wny/lmax
       if(ny .eq. 1) maxl=1.e32

!      Laplacian dissipation/diffusion, i.e. DNS
!      Maximum radius in wavenumber space mapped to unit sphere
!      This was, but is no longer, a user-specified parameter
       kmax =  (8.d0/9.d0)
    
	   do k = 1,n3
	    do j = 1,ny    
	     do i = 1,locnx

!          squared magnitude of wavenumber vector
            r2 = (wnx(i))**2 + (wny(j))**2 + (wnz(k))**2
            r2_scaled = (wnx(i)/maxk)**2 + (wny(j)/maxl)**2 + (wnz(k)/maxm)**2
            
	       if( r2_scaled .le. kmax**2 ) then
             amp(i,j,k,1) = exp(-dt*f1*r2) 
             amp(i,j,k,2) = exp(-dt*f2*r2)
             amp(i,j,k,3) = exp(-dt*f3*r2)
             
	       elseif ( r2_scaled .gt. kmax**2 .and. r2_scaled .le. 1 ) then 
             taper = cos( (pi/2.d0)*(sqrt(r2_scaled)-kmax)/(1.d0-kmax) )**4 
             ! 1 at r2_scaled=8/9 ---> 0 at r2_scaled=8/9=1.0
             amp(i,j,k,1) = exp(-dt*f1*r2)*taper 
             amp(i,j,k,2) = exp(-dt*f2*r2)*taper
             amp(i,j,k,3) = exp(-dt*f3*r2)*taper
             
	       elseif( r2_scaled .gt. 1.0 ) then
             amp(i,j,k,1) = 0.d0
             amp(i,j,k,2) = 0.d0
             amp(i,j,k,3) = 0.d0
	       endif

          enddo
         enddo
        enddo


	   return
	   end subroutine fill_wave_vectors


  subroutine write_results
  use flags_counters_etc
  use mpi_params
  use io_params
  use particles
  
   implicit none
   integer       :: ifile
  
   if( ntrajectories > 0 ) then
    call write_particle_data
   endif
   
   do ifile=1,num_netcdf_files
   
    if( mod( (istep-istart), nsteps(ifile) ) == 0    &
        .or. istep == istart ) then

! xlv     
      if ( ifile == 1 ) write(0,*) 'Call write_results, istep=',istep 

      if( trim(mode(ifile)) == 'new' .or. istep == istart) then      
        call init_netcdf(ifile) 
        call write_netcdf(ifile)         
      elseif( trim(mode(ifile)) == 'append') then         
        call write_netcdf(ifile)
      endif
               
    endif
    
   enddo
       
   call mpi_barrier(comm,ierr)
  return
 end subroutine write_results
  
  
 subroutine init_netcdf(fid)
 use flags_counters_etc
 use state_variables
 use intermediate_variables
 use independent_variables
 use dimensional_params
 use mpi_params
 use io_params
      
 implicit none
 include 'netcdf.inc'
 include 'mpif.h'

 character(len=80)         :: filename
 integer                   :: nbelow,ndims,All_nD_VarSIZE(4)
 integer                   :: start1D(1),count1D(1),iwrk
 integer                   :: k0,kend,kinc,kcount,counter
 real(kind=8),allocatable  :: scratch(:)
     
 integer      :: fid,iid,jid,kid,timeid     
 integer      :: xVarID,yVarID,zVarID,tVarID,s1VarID,s2VarID   
 integer      :: vort_xVarID,vort_yVarID,vort_zVarID
 integer      :: rhoVarID,uVarID,vVarID,wVarID,pVarID

 character(len=80)   :: s1_name, s2_name, s1_units, s2_units
 integer             :: s1_name_len, s2_name_len, s1_units_len, s2_units_len
 
 integer             :: count_vals
 external  count_vals
!!============================================================================ 
  
  
  do_x(fid)='no'  ; do_y(fid)='no'  ; do_z(fid)='no'  ;
  nbelow=numzplanes_below(myid)
            
  idim=1  !! count up for spatially varying dimensions
  jdim=4  !! count down for dimensions w/ indices held constant
  count(:,fid)=1
  nspace(fid)=0    !! count the space dimensions for this file
     
   if( ilocs(2,fid).ne.ilocs(1,fid) ) then
    do_x(fid)='yes'
    nspace(fid)=nspace(fid)+1
    dimid_x(fid) = idim
    idim = idim + 1
   else
    dimid_x(fid) = jdim
    jdim = jdim - 1
   endif
   
   if( jlocs(2,fid).ne.jlocs(1,fid) ) then
    do_y(fid)='yes'
    nspace(fid)=nspace(fid)+1
    dimid_y(fid) = idim
    idim = idim + 1
   else
    dimid_y(fid) = jdim
    jdim = jdim - 1
   endif
   
   if( klocs(2,fid).ne.klocs(1,fid) ) then
    do_z(fid)='yes'
    nspace(fid)=nspace(fid)+1
    dimid_z(fid) = idim
    idim = idim + 1
   else
    dimid_z(fid) = jdim
    jdim = jdim - 1
   endif
   dimid_t(fid) = idim
   ndims=idim   !! time + counted number of space dims
   
   !! given # of space dimensions, decide which directory to write to
   if( ndims-1==0 ) then
    dimstr='TS/'
   elseif( ndims-1==1 ) then
    dimstr='1D/'
   elseif( ndims-1==2 ) then
    dimstr='2D/'
   elseif( ndims-1==3 ) then
    dimstr='3D/'
   endif
   
   !!Create the filename
    write(unit = cnum, fmt = 100) istep
100 format(I6.6)
    write(unit = cid, fmt = 101) myid
101 format(I3.3)

  if( trim(mode(fid)) == 'new' ) then
    if(numprocs .gt. 1) then
      filename = 'output/'//dimstr//trim(filename_root(fid))//'_'//cnum//'_'//cid//'.nc'
    else
      filename = 'output/'//dimstr//trim(filename_root(fid))//'_'//cnum//'.nc'
    endif
  elseif( trim(mode(fid)) == 'append' ) then
    if(numprocs .gt. 1) then
      filename = 'output/'//dimstr//trim(filename_root(fid))//'_'//cid//'.nc'
    else
      filename = 'output/'//dimstr//trim(filename_root(fid))//'.nc'
    endif
  endif
  !!save the current full filename
  fullname(fid)=filename
  
  
   !! Store start, end and inc for myid, 
   !! for this file, and for each direction
   start(dimid_x(fid),fid)=ilocs(1,fid)
   count(dimid_x(fid),fid)=count_vals(ilocs(1,fid),ilocs(2,fid),ilocs(3,fid))
   
   start(dimid_y(fid),fid)=jlocs(1,fid)                     
   count(dimid_y(fid),fid)=count_vals(jlocs(1,fid),jlocs(2,fid),jlocs(3,fid))

   !! Calculate k start&ends for
   !! each processor using Local indexing
   
   !! requested data COMPLETELY EXCLUDES myid
   if( klocs(2,fid) <= nbelow ) then  !! request ends below my bottom slice
    my_kstart(fid)=-999  !! use as indicator that myid just waits
    my_kend(fid)=-999
    my_kinc(fid)=-999
    goto 999
   endif
   if( klocs(1,fid) > numzplanes(myid)+nbelow ) then  !! request starts above my top slice
    my_kstart(fid)=-999  !! use as indicator that myid just waits
    my_kend(fid)=-999
    my_kinc(fid)=-999
    goto 999
   endif

   !! if I get this far, I know myid has SOME data, 
   !! several possibilities for this, deal with them individually
   
   !! requested data SPANS all of myid
   if( klocs(1,fid) <= nbelow .and. klocs(2,fid) > nbelow+numzplanes(myid) ) then
    k0=1
    kend=numzplanes(myid)   !! local indexing for k0/k1
    kinc=klocs(3,fid)
    kcount=count_vals(k0,kend,kinc)
    start(dimid_z(fid),fid)=k0  !! local indexing
    count(dimid_z(fid),fid)=kcount
    
   !! requested data STARTS on myid and ends above (already know myid not excluded)
   elseif( klocs(1,fid) > nbelow .and. klocs(2,fid) > nbelow+numzplanes(myid) ) then
    k0=klocs(1,fid)-nbelow
    kend=numzplanes(myid)   !! local indexing for k0/k1
    kinc=klocs(3,fid)
    kcount=count_vals(k0,kend,kinc)
    start(dimid_z(fid),fid)=k0  !! local indexing
    count(dimid_z(fid),fid)=kcount

   !! requested data ENDS on myid and STARTS somewhere below
   elseif( klocs(1,fid) < nbelow .and. klocs(2,fid) <= nbelow+numzplanes(myid) ) then
    k0=1
    kend=numzplanes(myid)   !! local indexing for k0/k1
    kinc=klocs(3,fid)
    kcount=count_vals(k0,kend,kinc)
    start(dimid_z(fid),fid)=k0  !! local indexing
    count(dimid_z(fid),fid)=kcount

   !! requested data STARTS on myid and ENDS on myid
   elseif( klocs(1,fid) > nbelow .and. klocs(2,fid) <= nbelow+numzplanes(myid) ) then
    k0=(klocs(1,fid)-nbelow)
    kend=(klocs(2,fid)-nbelow)
    kinc=klocs(3,fid)
    kcount=count_vals(k0,kend,kinc)
    start(dimid_z(fid),fid)=k0  !! local indexing
    count(dimid_z(fid),fid)=kcount
   endif
  
  !! keep track of start,end,inc 
  !! index k LOCALLY i.e. in [1,numzplanes(myid)]
   my_kstart(fid)=k0
   my_kend(fid)=kend
   my_kinc(fid)=kinc

   iwrk=maxval( count(:,fid) )
   allocate( scratch(iwrk+1) )
  
!! Open a netcdf file
  rcode=NF_CREATE(filename,NF_NOCLOBBER,ncid)
  if(rcode.ne.NF_NOERR) print *,myid,'ERROR OPENING NETCDF FILE: ',rcode

!! Spatial Dimensions Definitions
  if(do_x(fid)=='yes') then
   rcode=NF_DEF_DIM(ncid,'idimension',count(dimid_x(fid),fid),iid)
   if (rcode.ne.NF_NOERR) print *,myid,  &
           ': NetCDF Error: NF_DEF_DIM: idimension', rcode
    All_nD_VarSIZE(dimid_x(fid))=iid
   endif
  if(do_y(fid)=='yes') then
   rcode=NF_DEF_DIM(ncid,'jdimension',count(dimid_y(fid),fid),jid)
   if (rcode.ne.NF_NOERR) print *,myid,  &
           ': NetCDF Error: NF_DEF_DIM: jdimension', rcode
   All_nD_VarSIZE(dimid_y(fid))=jid
  endif
  if(do_z(fid)=='yes') then  
   rcode=NF_DEF_DIM(ncid,'kdimension',count(dimid_z(fid),fid),kid)
   if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_DIM: kdimension', rcode
   All_nD_VarSIZE(dimid_z(fid))=kid
  endif
   
  rcode=NF_DEF_DIM(ncid,'timedimension',NF_UNLIMITED,timeid)
  if (rcode.ne.NF_NOERR) print *,myid,   &
          ': NetCDF Error: NF_DEF_DIM: timedimension', rcode
  All_nD_VarSIZE(dimid_t(fid))=timeid


!! Define x,y,z grid position  and time variables

!!**X*****
     if( do_x(fid) == 'yes' ) then
      rcode=NF_DEF_VAR(ncid,'xVar',NF_DOUBLE,1,iid,xVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: xVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,xVarID,'long_name',1,'x')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: xVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,xVarID,'units',1,'m')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: xVar', rcode
     endif

!!**Y*****
     if( do_y(fid) == 'yes' ) then
      rcode=NF_DEF_VAR(ncid,'yVar',NF_DOUBLE,1,jid,yVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: yVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,yVarID,'long_name',1,'y')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: yVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,yVarID,'units',1,'m')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: yVar', rcode
      endif
!!**Z*****
     if( do_z(fid) == 'yes' ) then
      rcode=NF_DEF_VAR(ncid,'zVar',NF_DOUBLE,1,kid,zVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: zVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,zVarID,'long_name',1,'z')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: zVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,zVarID,'units',1,'m')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: zVar', rcode
     endif
!!**time*****
      rcode=NF_DEF_VAR(ncid,'tVar',NF_DOUBLE,1,timeid,tVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: tVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,tVarID,'long_name',4,'time')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: tVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,tVarID,'units',7,'seconds')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: tVar', rcode     


!!     Set the appropriate labels for s1 and s2.
      if(trim(scalars) == 'ts') then
        s1_name = 'Temperature'
        s1_name_len = 11
        s1_units = 'deg C'
        s1_units_len = 5

        s2_name = 'Salinity'
        s2_name_len = 8
        s2_units = 'psu'
        s2_units_len = 3
      elseif(trim(scalars) == 'tp') then
        s1_name = 'Temperature'
        s1_name_len = 11
        s1_units = 'deg C'
        s1_units_len = 5

        s2_name = 'Passive Tracer'
        s2_name_len = 14
        s2_units = 'concentration'
        s2_units_len = 13
      elseif(trim(scalars) == 'sp') then
        s1_name = 'Salinity'
        s1_name_len = 8
        s1_units = 'psu'
        s1_units_len = 8

        s2_name = 'Passive Tracer'
        s2_name_len = 14
        s2_units = 'concentration'
        s2_units_len = 13
      elseif(trim(scalars) == 'rp') then
        s1_name = 'Perturbation Density'
        s1_name_len = 20
        s1_units = 'kg/m3'
        s1_units_len = 5

        s2_name = 'Passive Tracer'
        s2_name_len = 14
        s2_units = 'concentration'
        s2_units_len = 13
      elseif(trim(scalars) == 'pp') then
        s1_name = 'Passive Tracer'
        s1_name_len = 14
        s1_units = 'concentration'
        s1_units_len = 13 

        s2_name = 'Passive Tracer'
        s2_name_len = 14
        s2_units = 'concentration'
        s2_units_len = 13
      else
        write(0,*) "scalars value is not set properly: scalars -> ", scalars
        stop
      endif

!!
! scalar (1)
!!   
    if( variable_key(4,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'s1Var',NF_DOUBLE,ndims,All_nD_VarSIZE(1:ndims),s1VarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: s1Var', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,s1VarID,'long_name',s1_name_len,s1_name)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: s1Var', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,s1VarID,'units',s1_units_len,s1_units)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: s1Var', rcode
    endif

!!
!scalar (2)
!!
    if( variable_key(5,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'s2Var',NF_DOUBLE,ndims,All_nD_VarSIZE,s2VarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: s2Var', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,s2VarID,'long_name',s2_name_len,s2_name)                           
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: s2Var', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,s2VarID,'units',s2_units_len,s2_units)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: s2Var', rcode
     endif

!!
!x velocity = uVar
!!
    if( variable_key(1,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'uVar',NF_DOUBLE,ndims,All_nD_VarSIZE,uVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: uVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,uVarID,'long_name',1,'u')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: uVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,uVarID,'units',3,'m/s')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: uVar', rcode
    endif
!!
!y velocity = vVar
!! 
    if( variable_key(2,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'vVar',NF_DOUBLE,ndims,All_nD_VarSIZE,vVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: vVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vVarID,'long_name',1,'v')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vVarID,'units',3,'m/s')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vVar', rcode
    endif
!!
!z velocity = wVar
!!
    if( variable_key(3,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'wVar',NF_DOUBLE,ndims,All_nD_VarSIZE,wVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: wVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,wVarID,'long_name',1,'w')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: wVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,wVarID,'units',3,'m/s')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: wVar', rcode
    endif
    
!!
!pressure = pVar
!! 
    if( variable_key(9,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'pVar',NF_DOUBLE,ndims,All_nD_VarSIZE,pVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: pVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,pVarID,'long_name',8,'pressure')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: pVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,pVarID,'units',6,'kg/ms2')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: pVar', rcode
    endif
    
!!
!x vorticity = vort_xVar
!! 
    if( variable_key(6,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'vort_xVar',NF_DOUBLE,ndims,All_nD_VarSIZE,vort_xVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: vort_xVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vort_xVarID,'long_name',11,'x vorticity')                            
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vort_xVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vort_xVarID,'units',3,'1/s')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vort_xVar', rcode
    endif
    
!!
!y vorticity = vort_yVar
!! 
    if( variable_key(7,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'vort_yVar',NF_DOUBLE,ndims,All_nD_VarSIZE,vort_yVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: vort_yVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vort_yVarID,'long_name',11,'y vorticity')                            
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vort_yVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vort_yVarID,'units',3,'1/s')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vort_yVar', rcode
    endif
    
!!
!z vorticity = vort_zVar
!! 
    if( variable_key(8,fid) == 1 ) then
      rcode=NF_DEF_VAR(ncid,'vort_zVar',NF_DOUBLE,ndims,All_nD_VarSIZE,vort_zVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_DEF_VAR: vort_zVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vort_zVarID,'long_name',11,'z vorticity')                            
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vort_zVar', rcode
      rcode=NF_PUT_ATT_TEXT(ncid,vort_zVarID,'units',3,'1/s')
       if (rcode.ne.NF_NOERR) print *,myid,   &
           ': NetCDF Error: NF_PUT_ATT_TEXT: vort_zVar', rcode
    endif
    
!!
!End define mode
!!
    rcode=NF_ENDDEF(ncid)
    if(rcode.ne.NF_NOERR) print *,myid,'ERROR  LEAVING DEFINE MODE'


    if(do_x(fid)=='yes') then 
!    store x values in temp array
      counter = 1
      do i = ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
       scratch(counter) = x(i) ! [m]
       counter = counter + 1
      enddo

!    Put dimension data in netcdf file
      start1D(1) = 1
      count1D(1) = count_vals(ilocs(1,fid),ilocs(2,fid),ilocs(3,fid))
      rcode=NF_PUT_VARA_DOUBLE(ncid,xVarID,start1D,count1D,scratch)
      if(rcode.ne.NF_NOERR) print *,myid,'NetCDF ERROR xVarID: ',rcode
     endif

    if(do_y(fid)=='yes') then 
!    store y values in temp array
      counter = 1
      do j = jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
       scratch(counter) = y(j) ! [m]
       counter = counter + 1
      enddo

!    Put dimension data in netcdf file
      start1D(1) = 1
      count1D(1) = count_vals( jlocs(1,fid),jlocs(2,fid),jlocs(3,fid) )
      rcode=NF_PUT_VARA_DOUBLE(ncid,yVarID,start1D,count1D,scratch)
      if(rcode.ne.NF_NOERR) print *,myid,'NetCDF ERROR yVarID: ',rcode
     endif
    
    if(do_z(fid)=='yes') then 
!    store z values in temp array
      counter = 1
      do k = k0,kend,kinc
       scratch(counter) = z(k) ! [m]
       counter = counter + 1
      enddo

!    Put dimension data in netcdf file
      start1D(1) = 1
      count1D(1) = count_vals(k0,kend,kinc)  !! had kcount
      rcode=NF_PUT_VARA_DOUBLE(ncid,zVarID,start1D,count1D,scratch)
      if(rcode.ne.NF_NOERR) print *,myid,'NetCDF ERROR zVarID: ',rcode
    endif
        
      rcode=NF_CLOSE(ncid)
      if(rcode.ne.NF_NOERR) print *,myid,'ERROR CLOSING NETCDF FILE'
      deallocate( scratch )
999   call mpi_barrier(comm,ierr)
      return
      end

 subroutine write_netcdf(fid)
 
!     if pressure_flag = 'do_pressure_only' then the u field actually
!     contains dimensionless pressure, otherwise ignore pressure
 use flags_counters_etc
 use state_variables
 use intermediate_variables
 use independent_variables
 use dimensional_params
 use mpi_params
 use io_params
 use user_module
      
 implicit none
 include 'netcdf.inc'
 include 'mpif.h'

 character(len=80)         :: filename,err_msg
 integer                   :: Nd_start(4)=0,Nd_count(4)=0
 integer                   :: ndims,ii,jj,kk,kend,kinc
 integer                   :: kstart
 real(kind=8),allocatable  :: scratch(:,:,:)
     
 integer      :: fid,iid,jid,kid,timeid   
 integer      :: xVarID,yVarID,zVarID,tVarID,s1VarID,s2VarID   
 integer      :: vort_xVarID,vort_yVarID,vort_zVarID
 integer      :: rhoVarID,uVarID,vVarID,wVarID,pVarID

 if( my_kstart(fid) == -999 ) then  !! no data to write
  goto 999
 endif
 
 !!Create the filename, init overwrote filename_root(fid) w/ full name
 filename = fullname(fid)
 
! Each processor opens netcdf file and gets variable id's.
  if( myid .eq. 0 .and. pressure_flag .ne. 'do_pressure_only' ) then
  !write(0,*)'Adding simulation data to file: ' , filename
  endif

  rcode=NF_OPEN(filename,NF_WRITE,ncid)
   if(rcode.ne.NF_NOERR) print *,myid,   &
      'NetCDF ERROR OPENING FILE in write3D_mpi_netcdf: ',rcode,filename

   if( pressure_flag .ne. 'do_pressure_only' ) then
   
       rcode=NF_INQ_VARID(ncid,'tVar',tVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> tVar: ',rcode

      if( variable_key(4,fid) == 1 ) then
       rcode=NF_INQ_VARID(ncid,'s1Var',s1VarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> s1Var: ',rcode
      endif
      
      if( variable_key(5,fid) == 1 ) then
       rcode=NF_INQ_VARID(ncid,'s2Var',s2VarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> s2Var: ',rcode
      endif
      
      if( variable_key(1,fid) == 1 ) then
       rcode=NF_INQ_VARID(ncid,'uVar',uVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> uVar: ',rcode
      endif
      
      if( variable_key(2,fid) == 1 ) then
       rcode=NF_INQ_VARID(ncid,'vVar',vVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> vVar: ',rcode
      endif
      
      if( variable_key(3,fid) == 1 ) then
       rcode=NF_INQ_VARID(ncid,'wVar',wVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> wVar: ',rcode
      endif
      
      if( variable_key(6,fid) == 1 ) then 
       if( vort_flag .eq. 'yes' ) then       
        rcode=NF_INQ_VARID(ncid,'vort_xVar',vort_xVarID)
        if (rcode.ne.NF_NOERR) print *,myid,   &
          'NetCDF ERROR INQ_VARID -> vort_xVar: ',rcode
      endif
      
      if( variable_key(7,fid) == 1 ) then        
        rcode=NF_INQ_VARID(ncid,'vort_yVar',vort_yVarID)
        if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> vort_yVar: ',rcode
      endif
      
      if( variable_key(8,fid) == 1 ) then        
        rcode=NF_INQ_VARID(ncid,'vort_zVar',vort_zVarID)
        if (rcode.ne.NF_NOERR) print *,myid,   &
          'NetCDF ERROR INQ_VARID -> vort_zVar: ',rcode
       endif
      endif
        
      elseif( pressure_flag .eq. 'do_pressure_only' ) then
      if( variable_key(9,fid) == 1 ) then
       rcode=NF_INQ_VARID(ncid,'pVar',pVarID)
       if (rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR INQ_VARID -> pVar: ',rcode

       endif
      endif 
       
      
!     Write data to netcdf file. *******************************
       Nd_start(dimid_x(fid))=1
       Nd_count(dimid_x(fid))=count(dimid_x(fid),fid)
       
       Nd_start(dimid_y(fid))=1
       Nd_count(dimid_y(fid))=count(dimid_y(fid),fid)
       
       Nd_start(dimid_z(fid))=1
       Nd_count(dimid_z(fid))=count(dimid_z(fid),fid)
       
       Nd_start(dimid_t(fid))=time_counter(fid)
       Nd_count(dimid_t(fid))=1   !! always 1 time slice written per call

!     time in seconds
      rcode=NF_PUT_VARA_DOUBLE(ncid,tVarID,Nd_start(dimid_t(fid)),        &
                                           Nd_count(dimid_t(fid)),t_secs)           
      if(rcode.ne.NF_NOERR) print *,myid,'NetCDF ERROR tVarID: ',rcode
      if( trim(mode(fid)) == 'append' ) time_counter(fid)=time_counter(fid)+1
      
      allocate( scratch( Nd_count(dimid_x(fid)), &
                         Nd_count(dimid_y(fid)), &
                         Nd_count(dimid_z(fid)) ) )
      if( pressure_flag .ne. 'do_pressure_only' ) then
      
      !!write(0,*) '========================='
      !!write(0,*) '========================='
      !!write(0,*) 'dims for scratch: ',Nd_count(dimid_x(fid)),Nd_count(dimid_y(fid)),Nd_count(dimid_z(fid))
      !!write(0,*) 'Nd_start  ',Nd_start
      !!write(0,*) 'Nd_count  ',Nd_count
      !!write(0,*) ' ilocs     ',ilocs(:,fid)
      !!write(0,*) ' jlocs     ',jlocs(:,fid)
      !!write(0,*) ' my_ks     ',my_kstart(fid),my_kend(fid),my_kinc(fid)

!     Scalar 1  in dimensional units.
      if( variable_key(4,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
                  
          call eval_s1_bar(z(k),Lz,s1_bar,s1_bar_z,s1_bar_zz)  
          call nondimensionalize_s1(s1_bar,s1_bar_z,s1_bar_zz)
          scratch(ii,jj,kk) = (s1(i,j,k,N)+s1_bar)*s1_scale
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,s1VarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) then
         print *,myid,'NetCDF ERROR: PUT_VARA_DOUBLE scalar (1)',rcode
         err_msg=NF_STRERROR(rcode)
         print*,myid, err_msg
        endif
      endif
      
!     Scalar 2  in dimensional units.
      if( variable_key(5,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          call eval_s2_bar(z(k),Lz,s2_bar,s2_bar_z,s2_bar_zz)
          call nondimensionalize_s2(s2_bar,s2_bar_z,s2_bar_zz)
          scratch(ii,jj,kk) = (s2(i,j,k,N)+s2_bar)*s2_scale
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,s2VarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) then
        print *,myid,'NetCDF ERROR: PUT_VARA_DOUBLE scalar (2)',rcode
        err_msg=NF_STRERROR(rcode)
        print*,myid, err_msg
       endif
      endif
      
!     u velocity component in dimensional units.
      if( variable_key(1,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = u(i,j,k,N)*U0
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,uVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR: PUT_VARA_DOUBLE u',rcode
      endif
      
!      v velocity component in dimensional units.
      if( variable_key(2,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = v(i,j,k,N)*U0
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,vVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR: PUT_VARA_DOUBLE v',rcode
      endif
      
!      w velocity component in dimensional units.
      if( variable_key(3,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = w(i,j,k,N)*U0
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,wVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR PUT_VARA_DOUBLE: w ',rcode
      endif

!      x vorticity in dimensional units.
      if( variable_key(6,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = zetax(i,j,k)*U0/Lz
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,vort_xVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR PUT_VARA_DOUBLE: vort_x ',rcode
      endif

!      y vorticity in dimensional units.
      if( variable_key(7,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = zetay(i,j,k)*U0/Lz
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,vort_yVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) write(0,*) myid,   &
         'NetCDF ERROR PUT_VARA_DOUBLE: vort_y ',rcode
      endif
      
!      z vorticity in dimensional units.
      if( variable_key(8,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = zetaz(i,j,k)*U0/Lz
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,vort_zVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR PUT_VARA_DOUBLE: vort_z ',rcode
      endif
      
     elseif( pressure_flag .eq. 'do_pressure_only' ) then

!      pressure in dimensional units. ( pressure stored in u array )
      if( variable_key(9,fid)==1 ) then
       ii=1 ;  
       do i=ilocs(1,fid),ilocs(2,fid),ilocs(3,fid)
        jj=1;
        do j=jlocs(1,fid),jlocs(2,fid),jlocs(3,fid)
         kk=1;
         do k=my_kstart(fid),my_kend(fid),my_kinc(fid)
         
          scratch(ii,jj,kk) = 0.0*rho_0*U0**2
          
          kk=kk+1
         enddo
         jj=jj+1
        enddo
        ii=ii+1
       enddo

!      Each processor writes its local data to 
!      the appropriate locations in the netcdf file.
       rcode=NF_PUT_VARA_DOUBLE(ncid,pVarID,Nd_start,Nd_count,scratch)
       if(rcode.ne.NF_NOERR) print *,myid,   &
        'NetCDF ERROR PUT_VARA_DOUBLE: pressure ',rcode
      endif

     endif


!      Each processor closes the netcdf file.
       rcode=NF_CLOSE(ncid)
       if(rcode.ne.NF_NOERR) print *,myid,   &
         'NetCDF ERROR: closing file in write3D_mpi_netcdf',rcode

      deallocate( scratch )
999   call mpi_barrier(comm,ierr)
      !!write(0,*) 'Leaving write netcdf file, data written to:  ',trim(filename)
      return
      end


 subroutine write_particle_data
  use particles
  use independent_variables, only: t_secs
  use dimensional_params,    only: Lz
  use mpi_params,            only: myid,comm
      
  implicit none
  include 'netcdf.inc'
  
  character(len=80)  :: filename
  integer            :: iid  !!dimension id for particles
  integer            :: jid  !!dimension id for time
  integer            :: idimensionSIZE(1) !!arrays for  dimension sizes
  integer            :: jdimensionSIZE(1) 
  integer            :: idimensionID !! for dimension variable "float id"
  integer            :: jdimensionID !! for dimension variable "time"
  integer            :: rcode, ierr
  integer            :: ncid
      
!!Variables: x(particle id, time),y(particle id, time),z(particle id, time)
  integer                  :: All2DVarSIZE(2),start2D(2),count2D(2)
  integer                  :: xVarID,yVarID,zVarID   !! 2D Variable ids
  integer                  :: istart,jstart,icount,jcount,i
  integer,allocatable      :: idata(:)
  real(kind=8),allocatable :: data(:)
  
  call mpi_barrier(comm,ierr)
  !!write(0,*) 'INSIDE WRITE PARTICLES',myid
  
  if(myid>0) goto 999  !! and wait there
  
  allocate( idata(ntrajectories) )
  allocate( data(ntrajectories) )
  filename='output/particle_positions.nc'  
  
  !!write(0,*) 'ENTERING write_particle_data t,flag: ',t_secs,init_particle_file

 if(init_particle_file == 'yes' ) then
 
  !!write(0,*) 'creating particle file ',t_secs,init_particle_file
  rcode = NF_CREATE(filename, NF_NOCLOBBER, ncid) 
  if(rcode.ne.NF_NOERR) stop 'ERROR CREATING NETCDF FILE FOR PARTICLES'

!!Dimensions Definition
  rcode = NF_DEF_DIM(ncid, 'idimension', ntrajectories, iid)
  if(rcode.ne.NF_NOERR) write(0,*) 'init for particles NetCDF ERROR: ',rcode

  rcode = NF_DEF_DIM(ncid, 'jdimension', NF_UNLIMITED, jid)
  if(rcode.ne.NF_NOERR) print *,'init for particles NetCDF ERROR: ',rcode

!!Initialize size arrays for dimensions
  idimensionSIZE(1) = iid
  jdimensionSIZE(1) = jid

!!Define dimension variables
  rcode = NF_DEF_VAR(ncid,'idimension',NF_INT,1,idimensionSIZE,idimensionID)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode
  rcode=NF_PUT_ATT_TEXT(ncid,idimensionID,'long_name',11,'particle id')
  rcode=NF_PUT_ATT_TEXT(ncid,idimensionID,'units',3,'[1]')

  rcode = NF_DEF_VAR(ncid,'jdimension',NF_DOUBLE,1,jdimensionSIZE,jdimensionID)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode
  rcode=NF_PUT_ATT_TEXT(ncid,jdimensionID,'long_name',4,'time')
  rcode=NF_PUT_ATT_TEXT(ncid,jdimensionID,'units',7,'seconds')

!! Define 2D position variables
!!Initialize size arrays for id/time varying variables
  All2DVarSIZE(1) = iid
  All2DVarSIZE(2) = jid


!! x particle position = xVar
  rcode = NF_DEF_VAR(ncid,'xVar',NF_DOUBLE,2,All2DVarSIZE,xVarID)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode
  rcode=NF_PUT_ATT_TEXT(ncid,xVarID,'long_name',19,'x particle position')                      
  rcode=NF_PUT_ATT_TEXT(ncid,xVarID,'units',1,'m')
  
!! y particle position = yVar
  rcode = NF_DEF_VAR(ncid,'yVar',NF_DOUBLE,2,All2DVarSIZE,yVarID)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode
  rcode=NF_PUT_ATT_TEXT(ncid,yVarID,'long_name',19,'y particle position')                      
  rcode=NF_PUT_ATT_TEXT(ncid,yVarID,'units',1,'m')
  
!! z particle position = zVar
  rcode = NF_DEF_VAR(ncid,'zVar',NF_DOUBLE,2,All2DVarSIZE,zVarID)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode
  rcode=NF_PUT_ATT_TEXT(ncid,zVarID,'long_name',19,'z particle position')                      
  rcode=NF_PUT_ATT_TEXT(ncid,zVarID,'units',1,'m')
     
!!End define mode
  rcode=NF_ENDDEF(ncid)
  if(rcode.ne.NF_NOERR) print *,'ERROR  LEAVING DEFINE MODE'


!!Write out dimension variable: particle id
!!store values in temp array 
  do i = 1,ntrajectories
   idata(i) = i   !!--> integer
  enddo

!! Put idimension data, i.e. particle id, into netcdf file
  istart = 1
  icount = ntrajectories
  rcode=NF_PUT_VARA_INT(ncid,idimensionID,istart,icount,idata)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR storing particle ids: ',rcode

!!Close netcdf file
  rcode=NF_CLOSE(ncid)
  if(rcode.ne.NF_NOERR) print *,'ERROR CLOSING particle NETCDF FILE'
  
  init_particle_file='no'

endif  ! end file initialization 


!!=========================================================================
!! Open the previously created netcdf data file
  rcode=NF_OPEN(filename,NF_WRITE,ncid)
  if(rcode.ne.NF_NOERR) print *,'ERROR OPENING NETCDF FILE: ',filename


!!Write out current value of dimension variable time
  jstart = j_particle_time
  jcount = 1
  rcode=NF_INQ_VARID(ncid,'jdimension',jdimensionID)
  rcode=NF_PUT_VARA_DOUBLE(ncid,jdimensionID,jstart,jcount,t_secs)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode

!!Write 2D data to netcdf file

!!Initialize size arrays
  start2D(1) = 1
  start2D(2) = j_particle_time
  count2D(1) = ntrajectories
  count2D(2) = 1

!!Store x float position variable xVar
  do  i=1,ntrajectories
   data(i) = positions(i,1)*Lz
  enddo

!!Write variable to netcdf file
  rcode=NF_INQ_VARID(ncid,'xVar',xVarID)
  rcode=NF_PUT_VARA_DOUBLE(ncid,xVarID,start2D,count2D,data)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode

!!Store y float position variable yVar
  do  i=1,ntrajectories
   data(i) = positions(i,2)*Lz
  enddo

!!Write variable to netcdf file
   rcode=NF_INQ_VARID(ncid,'yVar',yVarID)
   rcode=NF_PUT_VARA_DOUBLE(ncid,yVarID,start2D,count2D,data)
   if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode

!!Store z float position variable zVar
  do  i=1,ntrajectories
   data(i) = positions(i,3)*Lz
  enddo

!!Write variable to netcdf file
  rcode=NF_INQ_VARID(ncid,'zVar',zVarID)
  rcode=NF_PUT_VARA_DOUBLE(ncid,zVarID,start2D,count2D,data)
  if(rcode.ne.NF_NOERR) print *,'NetCDF ERROR: ',rcode

!! Close netcdf file
  rcode=NF_CLOSE(ncid)
  if(rcode.ne.0) print *,'ERROR CLOSING NETCDF FILE'

!!increment integer time index       
  j_particle_time = j_particle_time + 1 
  deallocate(data,idata)

999 call mpi_barrier(comm,ierr)  !! sync up here, only 0 did the writing
 return
end subroutine write_particle_data



subroutine nondimensionalize_s1(s1,s1_z,s1_zz)
 use dimensional_params
 implicit none
 real(kind=8)    :: s1,s1_z,s1_zz       
 s1=s1/s1_scale
 s1_z=s1_z/(s1_scale/Lz)
 s1_zz=s1_zz/(s1_scale/Lz**2)       
end subroutine nondimensionalize_s1
      
subroutine nondimensionalize_s2(s2,s2_z,s2_zz)
 use dimensional_params
 implicit none
 real(kind=8)    :: s2,s2_z,s2_zz       
 s2=s2/s2_scale
 s2_z=s2_z/(s2_scale/Lz)
 s2_zz=s2_zz/(s2_scale/Lz**2)       
end subroutine nondimensionalize_s2
      
subroutine toggle
 use flags_counters_etc
 use independent_variables
 use dimensional_params
 use mpi_params, only: myid
 implicit none
!toggle N,N-1 pointers (storage ids for dependent variables)
 itmp=NM1
 NM1=N
 N=itmp

!toggle M,M-1 & M-2 pointers (storage ids for transfer terms)
 itmp=MM2
 MM2=MM1
 MM1=M
 M=itmp       
 t_secs = t_secs + dt*time_scale
 if(myid==0 .and. mod(istep,10) == 0)  &
  write(0,*) 'step ',istep,'  ====>',100*istep/(iend-istart),'% done.'
end subroutine toggle
      
!=============================================================== 
!===============================================================
 subroutine get_ghosted_velocity
  use state_variables
  use flags_counters_etc
  use particles
  use mpi_params
  
  implicit none
#include "mpif.h"
  
  integer              :: status_array(mpi_status_size,3*maxprocs+1)
  integer              :: reqs(2*maxprocs)
  !!============================================================================
  !! all cases, do interior and nx,ny extensions
  !!============================================================================
   gvelocity(1:nx,1:ny,1:numzplanes(myid),1) = u(1:nx,1:ny,1:numzplanes(myid),N)
   gvelocity(1:nx,1:ny,1:numzplanes(myid),2) = v(1:nx,1:ny,1:numzplanes(myid),N)
   gvelocity(1:nx,1:ny,1:numzplanes(myid),3) = w(1:nx,1:ny,1:numzplanes(myid),N)
   
  !! e.g.  gvelocity(nx+1,1:ny,1:numzplanes(myid),1) <=====    u(1,:,:,N)
   gvelocity(nx+1,1:ny,1:numzplanes(myid),1) = u(1,1:ny,1:numzplanes(myid),N)
   gvelocity(nx+1,1:ny,1:numzplanes(myid),2) = v(1,1:ny,1:numzplanes(myid),N)
   gvelocity(nx+1,1:ny,1:numzplanes(myid),3) = w(1,1:ny,1:numzplanes(myid),N)
   
  !! ny+1 <== 1
   gvelocity(1:nx+1,ny+1,1:numzplanes(myid),1) =   &
                                gvelocity(1:nx+1,1,1:numzplanes(myid),1)
                                
   gvelocity(1:nx+1,ny+1,1:numzplanes(myid),2) =   &
                                gvelocity(1:nx+1,1,1:numzplanes(myid),2)
   
   gvelocity(1:nx+1,ny+1,1:numzplanes(myid),2) =   &
                                gvelocity(1:nx+1,1,1:numzplanes(myid),3)
  !!============================================================================
  !!============================================================================
   
   
  !!============================================================================
  !!Single processor cases======================================================
  !!============================================================================
  if( numprocs==1 .and. bc_flag=='zperiodic') then
   !! n3+1 <== 1
   gvelocity(1:nx+1,1:ny+1,numzplanes(myid)+1,1) =  &
                                                 gvelocity(1:nx+1,1:ny+1,1,1)
                                                 
   gvelocity(1:nx+1,1:ny+1,numzplanes(myid)+1,2) =  &
                                                 gvelocity(1:nx+1,1:ny+1,1,2)
   
   gvelocity(1:nx+1,1:ny+1,numzplanes(myid)+1,3) =  &
                                                 gvelocity(1:nx+1,1:ny+1,1,3)
   
  elseif( numprocs==1 .and. bc_flag=='zslip') then
   !! no z extensions needed
  endif
  !!============================================================================
  !!============================================================================
  
  
  !!============================================================================
  !!MULTI-PROCESSOR CASE: INTERIOR NODES  ======================================
  !!============================================================================
  if( numprocs>1 .and. myid > 0 .and. myid < numprocs-1 ) then  
   
   nreqs=0
   !for both bcs, get data from north neighbor
   call get_uvw_from_north( nreqs,reqs )
   nreqs=nreqs+3
   
   !for both bcs, send data to south
   call send_uvw_south( nreqs,reqs )
   nreqs=nreqs+3
   
   call mpi_waitall(nreqs,reqs,status_array,ierr)
  
  !!============================================================================
  !!MULTI-PROCESSOR CASE:  TOP NODE  ===========================================
  !!============================================================================
  elseif( numprocs>1 .and. myid==numprocs-1 ) then 
   
   nreqs=0
   if(bc_flag=='zperiodic') then  !! get plane 1 from bottom processor.
    call get_uvw_from_north( nreqs,reqs )!!yes, north. The bottom is logically north of top node. 
    nreqs=nreqs+3
   endif 
   
   call send_uvw_south( nreqs,reqs )   !!
   nreqs=nreqs+3
   
   call mpi_waitall(nreqs,reqs,status_array,ierr)
  
  !!============================================================================
  !!MULTI-PROCESSOR CASE:  BOTTOM NODE  ========================================
  !!============================================================================
  elseif( numprocs>1 .and. myid==0 ) then  !! bottom node
   
   nreqs=0
   call get_uvw_from_north( nreqs,reqs )  !!get data from north
   nreqs=nreqs+3
   
   if(bc_flag=='zperiodic') then  !! send plane 1 to top processor.
    call send_uvw_south( nreqs,reqs )   !! yes, south. The top is logically south of node 0.
    nreqs=nreqs+3   
   endif 
   
   call mpi_waitall(nreqs,reqs,status_array,ierr)
      
  endif  
  
 end subroutine get_ghosted_velocity
!=============================================================== 
!===============================================================
 
 subroutine find_my_particles(label,xvals,yvals,zvals,npts)
  use mpi_params,         only: myid
  use particles,          only: owners,positions,ntrajectories
  use flags_counters_etc, only: numprocs
  use dimensional_params, only: Lx,Ly,Lz
  implicit none
  
  integer       :: npts,i,j
  integer       :: label(ntrajectories)
  real(kind=8)  :: xvals(ntrajectories)
  real(kind=8)  :: yvals(ntrajectories)
  real(kind=8)  :: zvals(ntrajectories)
  
  if(numprocs==1) then
   label = (/ (i,i=1,ntrajectories) /)
   xvals = positions(:,1)
   yvals = positions(:,2)
   zvals = positions(:,3)     !! could I do this w/ pointers to avoid copying?????
   npts=ntrajectories
   return
  endif
    
  j=1
  do i=1,ntrajectories
   if( owners(i)==myid) then
    label(j)=i  !! id number of particle
    xvals(j)=positions(i,1)
    yvals(j)=positions(i,2)
    zvals(j)=positions(i,3)
    j=j+1         !! increment count of my particles
   else
    positions(i,:)=0.0  !! zero out data if particle resides elsewhere
   endif
  enddo
  npts=j-1
  
 end subroutine find_my_particles
 
!=============================================================== 
!=============================================================== 
 subroutine spline_interp(xpos,ypos,zpos,npts,index)
  use particles
  use dimensional_params,    only: Lx,Ly,Lz
  use independent_variables, only: dx,dy,dz
  use mpi_params,            only: myid
  use flags_counters_etc
  implicit none
  character(len=80)             :: pflag
  integer                       :: npts
  integer                       :: one=1
  integer                       :: m_order
  integer                       :: info
  integer                       :: i0,i1
  integer                       :: j0,j1
  integer                       :: k0,k1
  integer                       :: index(npts) 
  real(kind=8)                  :: xpos(npts)
  real(kind=8)                  :: ypos(npts)
  real(kind=8)                  :: zpos(npts)
  real(kind=8)                  :: h1,h2,h3
       
 !! index is the map from [1,2..npts]-->[1,2,...ntrajectories]
  if(istep < istart_trajectories) return
    
  pflag='no' !!==> don't use periodicity to do interpolation, use ghost pts instead
  m_order=0  !!==> interpolate function, not 1st or 2nd derivatives 
   
  !!First time through, factor the spline coefficient matrix & save
  if( do_factor=='yes' ) then
   lda=4
   allocate( L(lda),D(lda),U1(lda),U2(lda),ipiv(lda) )
   L = 1.d0/6.d0 
   D = 2.d0/3.d0 
   U1 = 1.d0/6.d0
   call DGTTRF( lda, L, D, U1, U2, ipiv, info )  
   if(info .ne. 0 ) stop 'x DGTTRF problem'
   do_factor='no'
  endif

  h1=dx/Lz
  h2=dy/Lz
  h3=dz/Lz
      
 !! do interpolations only within a 4x4x4 subdomain
 !! of gvelocity that contains the interpolation point
 !! use get_neighborhood ===> compute corner locations
  do i=1,npts
   info=-999
   call get_neighborhood(xpos(i),h1,UBOUND(gvelocity,dim=1),i0,i1,info)
   m1 = i1-i0+1    
   call get_neighborhood(ypos(i),h2,UBOUND(gvelocity,dim=2),j0,j1,info)
   m2 = j1-j0+1  
   call get_neighborhood(zpos(i),h3,UBOUND(gvelocity,dim=3),k0,k1,numzplanes_below(myid))
   m3 = k1-k0+1
      
   u4x4=gvelocity(i0:i1,j0:j1,k0:k1,1)
   v4x4=gvelocity(i0:i1,j0:j1,k0:k1,2)
   w4x4=gvelocity(i0:i1,j0:j1,k0:k1,3)
   !write(0,*) 'particle ',i
   !write(0,*) ' i ',i0,i1,xpos(i)
   !write(0,*) ' j ',j0,j1,ypos(i)
   !write(0,*) ' k ',k0,k1,zpos(i)
   
   !! shift positions so that small box corner is origin
   xpos(i)=xpos(i)-(i0-1)*h1
   ypos(i)=ypos(i)-(j0-1)*h2
   zpos(i)=zpos(i) - numzplanes_below(myid)*h3 - (k0-1)*h3
   !write(0,*) ' myid & positions w/in 4x4x4 cube: ',myid,'    ',xpos(i),ypos(i),zpos(i)
   !write(0,*) '     i ',i0,i1,xpos(i)
   !write(0,*) '     j ',j0,j1,ypos(i)
   !write(0,*) '     k ',k0,k1,zpos(i)
 
 !! get u,v,w velocity uvels at each of npts xyz locations "positions"
 !! need to pass in a set of corner indices w/ m1,m2,m3 about 4
 !! spline( f,g,x,y,z,Lx,Ly,Lz,m,pflag,nx,ny,nz,npts )

   !! the location of the n'th element in myids list of xpos..
   !! is stored in index(n)
   call spline( u4x4,uvels(index(i),M),    &
                xpos(i),ypos(i),zpos(i),h1,h2,h3,  &
                m_order,pflag,m1,m2,m3,one )
 
   call spline( v4x4,vvels(index(i),M),    &
                xpos(i),ypos(i),zpos(i),h1,h2,h3,  &
                m_order,pflag,m1,m2,m3,one )
 
   call spline( w4x4,wvels(index(i),M),    &
                xpos(i),ypos(i),zpos(i),h1,h2,h3,  &
                m_order,pflag,m1,m2,m3,one )
                                
   !!shift back to computational coordinates            
   xpos(i)=xpos(i)+(i0-1)*h1
   ypos(i)=ypos(i)+(j0-1)*h2
   zpos(i)=zpos(i) + numzplanes_below(myid)*h3 + (k0-1)*h3

  enddo
                  
 end subroutine spline_interp
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
subroutine get_neighborhood(x,dx,nx,i0,i1,ishift)
 implicit none
 real(kind=8)  :: x,dx
 integer       :: nx,i0,i1,ishift
 
 !!find index of left neighbor
 i0 = floor(x/dx) + 1
 i1 = i0+1             !! always ok
 
 if( ishift > 0 ) then
  i0=i0-ishift
  i1=i1-ishift
 endif
 
 !! so...
 !!   o x  o
 !!  i0    i1
 
 if( i0>1 ) then
  i0=i0-1          !!  o   o x  o
 else
  i1=i0+3          !!  o x  o   o   o      ==> might change
  goto 999         ! 1=i0           i1=4
 endif
 
 !!!! so...
 !!  o   o x  o
 !! i0       i1
 if( i1 < nx ) then
  i1=i1+1              !!  o   o x  o   o
 else                  !! i0           i1
 
  i0=i0-1              !!  o   o   o x  o
 endif                 !! i0           i1=nx   ==> might change
 
999 return
 
end subroutine get_neighborhood




!=============================================================== 
!===============================================================
subroutine SplineBasis(x,dx,n,NN,m,bn,npts,pflag)
!! This function determines the nth cubic spline basis function bn(x).
!! Fortran90 version of Satoshi Mitarai's .m files.
!! L:      length of domain e.g. such that dx=L/NN (Satoshi's N --> NN)
!! NN:     number of points in domain
!! m:      gives m-th derivative of basis function. m can be 0, 1 or 2,
!! pflag:  optional, if present, periodicity is imposed
 implicit none
 integer                              ::  n,npts,i
 integer                              ::  NN
 integer                              ::  m
 real(kind=8),intent(in)              ::  x(npts)
 real(kind=8)                         ::  bn(npts)
 real(kind=8),allocatable             ::  xi(:)
 real(kind=8)                         ::  L,dx
 character(len=80)                    ::  pflag  
 real(kind=8),external                ::  bn_function
 
 allocate( xi(npts) )
 bn(:)=0.d0
  
!! Determine the n-th basis function. The functional form is
!! written in terms of normalized coordinate xi. 
  
 xi = ( x - (n-1)*dx )/dx 
 do i=1,npts
  if( abs(xi(i)) <= 2 ) bn(i) = bn_function(xi(i),m)
 enddo
 
  !! chain rule: d/dx = d/dxi * dxi/dx.  
  bn = bn / (dx**m )
 deallocate(xi)
end subroutine SplineBasis
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
function bn_function(xi,m) result(bn)
implicit none
integer        :: m
real(kind=8)   :: xi
real(kind=8)   :: bn
real(kind=8)   :: onesixth=1.d0/6.d0
real(kind=8)   :: onehalf=1.d0/2.d0
real(kind=8)   :: tmp


select case(m)
 case(0)
 
  xi = abs(xi)
  tmp = (1.d0-xi)
  if( 1 <= xi .and. xi <= 2 ) bn = onesixth*(2.d0-xi)**3
  
  if( 0 <= xi .and. xi <= 1 ) bn = onesixth*(1.d0 + 3.*tmp       &
                                                  + 3.*tmp**2    &
                                                  - 3.*tmp**3 )     
 case(1)
 
  if( -2 <= xi .and. xi <= -1 ) bn = (1.d0/2.d0)*(2.d0+xi)**2
  
  if( -1 <= xi .and. xi <= 0 )  bn = (1.d0/6.d0)*(3.d0 + 6.*(1+xi)       &
                                                       - 9.*(1+xi)**2 )  
                                                          
  if( 0 <= xi .and. xi <= 1 )  bn = (1.d0/6.d0)*(-3.d0 - 6.*(1-xi)       &
                                                       + 9.*(1-xi)**2 ) 
                                                          
  if( 1 <= xi .and. xi <= 2 ) bn = -(1.d0/2.d0)*(2.d0-xi)**2                                                        
                                                           
 case(2)
 
  xi = abs(xi)
  if( 1 <= xi .and. xi <= 2 ) bn = (2.d0-xi)
  if( 0 <= xi .and. xi <= 1 ) bn = (-2.d0+3.d0*xi)
  
 end select
 
end function bn_function
!=============================================================== 
!===============================================================


!=============================================================== 
!===============================================================
subroutine SplineCoeff(f,pflag,n)
!! Overwrite input rhs f w/ soln
!! This function determines the
!! The amplitude e1,e2,. . .,eN
!!
!! Tij ej = bi (i,j = 1,2,. .,
!!
!! Generally, T is a tridiagonal matrix, but in 
!! the periodic case T is a circular matrix
!!
!!
 use particles   !! for prefactored matrices and sizes
 implicit none
 integer                                :: nmax
 integer                                :: n, info, ldb, i
 integer                                :: nrhs
 real(kind=8)                           :: f(n)  !!,e(n)
 real(kind=8),allocatable,dimension(:)  :: y,uu
 real(kind=8)                           :: gamma
 character(len=80)                      :: pflag
 character(len=1)                       :: trans='N'
 
 allocate( y(n),uu(n) )
 
 ldb = n
 nrhs=1
   
 call DGTTRS( trans, n, nrhs, L, D, U1, U2, ipiv, f, ldb, info )
 if(info .ne. 0 ) stop ' DGTTRS problem'
 !!e=f !! rhs overwritten with soln
  
 deallocate( y,uu )
end subroutine SplineCoeff
!=============================================================== 
!===============================================================


!=============================================================== 
!=============================================================== 
subroutine spline( f,g,x,y,z,dx,dy,dz,m,pflag,nx,ny,nz,npts )
!! This function interpolates three-dimensional discrete 
!! data at an arbitary point (x,y,z) given discrete
!! function values f(:,:,:)
!!
!! L: domain size L x L x L.
!! N: number of grid points in each direction. 
!! m: m-th derivative of f. m can be 0, 1 or 2.
!!
!! Required subroutines
!! SplineCoeff: obtains e ijk.
!! spline basis.m: obtains 1-D cubic spline basis function.
 implicit none
 character(len=80)                    :: pflag
 integer                              :: nx,ny,nz,npts
 integer                              :: i,j,k,m
 real(kind=8)                         :: dx,dy,dz
 real(kind=8),dimension(nx,ny,nz)     :: f
 real(kind=8),dimension(npts)         :: x,y,z,g
 real(kind=8),allocatable             :: bi(:),bj(:),bk(:),e(:,:,:) 
 real(kind=8),allocatable             :: tmpx(:),tmpy(:),tmpz(:)
 external                             :: SplineCoeff
 

 allocate( e(nx,ny,nz),tmpx(nx),tmpy(ny),tmpz(nz),bi(npts),bj(npts),bk(npts) )
 !! Compute coefficients of three-dimensional tensor basis functions.
 !! Requires no info about Lx dx etc
 e = 0.d0
 do j = 1,ny
  do k = 1,nz
    tmpx=f(:,j,k)
    call SplineCoeff( tmpx,pflag,nx ) !! overwrites input w/ soln
    e(:,j,k) = tmpx
  enddo
 enddo

 do i = 1,nx
  do k = 1,nz
    tmpy = e(i,:,k)
    call SplineCoeff( tmpy,pflag,ny )  !! overwrites input w/ soln
    e(i,:,k) = tmpy(:)
  enddo
 enddo
 
 do i = 1,nx
  do j = 1,ny
    tmpz = e(i,j,:)
    call SplineCoeff( tmpz,pflag,nz )  !! overwrites input w/ soln
    e(i,j,:) = tmpz
  enddo
 enddo

 !! Summation of three-dimensional tensor basis functions.
 !! get spline coeffs for the function, i.e. m=0
 m = 0 
  g = 0.d0
   do i = 1,nx
   call SplineBasis(x,dx,i,nx,m,bi,npts,pflag)
    do j = 1,ny
     call SplineBasis(y,dy,j,ny,m,bj,npts,pflag)
      do k = 1,nz
       call SplineBasis(z,dz,k,nz,m,bk,npts,pflag)
        g = g + bi*bj*bk*e(i,j,k) 
      enddo
    enddo  
  enddo
 deallocate(tmpx,tmpy,tmpz,e)
 deallocate(bi,bj,bk)
end subroutine spline
!=============================================================== 
!===============================================================

 integer function count_vals(i0,i1,inc) result(numvals)
  integer, intent(in)  :: i0,i1,inc
  integer              :: n
  numvals = floor( float(i1-i0)/inc ) + 1
 end function count_vals

 subroutine tic( start,rate )
  implicit none
  integer   :: start  ! current value of system clock
  integer   :: rate   ! system clock counts/sec
  call system_clock(start,rate)
 end subroutine tic
 
 subroutine toc( start,rate,sec )
  implicit none
  integer           :: start  ! original value of system clock
  integer           :: finish ! current value of system clock
  integer           :: rate   ! system clock counts/sec
  real(kind=8)      :: sec
  call system_clock(finish)
  sec=0.
  sec=float(finish-start)/float(rate)
 end subroutine toc

