      

  subroutine initialize_mpi(comm,myid,numprocs)
   implicit none
   integer   :: comm,myid,numprocs,ierr
#include "mpif.h"

    comm=MPI_COMM_WORLD
    call mpi_init(ierr)
    call mpi_comm_rank(comm,myid,ierr)
    call mpi_comm_size(comm,numprocs,ierr)
    if(ierr.ne.0) stop 'error initializing MPI'
    if(myid==0) then
     write(0,*) 'MPI initialized. Number of processors = ',numprocs
    endif
    call MPI_BARRIER(comm,ierr)
  end subroutine initialize_mpi

  subroutine quit_mpi
   use mpi_params
   implicit none
#include "mpif.h"
   call mpi_finalize(ierr)
   stop
  end subroutine quit_mpi


   subroutine create_MPI_data_type
!  Routine to create a specialized data type for MPI data swaps
!  between processors. See real and complex data maps for details.
   use mpi_params
   use flags_counters_etc
   
   implicit none
#include "mpif.h"
   integer      :: sizeofword

   if(myid==0) write(0,*) 'INITIALIZING DERIVED MPI DATA TYPES'
   call MPI_TYPE_EXTENT(MPI_DOUBLE_PRECISION,sizeofword,ierr)


!  data type for 1d section of basic data slice: 2*locnx
   call MPI_TYPE_VECTOR(2*locnx,1,1,MPI_DOUBLE_PRECISION,oneslice,ierr)

!  data type for 2d section of basic data slice: ny 1d sections
   call MPI_TYPE_HVECTOR(ny,1,(nx+2)*sizeofword,oneslice,twoslice,ierr)
   if( myid .eq. 0 ) then
    if( ierr .ne. 0 ) write(6,*) ' TYPE_HVECTOR twoslice failure'
   endif
   call MPI_TYPE_COMMIT(twoslice,ierr)

!  data type for 3d section of basic data slice: locnz 2d sections
   call MPI_TYPE_HVECTOR(numzplanes(myid),1,(nx+2)*(nyplanes)*sizeofword,  &
                         twoslice,subslice,ierr)    
                         
   if( myid .eq. 0 ) then
    if( ierr .ne. 0 ) write(6,*) ' TYPE_HVECTOR subslice failure'
   endif
   call MPI_TYPE_COMMIT(subslice,ierr)

   return
  end


  subroutine columns_to_slabs(x,xc)

!  Routine to swap data from "column" orientation to "slab"  orientation.
!  See data maps for details. Relies on the specialized data  structures
!  "subslice" and "twoslice" defined in "create_MPI_data_type".

   use mpi_params
   use flags_counters_etc

   implicit none
#include "mpif.h"
      
   integer              :: status_array(mpi_status_size,3*maxprocs+1)
   integer              :: reqs(2*maxprocs)
   integer              :: i_start,kstart
   real(kind=8)         :: x(nx+2,ny,locnz)     ! slab
   double complex       :: xc(locnx,ny,n3)      ! column
   
 
!  Ship portions of my local data "column" xc needed by other nodes
!  and recieve remote data required to construct my local data
!  "slab" x.
!  1st post the receives
   j=0 ! request counter, the number that have been done
   do iproc=0,numprocs-1
    source=iproc
    i_start=(iproc*2*locnx)+1    ! start index for recieving "real"  data x       
    tag = 2
    j=j+1
    call  MPI_IRECV(x(i_start,1,1),1,subslice,source,tag,comm,reqs(j),ierr)
   enddo
   call MPI_BARRIER(comm,ierr)   ! make sure all processor's receives are posted 
                                 ! so we can use ready sends to avoid buffering

!  All processors now initiate nonblocking ready-sends, writing to the pre-established buffers
   do iproc=0,numprocs-1
    dest=iproc
    kstart=(numzplanes_below(dest))+1 ! start index for sending "complex"  data xc
       
    numwords=2*locnx*(ny)*numzplanes(dest) ! num of real(kind=8) words to store in xc  w/ each send
    tag = 2
    j=j+1
    call MPI_IRSEND(xc(1,1,kstart),numwords,MPI_DOUBLE_PRECISION,dest,tag,comm,reqs(j),ierr)
   enddo

!  make sure all communications are done before proceeding
   call MPI_WAITALL(j,reqs,status_array,ierr)      
   return
  end


  subroutine slabs_to_columns(x,xc)

!  Routine to swap data from "slab" orientation to "column"  orientation.
!  See data maps for details. Relies on the specialized data  structures
!  "subslice" and "twoslice" defined in "create_MPI_data_type".

   use mpi_params
   use flags_counters_etc

   implicit none
#include "mpif.h"
      
   integer              :: status_array(MPI_STATUS_SIZE,2048)
   integer              :: reqs(2048)
   integer              :: i_start,kstart
   real(kind=8)         :: x(nx+2,ny,locnz)
   double complex       :: xc(locnx,ny,n3)

!  Ship portions of my local data "slab" x needed by other nodes
!  and recieve remote data required to construct my local data
!  "column" xc.
!  All processors initiate nonblocking receives, establishing  buffers for the sends
   j=0 ! request counter, the number that have been done
   do iproc=0,numprocs-1
    source=iproc
    kstart=(numzplanes_below(source))+1       ! start index for recieving "complex"  data xc      
    numwords=2*locnx*(ny)*numzplanes(iproc)  ! num of double precision words to store in xc  w/ each send
    tag = 1
    j=j+1
    call  MPI_IRECV(xc(1,1,kstart),numwords,MPI_DOUBLE_PRECISION,source,tag,comm,reqs(j),ierr )
   enddo
   call MPI_BARRIER(comm,ierr)   ! make sure all processor's receives are posted 
                                 ! so we can use ready sends to avoid buffering

!  All processors now initiate nonblocking ready-sends, writing to the pre-established buffers
   do iproc=0,numprocs-1
    dest=iproc
    i_start=(iproc*2*locnx)+1  ! start index for sending "real" data  x
    numwords=1                ! i.e. 1 word of type subslice
    tag = 1
    j=j+1
    call  MPI_IRSEND(x(i_start,1,1),numwords,subslice,dest,tag,comm,reqs(j),ierr)
   enddo

   call MPI_WAITALL(j,reqs,status_array,ierr)
   return
  end      
      
      
 subroutine get_uvw_from_north( numreqs,reqs )
  use mpi_params
  use flags_counters_etc
  use particles
       
  implicit none 
#include "mpif.h"
  integer    ::  reqs(6), numreqs, kloc
       
  if(numprocs==1) then
   write(0,*) 'myid  numprocs ',myid,numprocs
   write(0,*) 'I dont have a north neighbor!!'
   stop ' stopping in get_uvw_from_north'
  endif
       
  source=myid+1!! NORTH neighbor
  if(myid==numprocs-1) source=0  !! periodicity
   dest=myid   !! me
   numwords=(nx+1)*(ny+1)  !! 1 horiz plane extended by 1 in x,y
   tag=100
   kloc = numzplanes(myid)+1
       
   j=numreqs+1
   call MPI_IRECV(gvelocity(1,1,kloc,1),numwords,    &
                  MPI_DOUBLE_PRECISION,source,tag,comm,  &
                  reqs(j),ierr )
       
   j=j+1
   call MPI_IRECV(gvelocity(1,1,kloc,2),numwords,    &
                  MPI_DOUBLE_PRECISION,source,tag,comm,  &
                  reqs(j),ierr )
       
   j=j+1
   call MPI_IRECV(gvelocity(1,1,kloc,3),numwords,    &
                  MPI_DOUBLE_PRECISION,source,tag,comm,  &
                  reqs(j),ierr )
       
   return
  end subroutine get_uvw_from_north
      
  subroutine send_uvw_south( numreqs,reqs )
   use mpi_params
   use flags_counters_etc
   use particles
       
   implicit none  
#include "mpif.h"
   integer    ::  numreqs, reqs(6), kloc
       
   if(numprocs==1) then
    write(0,*) 'myid  numprocs ',myid,numprocs
    write(0,*) 'I dont have a south neighbor!!'
    stop ' stopping in send_uvw_south'
   endif
               
   source=myid  !! me
   dest=myid-1  !! SOUTH neighbor
   if(myid==0) dest=numprocs-1  !! periodicity
   numwords=(nx+1)*(ny+1)  !! 1 horiz plane extended by 1 in x,y
   tag=100
   kloc=1
       
   !! send lowest extended u data to south
   j=numreqs+1   !! nreqs is the number of previous requests for myid
   call MPI_ISEND(gvelocity(1,1,kloc,1),numwords,   &
                   MPI_DOUBLE_PRECISION,dest,tag,comm, &
                   reqs(j),ierr)
       
       !! send lowest extended v data to south
       j=j+1
       call MPI_ISEND(gvelocity(1,1,kloc,2),numwords,   &
                       MPI_DOUBLE_PRECISION,dest,tag,comm, &
                       reqs(j),ierr)
       
       !! send lowest extended w data to south
       j=j+1
       call MPI_ISEND(gvelocity(1,1,kloc,3),numwords,   &
                       MPI_DOUBLE_PRECISION,dest,tag,comm, &
                       reqs(j),ierr)
       
       return
      end subroutine send_uvw_south

      
      subroutine update_owners
       use mpi_params
       use flags_counters_etc
       use independent_variables
       use particles
       use dimensional_params
       
       implicit none
#include "mpif.h"
       call mpi_barrier(comm,ierr)
       tmp_owners=0.d0
       !!===========================================================
       !! where necessary, wrap positions using periodicity
       !!===========================================================
       do i=1,ntrajectories
        if(positions(i,1) > Lx/Lz) positions(i,1)=positions(i,1)-(Lx/Lz)
        if(positions(i,1) < 0 )    positions(i,1)=positions(i,1)+(Lx/Lz)
        if(positions(i,2) > Ly/Lz) positions(i,2)=positions(i,2)-(Ly/Lz)
        if(positions(i,2) < 0 )    positions(i,2)=positions(i,2)+(Ly/Lz)
        if( bc_flag == 'zperiodic' ) then
         if(positions(i,3) > 1 )    positions(i,3)=positions(i,3)-(1.d0)
         if(positions(i,3) < 0 )    positions(i,3)=positions(i,3)+(1.d0)
        endif
        if(positions(i,3) >= my_z_limits(1)   &
           .and. positions(i,3) < my_z_limits(2) ) tmp_owners(i)=myid
       enddo
       
       if(numprocs==1) return
       
       call  MPI_ALLREDUCE(tmp_owners, owners, ntrajectories,  &
                           MPI_INTEGER, MPI_SUM, comm, ierr)
       !!===========================================================
       !! owners(i) is now equal to the id of the 
       !! processor that owns its current location
       !!===========================================================       
      end subroutine update_owners
      
      subroutine globally_update_particle_data
       use particles
       use mpi_params
       implicit none
#include "mpif.h"
  
       integer :: i
       real(kind=8), allocatable  :: tmp_positions(:,:)
       real(kind=8), allocatable  :: tmp_uvels(:,:)
       real(kind=8), allocatable  :: tmp_vvels(:,:)
       real(kind=8), allocatable  :: tmp_wvels(:,:)
  
       call mpi_barrier(comm,ierr)
       
       allocate( tmp_positions(ntrajectories,3) )
       allocate( tmp_uvels(ntrajectories,3) )
       allocate( tmp_vvels(ntrajectories,3) )
       allocate( tmp_wvels(ntrajectories,3) )
  
       
       do i=1,ntrajectories
        if( owners(i) .ne. myid ) then
         tmp_positions(i,:)=0.d0
         tmp_uvels(i,:)=0.d0
         tmp_vvels(i,:)=0.d0
         tmp_wvels(i,:)=0.d0
        elseif( owners(i) == myid ) then
         tmp_positions(i,:)=positions(i,:)
         tmp_uvels(i,:)=uvels(i,:)
         tmp_vvels(i,:)=vvels(i,:)
         tmp_wvels(i,:)=wvels(i,:)
        else
         write(0,*) 'doh.. unclaimed particle !'
         call quit_mpi
        endif
       enddo
  
      !! processor myid has updated information in the tmp
      !! arrays for the particles currently residing on myid
      !! with zeros for data corresponding to all other particles
      !! combine by doing global adds
       positions=0.
       do i=1,3
        call  MPI_ALLREDUCE(tmp_positions(:,i), positions(:,i), ntrajectories,  &
                            MPI_DOUBLE_PRECISION, MPI_SUM , comm, ierr)
                            
        call  MPI_ALLREDUCE(tmp_uvels(:,i), uvels(:,i), ntrajectories,  &
                            MPI_DOUBLE_PRECISION, MPI_SUM, comm, ierr)
        
        call  MPI_ALLREDUCE(tmp_vvels(:,i), vvels(:,i), ntrajectories,  &
                            MPI_DOUBLE_PRECISION, MPI_SUM, comm, ierr)
                            
        call  MPI_ALLREDUCE(tmp_wvels(:,i), wvels(:,i), ntrajectories,  &
                            MPI_DOUBLE_PRECISION, MPI_SUM, comm, ierr)
       enddo
       
      !! all processors now have identical copies of all the particle data
       deallocate( tmp_positions,tmp_uvels,tmp_vvels,tmp_wvels )
 end subroutine globally_update_particle_data